#include <mycpp/util.h>
#include <mycpp/io.h>

#include "cpp_pargen.h"

using namespace MyCpp;
using namespace Pargen;

namespace Scruffy {

static void
print_whsp (File *file,
            Size num_spaces)
{
    for (Size i = 0; i < num_spaces; i++)
        file->print (" ");
}

static void
print_tab (File *file,
           Size nest_level)
{
    print_whsp (file, nest_level * 1);
}

static Ref<Grammar> create_cpp_Scruffy_NotABrace ();
static Ref<Grammar> create_cpp_Scruffy_NotAParenthesis ();
static Ref<Grammar> create_cpp_Scruffy_SkipCompoundStatement ();
static Ref<Grammar> create_cpp_Scruffy_SkipExpressionList ();
static Ref<Grammar> create_cpp_Scruffy_SkipExceptionDeclaration ();
static Ref<Grammar> create_cpp_Scruffy_GlobalScope ();
static Ref<Grammar> create_cpp_Literal ();
static Ref<Grammar> create_cpp_StringLiteral ();
static Ref<Grammar> create_cpp_Identifier ();
static Ref<Grammar> create_cpp_TranslationUnit ();
static Ref<Grammar> create_cpp_PrimaryExpression ();
static Ref<Grammar> create_cpp_IdExpression ();
static Ref<Grammar> create_cpp_UnqualifiedId ();
static Ref<Grammar> create_cpp_QualifiedId ();
static Ref<Grammar> create_cpp_NestedNameSpecifier ();
static Ref<Grammar> create_cpp_Scruffy_NestedNameSpecifier ();
static Ref<Grammar> create_cpp_ClassOrNamespaceName ();
static Ref<Grammar> create_cpp_PostfixExpression ();
static Ref<Grammar> create_cpp_ExpressionList ();
static Ref<Grammar> create_cpp_PseudoDestructorName ();
static Ref<Grammar> create_cpp_UnaryExpression ();
static Ref<Grammar> create_cpp_UnaryOperator ();
static Ref<Grammar> create_cpp_NewExpression ();
static Ref<Grammar> create_cpp_NewPlacement ();
static Ref<Grammar> create_cpp_NewTypeId ();
static Ref<Grammar> create_cpp_NewDeclarator ();
static Ref<Grammar> create_cpp_DirectNewDeclarator ();
static Ref<Grammar> create_cpp_NewInitializer ();
static Ref<Grammar> create_cpp_DeleteExpression ();
static Ref<Grammar> create_cpp_CastExpression ();
static Ref<Grammar> create_cpp_PmExpression ();
static Ref<Grammar> create_cpp_MultiplicativeExpression ();
static Ref<Grammar> create_cpp_AdditiveExpression ();
static Ref<Grammar> create_cpp_ShiftExpression ();
static Ref<Grammar> create_cpp_RelationalExpression ();
static Ref<Grammar> create_cpp_EqualityExpression ();
static Ref<Grammar> create_cpp_AndExpression ();
static Ref<Grammar> create_cpp_ExclusiveOrExpression ();
static Ref<Grammar> create_cpp_InclusiveOrExpression ();
static Ref<Grammar> create_cpp_LogicalAndExpression ();
static Ref<Grammar> create_cpp_LogicalOrExpression ();
static Ref<Grammar> create_cpp_ConditionalExpression ();
static Ref<Grammar> create_cpp_AssignmentExpression ();
static Ref<Grammar> create_cpp_AssignmentOperator ();
static Ref<Grammar> create_cpp_Expression ();
static Ref<Grammar> create_cpp_ConstantExpression ();
static Ref<Grammar> create_cpp_Statement ();
static Ref<Grammar> create_cpp_LabeledStatement ();
static Ref<Grammar> create_cpp_ExpressionStatement ();
static Ref<Grammar> create_cpp_CompoundStatement ();
static Ref<Grammar> create_cpp_SelectionStatement ();
static Ref<Grammar> create_cpp_Condition ();
static Ref<Grammar> create_cpp_IterationStatement ();
static Ref<Grammar> create_cpp_ForInitStatement ();
static Ref<Grammar> create_cpp_JumpStatement ();
static Ref<Grammar> create_cpp_DeclarationStatement ();
static Ref<Grammar> create_cpp_DeclarationSeq ();
static Ref<Grammar> create_cpp_Declaration ();
static Ref<Grammar> create_cpp_BlockDeclaration ();
static Ref<Grammar> create_cpp_SimpleDeclaration ();
static Ref<Grammar> create_cpp_DeclSpecifier ();
static Ref<Grammar> create_cpp_DeclSpecifierSeq ();
static Ref<Grammar> create_cpp_Scruffy_DeclSpecifierSeq ();
static Ref<Grammar> create_cpp_StorageClassSpecifier ();
static Ref<Grammar> create_cpp_FunctionSpecifier ();
static Ref<Grammar> create_cpp_TypedefName ();
static Ref<Grammar> create_cpp_TypeSpecifier ();
static Ref<Grammar> create_cpp_SimpleTypeSpecifier ();
static Ref<Grammar> create_cpp_TypeName ();
static Ref<Grammar> create_cpp_ElaboratedTypeSpecifier ();
static Ref<Grammar> create_cpp_EnumName ();
static Ref<Grammar> create_cpp_EnumSpecifier ();
static Ref<Grammar> create_cpp_EnumeratorList ();
static Ref<Grammar> create_cpp_EnumeratorDefinition ();
static Ref<Grammar> create_cpp_Enumerator ();
static Ref<Grammar> create_cpp_NamespaceName ();
static Ref<Grammar> create_cpp_NamespaceDefinition ();
static Ref<Grammar> create_cpp_NamedNamespaceDefinition ();
static Ref<Grammar> create_cpp_UnnamedNamespaceDefinition ();
static Ref<Grammar> create_cpp_NamespaceBody ();
static Ref<Grammar> create_cpp_NamespaceAliasDefinition ();
static Ref<Grammar> create_cpp_QualifiedNamespaceSpecifier ();
static Ref<Grammar> create_cpp_UsingDeclaration ();
static Ref<Grammar> create_cpp_UsingDirective ();
static Ref<Grammar> create_cpp_AsmDefinition ();
static Ref<Grammar> create_cpp_LinkageSpecification ();
static Ref<Grammar> create_cpp_InitDeclaratorList ();
static Ref<Grammar> create_cpp_InitDeclarator ();
static Ref<Grammar> create_cpp_Declarator ();
static Ref<Grammar> create_cpp_Scruffy_Declarator_Common ();
static Ref<Grammar> create_cpp_Scruffy_Declarator ();
static Ref<Grammar> create_cpp_Scruffy_AbstractDeclarator ();
static Ref<Grammar> create_cpp_Scruffy_AnyDeclarator ();
static Ref<Grammar> create_cpp_DirectDeclarator ();
static Ref<Grammar> create_cpp_PtrOperator ();
static Ref<Grammar> create_cpp_CvQualifierSeq ();
static Ref<Grammar> create_cpp_CvQualifier ();
static Ref<Grammar> create_cpp_DeclaratorId ();
static Ref<Grammar> create_cpp_TypeId ();
static Ref<Grammar> create_cpp_TypeSpecifierSeq ();
static Ref<Grammar> create_cpp_Scruffy_TypeSpecifierSeq ();
static Ref<Grammar> create_cpp_ParameterDeclarationClause ();
static Ref<Grammar> create_cpp_ParameterDeclarationList ();
static Ref<Grammar> create_cpp_ParameterDeclaration ();
static Ref<Grammar> create_cpp_FunctionDefinition ();
static Ref<Grammar> create_cpp_FunctionBody ();
static Ref<Grammar> create_cpp_Initializer ();
static Ref<Grammar> create_cpp_InitializerClause ();
static Ref<Grammar> create_cpp_InitializerList ();
static Ref<Grammar> create_cpp_ClassName ();
static Ref<Grammar> create_cpp_ClassNameIdentifier ();
static Ref<Grammar> create_cpp_ClassSpecifier ();
static Ref<Grammar> create_cpp_ClassHead ();
static Ref<Grammar> create_cpp_ClassKey ();
static Ref<Grammar> create_cpp_MemberSpecification ();
static Ref<Grammar> create_cpp_MemberDeclaration ();
static Ref<Grammar> create_cpp_MemberDeclaratorList ();
static Ref<Grammar> create_cpp_MemberDeclarator ();
static Ref<Grammar> create_cpp_PureSpecifier ();
static Ref<Grammar> create_cpp_ConstantInitializer ();
static Ref<Grammar> create_cpp_BaseClause ();
static Ref<Grammar> create_cpp_BaseSpecifierList ();
static Ref<Grammar> create_cpp_BaseSpecifier ();
static Ref<Grammar> create_cpp_AccessSpecifier ();
static Ref<Grammar> create_cpp_ConversionFunctionId ();
static Ref<Grammar> create_cpp_ConversionTypeId ();
static Ref<Grammar> create_cpp_ConversionDeclarator ();
static Ref<Grammar> create_cpp_CtorInitializer ();
static Ref<Grammar> create_cpp_MemInitializerList ();
static Ref<Grammar> create_cpp_MemInitializer ();
static Ref<Grammar> create_cpp_MemInitializerId ();
static Ref<Grammar> create_cpp_OperatorFunctionId ();
static Ref<Grammar> create_cpp_Operator ();
static Ref<Grammar> create_cpp_TemplateDeclaration ();
static Ref<Grammar> create_cpp_TemplateParameterList ();
static Ref<Grammar> create_cpp_TemplateParameter ();
static Ref<Grammar> create_cpp_TypeParameter ();
static Ref<Grammar> create_cpp_TemplateId ();
static Ref<Grammar> create_cpp_TemplateName ();
static Ref<Grammar> create_cpp_TemplateArgumentList ();
static Ref<Grammar> create_cpp_TemplateArgument ();
static Ref<Grammar> create_cpp_ExplicitInstantiation ();
static Ref<Grammar> create_cpp_ExplicitSpecialization ();
static Ref<Grammar> create_cpp_TryBlock ();
static Ref<Grammar> create_cpp_FunctionTryBlock ();
static Ref<Grammar> create_cpp_HandlerSeq ();
static Ref<Grammar> create_cpp_Handler ();
static Ref<Grammar> create_cpp_ExceptionDeclaration ();
static Ref<Grammar> create_cpp_ThrowExpression ();
static Ref<Grammar> create_cpp_ExceptionSpecification ();
static Ref<Grammar> create_cpp_TypeIdList ();

bool cpp_Scruffy_NotABrace_match_func (
        Cpp_Scruffy_NotABrace *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_Scruffy_NotABrace_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Scruffy_NotABrace);

    Cpp_Scruffy_NotABrace * const &el = static_cast <Cpp_Scruffy_NotABrace*> (_el);

    return cpp_Scruffy_NotABrace_match_func (el, parser_control, data);
}

static void
cpp_Scruffy_NotABrace_set_any_token (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_NotABrace);

    Cpp_Scruffy_NotABrace * const &self = static_cast <Cpp_Scruffy_NotABrace*> (__self);

    self->any_token = static_cast <ParserElement_Token*> (___el);
}

static ParserElement*
cpp_Scruffy_NotABrace_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_NotABrace))) Cpp_Scruffy_NotABrace;
}

static Ref<Grammar>
create_cpp_Scruffy_NotABrace ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_NotABrace_creation_func));
    grammar->name = String::forData ("Scruffy_NotABrace");
    grammar->match_func = __pargen_cpp_Scruffy_NotABrace_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (""));
        entry->grammar = grammar__immediate;
        entry->assignment_func = cpp_Scruffy_NotABrace_set_any_token;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_NotABrace (Cpp_Scruffy_NotABrace const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_NotABrace\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_NotABrace\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: \n");
}

bool cpp_Scruffy_NotAParenthesis_match_func (
        Cpp_Scruffy_NotAParenthesis *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_Scruffy_NotAParenthesis_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Scruffy_NotAParenthesis);

    Cpp_Scruffy_NotAParenthesis * const &el = static_cast <Cpp_Scruffy_NotAParenthesis*> (_el);

    return cpp_Scruffy_NotAParenthesis_match_func (el, parser_control, data);
}

static void
cpp_Scruffy_NotAParenthesis_set_any_token (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_NotAParenthesis);

    Cpp_Scruffy_NotAParenthesis * const &self = static_cast <Cpp_Scruffy_NotAParenthesis*> (__self);

    self->any_token = static_cast <ParserElement_Token*> (___el);
}

static ParserElement*
cpp_Scruffy_NotAParenthesis_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_NotAParenthesis))) Cpp_Scruffy_NotAParenthesis;
}

static Ref<Grammar>
create_cpp_Scruffy_NotAParenthesis ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_NotAParenthesis_creation_func));
    grammar->name = String::forData ("Scruffy_NotAParenthesis");
    grammar->match_func = __pargen_cpp_Scruffy_NotAParenthesis_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (""));
        entry->grammar = grammar__immediate;
        entry->assignment_func = cpp_Scruffy_NotAParenthesis_set_any_token;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_NotAParenthesis (Cpp_Scruffy_NotAParenthesis const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_NotAParenthesis\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_NotAParenthesis\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: \n");
}

void cpp_Scruffy_SkipCompoundStatement_accept_func (
        Cpp_Scruffy_SkipCompoundStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_Scruffy_SkipCompoundStatement_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_Scruffy_SkipCompoundStatement_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Scruffy_SkipCompoundStatement);

    Cpp_Scruffy_SkipCompoundStatement * const &el = static_cast <Cpp_Scruffy_SkipCompoundStatement*> (_el);

    cpp_Scruffy_SkipCompoundStatement_accept_func (el, parser_control, data);
}

static void
cpp_Scruffy_SkipCompoundStatement_A_set_scruffy_NotABrace (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_SkipCompoundStatement ||
             __el->cpp_element_type != CppElement::_Scruffy_NotABrace);

    Cpp_Scruffy_SkipCompoundStatement * const &_self = static_cast <Cpp_Scruffy_SkipCompoundStatement*> (__self);

    abortIf (_self->scruffy_skip_compound_statement_type != Cpp_Scruffy_SkipCompoundStatement::_A);

    Cpp_Scruffy_SkipCompoundStatement_A * const &self = static_cast <Cpp_Scruffy_SkipCompoundStatement_A*> (_self);

    self->scruffy_NotABraces.append (static_cast <Cpp_Scruffy_NotABrace*> (__el));
}

static ParserElement*
cpp_Scruffy_SkipCompoundStatement_A_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_SkipCompoundStatement_A))) Cpp_Scruffy_SkipCompoundStatement_A;
}

static Ref<Grammar>
create_cpp_Scruffy_SkipCompoundStatement_A ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_SkipCompoundStatement_A_creation_func));
    grammar->name = String::forData ("Scruffy_SkipCompoundStatement_A");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_NotABrace ();
        entry->assignment_func = cpp_Scruffy_SkipCompoundStatement_A_set_scruffy_NotABrace;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipCompoundStatement_A (Cpp_Scruffy_SkipCompoundStatement_A const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipCompoundStatement_A\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_SkipCompoundStatement_A\n");

    {
        List<Cpp_Scruffy_NotABrace*>::DataIterator sub_iter (el->scruffy_NotABraces);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_NotABrace *&sub = sub_iter.next ();
            dump_cpp_Scruffy_NotABrace (sub, nest_level + 1);
        }
    }

}

static void
cpp_Scruffy_SkipCompoundStatement_B_set_scruffy_SkipCompoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_SkipCompoundStatement ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipCompoundStatement);

    Cpp_Scruffy_SkipCompoundStatement * const &_self = static_cast <Cpp_Scruffy_SkipCompoundStatement*> (__self);

    abortIf (_self->scruffy_skip_compound_statement_type != Cpp_Scruffy_SkipCompoundStatement::_B);

    Cpp_Scruffy_SkipCompoundStatement_B * const &self = static_cast <Cpp_Scruffy_SkipCompoundStatement_B*> (_self);

    self->scruffy_SkipCompoundStatements.append (static_cast <Cpp_Scruffy_SkipCompoundStatement*> (__el));
}

static ParserElement*
cpp_Scruffy_SkipCompoundStatement_B_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_SkipCompoundStatement_B))) Cpp_Scruffy_SkipCompoundStatement_B;
}

static Ref<Grammar>
create_cpp_Scruffy_SkipCompoundStatement_B ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_SkipCompoundStatement_B_creation_func));
    grammar->name = String::forData ("Scruffy_SkipCompoundStatement_B");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_SkipCompoundStatement ();
        entry->assignment_func = cpp_Scruffy_SkipCompoundStatement_B_set_scruffy_SkipCompoundStatement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipCompoundStatement_B (Cpp_Scruffy_SkipCompoundStatement_B const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipCompoundStatement_B\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_SkipCompoundStatement_B\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    {
        List<Cpp_Scruffy_SkipCompoundStatement*>::DataIterator sub_iter (el->scruffy_SkipCompoundStatements);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_SkipCompoundStatement *&sub = sub_iter.next ();
            dump_cpp_Scruffy_SkipCompoundStatement (sub, nest_level + 1);
        }
    }

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static Ref<Grammar>
create_cpp_Scruffy_SkipCompoundStatement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Scruffy_SkipCompoundStatement");
    grammar->accept_func = __pargen_cpp_Scruffy_SkipCompoundStatement_accept_func;

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipCompoundStatement_A ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipCompoundStatement_B ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipCompoundStatement (Cpp_Scruffy_SkipCompoundStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipCompoundStatement\n");
        return;
    }

    switch (el->scruffy_skip_compound_statement_type) {
        case Cpp_Scruffy_SkipCompoundStatement::_A:
            dump_cpp_Scruffy_SkipCompoundStatement_A (static_cast <Cpp_Scruffy_SkipCompoundStatement_A const *> (el), nest_level + 0);
            break;
        case Cpp_Scruffy_SkipCompoundStatement::_B:
            dump_cpp_Scruffy_SkipCompoundStatement_B (static_cast <Cpp_Scruffy_SkipCompoundStatement_B const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Scruffy_SkipExpressionList_A_set_scruffy_NotAParenthesis (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_SkipExpressionList ||
             __el->cpp_element_type != CppElement::_Scruffy_NotAParenthesis);

    Cpp_Scruffy_SkipExpressionList * const &_self = static_cast <Cpp_Scruffy_SkipExpressionList*> (__self);

    abortIf (_self->scruffy_skip_expression_list_type != Cpp_Scruffy_SkipExpressionList::_A);

    Cpp_Scruffy_SkipExpressionList_A * const &self = static_cast <Cpp_Scruffy_SkipExpressionList_A*> (_self);

    self->scruffy_NotAParenthesiss.append (static_cast <Cpp_Scruffy_NotAParenthesis*> (__el));
}

static ParserElement*
cpp_Scruffy_SkipExpressionList_A_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_SkipExpressionList_A))) Cpp_Scruffy_SkipExpressionList_A;
}

static Ref<Grammar>
create_cpp_Scruffy_SkipExpressionList_A ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_SkipExpressionList_A_creation_func));
    grammar->name = String::forData ("Scruffy_SkipExpressionList_A");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_NotAParenthesis ();
        entry->assignment_func = cpp_Scruffy_SkipExpressionList_A_set_scruffy_NotAParenthesis;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipExpressionList_A (Cpp_Scruffy_SkipExpressionList_A const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipExpressionList_A\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_SkipExpressionList_A\n");

    {
        List<Cpp_Scruffy_NotAParenthesis*>::DataIterator sub_iter (el->scruffy_NotAParenthesiss);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_NotAParenthesis *&sub = sub_iter.next ();
            dump_cpp_Scruffy_NotAParenthesis (sub, nest_level + 1);
        }
    }

}

static void
cpp_Scruffy_SkipExpressionList_B_set_scruffy_SkipExpressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_SkipExpressionList ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipExpressionList);

    Cpp_Scruffy_SkipExpressionList * const &_self = static_cast <Cpp_Scruffy_SkipExpressionList*> (__self);

    abortIf (_self->scruffy_skip_expression_list_type != Cpp_Scruffy_SkipExpressionList::_B);

    Cpp_Scruffy_SkipExpressionList_B * const &self = static_cast <Cpp_Scruffy_SkipExpressionList_B*> (_self);

    self->scruffy_SkipExpressionLists.append (static_cast <Cpp_Scruffy_SkipExpressionList*> (__el));
}

static ParserElement*
cpp_Scruffy_SkipExpressionList_B_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_SkipExpressionList_B))) Cpp_Scruffy_SkipExpressionList_B;
}

static Ref<Grammar>
create_cpp_Scruffy_SkipExpressionList_B ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_SkipExpressionList_B_creation_func));
    grammar->name = String::forData ("Scruffy_SkipExpressionList_B");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_SkipExpressionList ();
        entry->assignment_func = cpp_Scruffy_SkipExpressionList_B_set_scruffy_SkipExpressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipExpressionList_B (Cpp_Scruffy_SkipExpressionList_B const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipExpressionList_B\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_SkipExpressionList_B\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    {
        List<Cpp_Scruffy_SkipExpressionList*>::DataIterator sub_iter (el->scruffy_SkipExpressionLists);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_SkipExpressionList *&sub = sub_iter.next ();
            dump_cpp_Scruffy_SkipExpressionList (sub, nest_level + 1);
        }
    }

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static Ref<Grammar>
create_cpp_Scruffy_SkipExpressionList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Scruffy_SkipExpressionList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipExpressionList_A ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipExpressionList_B ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipExpressionList (Cpp_Scruffy_SkipExpressionList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipExpressionList\n");
        return;
    }

    switch (el->scruffy_skip_expression_list_type) {
        case Cpp_Scruffy_SkipExpressionList::_A:
            dump_cpp_Scruffy_SkipExpressionList_A (static_cast <Cpp_Scruffy_SkipExpressionList_A const *> (el), nest_level + 0);
            break;
        case Cpp_Scruffy_SkipExpressionList::_B:
            dump_cpp_Scruffy_SkipExpressionList_B (static_cast <Cpp_Scruffy_SkipExpressionList_B const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_Scruffy_SkipExceptionDeclaration_accept_func (
        Cpp_Scruffy_SkipExceptionDeclaration *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_Scruffy_SkipExceptionDeclaration_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_Scruffy_SkipExceptionDeclaration_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Scruffy_SkipExceptionDeclaration);

    Cpp_Scruffy_SkipExceptionDeclaration * const &el = static_cast <Cpp_Scruffy_SkipExceptionDeclaration*> (_el);

    cpp_Scruffy_SkipExceptionDeclaration_accept_func (el, parser_control, data);
}

static void
cpp_Scruffy_SkipExceptionDeclaration_A_set_scruffy_NotAParenthesis (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_SkipExceptionDeclaration ||
             __el->cpp_element_type != CppElement::_Scruffy_NotAParenthesis);

    Cpp_Scruffy_SkipExceptionDeclaration * const &_self = static_cast <Cpp_Scruffy_SkipExceptionDeclaration*> (__self);

    abortIf (_self->scruffy_skip_exception_declaration_type != Cpp_Scruffy_SkipExceptionDeclaration::_A);

    Cpp_Scruffy_SkipExceptionDeclaration_A * const &self = static_cast <Cpp_Scruffy_SkipExceptionDeclaration_A*> (_self);

    self->scruffy_NotAParenthesiss.append (static_cast <Cpp_Scruffy_NotAParenthesis*> (__el));
}

static ParserElement*
cpp_Scruffy_SkipExceptionDeclaration_A_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_SkipExceptionDeclaration_A))) Cpp_Scruffy_SkipExceptionDeclaration_A;
}

static Ref<Grammar>
create_cpp_Scruffy_SkipExceptionDeclaration_A ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_SkipExceptionDeclaration_A_creation_func));
    grammar->name = String::forData ("Scruffy_SkipExceptionDeclaration_A");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_NotAParenthesis ();
        entry->assignment_func = cpp_Scruffy_SkipExceptionDeclaration_A_set_scruffy_NotAParenthesis;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipExceptionDeclaration_A (Cpp_Scruffy_SkipExceptionDeclaration_A const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipExceptionDeclaration_A\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_SkipExceptionDeclaration_A\n");

    {
        List<Cpp_Scruffy_NotAParenthesis*>::DataIterator sub_iter (el->scruffy_NotAParenthesiss);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_NotAParenthesis *&sub = sub_iter.next ();
            dump_cpp_Scruffy_NotAParenthesis (sub, nest_level + 1);
        }
    }

}

static void
cpp_Scruffy_SkipExceptionDeclaration_B_set_scruffy_SkipExceptionDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Scruffy_SkipExceptionDeclaration ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipExceptionDeclaration);

    Cpp_Scruffy_SkipExceptionDeclaration * const &_self = static_cast <Cpp_Scruffy_SkipExceptionDeclaration*> (__self);

    abortIf (_self->scruffy_skip_exception_declaration_type != Cpp_Scruffy_SkipExceptionDeclaration::_B);

    Cpp_Scruffy_SkipExceptionDeclaration_B * const &self = static_cast <Cpp_Scruffy_SkipExceptionDeclaration_B*> (_self);

    self->scruffy_SkipExceptionDeclarations.append (static_cast <Cpp_Scruffy_SkipExceptionDeclaration*> (__el));
}

static ParserElement*
cpp_Scruffy_SkipExceptionDeclaration_B_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_SkipExceptionDeclaration_B))) Cpp_Scruffy_SkipExceptionDeclaration_B;
}

static Ref<Grammar>
create_cpp_Scruffy_SkipExceptionDeclaration_B ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_SkipExceptionDeclaration_B_creation_func));
    grammar->name = String::forData ("Scruffy_SkipExceptionDeclaration_B");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_SkipExceptionDeclaration ();
        entry->assignment_func = cpp_Scruffy_SkipExceptionDeclaration_B_set_scruffy_SkipExceptionDeclaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipExceptionDeclaration_B (Cpp_Scruffy_SkipExceptionDeclaration_B const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipExceptionDeclaration_B\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_SkipExceptionDeclaration_B\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    {
        List<Cpp_Scruffy_SkipExceptionDeclaration*>::DataIterator sub_iter (el->scruffy_SkipExceptionDeclarations);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_SkipExceptionDeclaration *&sub = sub_iter.next ();
            dump_cpp_Scruffy_SkipExceptionDeclaration (sub, nest_level + 1);
        }
    }

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static Ref<Grammar>
create_cpp_Scruffy_SkipExceptionDeclaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Scruffy_SkipExceptionDeclaration");
    grammar->accept_func = __pargen_cpp_Scruffy_SkipExceptionDeclaration_accept_func;

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipExceptionDeclaration_A ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipExceptionDeclaration_B ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_SkipExceptionDeclaration (Cpp_Scruffy_SkipExceptionDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_SkipExceptionDeclaration\n");
        return;
    }

    switch (el->scruffy_skip_exception_declaration_type) {
        case Cpp_Scruffy_SkipExceptionDeclaration::_A:
            dump_cpp_Scruffy_SkipExceptionDeclaration_A (static_cast <Cpp_Scruffy_SkipExceptionDeclaration_A const *> (el), nest_level + 0);
            break;
        case Cpp_Scruffy_SkipExceptionDeclaration::_B:
            dump_cpp_Scruffy_SkipExceptionDeclaration_B (static_cast <Cpp_Scruffy_SkipExceptionDeclaration_B const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_Scruffy_GlobalScope_accept_func (
        Cpp_Scruffy_GlobalScope *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_Scruffy_GlobalScope_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_Scruffy_GlobalScope_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_Scruffy_GlobalScope * const &el = static_cast <Cpp_Scruffy_GlobalScope*> (_el);

    cpp_Scruffy_GlobalScope_accept_func (el, parser_control, data);
}

static ParserElement*
cpp_Scruffy_GlobalScope_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Scruffy_GlobalScope))) Cpp_Scruffy_GlobalScope;
}

static Ref<Grammar>
create_cpp_Scruffy_GlobalScope ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Scruffy_GlobalScope_creation_func));
    grammar->name = String::forData ("Scruffy_GlobalScope");
    grammar->accept_func = __pargen_cpp_Scruffy_GlobalScope_accept_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("::"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Scruffy_GlobalScope (Cpp_Scruffy_GlobalScope const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Scruffy_GlobalScope\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Scruffy_GlobalScope\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ::\n");
}

static void
cpp_Literal_set_any_token (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);

    abortIf (__self->cpp_element_type != CppElement::_Literal);

    Cpp_Literal * const &self = static_cast <Cpp_Literal*> (__self);

    self->any_token = static_cast <ParserElement_Token*> (___el);
}

static ParserElement*
cpp_Literal_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Literal))) Cpp_Literal;
}

static Ref<Grammar>
create_cpp_Literal ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Literal_creation_func));
    grammar->name = String::forData ("Literal");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (""));

        bool cpp_literal_token_match_func (
                ConstMemoryDesc const &token,
                void                  *token_user_ptr,
                void                  *user_data);

        grammar__immediate->token_match_cb = cpp_literal_token_match_func;
        grammar__immediate->token_match_cb_name = grab (new String ("cpp_literal_token_match_func"));

        entry->grammar = grammar__immediate;
        entry->assignment_func = cpp_Literal_set_any_token;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Literal (Cpp_Literal const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Literal\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Literal\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: \n");
}

static void
cpp_StringLiteral_set_any_token (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);

    abortIf (__self->cpp_element_type != CppElement::_StringLiteral);

    Cpp_StringLiteral * const &self = static_cast <Cpp_StringLiteral*> (__self);

    self->any_token = static_cast <ParserElement_Token*> (___el);
}

static ParserElement*
cpp_StringLiteral_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_StringLiteral))) Cpp_StringLiteral;
}

static Ref<Grammar>
create_cpp_StringLiteral ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_StringLiteral_creation_func));
    grammar->name = String::forData ("StringLiteral");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (""));
        entry->grammar = grammar__immediate;
        entry->assignment_func = cpp_StringLiteral_set_any_token;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StringLiteral (Cpp_StringLiteral const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StringLiteral\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("StringLiteral\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: \n");
}

static void
cpp_Identifier_set_any_token (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);

    abortIf (__self->cpp_element_type != CppElement::_Identifier);

    Cpp_Identifier * const &self = static_cast <Cpp_Identifier*> (__self);

    self->any_token = static_cast <ParserElement_Token*> (___el);
}

static ParserElement*
cpp_Identifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Identifier))) Cpp_Identifier;
}

static Ref<Grammar>
create_cpp_Identifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Identifier_creation_func));
    grammar->name = String::forData ("Identifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (""));

        bool cpp_identifier_token_match_func (
                ConstMemoryDesc const &token,
                void                  *token_user_ptr,
                void                  *user_data);

        grammar__immediate->token_match_cb = cpp_identifier_token_match_func;
        grammar__immediate->token_match_cb_name = grab (new String ("cpp_identifier_token_match_func"));

        entry->grammar = grammar__immediate;
        entry->assignment_func = cpp_Identifier_set_any_token;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Identifier (Cpp_Identifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Identifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Identifier\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: \n");
}

static void
cpp_TranslationUnit_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TranslationUnit ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_TranslationUnit * const &self = static_cast <Cpp_TranslationUnit*> (__self);

    self->declarations.append (static_cast <Cpp_Declaration*> (__el));
}

static ParserElement*
cpp_TranslationUnit_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TranslationUnit))) Cpp_TranslationUnit;
}

static Ref<Grammar>
create_cpp_TranslationUnit ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TranslationUnit_creation_func));
    grammar->name = String::forData ("TranslationUnit");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_TranslationUnit_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TranslationUnit (Cpp_TranslationUnit const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TranslationUnit\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TranslationUnit\n");

    {
        List<Cpp_Declaration*>::DataIterator sub_iter (el->declarations);
        while (!sub_iter.done ()) {
            Cpp_Declaration *&sub = sub_iter.next ();
            dump_cpp_Declaration (sub, nest_level + 1);
        }
    }

}

static void
cpp_PrimaryExpression_Literal_set_literal (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PrimaryExpression ||
             __el->cpp_element_type != CppElement::_Literal);

    Cpp_PrimaryExpression * const &_self = static_cast <Cpp_PrimaryExpression*> (__self);

    abortIf (_self->primary_expression_type != Cpp_PrimaryExpression::_Literal);

    Cpp_PrimaryExpression_Literal * const &self = static_cast <Cpp_PrimaryExpression_Literal*> (_self);

    self->literal = static_cast <Cpp_Literal*> (___el);
}

bool cpp_accept_primary_expression (
        Cpp_PrimaryExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_primary_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_PrimaryExpression);

    Cpp_PrimaryExpression * const &el =
            static_cast <Cpp_PrimaryExpression*> (_el);

    return cpp_accept_primary_expression (el, parser_control, data);
}

static ParserElement*
cpp_PrimaryExpression_Literal_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PrimaryExpression_Literal))) Cpp_PrimaryExpression_Literal;
}

static Ref<Grammar>
create_cpp_PrimaryExpression_Literal ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PrimaryExpression_Literal_creation_func));
    grammar->name = String::forData ("PrimaryExpression_Literal");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Literal ();
        entry->assignment_func = cpp_PrimaryExpression_Literal_set_literal;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_primary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PrimaryExpression_Literal (Cpp_PrimaryExpression_Literal const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PrimaryExpression_Literal\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PrimaryExpression_Literal\n");

    dump_cpp_Literal (el->literal, nest_level + 1);

}

static ParserElement*
cpp_PrimaryExpression_This_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PrimaryExpression_This))) Cpp_PrimaryExpression_This;
}

static Ref<Grammar>
create_cpp_PrimaryExpression_This ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PrimaryExpression_This_creation_func));
    grammar->name = String::forData ("PrimaryExpression_This");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("this"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_primary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PrimaryExpression_This (Cpp_PrimaryExpression_This const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PrimaryExpression_This\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PrimaryExpression_This\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: this\n");
}

static void
cpp_PrimaryExpression_Braces_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PrimaryExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PrimaryExpression * const &_self = static_cast <Cpp_PrimaryExpression*> (__self);

    abortIf (_self->primary_expression_type != Cpp_PrimaryExpression::_Braces);

    Cpp_PrimaryExpression_Braces * const &self = static_cast <Cpp_PrimaryExpression_Braces*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_PrimaryExpression_Braces_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PrimaryExpression_Braces))) Cpp_PrimaryExpression_Braces;
}

static Ref<Grammar>
create_cpp_PrimaryExpression_Braces ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PrimaryExpression_Braces_creation_func));
    grammar->name = String::forData ("PrimaryExpression_Braces");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PrimaryExpression_Braces_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_primary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PrimaryExpression_Braces (Cpp_PrimaryExpression_Braces const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PrimaryExpression_Braces\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PrimaryExpression_Braces\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

bool cpp_begin_id (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_id (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_id (_el, parser_control, data);
}

static void
cpp_PrimaryExpression_IdExpression_set_idExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PrimaryExpression ||
             __el->cpp_element_type != CppElement::_IdExpression);

    Cpp_PrimaryExpression * const &_self = static_cast <Cpp_PrimaryExpression*> (__self);

    abortIf (_self->primary_expression_type != Cpp_PrimaryExpression::_IdExpression);

    Cpp_PrimaryExpression_IdExpression * const &self = static_cast <Cpp_PrimaryExpression_IdExpression*> (_self);

    self->idExpression = static_cast <Cpp_IdExpression*> (___el);
}

static ParserElement*
cpp_PrimaryExpression_IdExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PrimaryExpression_IdExpression))) Cpp_PrimaryExpression_IdExpression;
}

static Ref<Grammar>
create_cpp_PrimaryExpression_IdExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PrimaryExpression_IdExpression_creation_func));
    grammar->name = String::forData ("PrimaryExpression_IdExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IdExpression ();
        entry->assignment_func = cpp_PrimaryExpression_IdExpression_set_idExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_primary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PrimaryExpression_IdExpression (Cpp_PrimaryExpression_IdExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PrimaryExpression_IdExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PrimaryExpression_IdExpression\n");

    dump_cpp_IdExpression (el->idExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_PrimaryExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("PrimaryExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PrimaryExpression_Literal ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PrimaryExpression_This ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PrimaryExpression_Braces ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PrimaryExpression_IdExpression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PrimaryExpression (Cpp_PrimaryExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PrimaryExpression\n");
        return;
    }

    switch (el->primary_expression_type) {
        case Cpp_PrimaryExpression::_Literal:
            dump_cpp_PrimaryExpression_Literal (static_cast <Cpp_PrimaryExpression_Literal const *> (el), nest_level + 0);
            break;
        case Cpp_PrimaryExpression::_This:
            dump_cpp_PrimaryExpression_This (static_cast <Cpp_PrimaryExpression_This const *> (el), nest_level + 0);
            break;
        case Cpp_PrimaryExpression::_Braces:
            dump_cpp_PrimaryExpression_Braces (static_cast <Cpp_PrimaryExpression_Braces const *> (el), nest_level + 0);
            break;
        case Cpp_PrimaryExpression::_IdExpression:
            dump_cpp_PrimaryExpression_IdExpression (static_cast <Cpp_PrimaryExpression_IdExpression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_IdExpression_QualifiedId_set_qualifiedId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IdExpression ||
             __el->cpp_element_type != CppElement::_QualifiedId);

    Cpp_IdExpression * const &_self = static_cast <Cpp_IdExpression*> (__self);

    abortIf (_self->id_expression_type != Cpp_IdExpression::_QualifiedId);

    Cpp_IdExpression_QualifiedId * const &self = static_cast <Cpp_IdExpression_QualifiedId*> (_self);

    self->qualifiedId = static_cast <Cpp_QualifiedId*> (___el);
}

static ParserElement*
cpp_IdExpression_QualifiedId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_IdExpression_QualifiedId))) Cpp_IdExpression_QualifiedId;
}

static Ref<Grammar>
create_cpp_IdExpression_QualifiedId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_IdExpression_QualifiedId_creation_func));
    grammar->name = String::forData ("IdExpression_QualifiedId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_QualifiedId ();
        entry->assignment_func = cpp_IdExpression_QualifiedId_set_qualifiedId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IdExpression_QualifiedId (Cpp_IdExpression_QualifiedId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IdExpression_QualifiedId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("IdExpression_QualifiedId\n");

    dump_cpp_QualifiedId (el->qualifiedId, nest_level + 1);

}

bool cpp_begin_nested_name (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_nested_name (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_nested_name (_el, parser_control, data);
}

static void
cpp_IdExpression_UnqualifiedId_set_unqualifiedId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IdExpression ||
             __el->cpp_element_type != CppElement::_UnqualifiedId);

    Cpp_IdExpression * const &_self = static_cast <Cpp_IdExpression*> (__self);

    abortIf (_self->id_expression_type != Cpp_IdExpression::_UnqualifiedId);

    Cpp_IdExpression_UnqualifiedId * const &self = static_cast <Cpp_IdExpression_UnqualifiedId*> (_self);

    self->unqualifiedId = static_cast <Cpp_UnqualifiedId*> (___el);
}

static ParserElement*
cpp_IdExpression_UnqualifiedId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_IdExpression_UnqualifiedId))) Cpp_IdExpression_UnqualifiedId;
}

static Ref<Grammar>
create_cpp_IdExpression_UnqualifiedId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_IdExpression_UnqualifiedId_creation_func));
    grammar->name = String::forData ("IdExpression_UnqualifiedId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId ();
        entry->assignment_func = cpp_IdExpression_UnqualifiedId_set_unqualifiedId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IdExpression_UnqualifiedId (Cpp_IdExpression_UnqualifiedId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IdExpression_UnqualifiedId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("IdExpression_UnqualifiedId\n");

    dump_cpp_UnqualifiedId (el->unqualifiedId, nest_level + 1);

}

static Ref<Grammar>
create_cpp_IdExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("IdExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_IdExpression_QualifiedId ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_IdExpression_UnqualifiedId ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IdExpression (Cpp_IdExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IdExpression\n");
        return;
    }

    switch (el->id_expression_type) {
        case Cpp_IdExpression::_QualifiedId:
            dump_cpp_IdExpression_QualifiedId (static_cast <Cpp_IdExpression_QualifiedId const *> (el), nest_level + 0);
            break;
        case Cpp_IdExpression::_UnqualifiedId:
            dump_cpp_IdExpression_UnqualifiedId (static_cast <Cpp_IdExpression_UnqualifiedId const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_UnqualifiedId_Destructor_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnqualifiedId ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_UnqualifiedId * const &_self = static_cast <Cpp_UnqualifiedId*> (__self);

    abortIf (_self->unqualified_id_type != Cpp_UnqualifiedId::_Destructor);

    Cpp_UnqualifiedId_Destructor * const &self = static_cast <Cpp_UnqualifiedId_Destructor*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_UnqualifiedId_Destructor_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnqualifiedId_Destructor))) Cpp_UnqualifiedId_Destructor;
}

static Ref<Grammar>
create_cpp_UnqualifiedId_Destructor ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnqualifiedId_Destructor_creation_func));
    grammar->name = String::forData ("UnqualifiedId_Destructor");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("~"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_UnqualifiedId_Destructor_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnqualifiedId_Destructor (Cpp_UnqualifiedId_Destructor const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnqualifiedId_Destructor\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnqualifiedId_Destructor\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ~\n");
    dump_cpp_ClassName (el->className, nest_level + 1);

}

static void
cpp_UnqualifiedId_OperatorFunction_set_operatorFunctionId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnqualifiedId ||
             __el->cpp_element_type != CppElement::_OperatorFunctionId);

    Cpp_UnqualifiedId * const &_self = static_cast <Cpp_UnqualifiedId*> (__self);

    abortIf (_self->unqualified_id_type != Cpp_UnqualifiedId::_OperatorFunction);

    Cpp_UnqualifiedId_OperatorFunction * const &self = static_cast <Cpp_UnqualifiedId_OperatorFunction*> (_self);

    self->operatorFunctionId = static_cast <Cpp_OperatorFunctionId*> (___el);
}

static ParserElement*
cpp_UnqualifiedId_OperatorFunction_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnqualifiedId_OperatorFunction))) Cpp_UnqualifiedId_OperatorFunction;
}

static Ref<Grammar>
create_cpp_UnqualifiedId_OperatorFunction ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnqualifiedId_OperatorFunction_creation_func));
    grammar->name = String::forData ("UnqualifiedId_OperatorFunction");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_OperatorFunctionId ();
        entry->assignment_func = cpp_UnqualifiedId_OperatorFunction_set_operatorFunctionId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnqualifiedId_OperatorFunction (Cpp_UnqualifiedId_OperatorFunction const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnqualifiedId_OperatorFunction\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnqualifiedId_OperatorFunction\n");

    dump_cpp_OperatorFunctionId (el->operatorFunctionId, nest_level + 1);

}

static void
cpp_UnqualifiedId_ConversionFunction_set_conversionFunctionId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnqualifiedId ||
             __el->cpp_element_type != CppElement::_ConversionFunctionId);

    Cpp_UnqualifiedId * const &_self = static_cast <Cpp_UnqualifiedId*> (__self);

    abortIf (_self->unqualified_id_type != Cpp_UnqualifiedId::_ConversionFunction);

    Cpp_UnqualifiedId_ConversionFunction * const &self = static_cast <Cpp_UnqualifiedId_ConversionFunction*> (_self);

    self->conversionFunctionId = static_cast <Cpp_ConversionFunctionId*> (___el);
}

static ParserElement*
cpp_UnqualifiedId_ConversionFunction_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnqualifiedId_ConversionFunction))) Cpp_UnqualifiedId_ConversionFunction;
}

static Ref<Grammar>
create_cpp_UnqualifiedId_ConversionFunction ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnqualifiedId_ConversionFunction_creation_func));
    grammar->name = String::forData ("UnqualifiedId_ConversionFunction");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConversionFunctionId ();
        entry->assignment_func = cpp_UnqualifiedId_ConversionFunction_set_conversionFunctionId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnqualifiedId_ConversionFunction (Cpp_UnqualifiedId_ConversionFunction const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnqualifiedId_ConversionFunction\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnqualifiedId_ConversionFunction\n");

    dump_cpp_ConversionFunctionId (el->conversionFunctionId, nest_level + 1);

}

static void
cpp_UnqualifiedId_Template_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnqualifiedId ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_UnqualifiedId * const &_self = static_cast <Cpp_UnqualifiedId*> (__self);

    abortIf (_self->unqualified_id_type != Cpp_UnqualifiedId::_Template);

    Cpp_UnqualifiedId_Template * const &self = static_cast <Cpp_UnqualifiedId_Template*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static ParserElement*
cpp_UnqualifiedId_Template_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnqualifiedId_Template))) Cpp_UnqualifiedId_Template;
}

static Ref<Grammar>
create_cpp_UnqualifiedId_Template ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnqualifiedId_Template_creation_func));
    grammar->name = String::forData ("UnqualifiedId_Template");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_UnqualifiedId_Template_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnqualifiedId_Template (Cpp_UnqualifiedId_Template const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnqualifiedId_Template\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnqualifiedId_Template\n");

    dump_cpp_TemplateId (el->templateId, nest_level + 1);

}

static void
cpp_UnqualifiedId_Identifier_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnqualifiedId ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_UnqualifiedId * const &_self = static_cast <Cpp_UnqualifiedId*> (__self);

    abortIf (_self->unqualified_id_type != Cpp_UnqualifiedId::_Identifier);

    Cpp_UnqualifiedId_Identifier * const &self = static_cast <Cpp_UnqualifiedId_Identifier*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_UnqualifiedId_Identifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnqualifiedId_Identifier))) Cpp_UnqualifiedId_Identifier;
}

static Ref<Grammar>
create_cpp_UnqualifiedId_Identifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnqualifiedId_Identifier_creation_func));
    grammar->name = String::forData ("UnqualifiedId_Identifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_UnqualifiedId_Identifier_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnqualifiedId_Identifier (Cpp_UnqualifiedId_Identifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnqualifiedId_Identifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnqualifiedId_Identifier\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static Ref<Grammar>
create_cpp_UnqualifiedId ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("UnqualifiedId");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId_Destructor ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId_OperatorFunction ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId_ConversionFunction ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId_Template ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId_Identifier ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnqualifiedId (Cpp_UnqualifiedId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnqualifiedId\n");
        return;
    }

    switch (el->unqualified_id_type) {
        case Cpp_UnqualifiedId::_Destructor:
            dump_cpp_UnqualifiedId_Destructor (static_cast <Cpp_UnqualifiedId_Destructor const *> (el), nest_level + 0);
            break;
        case Cpp_UnqualifiedId::_OperatorFunction:
            dump_cpp_UnqualifiedId_OperatorFunction (static_cast <Cpp_UnqualifiedId_OperatorFunction const *> (el), nest_level + 0);
            break;
        case Cpp_UnqualifiedId::_ConversionFunction:
            dump_cpp_UnqualifiedId_ConversionFunction (static_cast <Cpp_UnqualifiedId_ConversionFunction const *> (el), nest_level + 0);
            break;
        case Cpp_UnqualifiedId::_Template:
            dump_cpp_UnqualifiedId_Template (static_cast <Cpp_UnqualifiedId_Template const *> (el), nest_level + 0);
            break;
        case Cpp_UnqualifiedId::_Identifier:
            dump_cpp_UnqualifiedId_Identifier (static_cast <Cpp_UnqualifiedId_Identifier const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_QualifiedId_NestedNameSpecifier_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_NestedNameSpecifier);

    Cpp_QualifiedId_NestedNameSpecifier * const &self = static_cast <Cpp_QualifiedId_NestedNameSpecifier*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_QualifiedId_NestedNameSpecifier_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_NestedNameSpecifier);

    Cpp_QualifiedId_NestedNameSpecifier * const &self = static_cast <Cpp_QualifiedId_NestedNameSpecifier*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

bool cpp_accept_qualified_id_nested_name (
        Cpp_QualifiedId *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_qualified_id_nested_name (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_QualifiedId);

    Cpp_QualifiedId * const &el =
            static_cast <Cpp_QualifiedId*> (_el);

    return cpp_accept_qualified_id_nested_name (el, parser_control, data);
}

static void
cpp_QualifiedId_NestedNameSpecifier_set_unqualifiedId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_UnqualifiedId);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_NestedNameSpecifier);

    Cpp_QualifiedId_NestedNameSpecifier * const &self = static_cast <Cpp_QualifiedId_NestedNameSpecifier*> (_self);

    self->unqualifiedId = static_cast <Cpp_UnqualifiedId*> (___el);
}

static ParserElement*
cpp_QualifiedId_NestedNameSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_QualifiedId_NestedNameSpecifier))) Cpp_QualifiedId_NestedNameSpecifier;
}

static Ref<Grammar>
create_cpp_QualifiedId_NestedNameSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_QualifiedId_NestedNameSpecifier_creation_func));
    grammar->name = String::forData ("QualifiedId_NestedNameSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_QualifiedId_NestedNameSpecifier_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_QualifiedId_NestedNameSpecifier_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_qualified_id_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId ();
        entry->assignment_func = cpp_QualifiedId_NestedNameSpecifier_set_unqualifiedId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_QualifiedId_NestedNameSpecifier (Cpp_QualifiedId_NestedNameSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) QualifiedId_NestedNameSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("QualifiedId_NestedNameSpecifier\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: template_opt\n");
    dump_cpp_UnqualifiedId (el->unqualifiedId, nest_level + 1);

}

static void
cpp_QualifiedId_Identifier_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_Identifier);

    Cpp_QualifiedId_Identifier * const &self = static_cast <Cpp_QualifiedId_Identifier*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_QualifiedId_Identifier_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_Identifier);

    Cpp_QualifiedId_Identifier * const &self = static_cast <Cpp_QualifiedId_Identifier*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_QualifiedId_Identifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_QualifiedId_Identifier))) Cpp_QualifiedId_Identifier;
}

static Ref<Grammar>
create_cpp_QualifiedId_Identifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_QualifiedId_Identifier_creation_func));
    grammar->name = String::forData ("QualifiedId_Identifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_QualifiedId_Identifier_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_qualified_id_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_QualifiedId ();
        entry->jump_switch_grammar_index = 2;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_QualifiedId_Identifier_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_QualifiedId_Identifier (Cpp_QualifiedId_Identifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) QualifiedId_Identifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("QualifiedId_Identifier\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_QualifiedId_OperatorFunction_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_OperatorFunction);

    Cpp_QualifiedId_OperatorFunction * const &self = static_cast <Cpp_QualifiedId_OperatorFunction*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_QualifiedId_OperatorFunction_set_operatorFunctionId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_OperatorFunctionId);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_OperatorFunction);

    Cpp_QualifiedId_OperatorFunction * const &self = static_cast <Cpp_QualifiedId_OperatorFunction*> (_self);

    self->operatorFunctionId = static_cast <Cpp_OperatorFunctionId*> (___el);
}

static ParserElement*
cpp_QualifiedId_OperatorFunction_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_QualifiedId_OperatorFunction))) Cpp_QualifiedId_OperatorFunction;
}

static Ref<Grammar>
create_cpp_QualifiedId_OperatorFunction ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_QualifiedId_OperatorFunction_creation_func));
    grammar->name = String::forData ("QualifiedId_OperatorFunction");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_QualifiedId_OperatorFunction_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_qualified_id_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_QualifiedId ();
        entry->jump_switch_grammar_index = 3;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_OperatorFunctionId ();
        entry->assignment_func = cpp_QualifiedId_OperatorFunction_set_operatorFunctionId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_QualifiedId_OperatorFunction (Cpp_QualifiedId_OperatorFunction const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) QualifiedId_OperatorFunction\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("QualifiedId_OperatorFunction\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_OperatorFunctionId (el->operatorFunctionId, nest_level + 1);

}

static void
cpp_QualifiedId_Template_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_Template);

    Cpp_QualifiedId_Template * const &self = static_cast <Cpp_QualifiedId_Template*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_QualifiedId_Template_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedId ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_QualifiedId * const &_self = static_cast <Cpp_QualifiedId*> (__self);

    abortIf (_self->qualified_id_type != Cpp_QualifiedId::_Template);

    Cpp_QualifiedId_Template * const &self = static_cast <Cpp_QualifiedId_Template*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static ParserElement*
cpp_QualifiedId_Template_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_QualifiedId_Template))) Cpp_QualifiedId_Template;
}

static Ref<Grammar>
create_cpp_QualifiedId_Template ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_QualifiedId_Template_creation_func));
    grammar->name = String::forData ("QualifiedId_Template");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_QualifiedId_Template_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_qualified_id_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_QualifiedId_Template_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_QualifiedId_Template (Cpp_QualifiedId_Template const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) QualifiedId_Template\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("QualifiedId_Template\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_TemplateId (el->templateId, nest_level + 1);

}

static Ref<Grammar>
create_cpp_QualifiedId ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("QualifiedId");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_QualifiedId_NestedNameSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_QualifiedId_Identifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_QualifiedId_OperatorFunction ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_QualifiedId_Template ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_QualifiedId (Cpp_QualifiedId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) QualifiedId\n");
        return;
    }

    switch (el->qualified_id_type) {
        case Cpp_QualifiedId::_NestedNameSpecifier:
            dump_cpp_QualifiedId_NestedNameSpecifier (static_cast <Cpp_QualifiedId_NestedNameSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_QualifiedId::_Identifier:
            dump_cpp_QualifiedId_Identifier (static_cast <Cpp_QualifiedId_Identifier const *> (el), nest_level + 0);
            break;
        case Cpp_QualifiedId::_OperatorFunction:
            dump_cpp_QualifiedId_OperatorFunction (static_cast <Cpp_QualifiedId_OperatorFunction const *> (el), nest_level + 0);
            break;
        case Cpp_QualifiedId::_Template:
            dump_cpp_QualifiedId_Template (static_cast <Cpp_QualifiedId_Template const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_NestedNameSpecifier_Template_set_classOrNamespaceName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NestedNameSpecifier ||
             __el->cpp_element_type != CppElement::_ClassOrNamespaceName);

    Cpp_NestedNameSpecifier * const &_self = static_cast <Cpp_NestedNameSpecifier*> (__self);

    abortIf (_self->nested_name_specifier_type != Cpp_NestedNameSpecifier::_Template);

    Cpp_NestedNameSpecifier_Template * const &self = static_cast <Cpp_NestedNameSpecifier_Template*> (_self);

    self->classOrNamespaceName = static_cast <Cpp_ClassOrNamespaceName*> (___el);
}

bool cpp_NestedNameSpecifier_ClassOrNamespaceName_accept_func (
        Cpp_NestedNameSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_NestedNameSpecifier_ClassOrNamespaceName_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_NestedNameSpecifier * const &el =
            static_cast <Cpp_NestedNameSpecifier*> (_el);

    return cpp_NestedNameSpecifier_ClassOrNamespaceName_accept_func (el, parser_control, data);
}

static void
cpp_NestedNameSpecifier_Template_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NestedNameSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_NestedNameSpecifier * const &_self = static_cast <Cpp_NestedNameSpecifier*> (__self);

    abortIf (_self->nested_name_specifier_type != Cpp_NestedNameSpecifier::_Template);

    Cpp_NestedNameSpecifier_Template * const &self = static_cast <Cpp_NestedNameSpecifier_Template*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static ParserElement*
cpp_NestedNameSpecifier_Template_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NestedNameSpecifier_Template))) Cpp_NestedNameSpecifier_Template;
}

static Ref<Grammar>
create_cpp_NestedNameSpecifier_Template ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NestedNameSpecifier_Template_creation_func));
    grammar->name = String::forData ("NestedNameSpecifier_Template");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassOrNamespaceName ();
        entry->assignment_func = cpp_NestedNameSpecifier_Template_set_classOrNamespaceName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_NestedNameSpecifier ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("::"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_NestedNameSpecifier_ClassOrNamespaceName_accept_func;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_NestedNameSpecifier ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NestedNameSpecifier ();
        entry->assignment_func = cpp_NestedNameSpecifier_Template_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NestedNameSpecifier_Template (Cpp_NestedNameSpecifier_Template const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NestedNameSpecifier_Template\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NestedNameSpecifier_Template\n");

    dump_cpp_ClassOrNamespaceName (el->classOrNamespaceName, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ::\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

}

static void
cpp_NestedNameSpecifier_Generic_set_classOrNamespaceName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NestedNameSpecifier ||
             __el->cpp_element_type != CppElement::_ClassOrNamespaceName);

    Cpp_NestedNameSpecifier * const &_self = static_cast <Cpp_NestedNameSpecifier*> (__self);

    abortIf (_self->nested_name_specifier_type != Cpp_NestedNameSpecifier::_Generic);

    Cpp_NestedNameSpecifier_Generic * const &self = static_cast <Cpp_NestedNameSpecifier_Generic*> (_self);

    self->classOrNamespaceName = static_cast <Cpp_ClassOrNamespaceName*> (___el);
}

static void
cpp_NestedNameSpecifier_Generic_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NestedNameSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_NestedNameSpecifier * const &_self = static_cast <Cpp_NestedNameSpecifier*> (__self);

    abortIf (_self->nested_name_specifier_type != Cpp_NestedNameSpecifier::_Generic);

    Cpp_NestedNameSpecifier_Generic * const &self = static_cast <Cpp_NestedNameSpecifier_Generic*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static ParserElement*
cpp_NestedNameSpecifier_Generic_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NestedNameSpecifier_Generic))) Cpp_NestedNameSpecifier_Generic;
}

static Ref<Grammar>
create_cpp_NestedNameSpecifier_Generic ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NestedNameSpecifier_Generic_creation_func));
    grammar->name = String::forData ("NestedNameSpecifier_Generic");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassOrNamespaceName ();
        entry->assignment_func = cpp_NestedNameSpecifier_Generic_set_classOrNamespaceName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("::"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_NestedNameSpecifier_ClassOrNamespaceName_accept_func;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NestedNameSpecifier ();
        entry->assignment_func = cpp_NestedNameSpecifier_Generic_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NestedNameSpecifier_Generic (Cpp_NestedNameSpecifier_Generic const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NestedNameSpecifier_Generic\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NestedNameSpecifier_Generic\n");

    dump_cpp_ClassOrNamespaceName (el->classOrNamespaceName, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ::\n");
    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

}

static Ref<Grammar>
create_cpp_NestedNameSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("NestedNameSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NestedNameSpecifier_Template ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NestedNameSpecifier_Generic ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NestedNameSpecifier (Cpp_NestedNameSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NestedNameSpecifier\n");
        return;
    }

    switch (el->nested_name_specifier_type) {
        case Cpp_NestedNameSpecifier::_Template:
            dump_cpp_NestedNameSpecifier_Template (static_cast <Cpp_NestedNameSpecifier_Template const *> (el), nest_level + 0);
            break;
        case Cpp_NestedNameSpecifier::_Generic:
            dump_cpp_NestedNameSpecifier_Generic (static_cast <Cpp_NestedNameSpecifier_Generic const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static Ref<Grammar>
create_cpp_Scruffy_NestedNameSpecifier ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_NestedNameSpecifier");
    grammar->aliased_grammar = create_cpp_NestedNameSpecifier ();


    return grammar.ptr ();
}

static void
cpp_ClassOrNamespaceName_ClassName_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassOrNamespaceName ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_ClassOrNamespaceName * const &_self = static_cast <Cpp_ClassOrNamespaceName*> (__self);

    abortIf (_self->class_or_namespace_name_type != Cpp_ClassOrNamespaceName::_ClassName);

    Cpp_ClassOrNamespaceName_ClassName * const &self = static_cast <Cpp_ClassOrNamespaceName_ClassName*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_ClassOrNamespaceName_ClassName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassOrNamespaceName_ClassName))) Cpp_ClassOrNamespaceName_ClassName;
}

static Ref<Grammar>
create_cpp_ClassOrNamespaceName_ClassName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassOrNamespaceName_ClassName_creation_func));
    grammar->name = String::forData ("ClassOrNamespaceName_ClassName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_ClassOrNamespaceName_ClassName_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassOrNamespaceName_ClassName (Cpp_ClassOrNamespaceName_ClassName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassOrNamespaceName_ClassName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassOrNamespaceName_ClassName\n");

    dump_cpp_ClassName (el->className, nest_level + 1);

}

static void
cpp_ClassOrNamespaceName_NamespaceName_set_namespaceName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassOrNamespaceName ||
             __el->cpp_element_type != CppElement::_NamespaceName);

    Cpp_ClassOrNamespaceName * const &_self = static_cast <Cpp_ClassOrNamespaceName*> (__self);

    abortIf (_self->class_or_namespace_name_type != Cpp_ClassOrNamespaceName::_NamespaceName);

    Cpp_ClassOrNamespaceName_NamespaceName * const &self = static_cast <Cpp_ClassOrNamespaceName_NamespaceName*> (_self);

    self->namespaceName = static_cast <Cpp_NamespaceName*> (___el);
}

static ParserElement*
cpp_ClassOrNamespaceName_NamespaceName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassOrNamespaceName_NamespaceName))) Cpp_ClassOrNamespaceName_NamespaceName;
}

static Ref<Grammar>
create_cpp_ClassOrNamespaceName_NamespaceName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassOrNamespaceName_NamespaceName_creation_func));
    grammar->name = String::forData ("ClassOrNamespaceName_NamespaceName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceName ();
        entry->assignment_func = cpp_ClassOrNamespaceName_NamespaceName_set_namespaceName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassOrNamespaceName_NamespaceName (Cpp_ClassOrNamespaceName_NamespaceName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassOrNamespaceName_NamespaceName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassOrNamespaceName_NamespaceName\n");

    dump_cpp_NamespaceName (el->namespaceName, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ClassOrNamespaceName ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ClassOrNamespaceName");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassOrNamespaceName_ClassName ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassOrNamespaceName_NamespaceName ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassOrNamespaceName (Cpp_ClassOrNamespaceName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassOrNamespaceName\n");
        return;
    }

    switch (el->class_or_namespace_name_type) {
        case Cpp_ClassOrNamespaceName::_ClassName:
            dump_cpp_ClassOrNamespaceName_ClassName (static_cast <Cpp_ClassOrNamespaceName_ClassName const *> (el), nest_level + 0);
            break;
        case Cpp_ClassOrNamespaceName::_NamespaceName:
            dump_cpp_ClassOrNamespaceName_NamespaceName (static_cast <Cpp_ClassOrNamespaceName_NamespaceName const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_PostfixExpression_Subscripting_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_Subscripting);

    Cpp_PostfixExpression_Subscripting * const &self = static_cast <Cpp_PostfixExpression_Subscripting*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

bool cpp_begin_expression (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_expression (_el, parser_control, data);
}

static void
cpp_PostfixExpression_Subscripting_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_Subscripting);

    Cpp_PostfixExpression_Subscripting * const &self = static_cast <Cpp_PostfixExpression_Subscripting*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

bool cpp_accept_postfix_expression (
        Cpp_PostfixExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_postfix_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &el =
            static_cast <Cpp_PostfixExpression*> (_el);

    return cpp_accept_postfix_expression (el, parser_control, data);
}

static ParserElement*
cpp_PostfixExpression_Subscripting_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_Subscripting))) Cpp_PostfixExpression_Subscripting;
}

static Ref<Grammar>
create_cpp_PostfixExpression_Subscripting ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_Subscripting_creation_func));
    grammar->name = String::forData ("PostfixExpression_Subscripting");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_Subscripting_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PostfixExpression_Subscripting_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_Subscripting (Cpp_PostfixExpression_Subscripting const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_Subscripting\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_Subscripting\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static void
cpp_PostfixExpression_FunctionCall_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_FunctionCall);

    Cpp_PostfixExpression_FunctionCall * const &self = static_cast <Cpp_PostfixExpression_FunctionCall*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static void
cpp_PostfixExpression_FunctionCall_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_FunctionCall);

    Cpp_PostfixExpression_FunctionCall * const &self = static_cast <Cpp_PostfixExpression_FunctionCall*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_PostfixExpression_FunctionCall_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_FunctionCall))) Cpp_PostfixExpression_FunctionCall;
}

static Ref<Grammar>
create_cpp_PostfixExpression_FunctionCall ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_FunctionCall_creation_func));
    grammar->name = String::forData ("PostfixExpression_FunctionCall");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_FunctionCall_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_PostfixExpression_FunctionCall_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_FunctionCall (Cpp_PostfixExpression_FunctionCall const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_FunctionCall\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_FunctionCall\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

bool cpp_begin_declaration (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_declaration (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_declaration (_el, parser_control, data);
}

static void
cpp_PostfixExpression_TypeConversion_set_simpleTypeSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_SimpleTypeSpecifier);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypeConversion);

    Cpp_PostfixExpression_TypeConversion * const &self = static_cast <Cpp_PostfixExpression_TypeConversion*> (_self);

    self->simpleTypeSpecifier = static_cast <Cpp_SimpleTypeSpecifier*> (___el);
}

static void
cpp_PostfixExpression_TypeConversion_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypeConversion);

    Cpp_PostfixExpression_TypeConversion * const &self = static_cast <Cpp_PostfixExpression_TypeConversion*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_PostfixExpression_TypeConversion_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_TypeConversion))) Cpp_PostfixExpression_TypeConversion;
}

static Ref<Grammar>
create_cpp_PostfixExpression_TypeConversion ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_TypeConversion_creation_func));
    grammar->name = String::forData ("PostfixExpression_TypeConversion");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier ();
        entry->assignment_func = cpp_PostfixExpression_TypeConversion_set_simpleTypeSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_PostfixExpression_TypeConversion_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_TypeConversion (Cpp_PostfixExpression_TypeConversion const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_TypeConversion\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_TypeConversion\n");

    dump_cpp_SimpleTypeSpecifier (el->simpleTypeSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_TypenameTypeConversion_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypenameTypeConversion);

    Cpp_PostfixExpression_TypenameTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TypenameTypeConversion*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_PostfixExpression_TypenameTypeConversion_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypenameTypeConversion);

    Cpp_PostfixExpression_TypenameTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TypenameTypeConversion*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_PostfixExpression_TypenameTypeConversion_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypenameTypeConversion);

    Cpp_PostfixExpression_TypenameTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TypenameTypeConversion*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_PostfixExpression_TypenameTypeConversion_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypenameTypeConversion);

    Cpp_PostfixExpression_TypenameTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TypenameTypeConversion*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_PostfixExpression_TypenameTypeConversion_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_TypenameTypeConversion))) Cpp_PostfixExpression_TypenameTypeConversion;
}

static Ref<Grammar>
create_cpp_PostfixExpression_TypenameTypeConversion ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_TypenameTypeConversion_creation_func));
    grammar->name = String::forData ("PostfixExpression_TypenameTypeConversion");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_PostfixExpression_TypenameTypeConversion_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_PostfixExpression_TypenameTypeConversion_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_PostfixExpression ();
        entry->jump_switch_grammar_index = 4;
        entry->jump_compound_grammar_index = 4;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_PostfixExpression_TypenameTypeConversion_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_PostfixExpression_TypenameTypeConversion_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_TypenameTypeConversion (Cpp_PostfixExpression_TypenameTypeConversion const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_TypenameTypeConversion\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_TypenameTypeConversion\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typename\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_TemplateTypeConversion_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TemplateTypeConversion);

    Cpp_PostfixExpression_TemplateTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TemplateTypeConversion*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_PostfixExpression_TemplateTypeConversion_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TemplateTypeConversion);

    Cpp_PostfixExpression_TemplateTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TemplateTypeConversion*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_PostfixExpression_TemplateTypeConversion_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TemplateTypeConversion);

    Cpp_PostfixExpression_TemplateTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TemplateTypeConversion*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static void
cpp_PostfixExpression_TemplateTypeConversion_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TemplateTypeConversion);

    Cpp_PostfixExpression_TemplateTypeConversion * const &self = static_cast <Cpp_PostfixExpression_TemplateTypeConversion*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_PostfixExpression_TemplateTypeConversion_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_TemplateTypeConversion))) Cpp_PostfixExpression_TemplateTypeConversion;
}

static Ref<Grammar>
create_cpp_PostfixExpression_TemplateTypeConversion ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_TemplateTypeConversion_creation_func));
    grammar->name = String::forData ("PostfixExpression_TemplateTypeConversion");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_PostfixExpression_TemplateTypeConversion_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_PostfixExpression_TemplateTypeConversion_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_PostfixExpression_TemplateTypeConversion_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_PostfixExpression_TemplateTypeConversion_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_TemplateTypeConversion (Cpp_PostfixExpression_TemplateTypeConversion const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_TemplateTypeConversion\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_TemplateTypeConversion\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typename\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: template_opt\n");
    dump_cpp_TemplateId (el->templateId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_DotMemberAccess_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_DotMemberAccess);

    Cpp_PostfixExpression_DotMemberAccess * const &self = static_cast <Cpp_PostfixExpression_DotMemberAccess*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static void
cpp_PostfixExpression_DotMemberAccess_set_idExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_IdExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_DotMemberAccess);

    Cpp_PostfixExpression_DotMemberAccess * const &self = static_cast <Cpp_PostfixExpression_DotMemberAccess*> (_self);

    self->idExpression = static_cast <Cpp_IdExpression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_DotMemberAccess_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_DotMemberAccess))) Cpp_PostfixExpression_DotMemberAccess;
}

static Ref<Grammar>
create_cpp_PostfixExpression_DotMemberAccess ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_DotMemberAccess_creation_func));
    grammar->name = String::forData ("PostfixExpression_DotMemberAccess");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_DotMemberAccess_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("."));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_PostfixExpression ();
        entry->jump_switch_grammar_index = 7;
        entry->jump_compound_grammar_index = 2;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IdExpression ();
        entry->assignment_func = cpp_PostfixExpression_DotMemberAccess_set_idExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_DotMemberAccess (Cpp_PostfixExpression_DotMemberAccess const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_DotMemberAccess\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_DotMemberAccess\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: .\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: template_opt\n");
    dump_cpp_IdExpression (el->idExpression, nest_level + 1);

}

static void
cpp_PostfixExpression_ArrowMemberAccess_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ArrowMemberAccess);

    Cpp_PostfixExpression_ArrowMemberAccess * const &self = static_cast <Cpp_PostfixExpression_ArrowMemberAccess*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static void
cpp_PostfixExpression_ArrowMemberAccess_set_idExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_IdExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ArrowMemberAccess);

    Cpp_PostfixExpression_ArrowMemberAccess * const &self = static_cast <Cpp_PostfixExpression_ArrowMemberAccess*> (_self);

    self->idExpression = static_cast <Cpp_IdExpression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_ArrowMemberAccess_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_ArrowMemberAccess))) Cpp_PostfixExpression_ArrowMemberAccess;
}

static Ref<Grammar>
create_cpp_PostfixExpression_ArrowMemberAccess ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_ArrowMemberAccess_creation_func));
    grammar->name = String::forData ("PostfixExpression_ArrowMemberAccess");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_ArrowMemberAccess_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("->"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IdExpression ();
        entry->assignment_func = cpp_PostfixExpression_ArrowMemberAccess_set_idExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_ArrowMemberAccess (Cpp_PostfixExpression_ArrowMemberAccess const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_ArrowMemberAccess\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_ArrowMemberAccess\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ->\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: template_opt\n");
    dump_cpp_IdExpression (el->idExpression, nest_level + 1);

}

static void
cpp_PostfixExpression_DotPseudoDestructorCall_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_DotPseudoDestructorCall);

    Cpp_PostfixExpression_DotPseudoDestructorCall * const &self = static_cast <Cpp_PostfixExpression_DotPseudoDestructorCall*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static void
cpp_PostfixExpression_DotPseudoDestructorCall_set_pseudoDestructorName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PseudoDestructorName);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_DotPseudoDestructorCall);

    Cpp_PostfixExpression_DotPseudoDestructorCall * const &self = static_cast <Cpp_PostfixExpression_DotPseudoDestructorCall*> (_self);

    self->pseudoDestructorName = static_cast <Cpp_PseudoDestructorName*> (___el);
}

static ParserElement*
cpp_PostfixExpression_DotPseudoDestructorCall_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_DotPseudoDestructorCall))) Cpp_PostfixExpression_DotPseudoDestructorCall;
}

static Ref<Grammar>
create_cpp_PostfixExpression_DotPseudoDestructorCall ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_DotPseudoDestructorCall_creation_func));
    grammar->name = String::forData ("PostfixExpression_DotPseudoDestructorCall");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_DotPseudoDestructorCall_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("."));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PseudoDestructorName ();
        entry->assignment_func = cpp_PostfixExpression_DotPseudoDestructorCall_set_pseudoDestructorName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_DotPseudoDestructorCall (Cpp_PostfixExpression_DotPseudoDestructorCall const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_DotPseudoDestructorCall\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_DotPseudoDestructorCall\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: .\n");
    dump_cpp_PseudoDestructorName (el->pseudoDestructorName, nest_level + 1);

}

static void
cpp_PostfixExpression_ArrowPseudoDestructorCall_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ArrowPseudoDestructorCall);

    Cpp_PostfixExpression_ArrowPseudoDestructorCall * const &self = static_cast <Cpp_PostfixExpression_ArrowPseudoDestructorCall*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static void
cpp_PostfixExpression_ArrowPseudoDestructorCall_set_pseudoDestructorName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PseudoDestructorName);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ArrowPseudoDestructorCall);

    Cpp_PostfixExpression_ArrowPseudoDestructorCall * const &self = static_cast <Cpp_PostfixExpression_ArrowPseudoDestructorCall*> (_self);

    self->pseudoDestructorName = static_cast <Cpp_PseudoDestructorName*> (___el);
}

static ParserElement*
cpp_PostfixExpression_ArrowPseudoDestructorCall_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_ArrowPseudoDestructorCall))) Cpp_PostfixExpression_ArrowPseudoDestructorCall;
}

static Ref<Grammar>
create_cpp_PostfixExpression_ArrowPseudoDestructorCall ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_ArrowPseudoDestructorCall_creation_func));
    grammar->name = String::forData ("PostfixExpression_ArrowPseudoDestructorCall");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_ArrowPseudoDestructorCall_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("->"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PseudoDestructorName ();
        entry->assignment_func = cpp_PostfixExpression_ArrowPseudoDestructorCall_set_pseudoDestructorName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_ArrowPseudoDestructorCall (Cpp_PostfixExpression_ArrowPseudoDestructorCall const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_ArrowPseudoDestructorCall\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_ArrowPseudoDestructorCall\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ->\n");
    dump_cpp_PseudoDestructorName (el->pseudoDestructorName, nest_level + 1);

}

static void
cpp_PostfixExpression_PostfixIncrement_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_PostfixIncrement);

    Cpp_PostfixExpression_PostfixIncrement * const &self = static_cast <Cpp_PostfixExpression_PostfixIncrement*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_PostfixIncrement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_PostfixIncrement))) Cpp_PostfixExpression_PostfixIncrement;
}

static Ref<Grammar>
create_cpp_PostfixExpression_PostfixIncrement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_PostfixIncrement_creation_func));
    grammar->name = String::forData ("PostfixExpression_PostfixIncrement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_PostfixIncrement_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("++"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_PostfixIncrement (Cpp_PostfixExpression_PostfixIncrement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_PostfixIncrement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_PostfixIncrement\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ++\n");
}

static void
cpp_PostfixExpression_PostfixDecrement_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_PostfixDecrement);

    Cpp_PostfixExpression_PostfixDecrement * const &self = static_cast <Cpp_PostfixExpression_PostfixDecrement*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_PostfixDecrement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_PostfixDecrement))) Cpp_PostfixExpression_PostfixDecrement;
}

static Ref<Grammar>
create_cpp_PostfixExpression_PostfixDecrement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_PostfixDecrement_creation_func));
    grammar->name = String::forData ("PostfixExpression_PostfixDecrement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_PostfixExpression_PostfixDecrement_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("--"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_PostfixDecrement (Cpp_PostfixExpression_PostfixDecrement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_PostfixDecrement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_PostfixDecrement\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: --\n");
}

bool cpp_begin_typeid (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_typeid (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_typeid (_el, parser_control, data);
}

static void
cpp_PostfixExpression_DynamicCast_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_DynamicCast);

    Cpp_PostfixExpression_DynamicCast * const &self = static_cast <Cpp_PostfixExpression_DynamicCast*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_PostfixExpression_DynamicCast_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_DynamicCast);

    Cpp_PostfixExpression_DynamicCast * const &self = static_cast <Cpp_PostfixExpression_DynamicCast*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_DynamicCast_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_DynamicCast))) Cpp_PostfixExpression_DynamicCast;
}

static Ref<Grammar>
create_cpp_PostfixExpression_DynamicCast ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_DynamicCast_creation_func));
    grammar->name = String::forData ("PostfixExpression_DynamicCast");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("dynamic_cast"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_PostfixExpression_DynamicCast_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PostfixExpression_DynamicCast_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_DynamicCast (Cpp_PostfixExpression_DynamicCast const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_DynamicCast\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_DynamicCast\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: dynamic_cast\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_StaticCast_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_StaticCast);

    Cpp_PostfixExpression_StaticCast * const &self = static_cast <Cpp_PostfixExpression_StaticCast*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_PostfixExpression_StaticCast_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_StaticCast);

    Cpp_PostfixExpression_StaticCast * const &self = static_cast <Cpp_PostfixExpression_StaticCast*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_StaticCast_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_StaticCast))) Cpp_PostfixExpression_StaticCast;
}

static Ref<Grammar>
create_cpp_PostfixExpression_StaticCast ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_StaticCast_creation_func));
    grammar->name = String::forData ("PostfixExpression_StaticCast");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("static_cast"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_PostfixExpression_StaticCast_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PostfixExpression_StaticCast_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_StaticCast (Cpp_PostfixExpression_StaticCast const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_StaticCast\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_StaticCast\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: static_cast\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_ReinterpretCast_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ReinterpretCast);

    Cpp_PostfixExpression_ReinterpretCast * const &self = static_cast <Cpp_PostfixExpression_ReinterpretCast*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_PostfixExpression_ReinterpretCast_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ReinterpretCast);

    Cpp_PostfixExpression_ReinterpretCast * const &self = static_cast <Cpp_PostfixExpression_ReinterpretCast*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_ReinterpretCast_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_ReinterpretCast))) Cpp_PostfixExpression_ReinterpretCast;
}

static Ref<Grammar>
create_cpp_PostfixExpression_ReinterpretCast ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_ReinterpretCast_creation_func));
    grammar->name = String::forData ("PostfixExpression_ReinterpretCast");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("reinterpret_cast"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_PostfixExpression_ReinterpretCast_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PostfixExpression_ReinterpretCast_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_ReinterpretCast (Cpp_PostfixExpression_ReinterpretCast const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_ReinterpretCast\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_ReinterpretCast\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: reinterpret_cast\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_ConstCast_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ConstCast);

    Cpp_PostfixExpression_ConstCast * const &self = static_cast <Cpp_PostfixExpression_ConstCast*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_PostfixExpression_ConstCast_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_ConstCast);

    Cpp_PostfixExpression_ConstCast * const &self = static_cast <Cpp_PostfixExpression_ConstCast*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_ConstCast_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_ConstCast))) Cpp_PostfixExpression_ConstCast;
}

static Ref<Grammar>
create_cpp_PostfixExpression_ConstCast ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_ConstCast_creation_func));
    grammar->name = String::forData ("PostfixExpression_ConstCast");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("const_cast"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_PostfixExpression_ConstCast_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PostfixExpression_ConstCast_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_ConstCast (Cpp_PostfixExpression_ConstCast const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_ConstCast\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_ConstCast\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: const_cast\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_TypeidTypeid_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypeidTypeid);

    Cpp_PostfixExpression_TypeidTypeid * const &self = static_cast <Cpp_PostfixExpression_TypeidTypeid*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static ParserElement*
cpp_PostfixExpression_TypeidTypeid_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_TypeidTypeid))) Cpp_PostfixExpression_TypeidTypeid;
}

static Ref<Grammar>
create_cpp_PostfixExpression_TypeidTypeid ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_TypeidTypeid_creation_func));
    grammar->name = String::forData ("PostfixExpression_TypeidTypeid");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typeid"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_PostfixExpression_TypeidTypeid_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_TypeidTypeid (Cpp_PostfixExpression_TypeidTypeid const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_TypeidTypeid\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_TypeidTypeid\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typeid\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_TypeidExpression_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_TypeidExpression);

    Cpp_PostfixExpression_TypeidExpression * const &self = static_cast <Cpp_PostfixExpression_TypeidExpression*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_TypeidExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_TypeidExpression))) Cpp_PostfixExpression_TypeidExpression;
}

static Ref<Grammar>
create_cpp_PostfixExpression_TypeidExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_TypeidExpression_creation_func));
    grammar->name = String::forData ("PostfixExpression_TypeidExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typeid"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_PostfixExpression_TypeidExpression_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_postfix_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_TypeidExpression (Cpp_PostfixExpression_TypeidExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_TypeidExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_TypeidExpression\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typeid\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_PostfixExpression_Expression_set_primaryExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PostfixExpression ||
             __el->cpp_element_type != CppElement::_PrimaryExpression);

    Cpp_PostfixExpression * const &_self = static_cast <Cpp_PostfixExpression*> (__self);

    abortIf (_self->postfix_expression_type != Cpp_PostfixExpression::_Expression);

    Cpp_PostfixExpression_Expression * const &self = static_cast <Cpp_PostfixExpression_Expression*> (_self);

    self->primaryExpression = static_cast <Cpp_PrimaryExpression*> (___el);
}

static ParserElement*
cpp_PostfixExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PostfixExpression_Expression))) Cpp_PostfixExpression_Expression;
}

static Ref<Grammar>
create_cpp_PostfixExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PostfixExpression_Expression_creation_func));
    grammar->name = String::forData ("PostfixExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PrimaryExpression ();
        entry->assignment_func = cpp_PostfixExpression_Expression_set_primaryExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression_Expression (Cpp_PostfixExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PostfixExpression_Expression\n");

    dump_cpp_PrimaryExpression (el->primaryExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_PostfixExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("PostfixExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_Subscripting ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_FunctionCall ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_TypeConversion ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_TypenameTypeConversion ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_TemplateTypeConversion ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_DotMemberAccess ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_ArrowMemberAccess ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_DotPseudoDestructorCall ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_ArrowPseudoDestructorCall ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_PostfixIncrement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_PostfixDecrement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_DynamicCast ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_StaticCast ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_ReinterpretCast ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_ConstCast ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_TypeidTypeid ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_TypeidExpression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PostfixExpression (Cpp_PostfixExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PostfixExpression\n");
        return;
    }

    switch (el->postfix_expression_type) {
        case Cpp_PostfixExpression::_Subscripting:
            dump_cpp_PostfixExpression_Subscripting (static_cast <Cpp_PostfixExpression_Subscripting const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_FunctionCall:
            dump_cpp_PostfixExpression_FunctionCall (static_cast <Cpp_PostfixExpression_FunctionCall const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_TypeConversion:
            dump_cpp_PostfixExpression_TypeConversion (static_cast <Cpp_PostfixExpression_TypeConversion const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_TypenameTypeConversion:
            dump_cpp_PostfixExpression_TypenameTypeConversion (static_cast <Cpp_PostfixExpression_TypenameTypeConversion const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_TemplateTypeConversion:
            dump_cpp_PostfixExpression_TemplateTypeConversion (static_cast <Cpp_PostfixExpression_TemplateTypeConversion const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_DotMemberAccess:
            dump_cpp_PostfixExpression_DotMemberAccess (static_cast <Cpp_PostfixExpression_DotMemberAccess const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_ArrowMemberAccess:
            dump_cpp_PostfixExpression_ArrowMemberAccess (static_cast <Cpp_PostfixExpression_ArrowMemberAccess const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_DotPseudoDestructorCall:
            dump_cpp_PostfixExpression_DotPseudoDestructorCall (static_cast <Cpp_PostfixExpression_DotPseudoDestructorCall const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_ArrowPseudoDestructorCall:
            dump_cpp_PostfixExpression_ArrowPseudoDestructorCall (static_cast <Cpp_PostfixExpression_ArrowPseudoDestructorCall const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_PostfixIncrement:
            dump_cpp_PostfixExpression_PostfixIncrement (static_cast <Cpp_PostfixExpression_PostfixIncrement const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_PostfixDecrement:
            dump_cpp_PostfixExpression_PostfixDecrement (static_cast <Cpp_PostfixExpression_PostfixDecrement const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_DynamicCast:
            dump_cpp_PostfixExpression_DynamicCast (static_cast <Cpp_PostfixExpression_DynamicCast const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_StaticCast:
            dump_cpp_PostfixExpression_StaticCast (static_cast <Cpp_PostfixExpression_StaticCast const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_ReinterpretCast:
            dump_cpp_PostfixExpression_ReinterpretCast (static_cast <Cpp_PostfixExpression_ReinterpretCast const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_ConstCast:
            dump_cpp_PostfixExpression_ConstCast (static_cast <Cpp_PostfixExpression_ConstCast const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_TypeidTypeid:
            dump_cpp_PostfixExpression_TypeidTypeid (static_cast <Cpp_PostfixExpression_TypeidTypeid const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_TypeidExpression:
            dump_cpp_PostfixExpression_TypeidExpression (static_cast <Cpp_PostfixExpression_TypeidExpression const *> (el), nest_level + 0);
            break;
        case Cpp_PostfixExpression::_Expression:
            dump_cpp_PostfixExpression_Expression (static_cast <Cpp_PostfixExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ExpressionList_ExpressionList_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExpressionList ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_ExpressionList * const &_self = static_cast <Cpp_ExpressionList*> (__self);

    abortIf (_self->expression_list_type != Cpp_ExpressionList::_ExpressionList);

    Cpp_ExpressionList_ExpressionList * const &self = static_cast <Cpp_ExpressionList_ExpressionList*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static void
cpp_ExpressionList_ExpressionList_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExpressionList ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_ExpressionList * const &_self = static_cast <Cpp_ExpressionList*> (__self);

    abortIf (_self->expression_list_type != Cpp_ExpressionList::_ExpressionList);

    Cpp_ExpressionList_ExpressionList * const &self = static_cast <Cpp_ExpressionList_ExpressionList*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_ExpressionList_ExpressionList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExpressionList_ExpressionList))) Cpp_ExpressionList_ExpressionList;
}

static Ref<Grammar>
create_cpp_ExpressionList_ExpressionList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExpressionList_ExpressionList_creation_func));
    grammar->name = String::forData ("ExpressionList_ExpressionList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_ExpressionList_ExpressionList_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_ExpressionList ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_ExpressionList_ExpressionList_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExpressionList_ExpressionList (Cpp_ExpressionList_ExpressionList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExpressionList_ExpressionList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExpressionList_ExpressionList\n");

    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

}

static void
cpp_ExpressionList_Expression_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExpressionList ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_ExpressionList * const &_self = static_cast <Cpp_ExpressionList*> (__self);

    abortIf (_self->expression_list_type != Cpp_ExpressionList::_Expression);

    Cpp_ExpressionList_Expression * const &self = static_cast <Cpp_ExpressionList_Expression*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static ParserElement*
cpp_ExpressionList_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExpressionList_Expression))) Cpp_ExpressionList_Expression;
}

static Ref<Grammar>
create_cpp_ExpressionList_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExpressionList_Expression_creation_func));
    grammar->name = String::forData ("ExpressionList_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_ExpressionList_Expression_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExpressionList_Expression (Cpp_ExpressionList_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExpressionList_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExpressionList_Expression\n");

    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ExpressionList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ExpressionList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ExpressionList_ExpressionList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ExpressionList_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExpressionList (Cpp_ExpressionList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExpressionList\n");
        return;
    }

    switch (el->expression_list_type) {
        case Cpp_ExpressionList::_ExpressionList:
            dump_cpp_ExpressionList_ExpressionList (static_cast <Cpp_ExpressionList_ExpressionList const *> (el), nest_level + 0);
            break;
        case Cpp_ExpressionList::_Expression:
            dump_cpp_ExpressionList_Expression (static_cast <Cpp_ExpressionList_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_PseudoDestructorName_TypeName_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_TypeName);

    Cpp_PseudoDestructorName_TypeName * const &self = static_cast <Cpp_PseudoDestructorName_TypeName*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_PseudoDestructorName_TypeName_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_TypeName);

    Cpp_PseudoDestructorName_TypeName * const &self = static_cast <Cpp_PseudoDestructorName_TypeName*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_PseudoDestructorName_TypeName_set_typeName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_TypeName);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_TypeName);

    Cpp_PseudoDestructorName_TypeName * const &self = static_cast <Cpp_PseudoDestructorName_TypeName*> (_self);

    self->typeName = static_cast <Cpp_TypeName*> (___el);
}

static void
cpp_PseudoDestructorName_TypeName_set_destructorTypeName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_TypeName);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_TypeName);

    Cpp_PseudoDestructorName_TypeName * const &self = static_cast <Cpp_PseudoDestructorName_TypeName*> (_self);

    self->destructorTypeName = static_cast <Cpp_TypeName*> (___el);
}

static ParserElement*
cpp_PseudoDestructorName_TypeName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PseudoDestructorName_TypeName))) Cpp_PseudoDestructorName_TypeName;
}

static Ref<Grammar>
create_cpp_PseudoDestructorName_TypeName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PseudoDestructorName_TypeName_creation_func));
    grammar->name = String::forData ("PseudoDestructorName_TypeName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_PseudoDestructorName_TypeName_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_PseudoDestructorName_TypeName_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_PseudoDestructorName ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeName ();
        entry->assignment_func = cpp_PseudoDestructorName_TypeName_set_typeName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("::"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("~"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeName ();
        entry->assignment_func = cpp_PseudoDestructorName_TypeName_set_destructorTypeName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PseudoDestructorName_TypeName (Cpp_PseudoDestructorName_TypeName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PseudoDestructorName_TypeName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PseudoDestructorName_TypeName\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_TypeName (el->typeName, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ::\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ~\n");
    dump_cpp_TypeName (el->destructorTypeName, nest_level + 1);

}

static void
cpp_PseudoDestructorName_Template_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Template);

    Cpp_PseudoDestructorName_Template * const &self = static_cast <Cpp_PseudoDestructorName_Template*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_PseudoDestructorName_Template_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Template);

    Cpp_PseudoDestructorName_Template * const &self = static_cast <Cpp_PseudoDestructorName_Template*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_PseudoDestructorName_Template_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Template);

    Cpp_PseudoDestructorName_Template * const &self = static_cast <Cpp_PseudoDestructorName_Template*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static void
cpp_PseudoDestructorName_Template_set_typeName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_TypeName);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Template);

    Cpp_PseudoDestructorName_Template * const &self = static_cast <Cpp_PseudoDestructorName_Template*> (_self);

    self->typeName = static_cast <Cpp_TypeName*> (___el);
}

static ParserElement*
cpp_PseudoDestructorName_Template_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PseudoDestructorName_Template))) Cpp_PseudoDestructorName_Template;
}

static Ref<Grammar>
create_cpp_PseudoDestructorName_Template ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PseudoDestructorName_Template_creation_func));
    grammar->name = String::forData ("PseudoDestructorName_Template");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_PseudoDestructorName_Template_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_PseudoDestructorName_Template_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_PseudoDestructorName ();
        entry->jump_switch_grammar_index = 2;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_PseudoDestructorName_Template_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("::"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("~"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeName ();
        entry->assignment_func = cpp_PseudoDestructorName_Template_set_typeName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PseudoDestructorName_Template (Cpp_PseudoDestructorName_Template const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PseudoDestructorName_Template\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PseudoDestructorName_Template\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    dump_cpp_TemplateId (el->templateId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ::\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ~\n");
    dump_cpp_TypeName (el->typeName, nest_level + 1);

}

static void
cpp_PseudoDestructorName_Basic_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Basic);

    Cpp_PseudoDestructorName_Basic * const &self = static_cast <Cpp_PseudoDestructorName_Basic*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_PseudoDestructorName_Basic_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Basic);

    Cpp_PseudoDestructorName_Basic * const &self = static_cast <Cpp_PseudoDestructorName_Basic*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_PseudoDestructorName_Basic_set_typeName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PseudoDestructorName ||
             __el->cpp_element_type != CppElement::_TypeName);

    Cpp_PseudoDestructorName * const &_self = static_cast <Cpp_PseudoDestructorName*> (__self);

    abortIf (_self->pseudo_destructor_name_type != Cpp_PseudoDestructorName::_Basic);

    Cpp_PseudoDestructorName_Basic * const &self = static_cast <Cpp_PseudoDestructorName_Basic*> (_self);

    self->typeName = static_cast <Cpp_TypeName*> (___el);
}

static ParserElement*
cpp_PseudoDestructorName_Basic_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PseudoDestructorName_Basic))) Cpp_PseudoDestructorName_Basic;
}

static Ref<Grammar>
create_cpp_PseudoDestructorName_Basic ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PseudoDestructorName_Basic_creation_func));
    grammar->name = String::forData ("PseudoDestructorName_Basic");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_PseudoDestructorName_Basic_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_PseudoDestructorName_Basic_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("~"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeName ();
        entry->assignment_func = cpp_PseudoDestructorName_Basic_set_typeName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PseudoDestructorName_Basic (Cpp_PseudoDestructorName_Basic const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PseudoDestructorName_Basic\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PseudoDestructorName_Basic\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ~\n");
    dump_cpp_TypeName (el->typeName, nest_level + 1);

}

static Ref<Grammar>
create_cpp_PseudoDestructorName ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("PseudoDestructorName");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PseudoDestructorName_TypeName ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PseudoDestructorName_Template ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PseudoDestructorName_Basic ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PseudoDestructorName (Cpp_PseudoDestructorName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PseudoDestructorName\n");
        return;
    }

    switch (el->pseudo_destructor_name_type) {
        case Cpp_PseudoDestructorName::_TypeName:
            dump_cpp_PseudoDestructorName_TypeName (static_cast <Cpp_PseudoDestructorName_TypeName const *> (el), nest_level + 0);
            break;
        case Cpp_PseudoDestructorName::_Template:
            dump_cpp_PseudoDestructorName_Template (static_cast <Cpp_PseudoDestructorName_Template const *> (el), nest_level + 0);
            break;
        case Cpp_PseudoDestructorName::_Basic:
            dump_cpp_PseudoDestructorName_Basic (static_cast <Cpp_PseudoDestructorName_Basic const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_UnaryExpression_PrefixIncrement_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_PrefixIncrement);

    Cpp_UnaryExpression_PrefixIncrement * const &self = static_cast <Cpp_UnaryExpression_PrefixIncrement*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

bool cpp_accept_unary_expression (
        Cpp_UnaryExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_unary_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_UnaryExpression);

    Cpp_UnaryExpression * const &el =
            static_cast <Cpp_UnaryExpression*> (_el);

    return cpp_accept_unary_expression (el, parser_control, data);
}

static ParserElement*
cpp_UnaryExpression_PrefixIncrement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_PrefixIncrement))) Cpp_UnaryExpression_PrefixIncrement;
}

static Ref<Grammar>
create_cpp_UnaryExpression_PrefixIncrement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_PrefixIncrement_creation_func));
    grammar->name = String::forData ("UnaryExpression_PrefixIncrement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("++"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_UnaryExpression_PrefixIncrement_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_unary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_PrefixIncrement (Cpp_UnaryExpression_PrefixIncrement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_PrefixIncrement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_PrefixIncrement\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ++\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_UnaryExpression_PrefixDecrement_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_PrefixDecrement);

    Cpp_UnaryExpression_PrefixDecrement * const &self = static_cast <Cpp_UnaryExpression_PrefixDecrement*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_UnaryExpression_PrefixDecrement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_PrefixDecrement))) Cpp_UnaryExpression_PrefixDecrement;
}

static Ref<Grammar>
create_cpp_UnaryExpression_PrefixDecrement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_PrefixDecrement_creation_func));
    grammar->name = String::forData ("UnaryExpression_PrefixDecrement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("--"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_UnaryExpression_PrefixDecrement_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_unary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_PrefixDecrement (Cpp_UnaryExpression_PrefixDecrement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_PrefixDecrement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_PrefixDecrement\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: --\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_UnaryExpression_UnaryOperator_set_unaryOperator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_UnaryOperator);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_UnaryOperator);

    Cpp_UnaryExpression_UnaryOperator * const &self = static_cast <Cpp_UnaryExpression_UnaryOperator*> (_self);

    self->unaryOperator = static_cast <Cpp_UnaryOperator*> (___el);
}

static void
cpp_UnaryExpression_UnaryOperator_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_UnaryOperator);

    Cpp_UnaryExpression_UnaryOperator * const &self = static_cast <Cpp_UnaryExpression_UnaryOperator*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_UnaryExpression_UnaryOperator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_UnaryOperator))) Cpp_UnaryExpression_UnaryOperator;
}

static Ref<Grammar>
create_cpp_UnaryExpression_UnaryOperator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_UnaryOperator_creation_func));
    grammar->name = String::forData ("UnaryExpression_UnaryOperator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator ();
        entry->assignment_func = cpp_UnaryExpression_UnaryOperator_set_unaryOperator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_UnaryExpression_UnaryOperator_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_unary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_UnaryOperator (Cpp_UnaryExpression_UnaryOperator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_UnaryOperator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_UnaryOperator\n");

    dump_cpp_UnaryOperator (el->unaryOperator, nest_level + 1);

    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_UnaryExpression_SizeofExpression_set_unaryExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_UnaryExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_SizeofExpression);

    Cpp_UnaryExpression_SizeofExpression * const &self = static_cast <Cpp_UnaryExpression_SizeofExpression*> (_self);

    self->unaryExpression = static_cast <Cpp_UnaryExpression*> (___el);
}

static ParserElement*
cpp_UnaryExpression_SizeofExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_SizeofExpression))) Cpp_UnaryExpression_SizeofExpression;
}

static Ref<Grammar>
create_cpp_UnaryExpression_SizeofExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_SizeofExpression_creation_func));
    grammar->name = String::forData ("UnaryExpression_SizeofExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("sizeof"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression ();
        entry->assignment_func = cpp_UnaryExpression_SizeofExpression_set_unaryExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_unary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_SizeofExpression (Cpp_UnaryExpression_SizeofExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_SizeofExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_SizeofExpression\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: sizeof\n");
    dump_cpp_UnaryExpression (el->unaryExpression, nest_level + 1);

}

static void
cpp_UnaryExpression_SizeofTypeid_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_SizeofTypeid);

    Cpp_UnaryExpression_SizeofTypeid * const &self = static_cast <Cpp_UnaryExpression_SizeofTypeid*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static ParserElement*
cpp_UnaryExpression_SizeofTypeid_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_SizeofTypeid))) Cpp_UnaryExpression_SizeofTypeid;
}

static Ref<Grammar>
create_cpp_UnaryExpression_SizeofTypeid ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_SizeofTypeid_creation_func));
    grammar->name = String::forData ("UnaryExpression_SizeofTypeid");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("sizeof"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_UnaryExpression_SizeofTypeid_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_unary_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_SizeofTypeid (Cpp_UnaryExpression_SizeofTypeid const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_SizeofTypeid\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_SizeofTypeid\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: sizeof\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_UnaryExpression_NewExpression_set_newExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_NewExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_NewExpression);

    Cpp_UnaryExpression_NewExpression * const &self = static_cast <Cpp_UnaryExpression_NewExpression*> (_self);

    self->newExpression = static_cast <Cpp_NewExpression*> (___el);
}

static ParserElement*
cpp_UnaryExpression_NewExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_NewExpression))) Cpp_UnaryExpression_NewExpression;
}

static Ref<Grammar>
create_cpp_UnaryExpression_NewExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_NewExpression_creation_func));
    grammar->name = String::forData ("UnaryExpression_NewExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NewExpression ();
        entry->assignment_func = cpp_UnaryExpression_NewExpression_set_newExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_NewExpression (Cpp_UnaryExpression_NewExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_NewExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_NewExpression\n");

    dump_cpp_NewExpression (el->newExpression, nest_level + 1);

}

static void
cpp_UnaryExpression_DeleteExpression_set_deleteExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_DeleteExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_DeleteExpression);

    Cpp_UnaryExpression_DeleteExpression * const &self = static_cast <Cpp_UnaryExpression_DeleteExpression*> (_self);

    self->deleteExpression = static_cast <Cpp_DeleteExpression*> (___el);
}

static ParserElement*
cpp_UnaryExpression_DeleteExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_DeleteExpression))) Cpp_UnaryExpression_DeleteExpression;
}

static Ref<Grammar>
create_cpp_UnaryExpression_DeleteExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_DeleteExpression_creation_func));
    grammar->name = String::forData ("UnaryExpression_DeleteExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DeleteExpression ();
        entry->assignment_func = cpp_UnaryExpression_DeleteExpression_set_deleteExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_DeleteExpression (Cpp_UnaryExpression_DeleteExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_DeleteExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_DeleteExpression\n");

    dump_cpp_DeleteExpression (el->deleteExpression, nest_level + 1);

}

static void
cpp_UnaryExpression_Expression_set_postfixExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnaryExpression ||
             __el->cpp_element_type != CppElement::_PostfixExpression);

    Cpp_UnaryExpression * const &_self = static_cast <Cpp_UnaryExpression*> (__self);

    abortIf (_self->unary_expression_type != Cpp_UnaryExpression::_Expression);

    Cpp_UnaryExpression_Expression * const &self = static_cast <Cpp_UnaryExpression_Expression*> (_self);

    self->postfixExpression = static_cast <Cpp_PostfixExpression*> (___el);
}

static ParserElement*
cpp_UnaryExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryExpression_Expression))) Cpp_UnaryExpression_Expression;
}

static Ref<Grammar>
create_cpp_UnaryExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryExpression_Expression_creation_func));
    grammar->name = String::forData ("UnaryExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PostfixExpression ();
        entry->assignment_func = cpp_UnaryExpression_Expression_set_postfixExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression_Expression (Cpp_UnaryExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryExpression_Expression\n");

    dump_cpp_PostfixExpression (el->postfixExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_UnaryExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("UnaryExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_PrefixIncrement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_PrefixDecrement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_UnaryOperator ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_SizeofExpression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_SizeofTypeid ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_NewExpression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_DeleteExpression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryExpression (Cpp_UnaryExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryExpression\n");
        return;
    }

    switch (el->unary_expression_type) {
        case Cpp_UnaryExpression::_PrefixIncrement:
            dump_cpp_UnaryExpression_PrefixIncrement (static_cast <Cpp_UnaryExpression_PrefixIncrement const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_PrefixDecrement:
            dump_cpp_UnaryExpression_PrefixDecrement (static_cast <Cpp_UnaryExpression_PrefixDecrement const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_UnaryOperator:
            dump_cpp_UnaryExpression_UnaryOperator (static_cast <Cpp_UnaryExpression_UnaryOperator const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_SizeofExpression:
            dump_cpp_UnaryExpression_SizeofExpression (static_cast <Cpp_UnaryExpression_SizeofExpression const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_SizeofTypeid:
            dump_cpp_UnaryExpression_SizeofTypeid (static_cast <Cpp_UnaryExpression_SizeofTypeid const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_NewExpression:
            dump_cpp_UnaryExpression_NewExpression (static_cast <Cpp_UnaryExpression_NewExpression const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_DeleteExpression:
            dump_cpp_UnaryExpression_DeleteExpression (static_cast <Cpp_UnaryExpression_DeleteExpression const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryExpression::_Expression:
            dump_cpp_UnaryExpression_Expression (static_cast <Cpp_UnaryExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static ParserElement*
cpp_UnaryOperator_Indirection_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryOperator_Indirection))) Cpp_UnaryOperator_Indirection;
}

static Ref<Grammar>
create_cpp_UnaryOperator_Indirection ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryOperator_Indirection_creation_func));
    grammar->name = String::forData ("UnaryOperator_Indirection");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator_Indirection (Cpp_UnaryOperator_Indirection const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator_Indirection\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryOperator_Indirection\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: *\n");
}

static ParserElement*
cpp_UnaryOperator_PointerTo_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryOperator_PointerTo))) Cpp_UnaryOperator_PointerTo;
}

static Ref<Grammar>
create_cpp_UnaryOperator_PointerTo ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryOperator_PointerTo_creation_func));
    grammar->name = String::forData ("UnaryOperator_PointerTo");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator_PointerTo (Cpp_UnaryOperator_PointerTo const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator_PointerTo\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryOperator_PointerTo\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: &\n");
}

static ParserElement*
cpp_UnaryOperator_UnaryPlus_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryOperator_UnaryPlus))) Cpp_UnaryOperator_UnaryPlus;
}

static Ref<Grammar>
create_cpp_UnaryOperator_UnaryPlus ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryOperator_UnaryPlus_creation_func));
    grammar->name = String::forData ("UnaryOperator_UnaryPlus");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("+"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator_UnaryPlus (Cpp_UnaryOperator_UnaryPlus const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator_UnaryPlus\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryOperator_UnaryPlus\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: +\n");
}

static ParserElement*
cpp_UnaryOperator_UnaryMinus_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryOperator_UnaryMinus))) Cpp_UnaryOperator_UnaryMinus;
}

static Ref<Grammar>
create_cpp_UnaryOperator_UnaryMinus ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryOperator_UnaryMinus_creation_func));
    grammar->name = String::forData ("UnaryOperator_UnaryMinus");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("-"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator_UnaryMinus (Cpp_UnaryOperator_UnaryMinus const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator_UnaryMinus\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryOperator_UnaryMinus\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: -\n");
}

static ParserElement*
cpp_UnaryOperator_LogicalNegation_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryOperator_LogicalNegation))) Cpp_UnaryOperator_LogicalNegation;
}

static Ref<Grammar>
create_cpp_UnaryOperator_LogicalNegation ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryOperator_LogicalNegation_creation_func));
    grammar->name = String::forData ("UnaryOperator_LogicalNegation");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("!"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator_LogicalNegation (Cpp_UnaryOperator_LogicalNegation const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator_LogicalNegation\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryOperator_LogicalNegation\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: !\n");
}

static ParserElement*
cpp_UnaryOperator_OnesComplement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnaryOperator_OnesComplement))) Cpp_UnaryOperator_OnesComplement;
}

static Ref<Grammar>
create_cpp_UnaryOperator_OnesComplement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnaryOperator_OnesComplement_creation_func));
    grammar->name = String::forData ("UnaryOperator_OnesComplement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("~"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator_OnesComplement (Cpp_UnaryOperator_OnesComplement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator_OnesComplement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnaryOperator_OnesComplement\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ~\n");
}

static Ref<Grammar>
create_cpp_UnaryOperator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("UnaryOperator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator_Indirection ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator_PointerTo ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator_UnaryPlus ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator_UnaryMinus ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator_LogicalNegation ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UnaryOperator_OnesComplement ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnaryOperator (Cpp_UnaryOperator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnaryOperator\n");
        return;
    }

    switch (el->unary_operator_type) {
        case Cpp_UnaryOperator::_Indirection:
            dump_cpp_UnaryOperator_Indirection (static_cast <Cpp_UnaryOperator_Indirection const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryOperator::_PointerTo:
            dump_cpp_UnaryOperator_PointerTo (static_cast <Cpp_UnaryOperator_PointerTo const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryOperator::_UnaryPlus:
            dump_cpp_UnaryOperator_UnaryPlus (static_cast <Cpp_UnaryOperator_UnaryPlus const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryOperator::_UnaryMinus:
            dump_cpp_UnaryOperator_UnaryMinus (static_cast <Cpp_UnaryOperator_UnaryMinus const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryOperator::_LogicalNegation:
            dump_cpp_UnaryOperator_LogicalNegation (static_cast <Cpp_UnaryOperator_LogicalNegation const *> (el), nest_level + 0);
            break;
        case Cpp_UnaryOperator::_OnesComplement:
            dump_cpp_UnaryOperator_OnesComplement (static_cast <Cpp_UnaryOperator_OnesComplement const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_NewExpression_Basic_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Basic);

    Cpp_NewExpression_Basic * const &self = static_cast <Cpp_NewExpression_Basic*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_NewExpression_Basic_set_newPlacement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_NewPlacement);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Basic);

    Cpp_NewExpression_Basic * const &self = static_cast <Cpp_NewExpression_Basic*> (_self);

    self->newPlacement = static_cast <Cpp_NewPlacement*> (___el);
}

static void
cpp_NewExpression_Basic_set_newTypeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_NewTypeId);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Basic);

    Cpp_NewExpression_Basic * const &self = static_cast <Cpp_NewExpression_Basic*> (_self);

    self->newTypeId = static_cast <Cpp_NewTypeId*> (___el);
}

static void
cpp_NewExpression_Basic_set_newInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_NewInitializer);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Basic);

    Cpp_NewExpression_Basic * const &self = static_cast <Cpp_NewExpression_Basic*> (_self);

    self->newInitializer = static_cast <Cpp_NewInitializer*> (___el);
}

static ParserElement*
cpp_NewExpression_Basic_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewExpression_Basic))) Cpp_NewExpression_Basic;
}

static Ref<Grammar>
create_cpp_NewExpression_Basic ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewExpression_Basic_creation_func));
    grammar->name = String::forData ("NewExpression_Basic");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_NewExpression_Basic_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("new"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NewPlacement ();
        entry->assignment_func = cpp_NewExpression_Basic_set_newPlacement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_NewExpression ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 4;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NewTypeId ();
        entry->assignment_func = cpp_NewExpression_Basic_set_newTypeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NewInitializer ();
        entry->assignment_func = cpp_NewExpression_Basic_set_newInitializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewExpression_Basic (Cpp_NewExpression_Basic const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewExpression_Basic\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewExpression_Basic\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: new\n");
    dump_cpp_NewPlacement (el->newPlacement, nest_level + 1);

    dump_cpp_NewTypeId (el->newTypeId, nest_level + 1);

    dump_cpp_NewInitializer (el->newInitializer, nest_level + 1);

}

static void
cpp_NewExpression_Braces_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Braces);

    Cpp_NewExpression_Braces * const &self = static_cast <Cpp_NewExpression_Braces*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_NewExpression_Braces_set_newPlacement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_NewPlacement);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Braces);

    Cpp_NewExpression_Braces * const &self = static_cast <Cpp_NewExpression_Braces*> (_self);

    self->newPlacement = static_cast <Cpp_NewPlacement*> (___el);
}

static void
cpp_NewExpression_Braces_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Braces);

    Cpp_NewExpression_Braces * const &self = static_cast <Cpp_NewExpression_Braces*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_NewExpression_Braces_set_newInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewExpression ||
             __el->cpp_element_type != CppElement::_NewInitializer);

    Cpp_NewExpression * const &_self = static_cast <Cpp_NewExpression*> (__self);

    abortIf (_self->new_expression_type != Cpp_NewExpression::_Braces);

    Cpp_NewExpression_Braces * const &self = static_cast <Cpp_NewExpression_Braces*> (_self);

    self->newInitializer = static_cast <Cpp_NewInitializer*> (___el);
}

static ParserElement*
cpp_NewExpression_Braces_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewExpression_Braces))) Cpp_NewExpression_Braces;
}

static Ref<Grammar>
create_cpp_NewExpression_Braces ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewExpression_Braces_creation_func));
    grammar->name = String::forData ("NewExpression_Braces");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_NewExpression_Braces_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("new"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NewPlacement ();
        entry->assignment_func = cpp_NewExpression_Braces_set_newPlacement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_NewExpression_Braces_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NewInitializer ();
        entry->assignment_func = cpp_NewExpression_Braces_set_newInitializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewExpression_Braces (Cpp_NewExpression_Braces const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewExpression_Braces\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewExpression_Braces\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: new\n");
    dump_cpp_NewPlacement (el->newPlacement, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_NewInitializer (el->newInitializer, nest_level + 1);

}

static Ref<Grammar>
create_cpp_NewExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("NewExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NewExpression_Basic ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NewExpression_Braces ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewExpression (Cpp_NewExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewExpression\n");
        return;
    }

    switch (el->new_expression_type) {
        case Cpp_NewExpression::_Basic:
            dump_cpp_NewExpression_Basic (static_cast <Cpp_NewExpression_Basic const *> (el), nest_level + 0);
            break;
        case Cpp_NewExpression::_Braces:
            dump_cpp_NewExpression_Braces (static_cast <Cpp_NewExpression_Braces const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_NewPlacement_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewPlacement ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_NewPlacement * const &self = static_cast <Cpp_NewPlacement*> (__self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_NewPlacement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewPlacement))) Cpp_NewPlacement;
}

static Ref<Grammar>
create_cpp_NewPlacement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewPlacement_creation_func));
    grammar->name = String::forData ("NewPlacement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_NewPlacement_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewPlacement (Cpp_NewPlacement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewPlacement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewPlacement\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_NewTypeId_set_typeSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewTypeId ||
             __el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_NewTypeId * const &self = static_cast <Cpp_NewTypeId*> (__self);

    self->typeSpecifierSeq = static_cast <Cpp_TypeSpecifierSeq*> (___el);
}

static void
cpp_NewTypeId_set_newDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewTypeId ||
             __el->cpp_element_type != CppElement::_NewDeclarator);

    Cpp_NewTypeId * const &self = static_cast <Cpp_NewTypeId*> (__self);

    self->newDeclarator = static_cast <Cpp_NewDeclarator*> (___el);
}

static ParserElement*
cpp_NewTypeId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewTypeId))) Cpp_NewTypeId;
}

static Ref<Grammar>
create_cpp_NewTypeId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewTypeId_creation_func));
    grammar->name = String::forData ("NewTypeId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_TypeSpecifierSeq ();
        entry->assignment_func = cpp_NewTypeId_set_typeSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NewDeclarator ();
        entry->assignment_func = cpp_NewTypeId_set_newDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewTypeId (Cpp_NewTypeId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewTypeId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewTypeId\n");

    dump_cpp_TypeSpecifierSeq (el->typeSpecifierSeq, nest_level + 1);

    dump_cpp_NewDeclarator (el->newDeclarator, nest_level + 1);

}

static void
cpp_NewDeclarator_PtrOperator_set_ptrOperator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewDeclarator ||
             __el->cpp_element_type != CppElement::_PtrOperator);

    Cpp_NewDeclarator * const &_self = static_cast <Cpp_NewDeclarator*> (__self);

    abortIf (_self->new_declarator_type != Cpp_NewDeclarator::_PtrOperator);

    Cpp_NewDeclarator_PtrOperator * const &self = static_cast <Cpp_NewDeclarator_PtrOperator*> (_self);

    self->ptrOperator = static_cast <Cpp_PtrOperator*> (___el);
}

static void
cpp_NewDeclarator_PtrOperator_set_newDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewDeclarator ||
             __el->cpp_element_type != CppElement::_NewDeclarator);

    Cpp_NewDeclarator * const &_self = static_cast <Cpp_NewDeclarator*> (__self);

    abortIf (_self->new_declarator_type != Cpp_NewDeclarator::_PtrOperator);

    Cpp_NewDeclarator_PtrOperator * const &self = static_cast <Cpp_NewDeclarator_PtrOperator*> (_self);

    self->newDeclarator = static_cast <Cpp_NewDeclarator*> (___el);
}

static ParserElement*
cpp_NewDeclarator_PtrOperator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewDeclarator_PtrOperator))) Cpp_NewDeclarator_PtrOperator;
}

static Ref<Grammar>
create_cpp_NewDeclarator_PtrOperator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewDeclarator_PtrOperator_creation_func));
    grammar->name = String::forData ("NewDeclarator_PtrOperator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PtrOperator ();
        entry->assignment_func = cpp_NewDeclarator_PtrOperator_set_ptrOperator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_NewDeclarator ();
        entry->assignment_func = cpp_NewDeclarator_PtrOperator_set_newDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewDeclarator_PtrOperator (Cpp_NewDeclarator_PtrOperator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewDeclarator_PtrOperator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewDeclarator_PtrOperator\n");

    dump_cpp_PtrOperator (el->ptrOperator, nest_level + 1);

    dump_cpp_NewDeclarator (el->newDeclarator, nest_level + 1);

}

static void
cpp_NewDeclarator_DirectNewDeclarator_set_directNewDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewDeclarator ||
             __el->cpp_element_type != CppElement::_DirectNewDeclarator);

    Cpp_NewDeclarator * const &_self = static_cast <Cpp_NewDeclarator*> (__self);

    abortIf (_self->new_declarator_type != Cpp_NewDeclarator::_DirectNewDeclarator);

    Cpp_NewDeclarator_DirectNewDeclarator * const &self = static_cast <Cpp_NewDeclarator_DirectNewDeclarator*> (_self);

    self->directNewDeclarator = static_cast <Cpp_DirectNewDeclarator*> (___el);
}

static ParserElement*
cpp_NewDeclarator_DirectNewDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewDeclarator_DirectNewDeclarator))) Cpp_NewDeclarator_DirectNewDeclarator;
}

static Ref<Grammar>
create_cpp_NewDeclarator_DirectNewDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewDeclarator_DirectNewDeclarator_creation_func));
    grammar->name = String::forData ("NewDeclarator_DirectNewDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DirectNewDeclarator ();
        entry->assignment_func = cpp_NewDeclarator_DirectNewDeclarator_set_directNewDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewDeclarator_DirectNewDeclarator (Cpp_NewDeclarator_DirectNewDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewDeclarator_DirectNewDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewDeclarator_DirectNewDeclarator\n");

    dump_cpp_DirectNewDeclarator (el->directNewDeclarator, nest_level + 1);

}

static Ref<Grammar>
create_cpp_NewDeclarator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("NewDeclarator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NewDeclarator_PtrOperator ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NewDeclarator_DirectNewDeclarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewDeclarator (Cpp_NewDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewDeclarator\n");
        return;
    }

    switch (el->new_declarator_type) {
        case Cpp_NewDeclarator::_PtrOperator:
            dump_cpp_NewDeclarator_PtrOperator (static_cast <Cpp_NewDeclarator_PtrOperator const *> (el), nest_level + 0);
            break;
        case Cpp_NewDeclarator::_DirectNewDeclarator:
            dump_cpp_NewDeclarator_DirectNewDeclarator (static_cast <Cpp_NewDeclarator_DirectNewDeclarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_DirectNewDeclarator_Expression_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectNewDeclarator ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_DirectNewDeclarator * const &_self = static_cast <Cpp_DirectNewDeclarator*> (__self);

    abortIf (_self->direct_new_declarator_type != Cpp_DirectNewDeclarator::_Expression);

    Cpp_DirectNewDeclarator_Expression * const &self = static_cast <Cpp_DirectNewDeclarator_Expression*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_DirectNewDeclarator_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DirectNewDeclarator_Expression))) Cpp_DirectNewDeclarator_Expression;
}

static Ref<Grammar>
create_cpp_DirectNewDeclarator_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DirectNewDeclarator_Expression_creation_func));
    grammar->name = String::forData ("DirectNewDeclarator_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_DirectNewDeclarator_Expression_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectNewDeclarator_Expression (Cpp_DirectNewDeclarator_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectNewDeclarator_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DirectNewDeclarator_Expression\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static void
cpp_DirectNewDeclarator_DirectNewDeclarator_set_directNewDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectNewDeclarator ||
             __el->cpp_element_type != CppElement::_DirectNewDeclarator);

    Cpp_DirectNewDeclarator * const &_self = static_cast <Cpp_DirectNewDeclarator*> (__self);

    abortIf (_self->direct_new_declarator_type != Cpp_DirectNewDeclarator::_DirectNewDeclarator);

    Cpp_DirectNewDeclarator_DirectNewDeclarator * const &self = static_cast <Cpp_DirectNewDeclarator_DirectNewDeclarator*> (_self);

    self->directNewDeclarator = static_cast <Cpp_DirectNewDeclarator*> (___el);
}

static void
cpp_DirectNewDeclarator_DirectNewDeclarator_set_constantExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectNewDeclarator ||
             __el->cpp_element_type != CppElement::_ConstantExpression);

    Cpp_DirectNewDeclarator * const &_self = static_cast <Cpp_DirectNewDeclarator*> (__self);

    abortIf (_self->direct_new_declarator_type != Cpp_DirectNewDeclarator::_DirectNewDeclarator);

    Cpp_DirectNewDeclarator_DirectNewDeclarator * const &self = static_cast <Cpp_DirectNewDeclarator_DirectNewDeclarator*> (_self);

    self->constantExpression = static_cast <Cpp_ConstantExpression*> (___el);
}

static ParserElement*
cpp_DirectNewDeclarator_DirectNewDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DirectNewDeclarator_DirectNewDeclarator))) Cpp_DirectNewDeclarator_DirectNewDeclarator;
}

static Ref<Grammar>
create_cpp_DirectNewDeclarator_DirectNewDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DirectNewDeclarator_DirectNewDeclarator_creation_func));
    grammar->name = String::forData ("DirectNewDeclarator_DirectNewDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DirectNewDeclarator ();
        entry->assignment_func = cpp_DirectNewDeclarator_DirectNewDeclarator_set_directNewDeclarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConstantExpression ();
        entry->assignment_func = cpp_DirectNewDeclarator_DirectNewDeclarator_set_constantExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectNewDeclarator_DirectNewDeclarator (Cpp_DirectNewDeclarator_DirectNewDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectNewDeclarator_DirectNewDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DirectNewDeclarator_DirectNewDeclarator\n");

    dump_cpp_DirectNewDeclarator (el->directNewDeclarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    dump_cpp_ConstantExpression (el->constantExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static Ref<Grammar>
create_cpp_DirectNewDeclarator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("DirectNewDeclarator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DirectNewDeclarator_Expression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DirectNewDeclarator_DirectNewDeclarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectNewDeclarator (Cpp_DirectNewDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectNewDeclarator\n");
        return;
    }

    switch (el->direct_new_declarator_type) {
        case Cpp_DirectNewDeclarator::_Expression:
            dump_cpp_DirectNewDeclarator_Expression (static_cast <Cpp_DirectNewDeclarator_Expression const *> (el), nest_level + 0);
            break;
        case Cpp_DirectNewDeclarator::_DirectNewDeclarator:
            dump_cpp_DirectNewDeclarator_DirectNewDeclarator (static_cast <Cpp_DirectNewDeclarator_DirectNewDeclarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_NewInitializer_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NewInitializer ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_NewInitializer * const &self = static_cast <Cpp_NewInitializer*> (__self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_NewInitializer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NewInitializer))) Cpp_NewInitializer;
}

static Ref<Grammar>
create_cpp_NewInitializer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NewInitializer_creation_func));
    grammar->name = String::forData ("NewInitializer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_NewInitializer_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NewInitializer (Cpp_NewInitializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NewInitializer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NewInitializer\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_DeleteExpression_Delete_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeleteExpression ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_DeleteExpression * const &_self = static_cast <Cpp_DeleteExpression*> (__self);

    abortIf (_self->delete_expression_type != Cpp_DeleteExpression::_Delete);

    Cpp_DeleteExpression_Delete * const &self = static_cast <Cpp_DeleteExpression_Delete*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_DeleteExpression_Delete_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeleteExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_DeleteExpression * const &_self = static_cast <Cpp_DeleteExpression*> (__self);

    abortIf (_self->delete_expression_type != Cpp_DeleteExpression::_Delete);

    Cpp_DeleteExpression_Delete * const &self = static_cast <Cpp_DeleteExpression_Delete*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_DeleteExpression_Delete_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeleteExpression_Delete))) Cpp_DeleteExpression_Delete;
}

static Ref<Grammar>
create_cpp_DeleteExpression_Delete ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeleteExpression_Delete_creation_func));
    grammar->name = String::forData ("DeleteExpression_Delete");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_DeleteExpression_Delete_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("delete"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_DeleteExpression ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_DeleteExpression_Delete_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeleteExpression_Delete (Cpp_DeleteExpression_Delete const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeleteExpression_Delete\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeleteExpression_Delete\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: delete\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_DeleteExpression_DeleteArray_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeleteExpression ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_DeleteExpression * const &_self = static_cast <Cpp_DeleteExpression*> (__self);

    abortIf (_self->delete_expression_type != Cpp_DeleteExpression::_DeleteArray);

    Cpp_DeleteExpression_DeleteArray * const &self = static_cast <Cpp_DeleteExpression_DeleteArray*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_DeleteExpression_DeleteArray_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeleteExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_DeleteExpression * const &_self = static_cast <Cpp_DeleteExpression*> (__self);

    abortIf (_self->delete_expression_type != Cpp_DeleteExpression::_DeleteArray);

    Cpp_DeleteExpression_DeleteArray * const &self = static_cast <Cpp_DeleteExpression_DeleteArray*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_DeleteExpression_DeleteArray_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeleteExpression_DeleteArray))) Cpp_DeleteExpression_DeleteArray;
}

static Ref<Grammar>
create_cpp_DeleteExpression_DeleteArray ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeleteExpression_DeleteArray_creation_func));
    grammar->name = String::forData ("DeleteExpression_DeleteArray");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_DeleteExpression_DeleteArray_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("delete"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_DeleteExpression_DeleteArray_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeleteExpression_DeleteArray (Cpp_DeleteExpression_DeleteArray const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeleteExpression_DeleteArray\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeleteExpression_DeleteArray\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: delete\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_DeleteExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("DeleteExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeleteExpression_Delete ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeleteExpression_DeleteArray ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeleteExpression (Cpp_DeleteExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeleteExpression\n");
        return;
    }

    switch (el->delete_expression_type) {
        case Cpp_DeleteExpression::_Delete:
            dump_cpp_DeleteExpression_Delete (static_cast <Cpp_DeleteExpression_Delete const *> (el), nest_level + 0);
            break;
        case Cpp_DeleteExpression::_DeleteArray:
            dump_cpp_DeleteExpression_DeleteArray (static_cast <Cpp_DeleteExpression_DeleteArray const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_CastExpression_Cast_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CastExpression ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_CastExpression * const &_self = static_cast <Cpp_CastExpression*> (__self);

    abortIf (_self->cast_expression_type != Cpp_CastExpression::_Cast);

    Cpp_CastExpression_Cast * const &self = static_cast <Cpp_CastExpression_Cast*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_CastExpression_Cast_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CastExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_CastExpression * const &_self = static_cast <Cpp_CastExpression*> (__self);

    abortIf (_self->cast_expression_type != Cpp_CastExpression::_Cast);

    Cpp_CastExpression_Cast * const &self = static_cast <Cpp_CastExpression_Cast*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

bool cpp_accept_cast_expression (
        Cpp_CastExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_cast_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_CastExpression);

    Cpp_CastExpression * const &el =
            static_cast <Cpp_CastExpression*> (_el);

    return cpp_accept_cast_expression (el, parser_control, data);
}

static ParserElement*
cpp_CastExpression_Cast_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CastExpression_Cast))) Cpp_CastExpression_Cast;
}

static Ref<Grammar>
create_cpp_CastExpression_Cast ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CastExpression_Cast_creation_func));
    grammar->name = String::forData ("CastExpression_Cast");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_CastExpression_Cast_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_CastExpression_Cast_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_cast_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CastExpression_Cast (Cpp_CastExpression_Cast const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CastExpression_Cast\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CastExpression_Cast\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_CastExpression_Expression_set_unaryExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CastExpression ||
             __el->cpp_element_type != CppElement::_UnaryExpression);

    Cpp_CastExpression * const &_self = static_cast <Cpp_CastExpression*> (__self);

    abortIf (_self->cast_expression_type != Cpp_CastExpression::_Expression);

    Cpp_CastExpression_Expression * const &self = static_cast <Cpp_CastExpression_Expression*> (_self);

    self->unaryExpression = static_cast <Cpp_UnaryExpression*> (___el);
}

static ParserElement*
cpp_CastExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CastExpression_Expression))) Cpp_CastExpression_Expression;
}

static Ref<Grammar>
create_cpp_CastExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CastExpression_Expression_creation_func));
    grammar->name = String::forData ("CastExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnaryExpression ();
        entry->assignment_func = cpp_CastExpression_Expression_set_unaryExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CastExpression_Expression (Cpp_CastExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CastExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CastExpression_Expression\n");

    dump_cpp_UnaryExpression (el->unaryExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_CastExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("CastExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_CastExpression_Cast ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_CastExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CastExpression (Cpp_CastExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CastExpression\n");
        return;
    }

    switch (el->cast_expression_type) {
        case Cpp_CastExpression::_Cast:
            dump_cpp_CastExpression_Cast (static_cast <Cpp_CastExpression_Cast const *> (el), nest_level + 0);
            break;
        case Cpp_CastExpression::_Expression:
            dump_cpp_CastExpression_Expression (static_cast <Cpp_CastExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_PmExpression_DotPm_set_pmExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PmExpression ||
             __el->cpp_element_type != CppElement::_PmExpression);

    Cpp_PmExpression * const &_self = static_cast <Cpp_PmExpression*> (__self);

    abortIf (_self->pm_expression_type != Cpp_PmExpression::_DotPm);

    Cpp_PmExpression_DotPm * const &self = static_cast <Cpp_PmExpression_DotPm*> (_self);

    self->pmExpression = static_cast <Cpp_PmExpression*> (___el);
}

static void
cpp_PmExpression_DotPm_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PmExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_PmExpression * const &_self = static_cast <Cpp_PmExpression*> (__self);

    abortIf (_self->pm_expression_type != Cpp_PmExpression::_DotPm);

    Cpp_PmExpression_DotPm * const &self = static_cast <Cpp_PmExpression_DotPm*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_PmExpression_DotPm_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PmExpression_DotPm))) Cpp_PmExpression_DotPm;
}

static Ref<Grammar>
create_cpp_PmExpression_DotPm ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PmExpression_DotPm_creation_func));
    grammar->name = String::forData ("PmExpression_DotPm");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PmExpression ();
        entry->assignment_func = cpp_PmExpression_DotPm_set_pmExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (".*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_PmExpression_DotPm_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PmExpression_DotPm (Cpp_PmExpression_DotPm const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PmExpression_DotPm\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PmExpression_DotPm\n");

    dump_cpp_PmExpression (el->pmExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: .*\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_PmExpression_ArrowPm_set_pmExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PmExpression ||
             __el->cpp_element_type != CppElement::_PmExpression);

    Cpp_PmExpression * const &_self = static_cast <Cpp_PmExpression*> (__self);

    abortIf (_self->pm_expression_type != Cpp_PmExpression::_ArrowPm);

    Cpp_PmExpression_ArrowPm * const &self = static_cast <Cpp_PmExpression_ArrowPm*> (_self);

    self->pmExpression = static_cast <Cpp_PmExpression*> (___el);
}

static void
cpp_PmExpression_ArrowPm_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PmExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_PmExpression * const &_self = static_cast <Cpp_PmExpression*> (__self);

    abortIf (_self->pm_expression_type != Cpp_PmExpression::_ArrowPm);

    Cpp_PmExpression_ArrowPm * const &self = static_cast <Cpp_PmExpression_ArrowPm*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_PmExpression_ArrowPm_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PmExpression_ArrowPm))) Cpp_PmExpression_ArrowPm;
}

static Ref<Grammar>
create_cpp_PmExpression_ArrowPm ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PmExpression_ArrowPm_creation_func));
    grammar->name = String::forData ("PmExpression_ArrowPm");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PmExpression ();
        entry->assignment_func = cpp_PmExpression_ArrowPm_set_pmExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("->*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_PmExpression_ArrowPm_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PmExpression_ArrowPm (Cpp_PmExpression_ArrowPm const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PmExpression_ArrowPm\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PmExpression_ArrowPm\n");

    dump_cpp_PmExpression (el->pmExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ->*\n");
    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static void
cpp_PmExpression_Expression_set_castExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PmExpression ||
             __el->cpp_element_type != CppElement::_CastExpression);

    Cpp_PmExpression * const &_self = static_cast <Cpp_PmExpression*> (__self);

    abortIf (_self->pm_expression_type != Cpp_PmExpression::_Expression);

    Cpp_PmExpression_Expression * const &self = static_cast <Cpp_PmExpression_Expression*> (_self);

    self->castExpression = static_cast <Cpp_CastExpression*> (___el);
}

static ParserElement*
cpp_PmExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PmExpression_Expression))) Cpp_PmExpression_Expression;
}

static Ref<Grammar>
create_cpp_PmExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PmExpression_Expression_creation_func));
    grammar->name = String::forData ("PmExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CastExpression ();
        entry->assignment_func = cpp_PmExpression_Expression_set_castExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PmExpression_Expression (Cpp_PmExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PmExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PmExpression_Expression\n");

    dump_cpp_CastExpression (el->castExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_PmExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("PmExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PmExpression_DotPm ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PmExpression_ArrowPm ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PmExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PmExpression (Cpp_PmExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PmExpression\n");
        return;
    }

    switch (el->pm_expression_type) {
        case Cpp_PmExpression::_DotPm:
            dump_cpp_PmExpression_DotPm (static_cast <Cpp_PmExpression_DotPm const *> (el), nest_level + 0);
            break;
        case Cpp_PmExpression::_ArrowPm:
            dump_cpp_PmExpression_ArrowPm (static_cast <Cpp_PmExpression_ArrowPm const *> (el), nest_level + 0);
            break;
        case Cpp_PmExpression::_Expression:
            dump_cpp_PmExpression_Expression (static_cast <Cpp_PmExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MultiplicativeExpression_Multiplication_set_multiplicativeExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Multiplication);

    Cpp_MultiplicativeExpression_Multiplication * const &self = static_cast <Cpp_MultiplicativeExpression_Multiplication*> (_self);

    self->multiplicativeExpression = static_cast <Cpp_MultiplicativeExpression*> (___el);
}

static void
cpp_MultiplicativeExpression_Multiplication_set_pmExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_PmExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Multiplication);

    Cpp_MultiplicativeExpression_Multiplication * const &self = static_cast <Cpp_MultiplicativeExpression_Multiplication*> (_self);

    self->pmExpression = static_cast <Cpp_PmExpression*> (___el);
}

bool cpp_accept_multiplicative_expression (
        Cpp_MultiplicativeExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_multiplicative_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_MultiplicativeExpression * const &el =
            static_cast <Cpp_MultiplicativeExpression*> (_el);

    return cpp_accept_multiplicative_expression (el, parser_control, data);
}

static ParserElement*
cpp_MultiplicativeExpression_Multiplication_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MultiplicativeExpression_Multiplication))) Cpp_MultiplicativeExpression_Multiplication;
}

static Ref<Grammar>
create_cpp_MultiplicativeExpression_Multiplication ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MultiplicativeExpression_Multiplication_creation_func));
    grammar->name = String::forData ("MultiplicativeExpression_Multiplication");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Multiplication_set_multiplicativeExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PmExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Multiplication_set_pmExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_multiplicative_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MultiplicativeExpression_Multiplication (Cpp_MultiplicativeExpression_Multiplication const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MultiplicativeExpression_Multiplication\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MultiplicativeExpression_Multiplication\n");

    dump_cpp_MultiplicativeExpression (el->multiplicativeExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: *\n");
    dump_cpp_PmExpression (el->pmExpression, nest_level + 1);

}

static void
cpp_MultiplicativeExpression_Division_set_multiplicativeExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Division);

    Cpp_MultiplicativeExpression_Division * const &self = static_cast <Cpp_MultiplicativeExpression_Division*> (_self);

    self->multiplicativeExpression = static_cast <Cpp_MultiplicativeExpression*> (___el);
}

static void
cpp_MultiplicativeExpression_Division_set_pmExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_PmExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Division);

    Cpp_MultiplicativeExpression_Division * const &self = static_cast <Cpp_MultiplicativeExpression_Division*> (_self);

    self->pmExpression = static_cast <Cpp_PmExpression*> (___el);
}

static ParserElement*
cpp_MultiplicativeExpression_Division_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MultiplicativeExpression_Division))) Cpp_MultiplicativeExpression_Division;
}

static Ref<Grammar>
create_cpp_MultiplicativeExpression_Division ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MultiplicativeExpression_Division_creation_func));
    grammar->name = String::forData ("MultiplicativeExpression_Division");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Division_set_multiplicativeExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("/"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PmExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Division_set_pmExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_multiplicative_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MultiplicativeExpression_Division (Cpp_MultiplicativeExpression_Division const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MultiplicativeExpression_Division\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MultiplicativeExpression_Division\n");

    dump_cpp_MultiplicativeExpression (el->multiplicativeExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: /\n");
    dump_cpp_PmExpression (el->pmExpression, nest_level + 1);

}

static void
cpp_MultiplicativeExpression_Remainder_set_multiplicativeExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Remainder);

    Cpp_MultiplicativeExpression_Remainder * const &self = static_cast <Cpp_MultiplicativeExpression_Remainder*> (_self);

    self->multiplicativeExpression = static_cast <Cpp_MultiplicativeExpression*> (___el);
}

static void
cpp_MultiplicativeExpression_Remainder_set_pmExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_PmExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Remainder);

    Cpp_MultiplicativeExpression_Remainder * const &self = static_cast <Cpp_MultiplicativeExpression_Remainder*> (_self);

    self->pmExpression = static_cast <Cpp_PmExpression*> (___el);
}

static ParserElement*
cpp_MultiplicativeExpression_Remainder_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MultiplicativeExpression_Remainder))) Cpp_MultiplicativeExpression_Remainder;
}

static Ref<Grammar>
create_cpp_MultiplicativeExpression_Remainder ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MultiplicativeExpression_Remainder_creation_func));
    grammar->name = String::forData ("MultiplicativeExpression_Remainder");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Remainder_set_multiplicativeExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("%"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PmExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Remainder_set_pmExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_multiplicative_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MultiplicativeExpression_Remainder (Cpp_MultiplicativeExpression_Remainder const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MultiplicativeExpression_Remainder\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MultiplicativeExpression_Remainder\n");

    dump_cpp_MultiplicativeExpression (el->multiplicativeExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: %\n");
    dump_cpp_PmExpression (el->pmExpression, nest_level + 1);

}

static void
cpp_MultiplicativeExpression_Expression_set_pmExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MultiplicativeExpression ||
             __el->cpp_element_type != CppElement::_PmExpression);

    Cpp_MultiplicativeExpression * const &_self = static_cast <Cpp_MultiplicativeExpression*> (__self);

    abortIf (_self->multiplicative_expression_type != Cpp_MultiplicativeExpression::_Expression);

    Cpp_MultiplicativeExpression_Expression * const &self = static_cast <Cpp_MultiplicativeExpression_Expression*> (_self);

    self->pmExpression = static_cast <Cpp_PmExpression*> (___el);
}

static ParserElement*
cpp_MultiplicativeExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MultiplicativeExpression_Expression))) Cpp_MultiplicativeExpression_Expression;
}

static Ref<Grammar>
create_cpp_MultiplicativeExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MultiplicativeExpression_Expression_creation_func));
    grammar->name = String::forData ("MultiplicativeExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PmExpression ();
        entry->assignment_func = cpp_MultiplicativeExpression_Expression_set_pmExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MultiplicativeExpression_Expression (Cpp_MultiplicativeExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MultiplicativeExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MultiplicativeExpression_Expression\n");

    dump_cpp_PmExpression (el->pmExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MultiplicativeExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MultiplicativeExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression_Multiplication ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression_Division ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression_Remainder ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MultiplicativeExpression (Cpp_MultiplicativeExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MultiplicativeExpression\n");
        return;
    }

    switch (el->multiplicative_expression_type) {
        case Cpp_MultiplicativeExpression::_Multiplication:
            dump_cpp_MultiplicativeExpression_Multiplication (static_cast <Cpp_MultiplicativeExpression_Multiplication const *> (el), nest_level + 0);
            break;
        case Cpp_MultiplicativeExpression::_Division:
            dump_cpp_MultiplicativeExpression_Division (static_cast <Cpp_MultiplicativeExpression_Division const *> (el), nest_level + 0);
            break;
        case Cpp_MultiplicativeExpression::_Remainder:
            dump_cpp_MultiplicativeExpression_Remainder (static_cast <Cpp_MultiplicativeExpression_Remainder const *> (el), nest_level + 0);
            break;
        case Cpp_MultiplicativeExpression::_Expression:
            dump_cpp_MultiplicativeExpression_Expression (static_cast <Cpp_MultiplicativeExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_AdditiveExpression_Addition_set_additiveExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AdditiveExpression ||
             __el->cpp_element_type != CppElement::_AdditiveExpression);

    Cpp_AdditiveExpression * const &_self = static_cast <Cpp_AdditiveExpression*> (__self);

    abortIf (_self->additive_expression_type != Cpp_AdditiveExpression::_Addition);

    Cpp_AdditiveExpression_Addition * const &self = static_cast <Cpp_AdditiveExpression_Addition*> (_self);

    self->additiveExpression = static_cast <Cpp_AdditiveExpression*> (___el);
}

static void
cpp_AdditiveExpression_Addition_set_multiplicativeExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AdditiveExpression ||
             __el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_AdditiveExpression * const &_self = static_cast <Cpp_AdditiveExpression*> (__self);

    abortIf (_self->additive_expression_type != Cpp_AdditiveExpression::_Addition);

    Cpp_AdditiveExpression_Addition * const &self = static_cast <Cpp_AdditiveExpression_Addition*> (_self);

    self->multiplicativeExpression = static_cast <Cpp_MultiplicativeExpression*> (___el);
}

bool cpp_accept_additive_expression (
        Cpp_AdditiveExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_additive_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_AdditiveExpression);

    Cpp_AdditiveExpression * const &el =
            static_cast <Cpp_AdditiveExpression*> (_el);

    return cpp_accept_additive_expression (el, parser_control, data);
}

static ParserElement*
cpp_AdditiveExpression_Addition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AdditiveExpression_Addition))) Cpp_AdditiveExpression_Addition;
}

static Ref<Grammar>
create_cpp_AdditiveExpression_Addition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AdditiveExpression_Addition_creation_func));
    grammar->name = String::forData ("AdditiveExpression_Addition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression ();
        entry->assignment_func = cpp_AdditiveExpression_Addition_set_additiveExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("+"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression ();
        entry->assignment_func = cpp_AdditiveExpression_Addition_set_multiplicativeExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_additive_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AdditiveExpression_Addition (Cpp_AdditiveExpression_Addition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AdditiveExpression_Addition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AdditiveExpression_Addition\n");

    dump_cpp_AdditiveExpression (el->additiveExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: +\n");
    dump_cpp_MultiplicativeExpression (el->multiplicativeExpression, nest_level + 1);

}

static void
cpp_AdditiveExpression_Subtraction_set_additiveExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AdditiveExpression ||
             __el->cpp_element_type != CppElement::_AdditiveExpression);

    Cpp_AdditiveExpression * const &_self = static_cast <Cpp_AdditiveExpression*> (__self);

    abortIf (_self->additive_expression_type != Cpp_AdditiveExpression::_Subtraction);

    Cpp_AdditiveExpression_Subtraction * const &self = static_cast <Cpp_AdditiveExpression_Subtraction*> (_self);

    self->additiveExpression = static_cast <Cpp_AdditiveExpression*> (___el);
}

static void
cpp_AdditiveExpression_Subtraction_set_multiplicativeExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AdditiveExpression ||
             __el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_AdditiveExpression * const &_self = static_cast <Cpp_AdditiveExpression*> (__self);

    abortIf (_self->additive_expression_type != Cpp_AdditiveExpression::_Subtraction);

    Cpp_AdditiveExpression_Subtraction * const &self = static_cast <Cpp_AdditiveExpression_Subtraction*> (_self);

    self->multiplicativeExpression = static_cast <Cpp_MultiplicativeExpression*> (___el);
}

static ParserElement*
cpp_AdditiveExpression_Subtraction_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AdditiveExpression_Subtraction))) Cpp_AdditiveExpression_Subtraction;
}

static Ref<Grammar>
create_cpp_AdditiveExpression_Subtraction ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AdditiveExpression_Subtraction_creation_func));
    grammar->name = String::forData ("AdditiveExpression_Subtraction");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression ();
        entry->assignment_func = cpp_AdditiveExpression_Subtraction_set_additiveExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("-"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression ();
        entry->assignment_func = cpp_AdditiveExpression_Subtraction_set_multiplicativeExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_additive_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AdditiveExpression_Subtraction (Cpp_AdditiveExpression_Subtraction const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AdditiveExpression_Subtraction\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AdditiveExpression_Subtraction\n");

    dump_cpp_AdditiveExpression (el->additiveExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: -\n");
    dump_cpp_MultiplicativeExpression (el->multiplicativeExpression, nest_level + 1);

}

static void
cpp_AdditiveExpression_Expression_set_multiplicativeExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AdditiveExpression ||
             __el->cpp_element_type != CppElement::_MultiplicativeExpression);

    Cpp_AdditiveExpression * const &_self = static_cast <Cpp_AdditiveExpression*> (__self);

    abortIf (_self->additive_expression_type != Cpp_AdditiveExpression::_Expression);

    Cpp_AdditiveExpression_Expression * const &self = static_cast <Cpp_AdditiveExpression_Expression*> (_self);

    self->multiplicativeExpression = static_cast <Cpp_MultiplicativeExpression*> (___el);
}

static ParserElement*
cpp_AdditiveExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AdditiveExpression_Expression))) Cpp_AdditiveExpression_Expression;
}

static Ref<Grammar>
create_cpp_AdditiveExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AdditiveExpression_Expression_creation_func));
    grammar->name = String::forData ("AdditiveExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MultiplicativeExpression ();
        entry->assignment_func = cpp_AdditiveExpression_Expression_set_multiplicativeExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AdditiveExpression_Expression (Cpp_AdditiveExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AdditiveExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AdditiveExpression_Expression\n");

    dump_cpp_MultiplicativeExpression (el->multiplicativeExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_AdditiveExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("AdditiveExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression_Addition ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression_Subtraction ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AdditiveExpression (Cpp_AdditiveExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AdditiveExpression\n");
        return;
    }

    switch (el->additive_expression_type) {
        case Cpp_AdditiveExpression::_Addition:
            dump_cpp_AdditiveExpression_Addition (static_cast <Cpp_AdditiveExpression_Addition const *> (el), nest_level + 0);
            break;
        case Cpp_AdditiveExpression::_Subtraction:
            dump_cpp_AdditiveExpression_Subtraction (static_cast <Cpp_AdditiveExpression_Subtraction const *> (el), nest_level + 0);
            break;
        case Cpp_AdditiveExpression::_Expression:
            dump_cpp_AdditiveExpression_Expression (static_cast <Cpp_AdditiveExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ShiftExpression_LeftShift_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ShiftExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_ShiftExpression * const &_self = static_cast <Cpp_ShiftExpression*> (__self);

    abortIf (_self->shift_expression_type != Cpp_ShiftExpression::_LeftShift);

    Cpp_ShiftExpression_LeftShift * const &self = static_cast <Cpp_ShiftExpression_LeftShift*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

static void
cpp_ShiftExpression_LeftShift_set_additiveExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ShiftExpression ||
             __el->cpp_element_type != CppElement::_AdditiveExpression);

    Cpp_ShiftExpression * const &_self = static_cast <Cpp_ShiftExpression*> (__self);

    abortIf (_self->shift_expression_type != Cpp_ShiftExpression::_LeftShift);

    Cpp_ShiftExpression_LeftShift * const &self = static_cast <Cpp_ShiftExpression_LeftShift*> (_self);

    self->additiveExpression = static_cast <Cpp_AdditiveExpression*> (___el);
}

bool cpp_accept_shift_expression (
        Cpp_ShiftExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_shift_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_ShiftExpression * const &el =
            static_cast <Cpp_ShiftExpression*> (_el);

    return cpp_accept_shift_expression (el, parser_control, data);
}

static ParserElement*
cpp_ShiftExpression_LeftShift_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ShiftExpression_LeftShift))) Cpp_ShiftExpression_LeftShift;
}

static Ref<Grammar>
create_cpp_ShiftExpression_LeftShift ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ShiftExpression_LeftShift_creation_func));
    grammar->name = String::forData ("ShiftExpression_LeftShift");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_ShiftExpression_LeftShift_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression ();
        entry->assignment_func = cpp_ShiftExpression_LeftShift_set_additiveExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_shift_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ShiftExpression_LeftShift (Cpp_ShiftExpression_LeftShift const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ShiftExpression_LeftShift\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ShiftExpression_LeftShift\n");

    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: <<\n");
    dump_cpp_AdditiveExpression (el->additiveExpression, nest_level + 1);

}

static void
cpp_ShiftExpression_RightShift_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ShiftExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_ShiftExpression * const &_self = static_cast <Cpp_ShiftExpression*> (__self);

    abortIf (_self->shift_expression_type != Cpp_ShiftExpression::_RightShift);

    Cpp_ShiftExpression_RightShift * const &self = static_cast <Cpp_ShiftExpression_RightShift*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

static void
cpp_ShiftExpression_RightShift_set_additiveExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ShiftExpression ||
             __el->cpp_element_type != CppElement::_AdditiveExpression);

    Cpp_ShiftExpression * const &_self = static_cast <Cpp_ShiftExpression*> (__self);

    abortIf (_self->shift_expression_type != Cpp_ShiftExpression::_RightShift);

    Cpp_ShiftExpression_RightShift * const &self = static_cast <Cpp_ShiftExpression_RightShift*> (_self);

    self->additiveExpression = static_cast <Cpp_AdditiveExpression*> (___el);
}

static ParserElement*
cpp_ShiftExpression_RightShift_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ShiftExpression_RightShift))) Cpp_ShiftExpression_RightShift;
}

static Ref<Grammar>
create_cpp_ShiftExpression_RightShift ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ShiftExpression_RightShift_creation_func));
    grammar->name = String::forData ("ShiftExpression_RightShift");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_ShiftExpression_RightShift_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">>"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression ();
        entry->assignment_func = cpp_ShiftExpression_RightShift_set_additiveExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_shift_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ShiftExpression_RightShift (Cpp_ShiftExpression_RightShift const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ShiftExpression_RightShift\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ShiftExpression_RightShift\n");

    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >>\n");
    dump_cpp_AdditiveExpression (el->additiveExpression, nest_level + 1);

}

static void
cpp_ShiftExpression_Expression_set_additiveExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ShiftExpression ||
             __el->cpp_element_type != CppElement::_AdditiveExpression);

    Cpp_ShiftExpression * const &_self = static_cast <Cpp_ShiftExpression*> (__self);

    abortIf (_self->shift_expression_type != Cpp_ShiftExpression::_Expression);

    Cpp_ShiftExpression_Expression * const &self = static_cast <Cpp_ShiftExpression_Expression*> (_self);

    self->additiveExpression = static_cast <Cpp_AdditiveExpression*> (___el);
}

static ParserElement*
cpp_ShiftExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ShiftExpression_Expression))) Cpp_ShiftExpression_Expression;
}

static Ref<Grammar>
create_cpp_ShiftExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ShiftExpression_Expression_creation_func));
    grammar->name = String::forData ("ShiftExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AdditiveExpression ();
        entry->assignment_func = cpp_ShiftExpression_Expression_set_additiveExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ShiftExpression_Expression (Cpp_ShiftExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ShiftExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ShiftExpression_Expression\n");

    dump_cpp_AdditiveExpression (el->additiveExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ShiftExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ShiftExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression_LeftShift ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression_RightShift ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ShiftExpression (Cpp_ShiftExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ShiftExpression\n");
        return;
    }

    switch (el->shift_expression_type) {
        case Cpp_ShiftExpression::_LeftShift:
            dump_cpp_ShiftExpression_LeftShift (static_cast <Cpp_ShiftExpression_LeftShift const *> (el), nest_level + 0);
            break;
        case Cpp_ShiftExpression::_RightShift:
            dump_cpp_ShiftExpression_RightShift (static_cast <Cpp_ShiftExpression_RightShift const *> (el), nest_level + 0);
            break;
        case Cpp_ShiftExpression::_Expression:
            dump_cpp_ShiftExpression_Expression (static_cast <Cpp_ShiftExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_RelationalExpression_Less_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_Less);

    Cpp_RelationalExpression_Less * const &self = static_cast <Cpp_RelationalExpression_Less*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

static void
cpp_RelationalExpression_Less_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_Less);

    Cpp_RelationalExpression_Less * const &self = static_cast <Cpp_RelationalExpression_Less*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

bool cpp_accept_relational_expression (
        Cpp_RelationalExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_relational_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_RelationalExpression * const &el =
            static_cast <Cpp_RelationalExpression*> (_el);

    return cpp_accept_relational_expression (el, parser_control, data);
}

static ParserElement*
cpp_RelationalExpression_Less_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_RelationalExpression_Less))) Cpp_RelationalExpression_Less;
}

static Ref<Grammar>
create_cpp_RelationalExpression_Less ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_RelationalExpression_Less_creation_func));
    grammar->name = String::forData ("RelationalExpression_Less");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_RelationalExpression_Less_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_RelationalExpression_Less_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_relational_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_RelationalExpression_Less (Cpp_RelationalExpression_Less const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) RelationalExpression_Less\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("RelationalExpression_Less\n");

    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

}

static void
cpp_RelationalExpression_Greater_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_Greater);

    Cpp_RelationalExpression_Greater * const &self = static_cast <Cpp_RelationalExpression_Greater*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

static void
cpp_RelationalExpression_Greater_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_Greater);

    Cpp_RelationalExpression_Greater * const &self = static_cast <Cpp_RelationalExpression_Greater*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

static ParserElement*
cpp_RelationalExpression_Greater_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_RelationalExpression_Greater))) Cpp_RelationalExpression_Greater;
}

static Ref<Grammar>
create_cpp_RelationalExpression_Greater ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_RelationalExpression_Greater_creation_func));
    grammar->name = String::forData ("RelationalExpression_Greater");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_RelationalExpression_Greater_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_RelationalExpression_Greater_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_relational_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_RelationalExpression_Greater (Cpp_RelationalExpression_Greater const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) RelationalExpression_Greater\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("RelationalExpression_Greater\n");

    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

}

static void
cpp_RelationalExpression_LessOrEqual_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_LessOrEqual);

    Cpp_RelationalExpression_LessOrEqual * const &self = static_cast <Cpp_RelationalExpression_LessOrEqual*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

static void
cpp_RelationalExpression_LessOrEqual_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_LessOrEqual);

    Cpp_RelationalExpression_LessOrEqual * const &self = static_cast <Cpp_RelationalExpression_LessOrEqual*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

static ParserElement*
cpp_RelationalExpression_LessOrEqual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_RelationalExpression_LessOrEqual))) Cpp_RelationalExpression_LessOrEqual;
}

static Ref<Grammar>
create_cpp_RelationalExpression_LessOrEqual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_RelationalExpression_LessOrEqual_creation_func));
    grammar->name = String::forData ("RelationalExpression_LessOrEqual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_RelationalExpression_LessOrEqual_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_RelationalExpression_LessOrEqual_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_relational_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_RelationalExpression_LessOrEqual (Cpp_RelationalExpression_LessOrEqual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) RelationalExpression_LessOrEqual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("RelationalExpression_LessOrEqual\n");

    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: <=\n");
    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

}

static void
cpp_RelationalExpression_GreaterOrEqual_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_GreaterOrEqual);

    Cpp_RelationalExpression_GreaterOrEqual * const &self = static_cast <Cpp_RelationalExpression_GreaterOrEqual*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

static void
cpp_RelationalExpression_GreaterOrEqual_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_GreaterOrEqual);

    Cpp_RelationalExpression_GreaterOrEqual * const &self = static_cast <Cpp_RelationalExpression_GreaterOrEqual*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

static ParserElement*
cpp_RelationalExpression_GreaterOrEqual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_RelationalExpression_GreaterOrEqual))) Cpp_RelationalExpression_GreaterOrEqual;
}

static Ref<Grammar>
create_cpp_RelationalExpression_GreaterOrEqual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_RelationalExpression_GreaterOrEqual_creation_func));
    grammar->name = String::forData ("RelationalExpression_GreaterOrEqual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_RelationalExpression_GreaterOrEqual_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_RelationalExpression_GreaterOrEqual_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_relational_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_RelationalExpression_GreaterOrEqual (Cpp_RelationalExpression_GreaterOrEqual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) RelationalExpression_GreaterOrEqual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("RelationalExpression_GreaterOrEqual\n");

    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >=\n");
    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

}

static void
cpp_RelationalExpression_Expression_set_shiftExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_RelationalExpression ||
             __el->cpp_element_type != CppElement::_ShiftExpression);

    Cpp_RelationalExpression * const &_self = static_cast <Cpp_RelationalExpression*> (__self);

    abortIf (_self->relational_expression_type != Cpp_RelationalExpression::_Expression);

    Cpp_RelationalExpression_Expression * const &self = static_cast <Cpp_RelationalExpression_Expression*> (_self);

    self->shiftExpression = static_cast <Cpp_ShiftExpression*> (___el);
}

static ParserElement*
cpp_RelationalExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_RelationalExpression_Expression))) Cpp_RelationalExpression_Expression;
}

static Ref<Grammar>
create_cpp_RelationalExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_RelationalExpression_Expression_creation_func));
    grammar->name = String::forData ("RelationalExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ShiftExpression ();
        entry->assignment_func = cpp_RelationalExpression_Expression_set_shiftExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_RelationalExpression_Expression (Cpp_RelationalExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) RelationalExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("RelationalExpression_Expression\n");

    dump_cpp_ShiftExpression (el->shiftExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_RelationalExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("RelationalExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression_Less ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression_Greater ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression_LessOrEqual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression_GreaterOrEqual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_RelationalExpression (Cpp_RelationalExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) RelationalExpression\n");
        return;
    }

    switch (el->relational_expression_type) {
        case Cpp_RelationalExpression::_Less:
            dump_cpp_RelationalExpression_Less (static_cast <Cpp_RelationalExpression_Less const *> (el), nest_level + 0);
            break;
        case Cpp_RelationalExpression::_Greater:
            dump_cpp_RelationalExpression_Greater (static_cast <Cpp_RelationalExpression_Greater const *> (el), nest_level + 0);
            break;
        case Cpp_RelationalExpression::_LessOrEqual:
            dump_cpp_RelationalExpression_LessOrEqual (static_cast <Cpp_RelationalExpression_LessOrEqual const *> (el), nest_level + 0);
            break;
        case Cpp_RelationalExpression::_GreaterOrEqual:
            dump_cpp_RelationalExpression_GreaterOrEqual (static_cast <Cpp_RelationalExpression_GreaterOrEqual const *> (el), nest_level + 0);
            break;
        case Cpp_RelationalExpression::_Expression:
            dump_cpp_RelationalExpression_Expression (static_cast <Cpp_RelationalExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_EqualityExpression_Equal_set_equalityExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EqualityExpression ||
             __el->cpp_element_type != CppElement::_EqualityExpression);

    Cpp_EqualityExpression * const &_self = static_cast <Cpp_EqualityExpression*> (__self);

    abortIf (_self->equality_expression_type != Cpp_EqualityExpression::_Equal);

    Cpp_EqualityExpression_Equal * const &self = static_cast <Cpp_EqualityExpression_Equal*> (_self);

    self->equalityExpression = static_cast <Cpp_EqualityExpression*> (___el);
}

static void
cpp_EqualityExpression_Equal_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EqualityExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_EqualityExpression * const &_self = static_cast <Cpp_EqualityExpression*> (__self);

    abortIf (_self->equality_expression_type != Cpp_EqualityExpression::_Equal);

    Cpp_EqualityExpression_Equal * const &self = static_cast <Cpp_EqualityExpression_Equal*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

bool cpp_accept_equality_expression (
        Cpp_EqualityExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_equality_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_EqualityExpression);

    Cpp_EqualityExpression * const &el =
            static_cast <Cpp_EqualityExpression*> (_el);

    return cpp_accept_equality_expression (el, parser_control, data);
}

static ParserElement*
cpp_EqualityExpression_Equal_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EqualityExpression_Equal))) Cpp_EqualityExpression_Equal;
}

static Ref<Grammar>
create_cpp_EqualityExpression_Equal ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EqualityExpression_Equal_creation_func));
    grammar->name = String::forData ("EqualityExpression_Equal");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression ();
        entry->assignment_func = cpp_EqualityExpression_Equal_set_equalityExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("=="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_EqualityExpression_Equal_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_equality_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EqualityExpression_Equal (Cpp_EqualityExpression_Equal const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EqualityExpression_Equal\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EqualityExpression_Equal\n");

    dump_cpp_EqualityExpression (el->equalityExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ==\n");
    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

}

static void
cpp_EqualityExpression_NotEqual_set_equalityExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EqualityExpression ||
             __el->cpp_element_type != CppElement::_EqualityExpression);

    Cpp_EqualityExpression * const &_self = static_cast <Cpp_EqualityExpression*> (__self);

    abortIf (_self->equality_expression_type != Cpp_EqualityExpression::_NotEqual);

    Cpp_EqualityExpression_NotEqual * const &self = static_cast <Cpp_EqualityExpression_NotEqual*> (_self);

    self->equalityExpression = static_cast <Cpp_EqualityExpression*> (___el);
}

static void
cpp_EqualityExpression_NotEqual_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EqualityExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_EqualityExpression * const &_self = static_cast <Cpp_EqualityExpression*> (__self);

    abortIf (_self->equality_expression_type != Cpp_EqualityExpression::_NotEqual);

    Cpp_EqualityExpression_NotEqual * const &self = static_cast <Cpp_EqualityExpression_NotEqual*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

static ParserElement*
cpp_EqualityExpression_NotEqual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EqualityExpression_NotEqual))) Cpp_EqualityExpression_NotEqual;
}

static Ref<Grammar>
create_cpp_EqualityExpression_NotEqual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EqualityExpression_NotEqual_creation_func));
    grammar->name = String::forData ("EqualityExpression_NotEqual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression ();
        entry->assignment_func = cpp_EqualityExpression_NotEqual_set_equalityExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("!="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_EqualityExpression_NotEqual_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_equality_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EqualityExpression_NotEqual (Cpp_EqualityExpression_NotEqual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EqualityExpression_NotEqual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EqualityExpression_NotEqual\n");

    dump_cpp_EqualityExpression (el->equalityExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: !=\n");
    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

}

static void
cpp_EqualityExpression_Expression_set_relationalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EqualityExpression ||
             __el->cpp_element_type != CppElement::_RelationalExpression);

    Cpp_EqualityExpression * const &_self = static_cast <Cpp_EqualityExpression*> (__self);

    abortIf (_self->equality_expression_type != Cpp_EqualityExpression::_Expression);

    Cpp_EqualityExpression_Expression * const &self = static_cast <Cpp_EqualityExpression_Expression*> (_self);

    self->relationalExpression = static_cast <Cpp_RelationalExpression*> (___el);
}

static ParserElement*
cpp_EqualityExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EqualityExpression_Expression))) Cpp_EqualityExpression_Expression;
}

static Ref<Grammar>
create_cpp_EqualityExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EqualityExpression_Expression_creation_func));
    grammar->name = String::forData ("EqualityExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_RelationalExpression ();
        entry->assignment_func = cpp_EqualityExpression_Expression_set_relationalExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EqualityExpression_Expression (Cpp_EqualityExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EqualityExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EqualityExpression_Expression\n");

    dump_cpp_RelationalExpression (el->relationalExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_EqualityExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("EqualityExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression_Equal ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression_NotEqual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EqualityExpression (Cpp_EqualityExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EqualityExpression\n");
        return;
    }

    switch (el->equality_expression_type) {
        case Cpp_EqualityExpression::_Equal:
            dump_cpp_EqualityExpression_Equal (static_cast <Cpp_EqualityExpression_Equal const *> (el), nest_level + 0);
            break;
        case Cpp_EqualityExpression::_NotEqual:
            dump_cpp_EqualityExpression_NotEqual (static_cast <Cpp_EqualityExpression_NotEqual const *> (el), nest_level + 0);
            break;
        case Cpp_EqualityExpression::_Expression:
            dump_cpp_EqualityExpression_Expression (static_cast <Cpp_EqualityExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_AndExpression_BitwiseAnd_set_andExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AndExpression ||
             __el->cpp_element_type != CppElement::_AndExpression);

    Cpp_AndExpression * const &_self = static_cast <Cpp_AndExpression*> (__self);

    abortIf (_self->and_expression_type != Cpp_AndExpression::_BitwiseAnd);

    Cpp_AndExpression_BitwiseAnd * const &self = static_cast <Cpp_AndExpression_BitwiseAnd*> (_self);

    self->andExpression = static_cast <Cpp_AndExpression*> (___el);
}

static void
cpp_AndExpression_BitwiseAnd_set_equalityExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AndExpression ||
             __el->cpp_element_type != CppElement::_EqualityExpression);

    Cpp_AndExpression * const &_self = static_cast <Cpp_AndExpression*> (__self);

    abortIf (_self->and_expression_type != Cpp_AndExpression::_BitwiseAnd);

    Cpp_AndExpression_BitwiseAnd * const &self = static_cast <Cpp_AndExpression_BitwiseAnd*> (_self);

    self->equalityExpression = static_cast <Cpp_EqualityExpression*> (___el);
}

bool cpp_accept_and_expression (
        Cpp_AndExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_and_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_AndExpression);

    Cpp_AndExpression * const &el =
            static_cast <Cpp_AndExpression*> (_el);

    return cpp_accept_and_expression (el, parser_control, data);
}

static ParserElement*
cpp_AndExpression_BitwiseAnd_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AndExpression_BitwiseAnd))) Cpp_AndExpression_BitwiseAnd;
}

static Ref<Grammar>
create_cpp_AndExpression_BitwiseAnd ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AndExpression_BitwiseAnd_creation_func));
    grammar->name = String::forData ("AndExpression_BitwiseAnd");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AndExpression ();
        entry->assignment_func = cpp_AndExpression_BitwiseAnd_set_andExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression ();
        entry->assignment_func = cpp_AndExpression_BitwiseAnd_set_equalityExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_and_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AndExpression_BitwiseAnd (Cpp_AndExpression_BitwiseAnd const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AndExpression_BitwiseAnd\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AndExpression_BitwiseAnd\n");

    dump_cpp_AndExpression (el->andExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: &\n");
    dump_cpp_EqualityExpression (el->equalityExpression, nest_level + 1);

}

static void
cpp_AndExpression_Expression_set_equalityExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AndExpression ||
             __el->cpp_element_type != CppElement::_EqualityExpression);

    Cpp_AndExpression * const &_self = static_cast <Cpp_AndExpression*> (__self);

    abortIf (_self->and_expression_type != Cpp_AndExpression::_Expression);

    Cpp_AndExpression_Expression * const &self = static_cast <Cpp_AndExpression_Expression*> (_self);

    self->equalityExpression = static_cast <Cpp_EqualityExpression*> (___el);
}

static ParserElement*
cpp_AndExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AndExpression_Expression))) Cpp_AndExpression_Expression;
}

static Ref<Grammar>
create_cpp_AndExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AndExpression_Expression_creation_func));
    grammar->name = String::forData ("AndExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EqualityExpression ();
        entry->assignment_func = cpp_AndExpression_Expression_set_equalityExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AndExpression_Expression (Cpp_AndExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AndExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AndExpression_Expression\n");

    dump_cpp_EqualityExpression (el->equalityExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_AndExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("AndExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AndExpression_BitwiseAnd ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AndExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AndExpression (Cpp_AndExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AndExpression\n");
        return;
    }

    switch (el->and_expression_type) {
        case Cpp_AndExpression::_BitwiseAnd:
            dump_cpp_AndExpression_BitwiseAnd (static_cast <Cpp_AndExpression_BitwiseAnd const *> (el), nest_level + 0);
            break;
        case Cpp_AndExpression::_Expression:
            dump_cpp_AndExpression_Expression (static_cast <Cpp_AndExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ExclusiveOrExpression_BitwiseExclusiveOr_set_exclusiveOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExclusiveOrExpression ||
             __el->cpp_element_type != CppElement::_ExclusiveOrExpression);

    Cpp_ExclusiveOrExpression * const &_self = static_cast <Cpp_ExclusiveOrExpression*> (__self);

    abortIf (_self->exclusive_or_expression_type != Cpp_ExclusiveOrExpression::_BitwiseExclusiveOr);

    Cpp_ExclusiveOrExpression_BitwiseExclusiveOr * const &self = static_cast <Cpp_ExclusiveOrExpression_BitwiseExclusiveOr*> (_self);

    self->exclusiveOrExpression = static_cast <Cpp_ExclusiveOrExpression*> (___el);
}

static void
cpp_ExclusiveOrExpression_BitwiseExclusiveOr_set_andExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExclusiveOrExpression ||
             __el->cpp_element_type != CppElement::_AndExpression);

    Cpp_ExclusiveOrExpression * const &_self = static_cast <Cpp_ExclusiveOrExpression*> (__self);

    abortIf (_self->exclusive_or_expression_type != Cpp_ExclusiveOrExpression::_BitwiseExclusiveOr);

    Cpp_ExclusiveOrExpression_BitwiseExclusiveOr * const &self = static_cast <Cpp_ExclusiveOrExpression_BitwiseExclusiveOr*> (_self);

    self->andExpression = static_cast <Cpp_AndExpression*> (___el);
}

bool cpp_accept_exclusive_or_expression (
        Cpp_ExclusiveOrExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_exclusive_or_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ExclusiveOrExpression);

    Cpp_ExclusiveOrExpression * const &el =
            static_cast <Cpp_ExclusiveOrExpression*> (_el);

    return cpp_accept_exclusive_or_expression (el, parser_control, data);
}

static ParserElement*
cpp_ExclusiveOrExpression_BitwiseExclusiveOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExclusiveOrExpression_BitwiseExclusiveOr))) Cpp_ExclusiveOrExpression_BitwiseExclusiveOr;
}

static Ref<Grammar>
create_cpp_ExclusiveOrExpression_BitwiseExclusiveOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExclusiveOrExpression_BitwiseExclusiveOr_creation_func));
    grammar->name = String::forData ("ExclusiveOrExpression_BitwiseExclusiveOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExclusiveOrExpression ();
        entry->assignment_func = cpp_ExclusiveOrExpression_BitwiseExclusiveOr_set_exclusiveOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("^"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AndExpression ();
        entry->assignment_func = cpp_ExclusiveOrExpression_BitwiseExclusiveOr_set_andExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_exclusive_or_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExclusiveOrExpression_BitwiseExclusiveOr (Cpp_ExclusiveOrExpression_BitwiseExclusiveOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExclusiveOrExpression_BitwiseExclusiveOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExclusiveOrExpression_BitwiseExclusiveOr\n");

    dump_cpp_ExclusiveOrExpression (el->exclusiveOrExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ^\n");
    dump_cpp_AndExpression (el->andExpression, nest_level + 1);

}

static void
cpp_ExclusiveOrExpression_Expression_set_andExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExclusiveOrExpression ||
             __el->cpp_element_type != CppElement::_AndExpression);

    Cpp_ExclusiveOrExpression * const &_self = static_cast <Cpp_ExclusiveOrExpression*> (__self);

    abortIf (_self->exclusive_or_expression_type != Cpp_ExclusiveOrExpression::_Expression);

    Cpp_ExclusiveOrExpression_Expression * const &self = static_cast <Cpp_ExclusiveOrExpression_Expression*> (_self);

    self->andExpression = static_cast <Cpp_AndExpression*> (___el);
}

static ParserElement*
cpp_ExclusiveOrExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExclusiveOrExpression_Expression))) Cpp_ExclusiveOrExpression_Expression;
}

static Ref<Grammar>
create_cpp_ExclusiveOrExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExclusiveOrExpression_Expression_creation_func));
    grammar->name = String::forData ("ExclusiveOrExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AndExpression ();
        entry->assignment_func = cpp_ExclusiveOrExpression_Expression_set_andExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExclusiveOrExpression_Expression (Cpp_ExclusiveOrExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExclusiveOrExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExclusiveOrExpression_Expression\n");

    dump_cpp_AndExpression (el->andExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ExclusiveOrExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ExclusiveOrExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ExclusiveOrExpression_BitwiseExclusiveOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ExclusiveOrExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExclusiveOrExpression (Cpp_ExclusiveOrExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExclusiveOrExpression\n");
        return;
    }

    switch (el->exclusive_or_expression_type) {
        case Cpp_ExclusiveOrExpression::_BitwiseExclusiveOr:
            dump_cpp_ExclusiveOrExpression_BitwiseExclusiveOr (static_cast <Cpp_ExclusiveOrExpression_BitwiseExclusiveOr const *> (el), nest_level + 0);
            break;
        case Cpp_ExclusiveOrExpression::_Expression:
            dump_cpp_ExclusiveOrExpression_Expression (static_cast <Cpp_ExclusiveOrExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_InclusiveOrExpression_BitwiseInclusiveOr_set_inclusiveOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InclusiveOrExpression ||
             __el->cpp_element_type != CppElement::_InclusiveOrExpression);

    Cpp_InclusiveOrExpression * const &_self = static_cast <Cpp_InclusiveOrExpression*> (__self);

    abortIf (_self->inclusive_or_expression_type != Cpp_InclusiveOrExpression::_BitwiseInclusiveOr);

    Cpp_InclusiveOrExpression_BitwiseInclusiveOr * const &self = static_cast <Cpp_InclusiveOrExpression_BitwiseInclusiveOr*> (_self);

    self->inclusiveOrExpression = static_cast <Cpp_InclusiveOrExpression*> (___el);
}

static void
cpp_InclusiveOrExpression_BitwiseInclusiveOr_set_exclusiveOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InclusiveOrExpression ||
             __el->cpp_element_type != CppElement::_ExclusiveOrExpression);

    Cpp_InclusiveOrExpression * const &_self = static_cast <Cpp_InclusiveOrExpression*> (__self);

    abortIf (_self->inclusive_or_expression_type != Cpp_InclusiveOrExpression::_BitwiseInclusiveOr);

    Cpp_InclusiveOrExpression_BitwiseInclusiveOr * const &self = static_cast <Cpp_InclusiveOrExpression_BitwiseInclusiveOr*> (_self);

    self->exclusiveOrExpression = static_cast <Cpp_ExclusiveOrExpression*> (___el);
}

bool cpp_accept_inclusive_or_expression (
        Cpp_InclusiveOrExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_inclusive_or_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_InclusiveOrExpression);

    Cpp_InclusiveOrExpression * const &el =
            static_cast <Cpp_InclusiveOrExpression*> (_el);

    return cpp_accept_inclusive_or_expression (el, parser_control, data);
}

static ParserElement*
cpp_InclusiveOrExpression_BitwiseInclusiveOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InclusiveOrExpression_BitwiseInclusiveOr))) Cpp_InclusiveOrExpression_BitwiseInclusiveOr;
}

static Ref<Grammar>
create_cpp_InclusiveOrExpression_BitwiseInclusiveOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InclusiveOrExpression_BitwiseInclusiveOr_creation_func));
    grammar->name = String::forData ("InclusiveOrExpression_BitwiseInclusiveOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InclusiveOrExpression ();
        entry->assignment_func = cpp_InclusiveOrExpression_BitwiseInclusiveOr_set_inclusiveOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("|"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExclusiveOrExpression ();
        entry->assignment_func = cpp_InclusiveOrExpression_BitwiseInclusiveOr_set_exclusiveOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_inclusive_or_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InclusiveOrExpression_BitwiseInclusiveOr (Cpp_InclusiveOrExpression_BitwiseInclusiveOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InclusiveOrExpression_BitwiseInclusiveOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InclusiveOrExpression_BitwiseInclusiveOr\n");

    dump_cpp_InclusiveOrExpression (el->inclusiveOrExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: |\n");
    dump_cpp_ExclusiveOrExpression (el->exclusiveOrExpression, nest_level + 1);

}

static void
cpp_InclusiveOrExpression_Expression_set_exclusiveOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InclusiveOrExpression ||
             __el->cpp_element_type != CppElement::_ExclusiveOrExpression);

    Cpp_InclusiveOrExpression * const &_self = static_cast <Cpp_InclusiveOrExpression*> (__self);

    abortIf (_self->inclusive_or_expression_type != Cpp_InclusiveOrExpression::_Expression);

    Cpp_InclusiveOrExpression_Expression * const &self = static_cast <Cpp_InclusiveOrExpression_Expression*> (_self);

    self->exclusiveOrExpression = static_cast <Cpp_ExclusiveOrExpression*> (___el);
}

static ParserElement*
cpp_InclusiveOrExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InclusiveOrExpression_Expression))) Cpp_InclusiveOrExpression_Expression;
}

static Ref<Grammar>
create_cpp_InclusiveOrExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InclusiveOrExpression_Expression_creation_func));
    grammar->name = String::forData ("InclusiveOrExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExclusiveOrExpression ();
        entry->assignment_func = cpp_InclusiveOrExpression_Expression_set_exclusiveOrExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InclusiveOrExpression_Expression (Cpp_InclusiveOrExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InclusiveOrExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InclusiveOrExpression_Expression\n");

    dump_cpp_ExclusiveOrExpression (el->exclusiveOrExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_InclusiveOrExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("InclusiveOrExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InclusiveOrExpression_BitwiseInclusiveOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InclusiveOrExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InclusiveOrExpression (Cpp_InclusiveOrExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InclusiveOrExpression\n");
        return;
    }

    switch (el->inclusive_or_expression_type) {
        case Cpp_InclusiveOrExpression::_BitwiseInclusiveOr:
            dump_cpp_InclusiveOrExpression_BitwiseInclusiveOr (static_cast <Cpp_InclusiveOrExpression_BitwiseInclusiveOr const *> (el), nest_level + 0);
            break;
        case Cpp_InclusiveOrExpression::_Expression:
            dump_cpp_InclusiveOrExpression_Expression (static_cast <Cpp_InclusiveOrExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_LogicalAndExpression_LogicalAnd_set_logicalAndExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LogicalAndExpression ||
             __el->cpp_element_type != CppElement::_LogicalAndExpression);

    Cpp_LogicalAndExpression * const &_self = static_cast <Cpp_LogicalAndExpression*> (__self);

    abortIf (_self->logical_and_expression_type != Cpp_LogicalAndExpression::_LogicalAnd);

    Cpp_LogicalAndExpression_LogicalAnd * const &self = static_cast <Cpp_LogicalAndExpression_LogicalAnd*> (_self);

    self->logicalAndExpression = static_cast <Cpp_LogicalAndExpression*> (___el);
}

static void
cpp_LogicalAndExpression_LogicalAnd_set_inclusiveOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LogicalAndExpression ||
             __el->cpp_element_type != CppElement::_InclusiveOrExpression);

    Cpp_LogicalAndExpression * const &_self = static_cast <Cpp_LogicalAndExpression*> (__self);

    abortIf (_self->logical_and_expression_type != Cpp_LogicalAndExpression::_LogicalAnd);

    Cpp_LogicalAndExpression_LogicalAnd * const &self = static_cast <Cpp_LogicalAndExpression_LogicalAnd*> (_self);

    self->inclusiveOrExpression = static_cast <Cpp_InclusiveOrExpression*> (___el);
}

bool cpp_accept_logical_and_expression (
        Cpp_LogicalAndExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_logical_and_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_LogicalAndExpression);

    Cpp_LogicalAndExpression * const &el =
            static_cast <Cpp_LogicalAndExpression*> (_el);

    return cpp_accept_logical_and_expression (el, parser_control, data);
}

static ParserElement*
cpp_LogicalAndExpression_LogicalAnd_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LogicalAndExpression_LogicalAnd))) Cpp_LogicalAndExpression_LogicalAnd;
}

static Ref<Grammar>
create_cpp_LogicalAndExpression_LogicalAnd ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LogicalAndExpression_LogicalAnd_creation_func));
    grammar->name = String::forData ("LogicalAndExpression_LogicalAnd");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalAndExpression ();
        entry->assignment_func = cpp_LogicalAndExpression_LogicalAnd_set_logicalAndExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&&"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InclusiveOrExpression ();
        entry->assignment_func = cpp_LogicalAndExpression_LogicalAnd_set_inclusiveOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_logical_and_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LogicalAndExpression_LogicalAnd (Cpp_LogicalAndExpression_LogicalAnd const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LogicalAndExpression_LogicalAnd\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LogicalAndExpression_LogicalAnd\n");

    dump_cpp_LogicalAndExpression (el->logicalAndExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: &&\n");
    dump_cpp_InclusiveOrExpression (el->inclusiveOrExpression, nest_level + 1);

}

static void
cpp_LogicalAndExpression_Expression_set_inclusiveOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LogicalAndExpression ||
             __el->cpp_element_type != CppElement::_InclusiveOrExpression);

    Cpp_LogicalAndExpression * const &_self = static_cast <Cpp_LogicalAndExpression*> (__self);

    abortIf (_self->logical_and_expression_type != Cpp_LogicalAndExpression::_Expression);

    Cpp_LogicalAndExpression_Expression * const &self = static_cast <Cpp_LogicalAndExpression_Expression*> (_self);

    self->inclusiveOrExpression = static_cast <Cpp_InclusiveOrExpression*> (___el);
}

static ParserElement*
cpp_LogicalAndExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LogicalAndExpression_Expression))) Cpp_LogicalAndExpression_Expression;
}

static Ref<Grammar>
create_cpp_LogicalAndExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LogicalAndExpression_Expression_creation_func));
    grammar->name = String::forData ("LogicalAndExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InclusiveOrExpression ();
        entry->assignment_func = cpp_LogicalAndExpression_Expression_set_inclusiveOrExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LogicalAndExpression_Expression (Cpp_LogicalAndExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LogicalAndExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LogicalAndExpression_Expression\n");

    dump_cpp_InclusiveOrExpression (el->inclusiveOrExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_LogicalAndExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("LogicalAndExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LogicalAndExpression_LogicalAnd ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LogicalAndExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LogicalAndExpression (Cpp_LogicalAndExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LogicalAndExpression\n");
        return;
    }

    switch (el->logical_and_expression_type) {
        case Cpp_LogicalAndExpression::_LogicalAnd:
            dump_cpp_LogicalAndExpression_LogicalAnd (static_cast <Cpp_LogicalAndExpression_LogicalAnd const *> (el), nest_level + 0);
            break;
        case Cpp_LogicalAndExpression::_Expression:
            dump_cpp_LogicalAndExpression_Expression (static_cast <Cpp_LogicalAndExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_LogicalOrExpression_LogicalOr_set_logicalOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LogicalOrExpression ||
             __el->cpp_element_type != CppElement::_LogicalOrExpression);

    Cpp_LogicalOrExpression * const &_self = static_cast <Cpp_LogicalOrExpression*> (__self);

    abortIf (_self->logical_or_expression_type != Cpp_LogicalOrExpression::_LogicalOr);

    Cpp_LogicalOrExpression_LogicalOr * const &self = static_cast <Cpp_LogicalOrExpression_LogicalOr*> (_self);

    self->logicalOrExpression = static_cast <Cpp_LogicalOrExpression*> (___el);
}

static void
cpp_LogicalOrExpression_LogicalOr_set_logicalAndExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LogicalOrExpression ||
             __el->cpp_element_type != CppElement::_LogicalAndExpression);

    Cpp_LogicalOrExpression * const &_self = static_cast <Cpp_LogicalOrExpression*> (__self);

    abortIf (_self->logical_or_expression_type != Cpp_LogicalOrExpression::_LogicalOr);

    Cpp_LogicalOrExpression_LogicalOr * const &self = static_cast <Cpp_LogicalOrExpression_LogicalOr*> (_self);

    self->logicalAndExpression = static_cast <Cpp_LogicalAndExpression*> (___el);
}

bool cpp_accept_logical_or_expression (
        Cpp_LogicalOrExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_logical_or_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_LogicalOrExpression);

    Cpp_LogicalOrExpression * const &el =
            static_cast <Cpp_LogicalOrExpression*> (_el);

    return cpp_accept_logical_or_expression (el, parser_control, data);
}

static ParserElement*
cpp_LogicalOrExpression_LogicalOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LogicalOrExpression_LogicalOr))) Cpp_LogicalOrExpression_LogicalOr;
}

static Ref<Grammar>
create_cpp_LogicalOrExpression_LogicalOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LogicalOrExpression_LogicalOr_creation_func));
    grammar->name = String::forData ("LogicalOrExpression_LogicalOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalOrExpression ();
        entry->assignment_func = cpp_LogicalOrExpression_LogicalOr_set_logicalOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("||"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalAndExpression ();
        entry->assignment_func = cpp_LogicalOrExpression_LogicalOr_set_logicalAndExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_logical_or_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LogicalOrExpression_LogicalOr (Cpp_LogicalOrExpression_LogicalOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LogicalOrExpression_LogicalOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LogicalOrExpression_LogicalOr\n");

    dump_cpp_LogicalOrExpression (el->logicalOrExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ||\n");
    dump_cpp_LogicalAndExpression (el->logicalAndExpression, nest_level + 1);

}

static void
cpp_LogicalOrExpression_Expression_set_logicalAndExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LogicalOrExpression ||
             __el->cpp_element_type != CppElement::_LogicalAndExpression);

    Cpp_LogicalOrExpression * const &_self = static_cast <Cpp_LogicalOrExpression*> (__self);

    abortIf (_self->logical_or_expression_type != Cpp_LogicalOrExpression::_Expression);

    Cpp_LogicalOrExpression_Expression * const &self = static_cast <Cpp_LogicalOrExpression_Expression*> (_self);

    self->logicalAndExpression = static_cast <Cpp_LogicalAndExpression*> (___el);
}

static ParserElement*
cpp_LogicalOrExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LogicalOrExpression_Expression))) Cpp_LogicalOrExpression_Expression;
}

static Ref<Grammar>
create_cpp_LogicalOrExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LogicalOrExpression_Expression_creation_func));
    grammar->name = String::forData ("LogicalOrExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalAndExpression ();
        entry->assignment_func = cpp_LogicalOrExpression_Expression_set_logicalAndExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LogicalOrExpression_Expression (Cpp_LogicalOrExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LogicalOrExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LogicalOrExpression_Expression\n");

    dump_cpp_LogicalAndExpression (el->logicalAndExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_LogicalOrExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("LogicalOrExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LogicalOrExpression_LogicalOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LogicalOrExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LogicalOrExpression (Cpp_LogicalOrExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LogicalOrExpression\n");
        return;
    }

    switch (el->logical_or_expression_type) {
        case Cpp_LogicalOrExpression::_LogicalOr:
            dump_cpp_LogicalOrExpression_LogicalOr (static_cast <Cpp_LogicalOrExpression_LogicalOr const *> (el), nest_level + 0);
            break;
        case Cpp_LogicalOrExpression::_Expression:
            dump_cpp_LogicalOrExpression_Expression (static_cast <Cpp_LogicalOrExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ConditionalExpression_Conditional_set_logicalOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConditionalExpression ||
             __el->cpp_element_type != CppElement::_LogicalOrExpression);

    Cpp_ConditionalExpression * const &_self = static_cast <Cpp_ConditionalExpression*> (__self);

    abortIf (_self->conditional_expression_type != Cpp_ConditionalExpression::_Conditional);

    Cpp_ConditionalExpression_Conditional * const &self = static_cast <Cpp_ConditionalExpression_Conditional*> (_self);

    self->logicalOrExpression = static_cast <Cpp_LogicalOrExpression*> (___el);
}

static void
cpp_ConditionalExpression_Conditional_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConditionalExpression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_ConditionalExpression * const &_self = static_cast <Cpp_ConditionalExpression*> (__self);

    abortIf (_self->conditional_expression_type != Cpp_ConditionalExpression::_Conditional);

    Cpp_ConditionalExpression_Conditional * const &self = static_cast <Cpp_ConditionalExpression_Conditional*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static void
cpp_ConditionalExpression_Conditional_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConditionalExpression ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_ConditionalExpression * const &_self = static_cast <Cpp_ConditionalExpression*> (__self);

    abortIf (_self->conditional_expression_type != Cpp_ConditionalExpression::_Conditional);

    Cpp_ConditionalExpression_Conditional * const &self = static_cast <Cpp_ConditionalExpression_Conditional*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static ParserElement*
cpp_ConditionalExpression_Conditional_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConditionalExpression_Conditional))) Cpp_ConditionalExpression_Conditional;
}

static Ref<Grammar>
create_cpp_ConditionalExpression_Conditional ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConditionalExpression_Conditional_creation_func));
    grammar->name = String::forData ("ConditionalExpression_Conditional");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalOrExpression ();
        entry->assignment_func = cpp_ConditionalExpression_Conditional_set_logicalOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_ConditionalExpression ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("?"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_ConditionalExpression_Conditional_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_ConditionalExpression_Conditional_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConditionalExpression_Conditional (Cpp_ConditionalExpression_Conditional const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConditionalExpression_Conditional\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConditionalExpression_Conditional\n");

    dump_cpp_LogicalOrExpression (el->logicalOrExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ?\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_ConditionalExpression_Expression_set_logicalOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConditionalExpression ||
             __el->cpp_element_type != CppElement::_LogicalOrExpression);

    Cpp_ConditionalExpression * const &_self = static_cast <Cpp_ConditionalExpression*> (__self);

    abortIf (_self->conditional_expression_type != Cpp_ConditionalExpression::_Expression);

    Cpp_ConditionalExpression_Expression * const &self = static_cast <Cpp_ConditionalExpression_Expression*> (_self);

    self->logicalOrExpression = static_cast <Cpp_LogicalOrExpression*> (___el);
}

static ParserElement*
cpp_ConditionalExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConditionalExpression_Expression))) Cpp_ConditionalExpression_Expression;
}

static Ref<Grammar>
create_cpp_ConditionalExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConditionalExpression_Expression_creation_func));
    grammar->name = String::forData ("ConditionalExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalOrExpression ();
        entry->assignment_func = cpp_ConditionalExpression_Expression_set_logicalOrExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConditionalExpression_Expression (Cpp_ConditionalExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConditionalExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConditionalExpression_Expression\n");

    dump_cpp_LogicalOrExpression (el->logicalOrExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ConditionalExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ConditionalExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ConditionalExpression_Conditional ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ConditionalExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConditionalExpression (Cpp_ConditionalExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConditionalExpression\n");
        return;
    }

    switch (el->conditional_expression_type) {
        case Cpp_ConditionalExpression::_Conditional:
            dump_cpp_ConditionalExpression_Conditional (static_cast <Cpp_ConditionalExpression_Conditional const *> (el), nest_level + 0);
            break;
        case Cpp_ConditionalExpression::_Expression:
            dump_cpp_ConditionalExpression_Expression (static_cast <Cpp_ConditionalExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_AssignmentExpression_Assignment_set_logicalOrExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AssignmentExpression ||
             __el->cpp_element_type != CppElement::_LogicalOrExpression);

    Cpp_AssignmentExpression * const &_self = static_cast <Cpp_AssignmentExpression*> (__self);

    abortIf (_self->assignment_expression_type != Cpp_AssignmentExpression::_Assignment);

    Cpp_AssignmentExpression_Assignment * const &self = static_cast <Cpp_AssignmentExpression_Assignment*> (_self);

    self->logicalOrExpression = static_cast <Cpp_LogicalOrExpression*> (___el);
}

static void
cpp_AssignmentExpression_Assignment_set_assignmentOperator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AssignmentExpression ||
             __el->cpp_element_type != CppElement::_AssignmentOperator);

    Cpp_AssignmentExpression * const &_self = static_cast <Cpp_AssignmentExpression*> (__self);

    abortIf (_self->assignment_expression_type != Cpp_AssignmentExpression::_Assignment);

    Cpp_AssignmentExpression_Assignment * const &self = static_cast <Cpp_AssignmentExpression_Assignment*> (_self);

    self->assignmentOperator = static_cast <Cpp_AssignmentOperator*> (___el);
}

static void
cpp_AssignmentExpression_Assignment_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AssignmentExpression ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_AssignmentExpression * const &_self = static_cast <Cpp_AssignmentExpression*> (__self);

    abortIf (_self->assignment_expression_type != Cpp_AssignmentExpression::_Assignment);

    Cpp_AssignmentExpression_Assignment * const &self = static_cast <Cpp_AssignmentExpression_Assignment*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

bool cpp_accept_assignment_expression (
        Cpp_AssignmentExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_assignment_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_AssignmentExpression * const &el =
            static_cast <Cpp_AssignmentExpression*> (_el);

    return cpp_accept_assignment_expression (el, parser_control, data);
}

static ParserElement*
cpp_AssignmentExpression_Assignment_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentExpression_Assignment))) Cpp_AssignmentExpression_Assignment;
}

static Ref<Grammar>
create_cpp_AssignmentExpression_Assignment ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentExpression_Assignment_creation_func));
    grammar->name = String::forData ("AssignmentExpression_Assignment");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LogicalOrExpression ();
        entry->assignment_func = cpp_AssignmentExpression_Assignment_set_logicalOrExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator ();
        entry->assignment_func = cpp_AssignmentExpression_Assignment_set_assignmentOperator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_AssignmentExpression_Assignment_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_assignment_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentExpression_Assignment (Cpp_AssignmentExpression_Assignment const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentExpression_Assignment\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentExpression_Assignment\n");

    dump_cpp_LogicalOrExpression (el->logicalOrExpression, nest_level + 1);

    dump_cpp_AssignmentOperator (el->assignmentOperator, nest_level + 1);

    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_AssignmentExpression_Throw_set_throwExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AssignmentExpression ||
             __el->cpp_element_type != CppElement::_ThrowExpression);

    Cpp_AssignmentExpression * const &_self = static_cast <Cpp_AssignmentExpression*> (__self);

    abortIf (_self->assignment_expression_type != Cpp_AssignmentExpression::_Throw);

    Cpp_AssignmentExpression_Throw * const &self = static_cast <Cpp_AssignmentExpression_Throw*> (_self);

    self->throwExpression = static_cast <Cpp_ThrowExpression*> (___el);
}

static ParserElement*
cpp_AssignmentExpression_Throw_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentExpression_Throw))) Cpp_AssignmentExpression_Throw;
}

static Ref<Grammar>
create_cpp_AssignmentExpression_Throw ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentExpression_Throw_creation_func));
    grammar->name = String::forData ("AssignmentExpression_Throw");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ThrowExpression ();
        entry->assignment_func = cpp_AssignmentExpression_Throw_set_throwExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentExpression_Throw (Cpp_AssignmentExpression_Throw const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentExpression_Throw\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentExpression_Throw\n");

    dump_cpp_ThrowExpression (el->throwExpression, nest_level + 1);

}

static void
cpp_AssignmentExpression_Expression_set_conditionalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AssignmentExpression ||
             __el->cpp_element_type != CppElement::_ConditionalExpression);

    Cpp_AssignmentExpression * const &_self = static_cast <Cpp_AssignmentExpression*> (__self);

    abortIf (_self->assignment_expression_type != Cpp_AssignmentExpression::_Expression);

    Cpp_AssignmentExpression_Expression * const &self = static_cast <Cpp_AssignmentExpression_Expression*> (_self);

    self->conditionalExpression = static_cast <Cpp_ConditionalExpression*> (___el);
}

static ParserElement*
cpp_AssignmentExpression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentExpression_Expression))) Cpp_AssignmentExpression_Expression;
}

static Ref<Grammar>
create_cpp_AssignmentExpression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentExpression_Expression_creation_func));
    grammar->name = String::forData ("AssignmentExpression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConditionalExpression ();
        entry->assignment_func = cpp_AssignmentExpression_Expression_set_conditionalExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentExpression_Expression (Cpp_AssignmentExpression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentExpression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentExpression_Expression\n");

    dump_cpp_ConditionalExpression (el->conditionalExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_AssignmentExpression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("AssignmentExpression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression_Assignment ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression_Throw ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentExpression (Cpp_AssignmentExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentExpression\n");
        return;
    }

    switch (el->assignment_expression_type) {
        case Cpp_AssignmentExpression::_Assignment:
            dump_cpp_AssignmentExpression_Assignment (static_cast <Cpp_AssignmentExpression_Assignment const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentExpression::_Throw:
            dump_cpp_AssignmentExpression_Throw (static_cast <Cpp_AssignmentExpression_Throw const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentExpression::_Expression:
            dump_cpp_AssignmentExpression_Expression (static_cast <Cpp_AssignmentExpression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static ParserElement*
cpp_AssignmentOperator_Generic_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_Generic))) Cpp_AssignmentOperator_Generic;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_Generic ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_Generic_creation_func));
    grammar->name = String::forData ("AssignmentOperator_Generic");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_Generic (Cpp_AssignmentOperator_Generic const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_Generic\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_Generic\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
}

static ParserElement*
cpp_AssignmentOperator_Multiplication_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_Multiplication))) Cpp_AssignmentOperator_Multiplication;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_Multiplication ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_Multiplication_creation_func));
    grammar->name = String::forData ("AssignmentOperator_Multiplication");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_Multiplication (Cpp_AssignmentOperator_Multiplication const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_Multiplication\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_Multiplication\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: *=\n");
}

static ParserElement*
cpp_AssignmentOperator_Division_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_Division))) Cpp_AssignmentOperator_Division;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_Division ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_Division_creation_func));
    grammar->name = String::forData ("AssignmentOperator_Division");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("/="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_Division (Cpp_AssignmentOperator_Division const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_Division\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_Division\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: /=\n");
}

static ParserElement*
cpp_AssignmentOperator_Remainder_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_Remainder))) Cpp_AssignmentOperator_Remainder;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_Remainder ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_Remainder_creation_func));
    grammar->name = String::forData ("AssignmentOperator_Remainder");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("%="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_Remainder (Cpp_AssignmentOperator_Remainder const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_Remainder\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_Remainder\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: %=\n");
}

static ParserElement*
cpp_AssignmentOperator_Addition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_Addition))) Cpp_AssignmentOperator_Addition;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_Addition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_Addition_creation_func));
    grammar->name = String::forData ("AssignmentOperator_Addition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("+="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_Addition (Cpp_AssignmentOperator_Addition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_Addition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_Addition\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: +=\n");
}

static ParserElement*
cpp_AssignmentOperator_Subtraction_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_Subtraction))) Cpp_AssignmentOperator_Subtraction;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_Subtraction ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_Subtraction_creation_func));
    grammar->name = String::forData ("AssignmentOperator_Subtraction");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("-="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_Subtraction (Cpp_AssignmentOperator_Subtraction const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_Subtraction\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_Subtraction\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: -=\n");
}

static ParserElement*
cpp_AssignmentOperator_RightShift_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_RightShift))) Cpp_AssignmentOperator_RightShift;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_RightShift ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_RightShift_creation_func));
    grammar->name = String::forData ("AssignmentOperator_RightShift");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">>="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_RightShift (Cpp_AssignmentOperator_RightShift const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_RightShift\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_RightShift\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: >>=\n");
}

static ParserElement*
cpp_AssignmentOperator_LeftShift_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_LeftShift))) Cpp_AssignmentOperator_LeftShift;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_LeftShift ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_LeftShift_creation_func));
    grammar->name = String::forData ("AssignmentOperator_LeftShift");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<<="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_LeftShift (Cpp_AssignmentOperator_LeftShift const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_LeftShift\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_LeftShift\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: <<=\n");
}

static ParserElement*
cpp_AssignmentOperator_BitwiseAnd_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_BitwiseAnd))) Cpp_AssignmentOperator_BitwiseAnd;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_BitwiseAnd ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_BitwiseAnd_creation_func));
    grammar->name = String::forData ("AssignmentOperator_BitwiseAnd");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_BitwiseAnd (Cpp_AssignmentOperator_BitwiseAnd const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_BitwiseAnd\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_BitwiseAnd\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: &=\n");
}

static ParserElement*
cpp_AssignmentOperator_BitwiseExclusiveOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_BitwiseExclusiveOr))) Cpp_AssignmentOperator_BitwiseExclusiveOr;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_BitwiseExclusiveOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_BitwiseExclusiveOr_creation_func));
    grammar->name = String::forData ("AssignmentOperator_BitwiseExclusiveOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("^="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_BitwiseExclusiveOr (Cpp_AssignmentOperator_BitwiseExclusiveOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_BitwiseExclusiveOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_BitwiseExclusiveOr\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ^=\n");
}

static ParserElement*
cpp_AssignmentOperator_BitwiseInclusiveOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AssignmentOperator_BitwiseInclusiveOr))) Cpp_AssignmentOperator_BitwiseInclusiveOr;
}

static Ref<Grammar>
create_cpp_AssignmentOperator_BitwiseInclusiveOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AssignmentOperator_BitwiseInclusiveOr_creation_func));
    grammar->name = String::forData ("AssignmentOperator_BitwiseInclusiveOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("|="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator_BitwiseInclusiveOr (Cpp_AssignmentOperator_BitwiseInclusiveOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator_BitwiseInclusiveOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AssignmentOperator_BitwiseInclusiveOr\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: |=\n");
}

static Ref<Grammar>
create_cpp_AssignmentOperator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("AssignmentOperator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_Generic ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_Multiplication ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_Division ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_Remainder ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_Addition ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_Subtraction ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_RightShift ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_LeftShift ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_BitwiseAnd ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_BitwiseExclusiveOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AssignmentOperator_BitwiseInclusiveOr ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AssignmentOperator (Cpp_AssignmentOperator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AssignmentOperator\n");
        return;
    }

    switch (el->assignment_operator_type) {
        case Cpp_AssignmentOperator::_Generic:
            dump_cpp_AssignmentOperator_Generic (static_cast <Cpp_AssignmentOperator_Generic const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_Multiplication:
            dump_cpp_AssignmentOperator_Multiplication (static_cast <Cpp_AssignmentOperator_Multiplication const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_Division:
            dump_cpp_AssignmentOperator_Division (static_cast <Cpp_AssignmentOperator_Division const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_Remainder:
            dump_cpp_AssignmentOperator_Remainder (static_cast <Cpp_AssignmentOperator_Remainder const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_Addition:
            dump_cpp_AssignmentOperator_Addition (static_cast <Cpp_AssignmentOperator_Addition const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_Subtraction:
            dump_cpp_AssignmentOperator_Subtraction (static_cast <Cpp_AssignmentOperator_Subtraction const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_RightShift:
            dump_cpp_AssignmentOperator_RightShift (static_cast <Cpp_AssignmentOperator_RightShift const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_LeftShift:
            dump_cpp_AssignmentOperator_LeftShift (static_cast <Cpp_AssignmentOperator_LeftShift const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_BitwiseAnd:
            dump_cpp_AssignmentOperator_BitwiseAnd (static_cast <Cpp_AssignmentOperator_BitwiseAnd const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_BitwiseExclusiveOr:
            dump_cpp_AssignmentOperator_BitwiseExclusiveOr (static_cast <Cpp_AssignmentOperator_BitwiseExclusiveOr const *> (el), nest_level + 0);
            break;
        case Cpp_AssignmentOperator::_BitwiseInclusiveOr:
            dump_cpp_AssignmentOperator_BitwiseInclusiveOr (static_cast <Cpp_AssignmentOperator_BitwiseInclusiveOr const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Expression_Comma_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Expression ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_Expression * const &_self = static_cast <Cpp_Expression*> (__self);

    abortIf (_self->expression_type != Cpp_Expression::_Comma);

    Cpp_Expression_Comma * const &self = static_cast <Cpp_Expression_Comma*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static void
cpp_Expression_Comma_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Expression ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_Expression * const &_self = static_cast <Cpp_Expression*> (__self);

    abortIf (_self->expression_type != Cpp_Expression::_Comma);

    Cpp_Expression_Comma * const &self = static_cast <Cpp_Expression_Comma*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

bool cpp_accept_comma_expression (
        Cpp_Expression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_comma_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Expression);

    Cpp_Expression * const &el =
            static_cast <Cpp_Expression*> (_el);

    return cpp_accept_comma_expression (el, parser_control, data);
}

static ParserElement*
cpp_Expression_Comma_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Expression_Comma))) Cpp_Expression_Comma;
}

static Ref<Grammar>
create_cpp_Expression_Comma ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Expression_Comma_creation_func));
    grammar->name = String::forData ("Expression_Comma");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_Expression_Comma_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_Expression_Comma_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_comma_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Expression_Comma (Cpp_Expression_Comma const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Expression_Comma\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Expression_Comma\n");

    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_Expression_Expression_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Expression ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_Expression * const &_self = static_cast <Cpp_Expression*> (__self);

    abortIf (_self->expression_type != Cpp_Expression::_Expression);

    Cpp_Expression_Expression * const &self = static_cast <Cpp_Expression_Expression*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static ParserElement*
cpp_Expression_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Expression_Expression))) Cpp_Expression_Expression;
}

static Ref<Grammar>
create_cpp_Expression_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Expression_Expression_creation_func));
    grammar->name = String::forData ("Expression_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_Expression_Expression_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Expression_Expression (Cpp_Expression_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Expression_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Expression_Expression\n");

    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_Expression ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Expression");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Expression_Comma ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Expression_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Expression (Cpp_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Expression\n");
        return;
    }

    switch (el->expression_type) {
        case Cpp_Expression::_Comma:
            dump_cpp_Expression_Comma (static_cast <Cpp_Expression_Comma const *> (el), nest_level + 0);
            break;
        case Cpp_Expression::_Expression:
            dump_cpp_Expression_Expression (static_cast <Cpp_Expression_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ConstantExpression_set_conditionalExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConstantExpression ||
             __el->cpp_element_type != CppElement::_ConditionalExpression);

    Cpp_ConstantExpression * const &self = static_cast <Cpp_ConstantExpression*> (__self);

    self->conditionalExpression = static_cast <Cpp_ConditionalExpression*> (___el);
}

static ParserElement*
cpp_ConstantExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConstantExpression))) Cpp_ConstantExpression;
}

static Ref<Grammar>
create_cpp_ConstantExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConstantExpression_creation_func));
    grammar->name = String::forData ("ConstantExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConditionalExpression ();
        entry->assignment_func = cpp_ConstantExpression_set_conditionalExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConstantExpression (Cpp_ConstantExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConstantExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConstantExpression\n");

    dump_cpp_ConditionalExpression (el->conditionalExpression, nest_level + 1);

}

static void
cpp_Statement_LabeledStatement_set_labeledStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_LabeledStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_LabeledStatement);

    Cpp_Statement_LabeledStatement * const &self = static_cast <Cpp_Statement_LabeledStatement*> (_self);

    self->labeledStatement = static_cast <Cpp_LabeledStatement*> (___el);
}

static ParserElement*
cpp_Statement_LabeledStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_LabeledStatement))) Cpp_Statement_LabeledStatement;
}

static Ref<Grammar>
create_cpp_Statement_LabeledStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_LabeledStatement_creation_func));
    grammar->name = String::forData ("Statement_LabeledStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LabeledStatement ();
        entry->assignment_func = cpp_Statement_LabeledStatement_set_labeledStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_LabeledStatement (Cpp_Statement_LabeledStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_LabeledStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_LabeledStatement\n");

    dump_cpp_LabeledStatement (el->labeledStatement, nest_level + 1);

}

static void
cpp_Statement_DeclarationStatement_set_declarationStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_DeclarationStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_DeclarationStatement);

    Cpp_Statement_DeclarationStatement * const &self = static_cast <Cpp_Statement_DeclarationStatement*> (_self);

    self->declarationStatement = static_cast <Cpp_DeclarationStatement*> (___el);
}

static ParserElement*
cpp_Statement_DeclarationStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_DeclarationStatement))) Cpp_Statement_DeclarationStatement;
}

static Ref<Grammar>
create_cpp_Statement_DeclarationStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_DeclarationStatement_creation_func));
    grammar->name = String::forData ("Statement_DeclarationStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DeclarationStatement ();
        entry->assignment_func = cpp_Statement_DeclarationStatement_set_declarationStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_DeclarationStatement (Cpp_Statement_DeclarationStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_DeclarationStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_DeclarationStatement\n");

    dump_cpp_DeclarationStatement (el->declarationStatement, nest_level + 1);

}

static void
cpp_Statement_ExpressionStatement_set_expressionStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_ExpressionStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_ExpressionStatement);

    Cpp_Statement_ExpressionStatement * const &self = static_cast <Cpp_Statement_ExpressionStatement*> (_self);

    self->expressionStatement = static_cast <Cpp_ExpressionStatement*> (___el);
}

static ParserElement*
cpp_Statement_ExpressionStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_ExpressionStatement))) Cpp_Statement_ExpressionStatement;
}

static Ref<Grammar>
create_cpp_Statement_ExpressionStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_ExpressionStatement_creation_func));
    grammar->name = String::forData ("Statement_ExpressionStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExpressionStatement ();
        entry->assignment_func = cpp_Statement_ExpressionStatement_set_expressionStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_ExpressionStatement (Cpp_Statement_ExpressionStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_ExpressionStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_ExpressionStatement\n");

    dump_cpp_ExpressionStatement (el->expressionStatement, nest_level + 1);

}

static void
cpp_Statement_CompoundStatement_set_compoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_CompoundStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_CompoundStatement);

    Cpp_Statement_CompoundStatement * const &self = static_cast <Cpp_Statement_CompoundStatement*> (_self);

    self->compoundStatement = static_cast <Cpp_CompoundStatement*> (___el);
}

static ParserElement*
cpp_Statement_CompoundStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_CompoundStatement))) Cpp_Statement_CompoundStatement;
}

static Ref<Grammar>
create_cpp_Statement_CompoundStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_CompoundStatement_creation_func));
    grammar->name = String::forData ("Statement_CompoundStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CompoundStatement ();
        entry->assignment_func = cpp_Statement_CompoundStatement_set_compoundStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_CompoundStatement (Cpp_Statement_CompoundStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_CompoundStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_CompoundStatement\n");

    dump_cpp_CompoundStatement (el->compoundStatement, nest_level + 1);

}

static void
cpp_Statement_SelectionStatement_set_selectionStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_SelectionStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_SelectionStatement);

    Cpp_Statement_SelectionStatement * const &self = static_cast <Cpp_Statement_SelectionStatement*> (_self);

    self->selectionStatement = static_cast <Cpp_SelectionStatement*> (___el);
}

static ParserElement*
cpp_Statement_SelectionStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_SelectionStatement))) Cpp_Statement_SelectionStatement;
}

static Ref<Grammar>
create_cpp_Statement_SelectionStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_SelectionStatement_creation_func));
    grammar->name = String::forData ("Statement_SelectionStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_SelectionStatement ();
        entry->assignment_func = cpp_Statement_SelectionStatement_set_selectionStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_SelectionStatement (Cpp_Statement_SelectionStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_SelectionStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_SelectionStatement\n");

    dump_cpp_SelectionStatement (el->selectionStatement, nest_level + 1);

}

static void
cpp_Statement_IterationStatement_set_iterationStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_IterationStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_IterationStatement);

    Cpp_Statement_IterationStatement * const &self = static_cast <Cpp_Statement_IterationStatement*> (_self);

    self->iterationStatement = static_cast <Cpp_IterationStatement*> (___el);
}

static ParserElement*
cpp_Statement_IterationStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_IterationStatement))) Cpp_Statement_IterationStatement;
}

static Ref<Grammar>
create_cpp_Statement_IterationStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_IterationStatement_creation_func));
    grammar->name = String::forData ("Statement_IterationStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IterationStatement ();
        entry->assignment_func = cpp_Statement_IterationStatement_set_iterationStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_IterationStatement (Cpp_Statement_IterationStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_IterationStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_IterationStatement\n");

    dump_cpp_IterationStatement (el->iterationStatement, nest_level + 1);

}

static void
cpp_Statement_JumpStatement_set_jumpStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_JumpStatement);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_JumpStatement);

    Cpp_Statement_JumpStatement * const &self = static_cast <Cpp_Statement_JumpStatement*> (_self);

    self->jumpStatement = static_cast <Cpp_JumpStatement*> (___el);
}

static ParserElement*
cpp_Statement_JumpStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_JumpStatement))) Cpp_Statement_JumpStatement;
}

static Ref<Grammar>
create_cpp_Statement_JumpStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_JumpStatement_creation_func));
    grammar->name = String::forData ("Statement_JumpStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_JumpStatement ();
        entry->assignment_func = cpp_Statement_JumpStatement_set_jumpStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_JumpStatement (Cpp_Statement_JumpStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_JumpStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_JumpStatement\n");

    dump_cpp_JumpStatement (el->jumpStatement, nest_level + 1);

}

static void
cpp_Statement_TryBlock_set_tryBlock (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Statement ||
             __el->cpp_element_type != CppElement::_TryBlock);

    Cpp_Statement * const &_self = static_cast <Cpp_Statement*> (__self);

    abortIf (_self->statement_type != Cpp_Statement::_TryBlock);

    Cpp_Statement_TryBlock * const &self = static_cast <Cpp_Statement_TryBlock*> (_self);

    self->tryBlock = static_cast <Cpp_TryBlock*> (___el);
}

static ParserElement*
cpp_Statement_TryBlock_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Statement_TryBlock))) Cpp_Statement_TryBlock;
}

static Ref<Grammar>
create_cpp_Statement_TryBlock ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Statement_TryBlock_creation_func));
    grammar->name = String::forData ("Statement_TryBlock");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TryBlock ();
        entry->assignment_func = cpp_Statement_TryBlock_set_tryBlock;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement_TryBlock (Cpp_Statement_TryBlock const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement_TryBlock\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Statement_TryBlock\n");

    dump_cpp_TryBlock (el->tryBlock, nest_level + 1);

}

static Ref<Grammar>
create_cpp_Statement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Statement");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_LabeledStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_DeclarationStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_ExpressionStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_CompoundStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_SelectionStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_IterationStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_JumpStatement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Statement_TryBlock ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Statement (Cpp_Statement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Statement\n");
        return;
    }

    switch (el->statement_type) {
        case Cpp_Statement::_LabeledStatement:
            dump_cpp_Statement_LabeledStatement (static_cast <Cpp_Statement_LabeledStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_DeclarationStatement:
            dump_cpp_Statement_DeclarationStatement (static_cast <Cpp_Statement_DeclarationStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_ExpressionStatement:
            dump_cpp_Statement_ExpressionStatement (static_cast <Cpp_Statement_ExpressionStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_CompoundStatement:
            dump_cpp_Statement_CompoundStatement (static_cast <Cpp_Statement_CompoundStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_SelectionStatement:
            dump_cpp_Statement_SelectionStatement (static_cast <Cpp_Statement_SelectionStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_IterationStatement:
            dump_cpp_Statement_IterationStatement (static_cast <Cpp_Statement_IterationStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_JumpStatement:
            dump_cpp_Statement_JumpStatement (static_cast <Cpp_Statement_JumpStatement const *> (el), nest_level + 0);
            break;
        case Cpp_Statement::_TryBlock:
            dump_cpp_Statement_TryBlock (static_cast <Cpp_Statement_TryBlock const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_LabeledStatement_Label_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LabeledStatement ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_LabeledStatement * const &_self = static_cast <Cpp_LabeledStatement*> (__self);

    abortIf (_self->labeled_statement_type != Cpp_LabeledStatement::_Label);

    Cpp_LabeledStatement_Label * const &self = static_cast <Cpp_LabeledStatement_Label*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

bool cpp_LabeledStatement_accept_func (
        Cpp_LabeledStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_LabeledStatement_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_LabeledStatement);

    Cpp_LabeledStatement * const &el =
            static_cast <Cpp_LabeledStatement*> (_el);

    return cpp_LabeledStatement_accept_func (el, parser_control, data);
}

bool cpp_begin_statement (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_statement (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_statement (_el, parser_control, data);
}

static void
cpp_LabeledStatement_Label_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LabeledStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_LabeledStatement * const &_self = static_cast <Cpp_LabeledStatement*> (__self);

    abortIf (_self->labeled_statement_type != Cpp_LabeledStatement::_Label);

    Cpp_LabeledStatement_Label * const &self = static_cast <Cpp_LabeledStatement_Label*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static ParserElement*
cpp_LabeledStatement_Label_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LabeledStatement_Label))) Cpp_LabeledStatement_Label;
}

static Ref<Grammar>
create_cpp_LabeledStatement_Label ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LabeledStatement_Label_creation_func));
    grammar->name = String::forData ("LabeledStatement_Label");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_LabeledStatement_Label_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_LabeledStatement_accept_func;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_LabeledStatement_Label_set_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LabeledStatement_Label (Cpp_LabeledStatement_Label const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LabeledStatement_Label\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LabeledStatement_Label\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static void
cpp_LabeledStatement_Case_set_constantExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LabeledStatement ||
             __el->cpp_element_type != CppElement::_ConstantExpression);

    Cpp_LabeledStatement * const &_self = static_cast <Cpp_LabeledStatement*> (__self);

    abortIf (_self->labeled_statement_type != Cpp_LabeledStatement::_Case);

    Cpp_LabeledStatement_Case * const &self = static_cast <Cpp_LabeledStatement_Case*> (_self);

    self->constantExpression = static_cast <Cpp_ConstantExpression*> (___el);
}

static void
cpp_LabeledStatement_Case_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LabeledStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_LabeledStatement * const &_self = static_cast <Cpp_LabeledStatement*> (__self);

    abortIf (_self->labeled_statement_type != Cpp_LabeledStatement::_Case);

    Cpp_LabeledStatement_Case * const &self = static_cast <Cpp_LabeledStatement_Case*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static ParserElement*
cpp_LabeledStatement_Case_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LabeledStatement_Case))) Cpp_LabeledStatement_Case;
}

static Ref<Grammar>
create_cpp_LabeledStatement_Case ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LabeledStatement_Case_creation_func));
    grammar->name = String::forData ("LabeledStatement_Case");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("case"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConstantExpression ();
        entry->assignment_func = cpp_LabeledStatement_Case_set_constantExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_LabeledStatement_accept_func;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_LabeledStatement_Case_set_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LabeledStatement_Case (Cpp_LabeledStatement_Case const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LabeledStatement_Case\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LabeledStatement_Case\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: case\n");
    dump_cpp_ConstantExpression (el->constantExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static void
cpp_LabeledStatement_Default_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LabeledStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_LabeledStatement * const &_self = static_cast <Cpp_LabeledStatement*> (__self);

    abortIf (_self->labeled_statement_type != Cpp_LabeledStatement::_Default);

    Cpp_LabeledStatement_Default * const &self = static_cast <Cpp_LabeledStatement_Default*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static ParserElement*
cpp_LabeledStatement_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LabeledStatement_Default))) Cpp_LabeledStatement_Default;
}

static Ref<Grammar>
create_cpp_LabeledStatement_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LabeledStatement_Default_creation_func));
    grammar->name = String::forData ("LabeledStatement_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("default"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_LabeledStatement_accept_func;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_LabeledStatement_Default_set_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LabeledStatement_Default (Cpp_LabeledStatement_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LabeledStatement_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LabeledStatement_Default\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: default\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static Ref<Grammar>
create_cpp_LabeledStatement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("LabeledStatement");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LabeledStatement_Label ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LabeledStatement_Case ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LabeledStatement_Default ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LabeledStatement (Cpp_LabeledStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LabeledStatement\n");
        return;
    }

    switch (el->labeled_statement_type) {
        case Cpp_LabeledStatement::_Label:
            dump_cpp_LabeledStatement_Label (static_cast <Cpp_LabeledStatement_Label const *> (el), nest_level + 0);
            break;
        case Cpp_LabeledStatement::_Case:
            dump_cpp_LabeledStatement_Case (static_cast <Cpp_LabeledStatement_Case const *> (el), nest_level + 0);
            break;
        case Cpp_LabeledStatement::_Default:
            dump_cpp_LabeledStatement_Default (static_cast <Cpp_LabeledStatement_Default const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_ExpressionStatement_accept_func (
        Cpp_ExpressionStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_ExpressionStatement_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_ExpressionStatement_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ExpressionStatement);

    Cpp_ExpressionStatement * const &el = static_cast <Cpp_ExpressionStatement*> (_el);

    cpp_ExpressionStatement_accept_func (el, parser_control, data);
}

static void
cpp_ExpressionStatement_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExpressionStatement ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_ExpressionStatement * const &self = static_cast <Cpp_ExpressionStatement*> (__self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_ExpressionStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExpressionStatement))) Cpp_ExpressionStatement;
}

static Ref<Grammar>
create_cpp_ExpressionStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExpressionStatement_creation_func));
    grammar->name = String::forData ("ExpressionStatement");
    grammar->accept_func = __pargen_cpp_ExpressionStatement_accept_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_ExpressionStatement_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExpressionStatement (Cpp_ExpressionStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExpressionStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExpressionStatement\n");

    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

bool cpp_begin_function_definition (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_function_definition (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_function_definition (_el, parser_control, data);
}

bool cpp_begin_temporal_namespace (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_temporal_namespace (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_temporal_namespace (_el, parser_control, data);
}

static void
cpp_CompoundStatement_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CompoundStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_CompoundStatement * const &self = static_cast <Cpp_CompoundStatement*> (__self);

    self->statements.append (static_cast <Cpp_Statement*> (__el));
}

bool cpp_accept_compound_statement (
        Cpp_CompoundStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_compound_statement (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_CompoundStatement);

    Cpp_CompoundStatement * const &el =
            static_cast <Cpp_CompoundStatement*> (_el);

    return cpp_accept_compound_statement (el, parser_control, data);
}

static ParserElement*
cpp_CompoundStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CompoundStatement))) Cpp_CompoundStatement;
}

static Ref<Grammar>
create_cpp_CompoundStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CompoundStatement_creation_func));
    grammar->name = String::forData ("CompoundStatement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_function_definition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_CompoundStatement_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_compound_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CompoundStatement (Cpp_CompoundStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CompoundStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CompoundStatement\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    {
        List<Cpp_Statement*>::DataIterator sub_iter (el->statements);
        while (!sub_iter.done ()) {
            Cpp_Statement *&sub = sub_iter.next ();
            dump_cpp_Statement (sub, nest_level + 1);
        }
    }

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

bool cpp_begin_condition (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_condition (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_condition (_el, parser_control, data);
}

static void
cpp_SelectionStatement_IfElse_set_condition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Condition);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_IfElse);

    Cpp_SelectionStatement_IfElse * const &self = static_cast <Cpp_SelectionStatement_IfElse*> (_self);

    self->condition = static_cast <Cpp_Condition*> (___el);
}

static void
cpp_SelectionStatement_IfElse_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_IfElse);

    Cpp_SelectionStatement_IfElse * const &self = static_cast <Cpp_SelectionStatement_IfElse*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static void
cpp_SelectionStatement_IfElse_set_elseStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_IfElse);

    Cpp_SelectionStatement_IfElse * const &self = static_cast <Cpp_SelectionStatement_IfElse*> (_self);

    self->elseStatement = static_cast <Cpp_Statement*> (___el);
}

bool cpp_accept_selection_statement (
        Cpp_SelectionStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_selection_statement (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_SelectionStatement);

    Cpp_SelectionStatement * const &el =
            static_cast <Cpp_SelectionStatement*> (_el);

    return cpp_accept_selection_statement (el, parser_control, data);
}

static ParserElement*
cpp_SelectionStatement_IfElse_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SelectionStatement_IfElse))) Cpp_SelectionStatement_IfElse;
}

static Ref<Grammar>
create_cpp_SelectionStatement_IfElse ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SelectionStatement_IfElse_creation_func));
    grammar->name = String::forData ("SelectionStatement_IfElse");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("if"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Condition ();
        entry->assignment_func = cpp_SelectionStatement_IfElse_set_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_SelectionStatement_IfElse_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_SelectionStatement ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 8;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("else"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_SelectionStatement_IfElse_set_elseStatement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_selection_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SelectionStatement_IfElse (Cpp_SelectionStatement_IfElse const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SelectionStatement_IfElse\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SelectionStatement_IfElse\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: if\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Condition (el->condition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: else\n");
    dump_cpp_Statement (el->elseStatement, nest_level + 1);

}

static void
cpp_SelectionStatement_If_set_condition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Condition);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_If);

    Cpp_SelectionStatement_If * const &self = static_cast <Cpp_SelectionStatement_If*> (_self);

    self->condition = static_cast <Cpp_Condition*> (___el);
}

static void
cpp_SelectionStatement_If_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_If);

    Cpp_SelectionStatement_If * const &self = static_cast <Cpp_SelectionStatement_If*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static ParserElement*
cpp_SelectionStatement_If_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SelectionStatement_If))) Cpp_SelectionStatement_If;
}

static Ref<Grammar>
create_cpp_SelectionStatement_If ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SelectionStatement_If_creation_func));
    grammar->name = String::forData ("SelectionStatement_If");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("if"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Condition ();
        entry->assignment_func = cpp_SelectionStatement_If_set_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_SelectionStatement_If_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_selection_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SelectionStatement_If (Cpp_SelectionStatement_If const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SelectionStatement_If\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SelectionStatement_If\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: if\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Condition (el->condition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static void
cpp_SelectionStatement_Switch_set_condition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Condition);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_Switch);

    Cpp_SelectionStatement_Switch * const &self = static_cast <Cpp_SelectionStatement_Switch*> (_self);

    self->condition = static_cast <Cpp_Condition*> (___el);
}

static void
cpp_SelectionStatement_Switch_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SelectionStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_SelectionStatement * const &_self = static_cast <Cpp_SelectionStatement*> (__self);

    abortIf (_self->selection_statement_type != Cpp_SelectionStatement::_Switch);

    Cpp_SelectionStatement_Switch * const &self = static_cast <Cpp_SelectionStatement_Switch*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static ParserElement*
cpp_SelectionStatement_Switch_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SelectionStatement_Switch))) Cpp_SelectionStatement_Switch;
}

static Ref<Grammar>
create_cpp_SelectionStatement_Switch ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SelectionStatement_Switch_creation_func));
    grammar->name = String::forData ("SelectionStatement_Switch");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("switch"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Condition ();
        entry->assignment_func = cpp_SelectionStatement_Switch_set_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_SelectionStatement_Switch_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_selection_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SelectionStatement_Switch (Cpp_SelectionStatement_Switch const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SelectionStatement_Switch\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SelectionStatement_Switch\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: switch\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Condition (el->condition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static Ref<Grammar>
create_cpp_SelectionStatement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("SelectionStatement");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SelectionStatement_IfElse ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SelectionStatement_If ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SelectionStatement_Switch ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SelectionStatement (Cpp_SelectionStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SelectionStatement\n");
        return;
    }

    switch (el->selection_statement_type) {
        case Cpp_SelectionStatement::_IfElse:
            dump_cpp_SelectionStatement_IfElse (static_cast <Cpp_SelectionStatement_IfElse const *> (el), nest_level + 0);
            break;
        case Cpp_SelectionStatement::_If:
            dump_cpp_SelectionStatement_If (static_cast <Cpp_SelectionStatement_If const *> (el), nest_level + 0);
            break;
        case Cpp_SelectionStatement::_Switch:
            dump_cpp_SelectionStatement_Switch (static_cast <Cpp_SelectionStatement_Switch const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Condition_Expression_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Condition ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_Condition * const &_self = static_cast <Cpp_Condition*> (__self);

    abortIf (_self->condition_type != Cpp_Condition::_Expression);

    Cpp_Condition_Expression * const &self = static_cast <Cpp_Condition_Expression*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

bool cpp_accept_condition (
        Cpp_Condition *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_condition (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Condition);

    Cpp_Condition * const &el =
            static_cast <Cpp_Condition*> (_el);

    return cpp_accept_condition (el, parser_control, data);
}

static ParserElement*
cpp_Condition_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Condition_Expression))) Cpp_Condition_Expression;
}

static Ref<Grammar>
create_cpp_Condition_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Condition_Expression_creation_func));
    grammar->name = String::forData ("Condition_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_Condition_Expression_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_condition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Condition_Expression (Cpp_Condition_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Condition_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Condition_Expression\n");

    dump_cpp_Expression (el->expression, nest_level + 1);

}

static void
cpp_Condition_Declarator_set_typeSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Condition ||
             __el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_Condition * const &_self = static_cast <Cpp_Condition*> (__self);

    abortIf (_self->condition_type != Cpp_Condition::_Declarator);

    Cpp_Condition_Declarator * const &self = static_cast <Cpp_Condition_Declarator*> (_self);

    self->typeSpecifierSeq = static_cast <Cpp_TypeSpecifierSeq*> (___el);
}

bool cpp_begin_dummy_constructor_ambiguity (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_dummy_constructor_ambiguity (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_dummy_constructor_ambiguity (_el, parser_control, data);
}

static void
cpp_Condition_Declarator_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Condition ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_Condition * const &_self = static_cast <Cpp_Condition*> (__self);

    abortIf (_self->condition_type != Cpp_Condition::_Declarator);

    Cpp_Condition_Declarator * const &self = static_cast <Cpp_Condition_Declarator*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static void
cpp_Condition_Declarator_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Condition ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_Condition * const &_self = static_cast <Cpp_Condition*> (__self);

    abortIf (_self->condition_type != Cpp_Condition::_Declarator);

    Cpp_Condition_Declarator * const &self = static_cast <Cpp_Condition_Declarator*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static ParserElement*
cpp_Condition_Declarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Condition_Declarator))) Cpp_Condition_Declarator;
}

static Ref<Grammar>
create_cpp_Condition_Declarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Condition_Declarator_creation_func));
    grammar->name = String::forData ("Condition_Declarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_TypeSpecifierSeq ();
        entry->assignment_func = cpp_Condition_Declarator_set_typeSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_dummy_constructor_ambiguity;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_Declarator ();
        entry->assignment_func = cpp_Condition_Declarator_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_Condition_Declarator_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_condition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Condition_Declarator (Cpp_Condition_Declarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Condition_Declarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Condition_Declarator\n");

    dump_cpp_TypeSpecifierSeq (el->typeSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_Condition ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Condition");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Condition_Expression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Condition_Declarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Condition (Cpp_Condition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Condition\n");
        return;
    }

    switch (el->condition_type) {
        case Cpp_Condition::_Expression:
            dump_cpp_Condition_Expression (static_cast <Cpp_Condition_Expression const *> (el), nest_level + 0);
            break;
        case Cpp_Condition::_Declarator:
            dump_cpp_Condition_Declarator (static_cast <Cpp_Condition_Declarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_IterationStatement_While_set_condition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Condition);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_While);

    Cpp_IterationStatement_While * const &self = static_cast <Cpp_IterationStatement_While*> (_self);

    self->condition = static_cast <Cpp_Condition*> (___el);
}

static void
cpp_IterationStatement_While_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_While);

    Cpp_IterationStatement_While * const &self = static_cast <Cpp_IterationStatement_While*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

bool cpp_accept_iteration_statement (
        Cpp_IterationStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_iteration_statement (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_IterationStatement);

    Cpp_IterationStatement * const &el =
            static_cast <Cpp_IterationStatement*> (_el);

    return cpp_accept_iteration_statement (el, parser_control, data);
}

static ParserElement*
cpp_IterationStatement_While_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_IterationStatement_While))) Cpp_IterationStatement_While;
}

static Ref<Grammar>
create_cpp_IterationStatement_While ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_IterationStatement_While_creation_func));
    grammar->name = String::forData ("IterationStatement_While");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("while"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Condition ();
        entry->assignment_func = cpp_IterationStatement_While_set_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_IterationStatement_While_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_iteration_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IterationStatement_While (Cpp_IterationStatement_While const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IterationStatement_While\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("IterationStatement_While\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: while\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Condition (el->condition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static void
cpp_IterationStatement_DoWhile_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_DoWhile);

    Cpp_IterationStatement_DoWhile * const &self = static_cast <Cpp_IterationStatement_DoWhile*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static void
cpp_IterationStatement_DoWhile_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_DoWhile);

    Cpp_IterationStatement_DoWhile * const &self = static_cast <Cpp_IterationStatement_DoWhile*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_IterationStatement_DoWhile_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_IterationStatement_DoWhile))) Cpp_IterationStatement_DoWhile;
}

static Ref<Grammar>
create_cpp_IterationStatement_DoWhile ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_IterationStatement_DoWhile_creation_func));
    grammar->name = String::forData ("IterationStatement_DoWhile");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("do"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_IterationStatement_DoWhile_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("while"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_IterationStatement_DoWhile_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_iteration_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IterationStatement_DoWhile (Cpp_IterationStatement_DoWhile const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IterationStatement_DoWhile\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("IterationStatement_DoWhile\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: do\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: while\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

bool cpp_begin_for_init_statement (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_for_init_statement (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_for_init_statement (_el, parser_control, data);
}

static void
cpp_IterationStatement_For_set_forInitStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_ForInitStatement);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_For);

    Cpp_IterationStatement_For * const &self = static_cast <Cpp_IterationStatement_For*> (_self);

    self->forInitStatement = static_cast <Cpp_ForInitStatement*> (___el);
}

static void
cpp_IterationStatement_For_set_condition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Condition);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_For);

    Cpp_IterationStatement_For * const &self = static_cast <Cpp_IterationStatement_For*> (_self);

    self->condition = static_cast <Cpp_Condition*> (___el);
}

static void
cpp_IterationStatement_For_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_For);

    Cpp_IterationStatement_For * const &self = static_cast <Cpp_IterationStatement_For*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static void
cpp_IterationStatement_For_set_statement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_IterationStatement ||
             __el->cpp_element_type != CppElement::_Statement);

    Cpp_IterationStatement * const &_self = static_cast <Cpp_IterationStatement*> (__self);

    abortIf (_self->iteration_statement_type != Cpp_IterationStatement::_For);

    Cpp_IterationStatement_For * const &self = static_cast <Cpp_IterationStatement_For*> (_self);

    self->statement = static_cast <Cpp_Statement*> (___el);
}

static ParserElement*
cpp_IterationStatement_For_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_IterationStatement_For))) Cpp_IterationStatement_For;
}

static Ref<Grammar>
create_cpp_IterationStatement_For ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_IterationStatement_For_creation_func));
    grammar->name = String::forData ("IterationStatement_For");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("for"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_temporal_namespace;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_for_init_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ForInitStatement ();
        entry->assignment_func = cpp_IterationStatement_For_set_forInitStatement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Condition ();
        entry->assignment_func = cpp_IterationStatement_For_set_condition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_IterationStatement_For_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Statement ();
        entry->assignment_func = cpp_IterationStatement_For_set_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_iteration_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IterationStatement_For (Cpp_IterationStatement_For const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IterationStatement_For\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("IterationStatement_For\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: for\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ForInitStatement (el->forInitStatement, nest_level + 1);

    dump_cpp_Condition (el->condition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_Statement (el->statement, nest_level + 1);

}

static Ref<Grammar>
create_cpp_IterationStatement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("IterationStatement");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_IterationStatement_While ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_IterationStatement_DoWhile ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_IterationStatement_For ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_IterationStatement (Cpp_IterationStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) IterationStatement\n");
        return;
    }

    switch (el->iteration_statement_type) {
        case Cpp_IterationStatement::_While:
            dump_cpp_IterationStatement_While (static_cast <Cpp_IterationStatement_While const *> (el), nest_level + 0);
            break;
        case Cpp_IterationStatement::_DoWhile:
            dump_cpp_IterationStatement_DoWhile (static_cast <Cpp_IterationStatement_DoWhile const *> (el), nest_level + 0);
            break;
        case Cpp_IterationStatement::_For:
            dump_cpp_IterationStatement_For (static_cast <Cpp_IterationStatement_For const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ForInitStatement_Declaration_set_simpleDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ForInitStatement ||
             __el->cpp_element_type != CppElement::_SimpleDeclaration);

    Cpp_ForInitStatement * const &_self = static_cast <Cpp_ForInitStatement*> (__self);

    abortIf (_self->for_init_statement_type != Cpp_ForInitStatement::_Declaration);

    Cpp_ForInitStatement_Declaration * const &self = static_cast <Cpp_ForInitStatement_Declaration*> (_self);

    self->simpleDeclaration = static_cast <Cpp_SimpleDeclaration*> (___el);
}

bool cpp_accept_for_init_statement (
        Cpp_ForInitStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_for_init_statement (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ForInitStatement);

    Cpp_ForInitStatement * const &el =
            static_cast <Cpp_ForInitStatement*> (_el);

    return cpp_accept_for_init_statement (el, parser_control, data);
}

static ParserElement*
cpp_ForInitStatement_Declaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ForInitStatement_Declaration))) Cpp_ForInitStatement_Declaration;
}

static Ref<Grammar>
create_cpp_ForInitStatement_Declaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ForInitStatement_Declaration_creation_func));
    grammar->name = String::forData ("ForInitStatement_Declaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_SimpleDeclaration ();
        entry->assignment_func = cpp_ForInitStatement_Declaration_set_simpleDeclaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_for_init_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ForInitStatement_Declaration (Cpp_ForInitStatement_Declaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ForInitStatement_Declaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ForInitStatement_Declaration\n");

    dump_cpp_SimpleDeclaration (el->simpleDeclaration, nest_level + 1);

}

static void
cpp_ForInitStatement_Expression_set_expressionStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ForInitStatement ||
             __el->cpp_element_type != CppElement::_ExpressionStatement);

    Cpp_ForInitStatement * const &_self = static_cast <Cpp_ForInitStatement*> (__self);

    abortIf (_self->for_init_statement_type != Cpp_ForInitStatement::_Expression);

    Cpp_ForInitStatement_Expression * const &self = static_cast <Cpp_ForInitStatement_Expression*> (_self);

    self->expressionStatement = static_cast <Cpp_ExpressionStatement*> (___el);
}

static ParserElement*
cpp_ForInitStatement_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ForInitStatement_Expression))) Cpp_ForInitStatement_Expression;
}

static Ref<Grammar>
create_cpp_ForInitStatement_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ForInitStatement_Expression_creation_func));
    grammar->name = String::forData ("ForInitStatement_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExpressionStatement ();
        entry->assignment_func = cpp_ForInitStatement_Expression_set_expressionStatement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_for_init_statement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ForInitStatement_Expression (Cpp_ForInitStatement_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ForInitStatement_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ForInitStatement_Expression\n");

    dump_cpp_ExpressionStatement (el->expressionStatement, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ForInitStatement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ForInitStatement");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ForInitStatement_Declaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ForInitStatement_Expression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ForInitStatement (Cpp_ForInitStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ForInitStatement\n");
        return;
    }

    switch (el->for_init_statement_type) {
        case Cpp_ForInitStatement::_Declaration:
            dump_cpp_ForInitStatement_Declaration (static_cast <Cpp_ForInitStatement_Declaration const *> (el), nest_level + 0);
            break;
        case Cpp_ForInitStatement::_Expression:
            dump_cpp_ForInitStatement_Expression (static_cast <Cpp_ForInitStatement_Expression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_JumpStatement_accept_func (
        Cpp_JumpStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_JumpStatement_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_JumpStatement_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_JumpStatement);

    Cpp_JumpStatement * const &el = static_cast <Cpp_JumpStatement*> (_el);

    cpp_JumpStatement_accept_func (el, parser_control, data);
}

static ParserElement*
cpp_JumpStatement_Break_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_JumpStatement_Break))) Cpp_JumpStatement_Break;
}

static Ref<Grammar>
create_cpp_JumpStatement_Break ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_JumpStatement_Break_creation_func));
    grammar->name = String::forData ("JumpStatement_Break");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("break"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_JumpStatement_Break (Cpp_JumpStatement_Break const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) JumpStatement_Break\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("JumpStatement_Break\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: break\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static ParserElement*
cpp_JumpStatement_Continue_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_JumpStatement_Continue))) Cpp_JumpStatement_Continue;
}

static Ref<Grammar>
create_cpp_JumpStatement_Continue ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_JumpStatement_Continue_creation_func));
    grammar->name = String::forData ("JumpStatement_Continue");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("continue"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_JumpStatement_Continue (Cpp_JumpStatement_Continue const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) JumpStatement_Continue\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("JumpStatement_Continue\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: continue\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_JumpStatement_Return_set_expression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_JumpStatement ||
             __el->cpp_element_type != CppElement::_Expression);

    Cpp_JumpStatement * const &_self = static_cast <Cpp_JumpStatement*> (__self);

    abortIf (_self->jump_statement_type != Cpp_JumpStatement::_Return);

    Cpp_JumpStatement_Return * const &self = static_cast <Cpp_JumpStatement_Return*> (_self);

    self->expression = static_cast <Cpp_Expression*> (___el);
}

static ParserElement*
cpp_JumpStatement_Return_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_JumpStatement_Return))) Cpp_JumpStatement_Return;
}

static Ref<Grammar>
create_cpp_JumpStatement_Return ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_JumpStatement_Return_creation_func));
    grammar->name = String::forData ("JumpStatement_Return");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("return"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Expression ();
        entry->assignment_func = cpp_JumpStatement_Return_set_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_JumpStatement_Return (Cpp_JumpStatement_Return const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) JumpStatement_Return\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("JumpStatement_Return\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: return\n");
    dump_cpp_Expression (el->expression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_JumpStatement_Goto_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_JumpStatement ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_JumpStatement * const &_self = static_cast <Cpp_JumpStatement*> (__self);

    abortIf (_self->jump_statement_type != Cpp_JumpStatement::_Goto);

    Cpp_JumpStatement_Goto * const &self = static_cast <Cpp_JumpStatement_Goto*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_JumpStatement_Goto_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_JumpStatement_Goto))) Cpp_JumpStatement_Goto;
}

static Ref<Grammar>
create_cpp_JumpStatement_Goto ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_JumpStatement_Goto_creation_func));
    grammar->name = String::forData ("JumpStatement_Goto");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("goto"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_JumpStatement_Goto_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_JumpStatement_Goto (Cpp_JumpStatement_Goto const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) JumpStatement_Goto\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("JumpStatement_Goto\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: goto\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static Ref<Grammar>
create_cpp_JumpStatement ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("JumpStatement");
    grammar->accept_func = __pargen_cpp_JumpStatement_accept_func;

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_JumpStatement_Break ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_JumpStatement_Continue ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_JumpStatement_Return ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_JumpStatement_Goto ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_JumpStatement (Cpp_JumpStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) JumpStatement\n");
        return;
    }

    switch (el->jump_statement_type) {
        case Cpp_JumpStatement::_Break:
            dump_cpp_JumpStatement_Break (static_cast <Cpp_JumpStatement_Break const *> (el), nest_level + 0);
            break;
        case Cpp_JumpStatement::_Continue:
            dump_cpp_JumpStatement_Continue (static_cast <Cpp_JumpStatement_Continue const *> (el), nest_level + 0);
            break;
        case Cpp_JumpStatement::_Return:
            dump_cpp_JumpStatement_Return (static_cast <Cpp_JumpStatement_Return const *> (el), nest_level + 0);
            break;
        case Cpp_JumpStatement::_Goto:
            dump_cpp_JumpStatement_Goto (static_cast <Cpp_JumpStatement_Goto const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_DeclarationStatement_begin_func (
        void *data);

void cpp_DeclarationStatement_accept_func (
        Cpp_DeclarationStatement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_DeclarationStatement_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_DeclarationStatement_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclarationStatement);

    Cpp_DeclarationStatement * const &el = static_cast <Cpp_DeclarationStatement*> (_el);

    cpp_DeclarationStatement_accept_func (el, parser_control, data);
}

static void
cpp_DeclarationStatement_set_blockDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclarationStatement ||
             __el->cpp_element_type != CppElement::_BlockDeclaration);

    Cpp_DeclarationStatement * const &self = static_cast <Cpp_DeclarationStatement*> (__self);

    self->blockDeclaration = static_cast <Cpp_BlockDeclaration*> (___el);
}

static ParserElement*
cpp_DeclarationStatement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclarationStatement))) Cpp_DeclarationStatement;
}

static Ref<Grammar>
create_cpp_DeclarationStatement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclarationStatement_creation_func));
    grammar->name = String::forData ("DeclarationStatement");
    grammar->begin_func = cpp_DeclarationStatement_begin_func;
    grammar->accept_func = __pargen_cpp_DeclarationStatement_accept_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration ();
        entry->assignment_func = cpp_DeclarationStatement_set_blockDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclarationStatement (Cpp_DeclarationStatement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclarationStatement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclarationStatement\n");

    dump_cpp_BlockDeclaration (el->blockDeclaration, nest_level + 1);

}

static void
cpp_DeclarationSeq_DeclarationSeq_set_declarationSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclarationSeq ||
             __el->cpp_element_type != CppElement::_DeclarationSeq);

    Cpp_DeclarationSeq * const &_self = static_cast <Cpp_DeclarationSeq*> (__self);

    abortIf (_self->declaration_seq_type != Cpp_DeclarationSeq::_DeclarationSeq);

    Cpp_DeclarationSeq_DeclarationSeq * const &self = static_cast <Cpp_DeclarationSeq_DeclarationSeq*> (_self);

    self->declarationSeq = static_cast <Cpp_DeclarationSeq*> (___el);
}

static void
cpp_DeclarationSeq_DeclarationSeq_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclarationSeq ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_DeclarationSeq * const &_self = static_cast <Cpp_DeclarationSeq*> (__self);

    abortIf (_self->declaration_seq_type != Cpp_DeclarationSeq::_DeclarationSeq);

    Cpp_DeclarationSeq_DeclarationSeq * const &self = static_cast <Cpp_DeclarationSeq_DeclarationSeq*> (_self);

    self->declaration = static_cast <Cpp_Declaration*> (___el);
}

static ParserElement*
cpp_DeclarationSeq_DeclarationSeq_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclarationSeq_DeclarationSeq))) Cpp_DeclarationSeq_DeclarationSeq;
}

static Ref<Grammar>
create_cpp_DeclarationSeq_DeclarationSeq ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclarationSeq_DeclarationSeq_creation_func));
    grammar->name = String::forData ("DeclarationSeq_DeclarationSeq");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DeclarationSeq ();
        entry->assignment_func = cpp_DeclarationSeq_DeclarationSeq_set_declarationSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_DeclarationSeq_DeclarationSeq_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclarationSeq_DeclarationSeq (Cpp_DeclarationSeq_DeclarationSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclarationSeq_DeclarationSeq\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclarationSeq_DeclarationSeq\n");

    dump_cpp_DeclarationSeq (el->declarationSeq, nest_level + 1);

    dump_cpp_Declaration (el->declaration, nest_level + 1);

}

static void
cpp_DeclarationSeq_Declaration_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclarationSeq ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_DeclarationSeq * const &_self = static_cast <Cpp_DeclarationSeq*> (__self);

    abortIf (_self->declaration_seq_type != Cpp_DeclarationSeq::_Declaration);

    Cpp_DeclarationSeq_Declaration * const &self = static_cast <Cpp_DeclarationSeq_Declaration*> (_self);

    self->declaration = static_cast <Cpp_Declaration*> (___el);
}

static ParserElement*
cpp_DeclarationSeq_Declaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclarationSeq_Declaration))) Cpp_DeclarationSeq_Declaration;
}

static Ref<Grammar>
create_cpp_DeclarationSeq_Declaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclarationSeq_Declaration_creation_func));
    grammar->name = String::forData ("DeclarationSeq_Declaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_DeclarationSeq_Declaration_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclarationSeq_Declaration (Cpp_DeclarationSeq_Declaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclarationSeq_Declaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclarationSeq_Declaration\n");

    dump_cpp_Declaration (el->declaration, nest_level + 1);

}

static Ref<Grammar>
create_cpp_DeclarationSeq ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("DeclarationSeq");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclarationSeq_DeclarationSeq ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclarationSeq_Declaration ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclarationSeq (Cpp_DeclarationSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclarationSeq\n");
        return;
    }

    switch (el->declaration_seq_type) {
        case Cpp_DeclarationSeq::_DeclarationSeq:
            dump_cpp_DeclarationSeq_DeclarationSeq (static_cast <Cpp_DeclarationSeq_DeclarationSeq const *> (el), nest_level + 0);
            break;
        case Cpp_DeclarationSeq::_Declaration:
            dump_cpp_DeclarationSeq_Declaration (static_cast <Cpp_DeclarationSeq_Declaration const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Declaration_BlockDeclaration_set_blockDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_BlockDeclaration);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_BlockDeclaration);

    Cpp_Declaration_BlockDeclaration * const &self = static_cast <Cpp_Declaration_BlockDeclaration*> (_self);

    self->blockDeclaration = static_cast <Cpp_BlockDeclaration*> (___el);
}

static ParserElement*
cpp_Declaration_BlockDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_BlockDeclaration))) Cpp_Declaration_BlockDeclaration;
}

static Ref<Grammar>
create_cpp_Declaration_BlockDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_BlockDeclaration_creation_func));
    grammar->name = String::forData ("Declaration_BlockDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration ();
        entry->assignment_func = cpp_Declaration_BlockDeclaration_set_blockDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_BlockDeclaration (Cpp_Declaration_BlockDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_BlockDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_BlockDeclaration\n");

    dump_cpp_BlockDeclaration (el->blockDeclaration, nest_level + 1);

}

static void
cpp_Declaration_FunctionDefinition_set_functionDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_FunctionDefinition);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_FunctionDefinition);

    Cpp_Declaration_FunctionDefinition * const &self = static_cast <Cpp_Declaration_FunctionDefinition*> (_self);

    self->functionDefinition = static_cast <Cpp_FunctionDefinition*> (___el);
}

static ParserElement*
cpp_Declaration_FunctionDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_FunctionDefinition))) Cpp_Declaration_FunctionDefinition;
}

static Ref<Grammar>
create_cpp_Declaration_FunctionDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_FunctionDefinition_creation_func));
    grammar->name = String::forData ("Declaration_FunctionDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_FunctionDefinition ();
        entry->assignment_func = cpp_Declaration_FunctionDefinition_set_functionDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_FunctionDefinition (Cpp_Declaration_FunctionDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_FunctionDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_FunctionDefinition\n");

    dump_cpp_FunctionDefinition (el->functionDefinition, nest_level + 1);

}

static void
cpp_Declaration_TemplateDeclaration_set_templateDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_TemplateDeclaration);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_TemplateDeclaration);

    Cpp_Declaration_TemplateDeclaration * const &self = static_cast <Cpp_Declaration_TemplateDeclaration*> (_self);

    self->templateDeclaration = static_cast <Cpp_TemplateDeclaration*> (___el);
}

static ParserElement*
cpp_Declaration_TemplateDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_TemplateDeclaration))) Cpp_Declaration_TemplateDeclaration;
}

static Ref<Grammar>
create_cpp_Declaration_TemplateDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_TemplateDeclaration_creation_func));
    grammar->name = String::forData ("Declaration_TemplateDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateDeclaration ();
        entry->assignment_func = cpp_Declaration_TemplateDeclaration_set_templateDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_TemplateDeclaration (Cpp_Declaration_TemplateDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_TemplateDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_TemplateDeclaration\n");

    dump_cpp_TemplateDeclaration (el->templateDeclaration, nest_level + 1);

}

static void
cpp_Declaration_ExplicitInstantiation_set_explicitInstantiation (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_ExplicitInstantiation);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_ExplicitInstantiation);

    Cpp_Declaration_ExplicitInstantiation * const &self = static_cast <Cpp_Declaration_ExplicitInstantiation*> (_self);

    self->explicitInstantiation = static_cast <Cpp_ExplicitInstantiation*> (___el);
}

static ParserElement*
cpp_Declaration_ExplicitInstantiation_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_ExplicitInstantiation))) Cpp_Declaration_ExplicitInstantiation;
}

static Ref<Grammar>
create_cpp_Declaration_ExplicitInstantiation ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_ExplicitInstantiation_creation_func));
    grammar->name = String::forData ("Declaration_ExplicitInstantiation");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExplicitInstantiation ();
        entry->assignment_func = cpp_Declaration_ExplicitInstantiation_set_explicitInstantiation;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_ExplicitInstantiation (Cpp_Declaration_ExplicitInstantiation const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_ExplicitInstantiation\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_ExplicitInstantiation\n");

    dump_cpp_ExplicitInstantiation (el->explicitInstantiation, nest_level + 1);

}

static void
cpp_Declaration_ExplicitSpecialization_set_explicitSpecialization (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_ExplicitSpecialization);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_ExplicitSpecialization);

    Cpp_Declaration_ExplicitSpecialization * const &self = static_cast <Cpp_Declaration_ExplicitSpecialization*> (_self);

    self->explicitSpecialization = static_cast <Cpp_ExplicitSpecialization*> (___el);
}

static ParserElement*
cpp_Declaration_ExplicitSpecialization_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_ExplicitSpecialization))) Cpp_Declaration_ExplicitSpecialization;
}

static Ref<Grammar>
create_cpp_Declaration_ExplicitSpecialization ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_ExplicitSpecialization_creation_func));
    grammar->name = String::forData ("Declaration_ExplicitSpecialization");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExplicitSpecialization ();
        entry->assignment_func = cpp_Declaration_ExplicitSpecialization_set_explicitSpecialization;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_ExplicitSpecialization (Cpp_Declaration_ExplicitSpecialization const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_ExplicitSpecialization\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_ExplicitSpecialization\n");

    dump_cpp_ExplicitSpecialization (el->explicitSpecialization, nest_level + 1);

}

static void
cpp_Declaration_LinkageSpecification_set_linkageSpecification (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_LinkageSpecification);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_LinkageSpecification);

    Cpp_Declaration_LinkageSpecification * const &self = static_cast <Cpp_Declaration_LinkageSpecification*> (_self);

    self->linkageSpecification = static_cast <Cpp_LinkageSpecification*> (___el);
}

static ParserElement*
cpp_Declaration_LinkageSpecification_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_LinkageSpecification))) Cpp_Declaration_LinkageSpecification;
}

static Ref<Grammar>
create_cpp_Declaration_LinkageSpecification ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_LinkageSpecification_creation_func));
    grammar->name = String::forData ("Declaration_LinkageSpecification");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_LinkageSpecification ();
        entry->assignment_func = cpp_Declaration_LinkageSpecification_set_linkageSpecification;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_LinkageSpecification (Cpp_Declaration_LinkageSpecification const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_LinkageSpecification\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_LinkageSpecification\n");

    dump_cpp_LinkageSpecification (el->linkageSpecification, nest_level + 1);

}

static void
cpp_Declaration_NamespaceDefinition_set_namespaceDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declaration ||
             __el->cpp_element_type != CppElement::_NamespaceDefinition);

    Cpp_Declaration * const &_self = static_cast <Cpp_Declaration*> (__self);

    abortIf (_self->declaration_type != Cpp_Declaration::_NamespaceDefinition);

    Cpp_Declaration_NamespaceDefinition * const &self = static_cast <Cpp_Declaration_NamespaceDefinition*> (_self);

    self->namespaceDefinition = static_cast <Cpp_NamespaceDefinition*> (___el);
}

static ParserElement*
cpp_Declaration_NamespaceDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declaration_NamespaceDefinition))) Cpp_Declaration_NamespaceDefinition;
}

static Ref<Grammar>
create_cpp_Declaration_NamespaceDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declaration_NamespaceDefinition_creation_func));
    grammar->name = String::forData ("Declaration_NamespaceDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceDefinition ();
        entry->assignment_func = cpp_Declaration_NamespaceDefinition_set_namespaceDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration_NamespaceDefinition (Cpp_Declaration_NamespaceDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration_NamespaceDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declaration_NamespaceDefinition\n");

    dump_cpp_NamespaceDefinition (el->namespaceDefinition, nest_level + 1);

}

static Ref<Grammar>
create_cpp_Declaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Declaration");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_BlockDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_FunctionDefinition ();
    entry->variants.append (String::forData ("AST"));

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_TemplateDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_ExplicitInstantiation ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_ExplicitSpecialization ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_LinkageSpecification ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declaration_NamespaceDefinition ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declaration (Cpp_Declaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declaration\n");
        return;
    }

    switch (el->declaration_type) {
        case Cpp_Declaration::_BlockDeclaration:
            dump_cpp_Declaration_BlockDeclaration (static_cast <Cpp_Declaration_BlockDeclaration const *> (el), nest_level + 0);
            break;
        case Cpp_Declaration::_FunctionDefinition:
            dump_cpp_Declaration_FunctionDefinition (static_cast <Cpp_Declaration_FunctionDefinition const *> (el), nest_level + 0);
            break;
        case Cpp_Declaration::_TemplateDeclaration:
            dump_cpp_Declaration_TemplateDeclaration (static_cast <Cpp_Declaration_TemplateDeclaration const *> (el), nest_level + 0);
            break;
        case Cpp_Declaration::_ExplicitInstantiation:
            dump_cpp_Declaration_ExplicitInstantiation (static_cast <Cpp_Declaration_ExplicitInstantiation const *> (el), nest_level + 0);
            break;
        case Cpp_Declaration::_ExplicitSpecialization:
            dump_cpp_Declaration_ExplicitSpecialization (static_cast <Cpp_Declaration_ExplicitSpecialization const *> (el), nest_level + 0);
            break;
        case Cpp_Declaration::_LinkageSpecification:
            dump_cpp_Declaration_LinkageSpecification (static_cast <Cpp_Declaration_LinkageSpecification const *> (el), nest_level + 0);
            break;
        case Cpp_Declaration::_NamespaceDefinition:
            dump_cpp_Declaration_NamespaceDefinition (static_cast <Cpp_Declaration_NamespaceDefinition const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_BlockDeclaration_SimpleDeclaration_set_simpleDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BlockDeclaration ||
             __el->cpp_element_type != CppElement::_SimpleDeclaration);

    Cpp_BlockDeclaration * const &_self = static_cast <Cpp_BlockDeclaration*> (__self);

    abortIf (_self->block_declaration_type != Cpp_BlockDeclaration::_SimpleDeclaration);

    Cpp_BlockDeclaration_SimpleDeclaration * const &self = static_cast <Cpp_BlockDeclaration_SimpleDeclaration*> (_self);

    self->simpleDeclaration = static_cast <Cpp_SimpleDeclaration*> (___el);
}

static ParserElement*
cpp_BlockDeclaration_SimpleDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BlockDeclaration_SimpleDeclaration))) Cpp_BlockDeclaration_SimpleDeclaration;
}

static Ref<Grammar>
create_cpp_BlockDeclaration_SimpleDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BlockDeclaration_SimpleDeclaration_creation_func));
    grammar->name = String::forData ("BlockDeclaration_SimpleDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_SimpleDeclaration ();
        entry->assignment_func = cpp_BlockDeclaration_SimpleDeclaration_set_simpleDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BlockDeclaration_SimpleDeclaration (Cpp_BlockDeclaration_SimpleDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BlockDeclaration_SimpleDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BlockDeclaration_SimpleDeclaration\n");

    dump_cpp_SimpleDeclaration (el->simpleDeclaration, nest_level + 1);

}

static void
cpp_BlockDeclaration_AsmDefinition_set_asmDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BlockDeclaration ||
             __el->cpp_element_type != CppElement::_AsmDefinition);

    Cpp_BlockDeclaration * const &_self = static_cast <Cpp_BlockDeclaration*> (__self);

    abortIf (_self->block_declaration_type != Cpp_BlockDeclaration::_AsmDefinition);

    Cpp_BlockDeclaration_AsmDefinition * const &self = static_cast <Cpp_BlockDeclaration_AsmDefinition*> (_self);

    self->asmDefinition = static_cast <Cpp_AsmDefinition*> (___el);
}

static ParserElement*
cpp_BlockDeclaration_AsmDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BlockDeclaration_AsmDefinition))) Cpp_BlockDeclaration_AsmDefinition;
}

static Ref<Grammar>
create_cpp_BlockDeclaration_AsmDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BlockDeclaration_AsmDefinition_creation_func));
    grammar->name = String::forData ("BlockDeclaration_AsmDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AsmDefinition ();
        entry->assignment_func = cpp_BlockDeclaration_AsmDefinition_set_asmDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BlockDeclaration_AsmDefinition (Cpp_BlockDeclaration_AsmDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BlockDeclaration_AsmDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BlockDeclaration_AsmDefinition\n");

    dump_cpp_AsmDefinition (el->asmDefinition, nest_level + 1);

}

static void
cpp_BlockDeclaration_NamespaceAliasDefinition_set_namespaceAliasDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BlockDeclaration ||
             __el->cpp_element_type != CppElement::_NamespaceAliasDefinition);

    Cpp_BlockDeclaration * const &_self = static_cast <Cpp_BlockDeclaration*> (__self);

    abortIf (_self->block_declaration_type != Cpp_BlockDeclaration::_NamespaceAliasDefinition);

    Cpp_BlockDeclaration_NamespaceAliasDefinition * const &self = static_cast <Cpp_BlockDeclaration_NamespaceAliasDefinition*> (_self);

    self->namespaceAliasDefinition = static_cast <Cpp_NamespaceAliasDefinition*> (___el);
}

static ParserElement*
cpp_BlockDeclaration_NamespaceAliasDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BlockDeclaration_NamespaceAliasDefinition))) Cpp_BlockDeclaration_NamespaceAliasDefinition;
}

static Ref<Grammar>
create_cpp_BlockDeclaration_NamespaceAliasDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BlockDeclaration_NamespaceAliasDefinition_creation_func));
    grammar->name = String::forData ("BlockDeclaration_NamespaceAliasDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceAliasDefinition ();
        entry->assignment_func = cpp_BlockDeclaration_NamespaceAliasDefinition_set_namespaceAliasDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BlockDeclaration_NamespaceAliasDefinition (Cpp_BlockDeclaration_NamespaceAliasDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BlockDeclaration_NamespaceAliasDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BlockDeclaration_NamespaceAliasDefinition\n");

    dump_cpp_NamespaceAliasDefinition (el->namespaceAliasDefinition, nest_level + 1);

}

static void
cpp_BlockDeclaration_UsingDeclaration_set_usingDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BlockDeclaration ||
             __el->cpp_element_type != CppElement::_UsingDeclaration);

    Cpp_BlockDeclaration * const &_self = static_cast <Cpp_BlockDeclaration*> (__self);

    abortIf (_self->block_declaration_type != Cpp_BlockDeclaration::_UsingDeclaration);

    Cpp_BlockDeclaration_UsingDeclaration * const &self = static_cast <Cpp_BlockDeclaration_UsingDeclaration*> (_self);

    self->usingDeclaration = static_cast <Cpp_UsingDeclaration*> (___el);
}

static ParserElement*
cpp_BlockDeclaration_UsingDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BlockDeclaration_UsingDeclaration))) Cpp_BlockDeclaration_UsingDeclaration;
}

static Ref<Grammar>
create_cpp_BlockDeclaration_UsingDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BlockDeclaration_UsingDeclaration_creation_func));
    grammar->name = String::forData ("BlockDeclaration_UsingDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UsingDeclaration ();
        entry->assignment_func = cpp_BlockDeclaration_UsingDeclaration_set_usingDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BlockDeclaration_UsingDeclaration (Cpp_BlockDeclaration_UsingDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BlockDeclaration_UsingDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BlockDeclaration_UsingDeclaration\n");

    dump_cpp_UsingDeclaration (el->usingDeclaration, nest_level + 1);

}

static void
cpp_BlockDeclaration_UsingDirective_set_usingDirective (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BlockDeclaration ||
             __el->cpp_element_type != CppElement::_UsingDirective);

    Cpp_BlockDeclaration * const &_self = static_cast <Cpp_BlockDeclaration*> (__self);

    abortIf (_self->block_declaration_type != Cpp_BlockDeclaration::_UsingDirective);

    Cpp_BlockDeclaration_UsingDirective * const &self = static_cast <Cpp_BlockDeclaration_UsingDirective*> (_self);

    self->usingDirective = static_cast <Cpp_UsingDirective*> (___el);
}

static ParserElement*
cpp_BlockDeclaration_UsingDirective_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BlockDeclaration_UsingDirective))) Cpp_BlockDeclaration_UsingDirective;
}

static Ref<Grammar>
create_cpp_BlockDeclaration_UsingDirective ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BlockDeclaration_UsingDirective_creation_func));
    grammar->name = String::forData ("BlockDeclaration_UsingDirective");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UsingDirective ();
        entry->assignment_func = cpp_BlockDeclaration_UsingDirective_set_usingDirective;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BlockDeclaration_UsingDirective (Cpp_BlockDeclaration_UsingDirective const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BlockDeclaration_UsingDirective\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BlockDeclaration_UsingDirective\n");

    dump_cpp_UsingDirective (el->usingDirective, nest_level + 1);

}

static Ref<Grammar>
create_cpp_BlockDeclaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("BlockDeclaration");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration_SimpleDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration_AsmDefinition ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration_NamespaceAliasDefinition ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration_UsingDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BlockDeclaration_UsingDirective ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BlockDeclaration (Cpp_BlockDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BlockDeclaration\n");
        return;
    }

    switch (el->block_declaration_type) {
        case Cpp_BlockDeclaration::_SimpleDeclaration:
            dump_cpp_BlockDeclaration_SimpleDeclaration (static_cast <Cpp_BlockDeclaration_SimpleDeclaration const *> (el), nest_level + 0);
            break;
        case Cpp_BlockDeclaration::_AsmDefinition:
            dump_cpp_BlockDeclaration_AsmDefinition (static_cast <Cpp_BlockDeclaration_AsmDefinition const *> (el), nest_level + 0);
            break;
        case Cpp_BlockDeclaration::_NamespaceAliasDefinition:
            dump_cpp_BlockDeclaration_NamespaceAliasDefinition (static_cast <Cpp_BlockDeclaration_NamespaceAliasDefinition const *> (el), nest_level + 0);
            break;
        case Cpp_BlockDeclaration::_UsingDeclaration:
            dump_cpp_BlockDeclaration_UsingDeclaration (static_cast <Cpp_BlockDeclaration_UsingDeclaration const *> (el), nest_level + 0);
            break;
        case Cpp_BlockDeclaration::_UsingDirective:
            dump_cpp_BlockDeclaration_UsingDirective (static_cast <Cpp_BlockDeclaration_UsingDirective const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_begin_constructor_ambiguity (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_constructor_ambiguity (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_constructor_ambiguity (_el, parser_control, data);
}

static void
cpp_SimpleDeclaration_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleDeclaration ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_SimpleDeclaration * const &self = static_cast <Cpp_SimpleDeclaration*> (__self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static void
cpp_SimpleDeclaration_set_initDeclaratorList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleDeclaration ||
             __el->cpp_element_type != CppElement::_InitDeclaratorList);

    Cpp_SimpleDeclaration * const &self = static_cast <Cpp_SimpleDeclaration*> (__self);

    self->initDeclaratorList = static_cast <Cpp_InitDeclaratorList*> (___el);
}

bool cpp_simple_declaration__function_definition_jump (
        ParserElement *parser_element,
        void *data);

static ParserElement*
cpp_SimpleDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleDeclaration))) Cpp_SimpleDeclaration;
}

static Ref<Grammar>
create_cpp_SimpleDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleDeclaration_creation_func));
    grammar->name = String::forData ("SimpleDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_constructor_ambiguity;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_DeclSpecifierSeq ();
        entry->assignment_func = cpp_SimpleDeclaration_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_InitDeclaratorList ();
        entry->assignment_func = cpp_SimpleDeclaration_set_initDeclaratorList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_FunctionDefinition ();
        entry->jump_cb = cpp_simple_declaration__function_definition_jump;
        entry->jump_switch_grammar_index = 0;
        entry->jump_compound_grammar_index = 5;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleDeclaration (Cpp_SimpleDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleDeclaration\n");

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

    dump_cpp_InitDeclaratorList (el->initDeclaratorList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

bool cpp_DeclSpecifier_match_func (
        Cpp_DeclSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_DeclSpecifier_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclSpecifier);

    Cpp_DeclSpecifier * const &el = static_cast <Cpp_DeclSpecifier*> (_el);

    return cpp_DeclSpecifier_match_func (el, parser_control, data);
}

static void
cpp_DeclSpecifier_StorageClassSpecifier_set_storageClassSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclSpecifier ||
             __el->cpp_element_type != CppElement::_StorageClassSpecifier);

    Cpp_DeclSpecifier * const &_self = static_cast <Cpp_DeclSpecifier*> (__self);

    abortIf (_self->decl_specifier_type != Cpp_DeclSpecifier::_StorageClassSpecifier);

    Cpp_DeclSpecifier_StorageClassSpecifier * const &self = static_cast <Cpp_DeclSpecifier_StorageClassSpecifier*> (_self);

    self->storageClassSpecifier = static_cast <Cpp_StorageClassSpecifier*> (___el);
}

bool cpp_accept_decl_specifier__storage_class_specifier (
        Cpp_DeclSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_decl_specifier__storage_class_specifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclSpecifier);

    Cpp_DeclSpecifier * const &el =
            static_cast <Cpp_DeclSpecifier*> (_el);

    return cpp_accept_decl_specifier__storage_class_specifier (el, parser_control, data);
}

static ParserElement*
cpp_DeclSpecifier_StorageClassSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclSpecifier_StorageClassSpecifier))) Cpp_DeclSpecifier_StorageClassSpecifier;
}

static Ref<Grammar>
create_cpp_DeclSpecifier_StorageClassSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclSpecifier_StorageClassSpecifier_creation_func));
    grammar->name = String::forData ("DeclSpecifier_StorageClassSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_StorageClassSpecifier ();
        entry->assignment_func = cpp_DeclSpecifier_StorageClassSpecifier_set_storageClassSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_decl_specifier__storage_class_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifier_StorageClassSpecifier (Cpp_DeclSpecifier_StorageClassSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifier_StorageClassSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclSpecifier_StorageClassSpecifier\n");

    dump_cpp_StorageClassSpecifier (el->storageClassSpecifier, nest_level + 1);

}

static void
cpp_DeclSpecifier_TypeSpecifier_set_typeSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclSpecifier ||
             __el->cpp_element_type != CppElement::_TypeSpecifier);

    Cpp_DeclSpecifier * const &_self = static_cast <Cpp_DeclSpecifier*> (__self);

    abortIf (_self->decl_specifier_type != Cpp_DeclSpecifier::_TypeSpecifier);

    Cpp_DeclSpecifier_TypeSpecifier * const &self = static_cast <Cpp_DeclSpecifier_TypeSpecifier*> (_self);

    self->typeSpecifier = static_cast <Cpp_TypeSpecifier*> (___el);
}

static ParserElement*
cpp_DeclSpecifier_TypeSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclSpecifier_TypeSpecifier))) Cpp_DeclSpecifier_TypeSpecifier;
}

static Ref<Grammar>
create_cpp_DeclSpecifier_TypeSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclSpecifier_TypeSpecifier_creation_func));
    grammar->name = String::forData ("DeclSpecifier_TypeSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier ();
        entry->assignment_func = cpp_DeclSpecifier_TypeSpecifier_set_typeSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifier_TypeSpecifier (Cpp_DeclSpecifier_TypeSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifier_TypeSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclSpecifier_TypeSpecifier\n");

    dump_cpp_TypeSpecifier (el->typeSpecifier, nest_level + 1);

}

static void
cpp_DeclSpecifier_FunctionSpecifier_set_functionSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclSpecifier ||
             __el->cpp_element_type != CppElement::_FunctionSpecifier);

    Cpp_DeclSpecifier * const &_self = static_cast <Cpp_DeclSpecifier*> (__self);

    abortIf (_self->decl_specifier_type != Cpp_DeclSpecifier::_FunctionSpecifier);

    Cpp_DeclSpecifier_FunctionSpecifier * const &self = static_cast <Cpp_DeclSpecifier_FunctionSpecifier*> (_self);

    self->functionSpecifier = static_cast <Cpp_FunctionSpecifier*> (___el);
}

bool cpp_accept_decl_specifier__function_specifier (
        Cpp_DeclSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_decl_specifier__function_specifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclSpecifier);

    Cpp_DeclSpecifier * const &el =
            static_cast <Cpp_DeclSpecifier*> (_el);

    return cpp_accept_decl_specifier__function_specifier (el, parser_control, data);
}

static ParserElement*
cpp_DeclSpecifier_FunctionSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclSpecifier_FunctionSpecifier))) Cpp_DeclSpecifier_FunctionSpecifier;
}

static Ref<Grammar>
create_cpp_DeclSpecifier_FunctionSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclSpecifier_FunctionSpecifier_creation_func));
    grammar->name = String::forData ("DeclSpecifier_FunctionSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_FunctionSpecifier ();
        entry->assignment_func = cpp_DeclSpecifier_FunctionSpecifier_set_functionSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_decl_specifier__function_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifier_FunctionSpecifier (Cpp_DeclSpecifier_FunctionSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifier_FunctionSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclSpecifier_FunctionSpecifier\n");

    dump_cpp_FunctionSpecifier (el->functionSpecifier, nest_level + 1);

}

bool cpp_accept_decl_specifier__friend (
        Cpp_DeclSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_decl_specifier__friend (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclSpecifier);

    Cpp_DeclSpecifier * const &el =
            static_cast <Cpp_DeclSpecifier*> (_el);

    return cpp_accept_decl_specifier__friend (el, parser_control, data);
}

static ParserElement*
cpp_DeclSpecifier_Friend_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclSpecifier_Friend))) Cpp_DeclSpecifier_Friend;
}

static Ref<Grammar>
create_cpp_DeclSpecifier_Friend ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclSpecifier_Friend_creation_func));
    grammar->name = String::forData ("DeclSpecifier_Friend");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("friend"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_decl_specifier__friend;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifier_Friend (Cpp_DeclSpecifier_Friend const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifier_Friend\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclSpecifier_Friend\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: friend\n");
}

bool cpp_accept_decl_specifier__typedef (
        Cpp_DeclSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_decl_specifier__typedef (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclSpecifier);

    Cpp_DeclSpecifier * const &el =
            static_cast <Cpp_DeclSpecifier*> (_el);

    return cpp_accept_decl_specifier__typedef (el, parser_control, data);
}

static ParserElement*
cpp_DeclSpecifier_Typedef_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclSpecifier_Typedef))) Cpp_DeclSpecifier_Typedef;
}

static Ref<Grammar>
create_cpp_DeclSpecifier_Typedef ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclSpecifier_Typedef_creation_func));
    grammar->name = String::forData ("DeclSpecifier_Typedef");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typedef"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_decl_specifier__typedef;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifier_Typedef (Cpp_DeclSpecifier_Typedef const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifier_Typedef\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclSpecifier_Typedef\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typedef\n");
}

static Ref<Grammar>
create_cpp_DeclSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("DeclSpecifier");
    grammar->match_func = __pargen_cpp_DeclSpecifier_match_func;

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclSpecifier_StorageClassSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclSpecifier_TypeSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclSpecifier_FunctionSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclSpecifier_Friend ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DeclSpecifier_Typedef ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifier (Cpp_DeclSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifier\n");
        return;
    }

    switch (el->decl_specifier_type) {
        case Cpp_DeclSpecifier::_StorageClassSpecifier:
            dump_cpp_DeclSpecifier_StorageClassSpecifier (static_cast <Cpp_DeclSpecifier_StorageClassSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_DeclSpecifier::_TypeSpecifier:
            dump_cpp_DeclSpecifier_TypeSpecifier (static_cast <Cpp_DeclSpecifier_TypeSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_DeclSpecifier::_FunctionSpecifier:
            dump_cpp_DeclSpecifier_FunctionSpecifier (static_cast <Cpp_DeclSpecifier_FunctionSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_DeclSpecifier::_Friend:
            dump_cpp_DeclSpecifier_Friend (static_cast <Cpp_DeclSpecifier_Friend const *> (el), nest_level + 0);
            break;
        case Cpp_DeclSpecifier::_Typedef:
            dump_cpp_DeclSpecifier_Typedef (static_cast <Cpp_DeclSpecifier_Typedef const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_DeclSpecifierSeq_set_declSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclSpecifierSeq ||
             __el->cpp_element_type != CppElement::_DeclSpecifier);

    Cpp_DeclSpecifierSeq * const &self = static_cast <Cpp_DeclSpecifierSeq*> (__self);

    self->declSpecifier = static_cast <Cpp_DeclSpecifier*> (___el);
}

static void
cpp_DeclSpecifierSeq_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclSpecifierSeq ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_DeclSpecifierSeq * const &self = static_cast <Cpp_DeclSpecifierSeq*> (__self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static ParserElement*
cpp_DeclSpecifierSeq_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclSpecifierSeq))) Cpp_DeclSpecifierSeq;
}

static Ref<Grammar>
create_cpp_DeclSpecifierSeq ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclSpecifierSeq_creation_func));
    grammar->name = String::forData ("DeclSpecifierSeq");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DeclSpecifier ();
        entry->assignment_func = cpp_DeclSpecifierSeq_set_declSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_DeclSpecifierSeq ();
        entry->assignment_func = cpp_DeclSpecifierSeq_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclSpecifierSeq (Cpp_DeclSpecifierSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclSpecifierSeq\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclSpecifierSeq\n");

    dump_cpp_DeclSpecifier (el->declSpecifier, nest_level + 1);

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

}

void cpp_Scruffy_DeclSpecifierSeq_begin_func (
        void *data);

void cpp_Scruffy_DeclSpecifierSeq_accept_func (
        Cpp_DeclSpecifierSeq *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_Scruffy_DeclSpecifierSeq_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_Scruffy_DeclSpecifierSeq_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_DeclSpecifierSeq * const &el = static_cast <Cpp_DeclSpecifierSeq*> (_el);

    cpp_Scruffy_DeclSpecifierSeq_accept_func (el, parser_control, data);
}

static Ref<Grammar>
create_cpp_Scruffy_DeclSpecifierSeq ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_DeclSpecifierSeq");
    grammar->aliased_grammar = create_cpp_DeclSpecifierSeq ();

    grammar->begin_func = cpp_Scruffy_DeclSpecifierSeq_begin_func;
    grammar->accept_func = __pargen_cpp_Scruffy_DeclSpecifierSeq_accept_func;

    return grammar.ptr ();
}

static ParserElement*
cpp_StorageClassSpecifier_Auto_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_StorageClassSpecifier_Auto))) Cpp_StorageClassSpecifier_Auto;
}

static Ref<Grammar>
create_cpp_StorageClassSpecifier_Auto ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_StorageClassSpecifier_Auto_creation_func));
    grammar->name = String::forData ("StorageClassSpecifier_Auto");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("auto"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StorageClassSpecifier_Auto (Cpp_StorageClassSpecifier_Auto const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StorageClassSpecifier_Auto\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("StorageClassSpecifier_Auto\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: auto\n");
}

static ParserElement*
cpp_StorageClassSpecifier_Register_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_StorageClassSpecifier_Register))) Cpp_StorageClassSpecifier_Register;
}

static Ref<Grammar>
create_cpp_StorageClassSpecifier_Register ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_StorageClassSpecifier_Register_creation_func));
    grammar->name = String::forData ("StorageClassSpecifier_Register");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("register"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StorageClassSpecifier_Register (Cpp_StorageClassSpecifier_Register const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StorageClassSpecifier_Register\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("StorageClassSpecifier_Register\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: register\n");
}

static ParserElement*
cpp_StorageClassSpecifier_Static_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_StorageClassSpecifier_Static))) Cpp_StorageClassSpecifier_Static;
}

static Ref<Grammar>
create_cpp_StorageClassSpecifier_Static ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_StorageClassSpecifier_Static_creation_func));
    grammar->name = String::forData ("StorageClassSpecifier_Static");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("static"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StorageClassSpecifier_Static (Cpp_StorageClassSpecifier_Static const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StorageClassSpecifier_Static\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("StorageClassSpecifier_Static\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: static\n");
}

static ParserElement*
cpp_StorageClassSpecifier_Extern_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_StorageClassSpecifier_Extern))) Cpp_StorageClassSpecifier_Extern;
}

static Ref<Grammar>
create_cpp_StorageClassSpecifier_Extern ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_StorageClassSpecifier_Extern_creation_func));
    grammar->name = String::forData ("StorageClassSpecifier_Extern");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("extern"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StorageClassSpecifier_Extern (Cpp_StorageClassSpecifier_Extern const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StorageClassSpecifier_Extern\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("StorageClassSpecifier_Extern\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: extern\n");
}

static ParserElement*
cpp_StorageClassSpecifier_Mutable_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_StorageClassSpecifier_Mutable))) Cpp_StorageClassSpecifier_Mutable;
}

static Ref<Grammar>
create_cpp_StorageClassSpecifier_Mutable ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_StorageClassSpecifier_Mutable_creation_func));
    grammar->name = String::forData ("StorageClassSpecifier_Mutable");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("mutable"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StorageClassSpecifier_Mutable (Cpp_StorageClassSpecifier_Mutable const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StorageClassSpecifier_Mutable\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("StorageClassSpecifier_Mutable\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: mutable\n");
}

static Ref<Grammar>
create_cpp_StorageClassSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("StorageClassSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_StorageClassSpecifier_Auto ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_StorageClassSpecifier_Register ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_StorageClassSpecifier_Static ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_StorageClassSpecifier_Extern ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_StorageClassSpecifier_Mutable ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_StorageClassSpecifier (Cpp_StorageClassSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) StorageClassSpecifier\n");
        return;
    }

    switch (el->storage_class_specifier_type) {
        case Cpp_StorageClassSpecifier::_Auto:
            dump_cpp_StorageClassSpecifier_Auto (static_cast <Cpp_StorageClassSpecifier_Auto const *> (el), nest_level + 0);
            break;
        case Cpp_StorageClassSpecifier::_Register:
            dump_cpp_StorageClassSpecifier_Register (static_cast <Cpp_StorageClassSpecifier_Register const *> (el), nest_level + 0);
            break;
        case Cpp_StorageClassSpecifier::_Static:
            dump_cpp_StorageClassSpecifier_Static (static_cast <Cpp_StorageClassSpecifier_Static const *> (el), nest_level + 0);
            break;
        case Cpp_StorageClassSpecifier::_Extern:
            dump_cpp_StorageClassSpecifier_Extern (static_cast <Cpp_StorageClassSpecifier_Extern const *> (el), nest_level + 0);
            break;
        case Cpp_StorageClassSpecifier::_Mutable:
            dump_cpp_StorageClassSpecifier_Mutable (static_cast <Cpp_StorageClassSpecifier_Mutable const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static ParserElement*
cpp_FunctionSpecifier_Inline_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionSpecifier_Inline))) Cpp_FunctionSpecifier_Inline;
}

static Ref<Grammar>
create_cpp_FunctionSpecifier_Inline ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionSpecifier_Inline_creation_func));
    grammar->name = String::forData ("FunctionSpecifier_Inline");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("inline"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionSpecifier_Inline (Cpp_FunctionSpecifier_Inline const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionSpecifier_Inline\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionSpecifier_Inline\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: inline\n");
}

static ParserElement*
cpp_FunctionSpecifier_Virtual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionSpecifier_Virtual))) Cpp_FunctionSpecifier_Virtual;
}

static Ref<Grammar>
create_cpp_FunctionSpecifier_Virtual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionSpecifier_Virtual_creation_func));
    grammar->name = String::forData ("FunctionSpecifier_Virtual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("virtual"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionSpecifier_Virtual (Cpp_FunctionSpecifier_Virtual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionSpecifier_Virtual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionSpecifier_Virtual\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: virtual\n");
}

static ParserElement*
cpp_FunctionSpecifier_Explicit_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionSpecifier_Explicit))) Cpp_FunctionSpecifier_Explicit;
}

static Ref<Grammar>
create_cpp_FunctionSpecifier_Explicit ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionSpecifier_Explicit_creation_func));
    grammar->name = String::forData ("FunctionSpecifier_Explicit");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("explicit"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionSpecifier_Explicit (Cpp_FunctionSpecifier_Explicit const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionSpecifier_Explicit\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionSpecifier_Explicit\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: explicit\n");
}

static Ref<Grammar>
create_cpp_FunctionSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("FunctionSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionSpecifier_Inline ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionSpecifier_Virtual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionSpecifier_Explicit ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionSpecifier (Cpp_FunctionSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionSpecifier\n");
        return;
    }

    switch (el->function_specifier_type) {
        case Cpp_FunctionSpecifier::_Inline:
            dump_cpp_FunctionSpecifier_Inline (static_cast <Cpp_FunctionSpecifier_Inline const *> (el), nest_level + 0);
            break;
        case Cpp_FunctionSpecifier::_Virtual:
            dump_cpp_FunctionSpecifier_Virtual (static_cast <Cpp_FunctionSpecifier_Virtual const *> (el), nest_level + 0);
            break;
        case Cpp_FunctionSpecifier::_Explicit:
            dump_cpp_FunctionSpecifier_Explicit (static_cast <Cpp_FunctionSpecifier_Explicit const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_TypedefName_match_func (
        Cpp_TypedefName *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_TypedefName_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TypedefName);

    Cpp_TypedefName * const &el = static_cast <Cpp_TypedefName*> (_el);

    return cpp_TypedefName_match_func (el, parser_control, data);
}

static void
cpp_TypedefName_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypedefName ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypedefName * const &self = static_cast <Cpp_TypedefName*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_TypedefName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypedefName))) Cpp_TypedefName;
}

static Ref<Grammar>
create_cpp_TypedefName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypedefName_creation_func));
    grammar->name = String::forData ("TypedefName");
    grammar->match_func = __pargen_cpp_TypedefName_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypedefName_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypedefName (Cpp_TypedefName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypedefName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypedefName\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_TypeSpecifier_SimpleTypeSpecifier_set_simpleTypeSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifier ||
             __el->cpp_element_type != CppElement::_SimpleTypeSpecifier);

    Cpp_TypeSpecifier * const &_self = static_cast <Cpp_TypeSpecifier*> (__self);

    abortIf (_self->type_specifier_type != Cpp_TypeSpecifier::_SimpleTypeSpecifier);

    Cpp_TypeSpecifier_SimpleTypeSpecifier * const &self = static_cast <Cpp_TypeSpecifier_SimpleTypeSpecifier*> (_self);

    self->simpleTypeSpecifier = static_cast <Cpp_SimpleTypeSpecifier*> (___el);
}

static ParserElement*
cpp_TypeSpecifier_SimpleTypeSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeSpecifier_SimpleTypeSpecifier))) Cpp_TypeSpecifier_SimpleTypeSpecifier;
}

static Ref<Grammar>
create_cpp_TypeSpecifier_SimpleTypeSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeSpecifier_SimpleTypeSpecifier_creation_func));
    grammar->name = String::forData ("TypeSpecifier_SimpleTypeSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier ();
        entry->assignment_func = cpp_TypeSpecifier_SimpleTypeSpecifier_set_simpleTypeSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifier_SimpleTypeSpecifier (Cpp_TypeSpecifier_SimpleTypeSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifier_SimpleTypeSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeSpecifier_SimpleTypeSpecifier\n");

    dump_cpp_SimpleTypeSpecifier (el->simpleTypeSpecifier, nest_level + 1);

}

static void
cpp_TypeSpecifier_ClassSpecifier_set_classSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifier ||
             __el->cpp_element_type != CppElement::_ClassSpecifier);

    Cpp_TypeSpecifier * const &_self = static_cast <Cpp_TypeSpecifier*> (__self);

    abortIf (_self->type_specifier_type != Cpp_TypeSpecifier::_ClassSpecifier);

    Cpp_TypeSpecifier_ClassSpecifier * const &self = static_cast <Cpp_TypeSpecifier_ClassSpecifier*> (_self);

    self->classSpecifier = static_cast <Cpp_ClassSpecifier*> (___el);
}

static ParserElement*
cpp_TypeSpecifier_ClassSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeSpecifier_ClassSpecifier))) Cpp_TypeSpecifier_ClassSpecifier;
}

static Ref<Grammar>
create_cpp_TypeSpecifier_ClassSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeSpecifier_ClassSpecifier_creation_func));
    grammar->name = String::forData ("TypeSpecifier_ClassSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassSpecifier ();
        entry->assignment_func = cpp_TypeSpecifier_ClassSpecifier_set_classSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifier_ClassSpecifier (Cpp_TypeSpecifier_ClassSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifier_ClassSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeSpecifier_ClassSpecifier\n");

    dump_cpp_ClassSpecifier (el->classSpecifier, nest_level + 1);

}

static void
cpp_TypeSpecifier_EnumSpecifier_set_enumSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifier ||
             __el->cpp_element_type != CppElement::_EnumSpecifier);

    Cpp_TypeSpecifier * const &_self = static_cast <Cpp_TypeSpecifier*> (__self);

    abortIf (_self->type_specifier_type != Cpp_TypeSpecifier::_EnumSpecifier);

    Cpp_TypeSpecifier_EnumSpecifier * const &self = static_cast <Cpp_TypeSpecifier_EnumSpecifier*> (_self);

    self->enumSpecifier = static_cast <Cpp_EnumSpecifier*> (___el);
}

static ParserElement*
cpp_TypeSpecifier_EnumSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeSpecifier_EnumSpecifier))) Cpp_TypeSpecifier_EnumSpecifier;
}

static Ref<Grammar>
create_cpp_TypeSpecifier_EnumSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeSpecifier_EnumSpecifier_creation_func));
    grammar->name = String::forData ("TypeSpecifier_EnumSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EnumSpecifier ();
        entry->assignment_func = cpp_TypeSpecifier_EnumSpecifier_set_enumSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifier_EnumSpecifier (Cpp_TypeSpecifier_EnumSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifier_EnumSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeSpecifier_EnumSpecifier\n");

    dump_cpp_EnumSpecifier (el->enumSpecifier, nest_level + 1);

}

static void
cpp_TypeSpecifier_ElaboratedTypeSpecifier_set_elaboratedTypeSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifier ||
             __el->cpp_element_type != CppElement::_ElaboratedTypeSpecifier);

    Cpp_TypeSpecifier * const &_self = static_cast <Cpp_TypeSpecifier*> (__self);

    abortIf (_self->type_specifier_type != Cpp_TypeSpecifier::_ElaboratedTypeSpecifier);

    Cpp_TypeSpecifier_ElaboratedTypeSpecifier * const &self = static_cast <Cpp_TypeSpecifier_ElaboratedTypeSpecifier*> (_self);

    self->elaboratedTypeSpecifier = static_cast <Cpp_ElaboratedTypeSpecifier*> (___el);
}

static ParserElement*
cpp_TypeSpecifier_ElaboratedTypeSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeSpecifier_ElaboratedTypeSpecifier))) Cpp_TypeSpecifier_ElaboratedTypeSpecifier;
}

static Ref<Grammar>
create_cpp_TypeSpecifier_ElaboratedTypeSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeSpecifier_ElaboratedTypeSpecifier_creation_func));
    grammar->name = String::forData ("TypeSpecifier_ElaboratedTypeSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ElaboratedTypeSpecifier ();
        entry->assignment_func = cpp_TypeSpecifier_ElaboratedTypeSpecifier_set_elaboratedTypeSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifier_ElaboratedTypeSpecifier (Cpp_TypeSpecifier_ElaboratedTypeSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifier_ElaboratedTypeSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeSpecifier_ElaboratedTypeSpecifier\n");

    dump_cpp_ElaboratedTypeSpecifier (el->elaboratedTypeSpecifier, nest_level + 1);

}

static void
cpp_TypeSpecifier_CvQualifier_set_cvQualifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifier ||
             __el->cpp_element_type != CppElement::_CvQualifier);

    Cpp_TypeSpecifier * const &_self = static_cast <Cpp_TypeSpecifier*> (__self);

    abortIf (_self->type_specifier_type != Cpp_TypeSpecifier::_CvQualifier);

    Cpp_TypeSpecifier_CvQualifier * const &self = static_cast <Cpp_TypeSpecifier_CvQualifier*> (_self);

    self->cvQualifier = static_cast <Cpp_CvQualifier*> (___el);
}

bool cpp_accept_type_specifier__cv_qualifier (
        Cpp_TypeSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_type_specifier__cv_qualifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TypeSpecifier);

    Cpp_TypeSpecifier * const &el =
            static_cast <Cpp_TypeSpecifier*> (_el);

    return cpp_accept_type_specifier__cv_qualifier (el, parser_control, data);
}

static ParserElement*
cpp_TypeSpecifier_CvQualifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeSpecifier_CvQualifier))) Cpp_TypeSpecifier_CvQualifier;
}

static Ref<Grammar>
create_cpp_TypeSpecifier_CvQualifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeSpecifier_CvQualifier_creation_func));
    grammar->name = String::forData ("TypeSpecifier_CvQualifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CvQualifier ();
        entry->assignment_func = cpp_TypeSpecifier_CvQualifier_set_cvQualifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_specifier__cv_qualifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifier_CvQualifier (Cpp_TypeSpecifier_CvQualifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifier_CvQualifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeSpecifier_CvQualifier\n");

    dump_cpp_CvQualifier (el->cvQualifier, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TypeSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TypeSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier_SimpleTypeSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier_ClassSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier_EnumSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier_ElaboratedTypeSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier_CvQualifier ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifier (Cpp_TypeSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifier\n");
        return;
    }

    switch (el->type_specifier_type) {
        case Cpp_TypeSpecifier::_SimpleTypeSpecifier:
            dump_cpp_TypeSpecifier_SimpleTypeSpecifier (static_cast <Cpp_TypeSpecifier_SimpleTypeSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_TypeSpecifier::_ClassSpecifier:
            dump_cpp_TypeSpecifier_ClassSpecifier (static_cast <Cpp_TypeSpecifier_ClassSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_TypeSpecifier::_EnumSpecifier:
            dump_cpp_TypeSpecifier_EnumSpecifier (static_cast <Cpp_TypeSpecifier_EnumSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_TypeSpecifier::_ElaboratedTypeSpecifier:
            dump_cpp_TypeSpecifier_ElaboratedTypeSpecifier (static_cast <Cpp_TypeSpecifier_ElaboratedTypeSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_TypeSpecifier::_CvQualifier:
            dump_cpp_TypeSpecifier_CvQualifier (static_cast <Cpp_TypeSpecifier_CvQualifier const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_accept_simple_type_specifier (
        Cpp_SimpleTypeSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_simple_type_specifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_SimpleTypeSpecifier);

    Cpp_SimpleTypeSpecifier * const &el =
            static_cast <Cpp_SimpleTypeSpecifier*> (_el);

    return cpp_accept_simple_type_specifier (el, parser_control, data);
}

static ParserElement*
cpp_SimpleTypeSpecifier_Void_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Void))) Cpp_SimpleTypeSpecifier_Void;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Void ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Void_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Void");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("void"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Void (Cpp_SimpleTypeSpecifier_Void const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Void\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Void\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: void\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Bool_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Bool))) Cpp_SimpleTypeSpecifier_Bool;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Bool ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Bool_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Bool");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("bool"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Bool (Cpp_SimpleTypeSpecifier_Bool const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Bool\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Bool\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: bool\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Int_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Int))) Cpp_SimpleTypeSpecifier_Int;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Int ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Int_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Int");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("int"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Int (Cpp_SimpleTypeSpecifier_Int const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Int\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Int\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: int\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Char_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Char))) Cpp_SimpleTypeSpecifier_Char;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Char ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Char_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Char");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("char"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Char (Cpp_SimpleTypeSpecifier_Char const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Char\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Char\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: char\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Unsigned_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Unsigned))) Cpp_SimpleTypeSpecifier_Unsigned;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Unsigned ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Unsigned_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Unsigned");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("unsigned"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Unsigned (Cpp_SimpleTypeSpecifier_Unsigned const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Unsigned\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Unsigned\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: unsigned\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Long_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Long))) Cpp_SimpleTypeSpecifier_Long;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Long ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Long_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Long");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("long"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Long (Cpp_SimpleTypeSpecifier_Long const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Long\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Long\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: long\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Float_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Float))) Cpp_SimpleTypeSpecifier_Float;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Float ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Float_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Float");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("float"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Float (Cpp_SimpleTypeSpecifier_Float const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Float\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Float\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: float\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Double_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Double))) Cpp_SimpleTypeSpecifier_Double;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Double ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Double_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Double");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("double"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Double (Cpp_SimpleTypeSpecifier_Double const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Double\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Double\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: double\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Short_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Short))) Cpp_SimpleTypeSpecifier_Short;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Short ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Short_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Short");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("short"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Short (Cpp_SimpleTypeSpecifier_Short const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Short\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Short\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: short\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_WcharT_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_WcharT))) Cpp_SimpleTypeSpecifier_WcharT;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_WcharT ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_WcharT_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_WcharT");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("wchar_t"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_WcharT (Cpp_SimpleTypeSpecifier_WcharT const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_WcharT\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_WcharT\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: wchar_t\n");
}

static ParserElement*
cpp_SimpleTypeSpecifier_Signed_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_Signed))) Cpp_SimpleTypeSpecifier_Signed;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_Signed ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_Signed_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_Signed");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("signed"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_Signed (Cpp_SimpleTypeSpecifier_Signed const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_Signed\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_Signed\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: signed\n");
}

static void
cpp_SimpleTypeSpecifier_TypeName_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_SimpleTypeSpecifier * const &_self = static_cast <Cpp_SimpleTypeSpecifier*> (__self);

    abortIf (_self->simple_type_specifier_type != Cpp_SimpleTypeSpecifier::_TypeName);

    Cpp_SimpleTypeSpecifier_TypeName * const &self = static_cast <Cpp_SimpleTypeSpecifier_TypeName*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_SimpleTypeSpecifier_TypeName_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleTypeSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_SimpleTypeSpecifier * const &_self = static_cast <Cpp_SimpleTypeSpecifier*> (__self);

    abortIf (_self->simple_type_specifier_type != Cpp_SimpleTypeSpecifier::_TypeName);

    Cpp_SimpleTypeSpecifier_TypeName * const &self = static_cast <Cpp_SimpleTypeSpecifier_TypeName*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

bool cpp_simple_type_specifier__template_id_jump (
        ParserElement *parser_element,
        void *data);

static void
cpp_SimpleTypeSpecifier_TypeName_set_typeName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleTypeSpecifier ||
             __el->cpp_element_type != CppElement::_TypeName);

    Cpp_SimpleTypeSpecifier * const &_self = static_cast <Cpp_SimpleTypeSpecifier*> (__self);

    abortIf (_self->simple_type_specifier_type != Cpp_SimpleTypeSpecifier::_TypeName);

    Cpp_SimpleTypeSpecifier_TypeName * const &self = static_cast <Cpp_SimpleTypeSpecifier_TypeName*> (_self);

    self->typeName = static_cast <Cpp_TypeName*> (___el);
}

static ParserElement*
cpp_SimpleTypeSpecifier_TypeName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_TypeName))) Cpp_SimpleTypeSpecifier_TypeName;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_TypeName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_TypeName_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_TypeName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_SimpleTypeSpecifier_TypeName_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_SimpleTypeSpecifier ();
        entry->jump_switch_grammar_index = 12;
        entry->jump_compound_grammar_index = 2;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_SimpleTypeSpecifier_TypeName_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_SimpleTypeSpecifier ();
        entry->jump_cb = cpp_simple_type_specifier__template_id_jump;
        entry->jump_switch_grammar_index = 12;
        entry->jump_compound_grammar_index = 3;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeName ();
        entry->assignment_func = cpp_SimpleTypeSpecifier_TypeName_set_typeName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_TypeName (Cpp_SimpleTypeSpecifier_TypeName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_TypeName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_TypeName\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_TypeName (el->typeName, nest_level + 1);

}

static void
cpp_SimpleTypeSpecifier_TemplateId_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_SimpleTypeSpecifier * const &_self = static_cast <Cpp_SimpleTypeSpecifier*> (__self);

    abortIf (_self->simple_type_specifier_type != Cpp_SimpleTypeSpecifier::_TemplateId);

    Cpp_SimpleTypeSpecifier_TemplateId * const &self = static_cast <Cpp_SimpleTypeSpecifier_TemplateId*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_SimpleTypeSpecifier_TemplateId_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleTypeSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_SimpleTypeSpecifier * const &_self = static_cast <Cpp_SimpleTypeSpecifier*> (__self);

    abortIf (_self->simple_type_specifier_type != Cpp_SimpleTypeSpecifier::_TemplateId);

    Cpp_SimpleTypeSpecifier_TemplateId * const &self = static_cast <Cpp_SimpleTypeSpecifier_TemplateId*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_SimpleTypeSpecifier_TemplateId_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_SimpleTypeSpecifier ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_SimpleTypeSpecifier * const &_self = static_cast <Cpp_SimpleTypeSpecifier*> (__self);

    abortIf (_self->simple_type_specifier_type != Cpp_SimpleTypeSpecifier::_TemplateId);

    Cpp_SimpleTypeSpecifier_TemplateId * const &self = static_cast <Cpp_SimpleTypeSpecifier_TemplateId*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static ParserElement*
cpp_SimpleTypeSpecifier_TemplateId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_SimpleTypeSpecifier_TemplateId))) Cpp_SimpleTypeSpecifier_TemplateId;
}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier_TemplateId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_SimpleTypeSpecifier_TemplateId_creation_func));
    grammar->name = String::forData ("SimpleTypeSpecifier_TemplateId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_SimpleTypeSpecifier_TemplateId_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_SimpleTypeSpecifier_TemplateId_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_SimpleTypeSpecifier_TemplateId_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_simple_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier_TemplateId (Cpp_SimpleTypeSpecifier_TemplateId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier_TemplateId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("SimpleTypeSpecifier_TemplateId\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    dump_cpp_TemplateId (el->templateId, nest_level + 1);

}

static Ref<Grammar>
create_cpp_SimpleTypeSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("SimpleTypeSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Void ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Bool ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Int ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Char ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Unsigned ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Long ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Float ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Double ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Short ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_WcharT ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_Signed ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_TypeName ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_SimpleTypeSpecifier_TemplateId ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_SimpleTypeSpecifier (Cpp_SimpleTypeSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) SimpleTypeSpecifier\n");
        return;
    }

    switch (el->simple_type_specifier_type) {
        case Cpp_SimpleTypeSpecifier::_Void:
            dump_cpp_SimpleTypeSpecifier_Void (static_cast <Cpp_SimpleTypeSpecifier_Void const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Bool:
            dump_cpp_SimpleTypeSpecifier_Bool (static_cast <Cpp_SimpleTypeSpecifier_Bool const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Int:
            dump_cpp_SimpleTypeSpecifier_Int (static_cast <Cpp_SimpleTypeSpecifier_Int const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Char:
            dump_cpp_SimpleTypeSpecifier_Char (static_cast <Cpp_SimpleTypeSpecifier_Char const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Unsigned:
            dump_cpp_SimpleTypeSpecifier_Unsigned (static_cast <Cpp_SimpleTypeSpecifier_Unsigned const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Long:
            dump_cpp_SimpleTypeSpecifier_Long (static_cast <Cpp_SimpleTypeSpecifier_Long const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Float:
            dump_cpp_SimpleTypeSpecifier_Float (static_cast <Cpp_SimpleTypeSpecifier_Float const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Double:
            dump_cpp_SimpleTypeSpecifier_Double (static_cast <Cpp_SimpleTypeSpecifier_Double const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Short:
            dump_cpp_SimpleTypeSpecifier_Short (static_cast <Cpp_SimpleTypeSpecifier_Short const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_WcharT:
            dump_cpp_SimpleTypeSpecifier_WcharT (static_cast <Cpp_SimpleTypeSpecifier_WcharT const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_Signed:
            dump_cpp_SimpleTypeSpecifier_Signed (static_cast <Cpp_SimpleTypeSpecifier_Signed const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_TypeName:
            dump_cpp_SimpleTypeSpecifier_TypeName (static_cast <Cpp_SimpleTypeSpecifier_TypeName const *> (el), nest_level + 0);
            break;
        case Cpp_SimpleTypeSpecifier::_TemplateId:
            dump_cpp_SimpleTypeSpecifier_TemplateId (static_cast <Cpp_SimpleTypeSpecifier_TemplateId const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_TypeName_ClassName_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeName ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_TypeName * const &_self = static_cast <Cpp_TypeName*> (__self);

    abortIf (_self->type_name_type != Cpp_TypeName::_ClassName);

    Cpp_TypeName_ClassName * const &self = static_cast <Cpp_TypeName_ClassName*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_TypeName_ClassName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeName_ClassName))) Cpp_TypeName_ClassName;
}

static Ref<Grammar>
create_cpp_TypeName_ClassName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeName_ClassName_creation_func));
    grammar->name = String::forData ("TypeName_ClassName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_TypeName_ClassName_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeName_ClassName (Cpp_TypeName_ClassName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeName_ClassName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeName_ClassName\n");

    dump_cpp_ClassName (el->className, nest_level + 1);

}

static void
cpp_TypeName_EnumName_set_enumName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeName ||
             __el->cpp_element_type != CppElement::_EnumName);

    Cpp_TypeName * const &_self = static_cast <Cpp_TypeName*> (__self);

    abortIf (_self->type_name_type != Cpp_TypeName::_EnumName);

    Cpp_TypeName_EnumName * const &self = static_cast <Cpp_TypeName_EnumName*> (_self);

    self->enumName = static_cast <Cpp_EnumName*> (___el);
}

static ParserElement*
cpp_TypeName_EnumName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeName_EnumName))) Cpp_TypeName_EnumName;
}

static Ref<Grammar>
create_cpp_TypeName_EnumName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeName_EnumName_creation_func));
    grammar->name = String::forData ("TypeName_EnumName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EnumName ();
        entry->assignment_func = cpp_TypeName_EnumName_set_enumName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeName_EnumName (Cpp_TypeName_EnumName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeName_EnumName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeName_EnumName\n");

    dump_cpp_EnumName (el->enumName, nest_level + 1);

}

static void
cpp_TypeName_TypedefName_set_typedefName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeName ||
             __el->cpp_element_type != CppElement::_TypedefName);

    Cpp_TypeName * const &_self = static_cast <Cpp_TypeName*> (__self);

    abortIf (_self->type_name_type != Cpp_TypeName::_TypedefName);

    Cpp_TypeName_TypedefName * const &self = static_cast <Cpp_TypeName_TypedefName*> (_self);

    self->typedefName = static_cast <Cpp_TypedefName*> (___el);
}

static ParserElement*
cpp_TypeName_TypedefName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeName_TypedefName))) Cpp_TypeName_TypedefName;
}

static Ref<Grammar>
create_cpp_TypeName_TypedefName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeName_TypedefName_creation_func));
    grammar->name = String::forData ("TypeName_TypedefName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypedefName ();
        entry->assignment_func = cpp_TypeName_TypedefName_set_typedefName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeName_TypedefName (Cpp_TypeName_TypedefName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeName_TypedefName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeName_TypedefName\n");

    dump_cpp_TypedefName (el->typedefName, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TypeName ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TypeName");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeName_ClassName ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeName_EnumName ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeName_TypedefName ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeName (Cpp_TypeName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeName\n");
        return;
    }

    switch (el->type_name_type) {
        case Cpp_TypeName::_ClassName:
            dump_cpp_TypeName_ClassName (static_cast <Cpp_TypeName_ClassName const *> (el), nest_level + 0);
            break;
        case Cpp_TypeName::_EnumName:
            dump_cpp_TypeName_EnumName (static_cast <Cpp_TypeName_EnumName const *> (el), nest_level + 0);
            break;
        case Cpp_TypeName::_TypedefName:
            dump_cpp_TypeName_TypedefName (static_cast <Cpp_TypeName_TypedefName const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ElaboratedTypeSpecifier_Class_set_classKey (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_ClassKey);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Class);

    Cpp_ElaboratedTypeSpecifier_Class * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Class*> (_self);

    self->classKey = static_cast <Cpp_ClassKey*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Class_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Class);

    Cpp_ElaboratedTypeSpecifier_Class * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Class*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Class_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Class);

    Cpp_ElaboratedTypeSpecifier_Class * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Class*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Class_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Class);

    Cpp_ElaboratedTypeSpecifier_Class * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Class*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

bool cpp_accept_elaborated_type_specifier (
        Cpp_ElaboratedTypeSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_elaborated_type_specifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ElaboratedTypeSpecifier);

    Cpp_ElaboratedTypeSpecifier * const &el =
            static_cast <Cpp_ElaboratedTypeSpecifier*> (_el);

    return cpp_accept_elaborated_type_specifier (el, parser_control, data);
}

static ParserElement*
cpp_ElaboratedTypeSpecifier_Class_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ElaboratedTypeSpecifier_Class))) Cpp_ElaboratedTypeSpecifier_Class;
}

static Ref<Grammar>
create_cpp_ElaboratedTypeSpecifier_Class ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ElaboratedTypeSpecifier_Class_creation_func));
    grammar->name = String::forData ("ElaboratedTypeSpecifier_Class");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassKey ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Class_set_classKey;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Class_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Class_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Class_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_elaborated_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ElaboratedTypeSpecifier_Class (Cpp_ElaboratedTypeSpecifier_Class const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ElaboratedTypeSpecifier_Class\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ElaboratedTypeSpecifier_Class\n");

    dump_cpp_ClassKey (el->classKey, nest_level + 1);

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_ElaboratedTypeSpecifier_Enum_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Enum);

    Cpp_ElaboratedTypeSpecifier_Enum * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Enum*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Enum_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Enum);

    Cpp_ElaboratedTypeSpecifier_Enum * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Enum*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Enum_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Enum);

    Cpp_ElaboratedTypeSpecifier_Enum * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Enum*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_ElaboratedTypeSpecifier_Enum_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ElaboratedTypeSpecifier_Enum))) Cpp_ElaboratedTypeSpecifier_Enum;
}

static Ref<Grammar>
create_cpp_ElaboratedTypeSpecifier_Enum ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ElaboratedTypeSpecifier_Enum_creation_func));
    grammar->name = String::forData ("ElaboratedTypeSpecifier_Enum");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("enum"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Enum_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Enum_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Enum_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_elaborated_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ElaboratedTypeSpecifier_Enum (Cpp_ElaboratedTypeSpecifier_Enum const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ElaboratedTypeSpecifier_Enum\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ElaboratedTypeSpecifier_Enum\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: enum\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_ElaboratedTypeSpecifier_Typename_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Typename);

    Cpp_ElaboratedTypeSpecifier_Typename * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Typename*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Typename_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Typename);

    Cpp_ElaboratedTypeSpecifier_Typename * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Typename*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_Typename_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_Typename);

    Cpp_ElaboratedTypeSpecifier_Typename * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_Typename*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_ElaboratedTypeSpecifier_Typename_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ElaboratedTypeSpecifier_Typename))) Cpp_ElaboratedTypeSpecifier_Typename;
}

static Ref<Grammar>
create_cpp_ElaboratedTypeSpecifier_Typename ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ElaboratedTypeSpecifier_Typename_creation_func));
    grammar->name = String::forData ("ElaboratedTypeSpecifier_Typename");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Typename_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Typename_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_ElaboratedTypeSpecifier ();
        entry->jump_switch_grammar_index = 3;
        entry->jump_compound_grammar_index = 4;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_Typename_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_elaborated_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ElaboratedTypeSpecifier_Typename (Cpp_ElaboratedTypeSpecifier_Typename const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ElaboratedTypeSpecifier_Typename\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ElaboratedTypeSpecifier_Typename\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typename\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_ElaboratedTypeSpecifier_TypenameTemplate_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_TypenameTemplate);

    Cpp_ElaboratedTypeSpecifier_TypenameTemplate * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_TypenameTemplate*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_TypenameTemplate_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_TypenameTemplate);

    Cpp_ElaboratedTypeSpecifier_TypenameTemplate * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_TypenameTemplate*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_ElaboratedTypeSpecifier_TypenameTemplate_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ElaboratedTypeSpecifier ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_ElaboratedTypeSpecifier * const &_self = static_cast <Cpp_ElaboratedTypeSpecifier*> (__self);

    abortIf (_self->elaborated_type_specifier_type != Cpp_ElaboratedTypeSpecifier::_TypenameTemplate);

    Cpp_ElaboratedTypeSpecifier_TypenameTemplate * const &self = static_cast <Cpp_ElaboratedTypeSpecifier_TypenameTemplate*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static ParserElement*
cpp_ElaboratedTypeSpecifier_TypenameTemplate_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ElaboratedTypeSpecifier_TypenameTemplate))) Cpp_ElaboratedTypeSpecifier_TypenameTemplate;
}

static Ref<Grammar>
create_cpp_ElaboratedTypeSpecifier_TypenameTemplate ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ElaboratedTypeSpecifier_TypenameTemplate_creation_func));
    grammar->name = String::forData ("ElaboratedTypeSpecifier_TypenameTemplate");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_TypenameTemplate_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_TypenameTemplate_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_ElaboratedTypeSpecifier_TypenameTemplate_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_elaborated_type_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ElaboratedTypeSpecifier_TypenameTemplate (Cpp_ElaboratedTypeSpecifier_TypenameTemplate const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ElaboratedTypeSpecifier_TypenameTemplate\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ElaboratedTypeSpecifier_TypenameTemplate\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typename\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: template_opt\n");
    dump_cpp_TemplateId (el->templateId, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ElaboratedTypeSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ElaboratedTypeSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ElaboratedTypeSpecifier_Class ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ElaboratedTypeSpecifier_Enum ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ElaboratedTypeSpecifier_Typename ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ElaboratedTypeSpecifier_TypenameTemplate ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ElaboratedTypeSpecifier (Cpp_ElaboratedTypeSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ElaboratedTypeSpecifier\n");
        return;
    }

    switch (el->elaborated_type_specifier_type) {
        case Cpp_ElaboratedTypeSpecifier::_Class:
            dump_cpp_ElaboratedTypeSpecifier_Class (static_cast <Cpp_ElaboratedTypeSpecifier_Class const *> (el), nest_level + 0);
            break;
        case Cpp_ElaboratedTypeSpecifier::_Enum:
            dump_cpp_ElaboratedTypeSpecifier_Enum (static_cast <Cpp_ElaboratedTypeSpecifier_Enum const *> (el), nest_level + 0);
            break;
        case Cpp_ElaboratedTypeSpecifier::_Typename:
            dump_cpp_ElaboratedTypeSpecifier_Typename (static_cast <Cpp_ElaboratedTypeSpecifier_Typename const *> (el), nest_level + 0);
            break;
        case Cpp_ElaboratedTypeSpecifier::_TypenameTemplate:
            dump_cpp_ElaboratedTypeSpecifier_TypenameTemplate (static_cast <Cpp_ElaboratedTypeSpecifier_TypenameTemplate const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_EnumName_match_func (
        Cpp_EnumName *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_EnumName_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_EnumName);

    Cpp_EnumName * const &el = static_cast <Cpp_EnumName*> (_el);

    return cpp_EnumName_match_func (el, parser_control, data);
}

static void
cpp_EnumName_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumName ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_EnumName * const &self = static_cast <Cpp_EnumName*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_EnumName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EnumName))) Cpp_EnumName;
}

static Ref<Grammar>
create_cpp_EnumName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EnumName_creation_func));
    grammar->name = String::forData ("EnumName");
    grammar->match_func = __pargen_cpp_EnumName_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_EnumName_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumName (Cpp_EnumName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EnumName\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_EnumSpecifier_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumSpecifier ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_EnumSpecifier * const &self = static_cast <Cpp_EnumSpecifier*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_EnumSpecifier_set_enumeratorList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumSpecifier ||
             __el->cpp_element_type != CppElement::_EnumeratorList);

    Cpp_EnumSpecifier * const &self = static_cast <Cpp_EnumSpecifier*> (__self);

    self->enumeratorList = static_cast <Cpp_EnumeratorList*> (___el);
}

bool cpp_accept_enum_specifier (
        Cpp_EnumSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_enum_specifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_EnumSpecifier);

    Cpp_EnumSpecifier * const &el =
            static_cast <Cpp_EnumSpecifier*> (_el);

    return cpp_accept_enum_specifier (el, parser_control, data);
}

static ParserElement*
cpp_EnumSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EnumSpecifier))) Cpp_EnumSpecifier;
}

static Ref<Grammar>
create_cpp_EnumSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EnumSpecifier_creation_func));
    grammar->name = String::forData ("EnumSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("enum"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_EnumSpecifier_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_EnumeratorList ();
        entry->assignment_func = cpp_EnumSpecifier_set_enumeratorList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_enum_specifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumSpecifier (Cpp_EnumSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EnumSpecifier\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: enum\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    dump_cpp_EnumeratorList (el->enumeratorList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static void
cpp_EnumeratorList_EnumeratorList_set_enumeratorDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumeratorList ||
             __el->cpp_element_type != CppElement::_EnumeratorDefinition);

    Cpp_EnumeratorList * const &_self = static_cast <Cpp_EnumeratorList*> (__self);

    abortIf (_self->enumerator_list_type != Cpp_EnumeratorList::_EnumeratorList);

    Cpp_EnumeratorList_EnumeratorList * const &self = static_cast <Cpp_EnumeratorList_EnumeratorList*> (_self);

    self->enumeratorDefinition = static_cast <Cpp_EnumeratorDefinition*> (___el);
}

static void
cpp_EnumeratorList_EnumeratorList_set_enumeratorList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumeratorList ||
             __el->cpp_element_type != CppElement::_EnumeratorList);

    Cpp_EnumeratorList * const &_self = static_cast <Cpp_EnumeratorList*> (__self);

    abortIf (_self->enumerator_list_type != Cpp_EnumeratorList::_EnumeratorList);

    Cpp_EnumeratorList_EnumeratorList * const &self = static_cast <Cpp_EnumeratorList_EnumeratorList*> (_self);

    self->enumeratorList = static_cast <Cpp_EnumeratorList*> (___el);
}

static ParserElement*
cpp_EnumeratorList_EnumeratorList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EnumeratorList_EnumeratorList))) Cpp_EnumeratorList_EnumeratorList;
}

static Ref<Grammar>
create_cpp_EnumeratorList_EnumeratorList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EnumeratorList_EnumeratorList_creation_func));
    grammar->name = String::forData ("EnumeratorList_EnumeratorList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorDefinition ();
        entry->assignment_func = cpp_EnumeratorList_EnumeratorList_set_enumeratorDefinition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_EnumeratorList ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorList ();
        entry->assignment_func = cpp_EnumeratorList_EnumeratorList_set_enumeratorList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumeratorList_EnumeratorList (Cpp_EnumeratorList_EnumeratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumeratorList_EnumeratorList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EnumeratorList_EnumeratorList\n");

    dump_cpp_EnumeratorDefinition (el->enumeratorDefinition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_EnumeratorList (el->enumeratorList, nest_level + 1);

}

static void
cpp_EnumeratorList_EnumeratorDefinition_set_enumeratorDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumeratorList ||
             __el->cpp_element_type != CppElement::_EnumeratorDefinition);

    Cpp_EnumeratorList * const &_self = static_cast <Cpp_EnumeratorList*> (__self);

    abortIf (_self->enumerator_list_type != Cpp_EnumeratorList::_EnumeratorDefinition);

    Cpp_EnumeratorList_EnumeratorDefinition * const &self = static_cast <Cpp_EnumeratorList_EnumeratorDefinition*> (_self);

    self->enumeratorDefinition = static_cast <Cpp_EnumeratorDefinition*> (___el);
}

static ParserElement*
cpp_EnumeratorList_EnumeratorDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EnumeratorList_EnumeratorDefinition))) Cpp_EnumeratorList_EnumeratorDefinition;
}

static Ref<Grammar>
create_cpp_EnumeratorList_EnumeratorDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EnumeratorList_EnumeratorDefinition_creation_func));
    grammar->name = String::forData ("EnumeratorList_EnumeratorDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorDefinition ();
        entry->assignment_func = cpp_EnumeratorList_EnumeratorDefinition_set_enumeratorDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumeratorList_EnumeratorDefinition (Cpp_EnumeratorList_EnumeratorDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumeratorList_EnumeratorDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EnumeratorList_EnumeratorDefinition\n");

    dump_cpp_EnumeratorDefinition (el->enumeratorDefinition, nest_level + 1);

}

static Ref<Grammar>
create_cpp_EnumeratorList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("EnumeratorList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorList_EnumeratorList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorList_EnumeratorDefinition ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumeratorList (Cpp_EnumeratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumeratorList\n");
        return;
    }

    switch (el->enumerator_list_type) {
        case Cpp_EnumeratorList::_EnumeratorList:
            dump_cpp_EnumeratorList_EnumeratorList (static_cast <Cpp_EnumeratorList_EnumeratorList const *> (el), nest_level + 0);
            break;
        case Cpp_EnumeratorList::_EnumeratorDefinition:
            dump_cpp_EnumeratorList_EnumeratorDefinition (static_cast <Cpp_EnumeratorList_EnumeratorDefinition const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_EnumeratorDefinition_Initializer_set_enumerator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumeratorDefinition ||
             __el->cpp_element_type != CppElement::_Enumerator);

    Cpp_EnumeratorDefinition * const &_self = static_cast <Cpp_EnumeratorDefinition*> (__self);

    abortIf (_self->enumerator_definition_type != Cpp_EnumeratorDefinition::_Initializer);

    Cpp_EnumeratorDefinition_Initializer * const &self = static_cast <Cpp_EnumeratorDefinition_Initializer*> (_self);

    self->enumerator = static_cast <Cpp_Enumerator*> (___el);
}

static void
cpp_EnumeratorDefinition_Initializer_set_constantExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumeratorDefinition ||
             __el->cpp_element_type != CppElement::_ConstantExpression);

    Cpp_EnumeratorDefinition * const &_self = static_cast <Cpp_EnumeratorDefinition*> (__self);

    abortIf (_self->enumerator_definition_type != Cpp_EnumeratorDefinition::_Initializer);

    Cpp_EnumeratorDefinition_Initializer * const &self = static_cast <Cpp_EnumeratorDefinition_Initializer*> (_self);

    self->constantExpression = static_cast <Cpp_ConstantExpression*> (___el);
}

static ParserElement*
cpp_EnumeratorDefinition_Initializer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EnumeratorDefinition_Initializer))) Cpp_EnumeratorDefinition_Initializer;
}

static Ref<Grammar>
create_cpp_EnumeratorDefinition_Initializer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EnumeratorDefinition_Initializer_creation_func));
    grammar->name = String::forData ("EnumeratorDefinition_Initializer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Enumerator ();
        entry->assignment_func = cpp_EnumeratorDefinition_Initializer_set_enumerator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_EnumeratorDefinition ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConstantExpression ();
        entry->assignment_func = cpp_EnumeratorDefinition_Initializer_set_constantExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumeratorDefinition_Initializer (Cpp_EnumeratorDefinition_Initializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumeratorDefinition_Initializer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EnumeratorDefinition_Initializer\n");

    dump_cpp_Enumerator (el->enumerator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_ConstantExpression (el->constantExpression, nest_level + 1);

}

static void
cpp_EnumeratorDefinition_Enumerator_set_enumerator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_EnumeratorDefinition ||
             __el->cpp_element_type != CppElement::_Enumerator);

    Cpp_EnumeratorDefinition * const &_self = static_cast <Cpp_EnumeratorDefinition*> (__self);

    abortIf (_self->enumerator_definition_type != Cpp_EnumeratorDefinition::_Enumerator);

    Cpp_EnumeratorDefinition_Enumerator * const &self = static_cast <Cpp_EnumeratorDefinition_Enumerator*> (_self);

    self->enumerator = static_cast <Cpp_Enumerator*> (___el);
}

static ParserElement*
cpp_EnumeratorDefinition_Enumerator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_EnumeratorDefinition_Enumerator))) Cpp_EnumeratorDefinition_Enumerator;
}

static Ref<Grammar>
create_cpp_EnumeratorDefinition_Enumerator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_EnumeratorDefinition_Enumerator_creation_func));
    grammar->name = String::forData ("EnumeratorDefinition_Enumerator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Enumerator ();
        entry->assignment_func = cpp_EnumeratorDefinition_Enumerator_set_enumerator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumeratorDefinition_Enumerator (Cpp_EnumeratorDefinition_Enumerator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumeratorDefinition_Enumerator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("EnumeratorDefinition_Enumerator\n");

    dump_cpp_Enumerator (el->enumerator, nest_level + 1);

}

static Ref<Grammar>
create_cpp_EnumeratorDefinition ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("EnumeratorDefinition");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorDefinition_Initializer ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_EnumeratorDefinition_Enumerator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_EnumeratorDefinition (Cpp_EnumeratorDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) EnumeratorDefinition\n");
        return;
    }

    switch (el->enumerator_definition_type) {
        case Cpp_EnumeratorDefinition::_Initializer:
            dump_cpp_EnumeratorDefinition_Initializer (static_cast <Cpp_EnumeratorDefinition_Initializer const *> (el), nest_level + 0);
            break;
        case Cpp_EnumeratorDefinition::_Enumerator:
            dump_cpp_EnumeratorDefinition_Enumerator (static_cast <Cpp_EnumeratorDefinition_Enumerator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Enumerator_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Enumerator ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_Enumerator * const &self = static_cast <Cpp_Enumerator*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_Enumerator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Enumerator))) Cpp_Enumerator;
}

static Ref<Grammar>
create_cpp_Enumerator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Enumerator_creation_func));
    grammar->name = String::forData ("Enumerator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_Enumerator_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Enumerator (Cpp_Enumerator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Enumerator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Enumerator\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

bool cpp_NamespaceName_match_func (
        Cpp_NamespaceName *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_NamespaceName_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_NamespaceName);

    Cpp_NamespaceName * const &el = static_cast <Cpp_NamespaceName*> (_el);

    return cpp_NamespaceName_match_func (el, parser_control, data);
}

static void
cpp_NamespaceName_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamespaceName ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_NamespaceName * const &self = static_cast <Cpp_NamespaceName*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_NamespaceName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NamespaceName))) Cpp_NamespaceName;
}

static Ref<Grammar>
create_cpp_NamespaceName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NamespaceName_creation_func));
    grammar->name = String::forData ("NamespaceName");
    grammar->match_func = __pargen_cpp_NamespaceName_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_NamespaceName_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamespaceName (Cpp_NamespaceName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamespaceName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NamespaceName\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_NamespaceDefinition_NamedNamespaceDefinition_set_namedNamespaceDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamespaceDefinition ||
             __el->cpp_element_type != CppElement::_NamedNamespaceDefinition);

    Cpp_NamespaceDefinition * const &_self = static_cast <Cpp_NamespaceDefinition*> (__self);

    abortIf (_self->namespace_definition_type != Cpp_NamespaceDefinition::_NamedNamespaceDefinition);

    Cpp_NamespaceDefinition_NamedNamespaceDefinition * const &self = static_cast <Cpp_NamespaceDefinition_NamedNamespaceDefinition*> (_self);

    self->namedNamespaceDefinition = static_cast <Cpp_NamedNamespaceDefinition*> (___el);
}

static ParserElement*
cpp_NamespaceDefinition_NamedNamespaceDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NamespaceDefinition_NamedNamespaceDefinition))) Cpp_NamespaceDefinition_NamedNamespaceDefinition;
}

static Ref<Grammar>
create_cpp_NamespaceDefinition_NamedNamespaceDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NamespaceDefinition_NamedNamespaceDefinition_creation_func));
    grammar->name = String::forData ("NamespaceDefinition_NamedNamespaceDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamedNamespaceDefinition ();
        entry->assignment_func = cpp_NamespaceDefinition_NamedNamespaceDefinition_set_namedNamespaceDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamespaceDefinition_NamedNamespaceDefinition (Cpp_NamespaceDefinition_NamedNamespaceDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamespaceDefinition_NamedNamespaceDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NamespaceDefinition_NamedNamespaceDefinition\n");

    dump_cpp_NamedNamespaceDefinition (el->namedNamespaceDefinition, nest_level + 1);

}

static void
cpp_NamespaceDefinition_UnnamedNamespaceDefinition_set_unnamedNamespaceDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamespaceDefinition ||
             __el->cpp_element_type != CppElement::_UnnamedNamespaceDefinition);

    Cpp_NamespaceDefinition * const &_self = static_cast <Cpp_NamespaceDefinition*> (__self);

    abortIf (_self->namespace_definition_type != Cpp_NamespaceDefinition::_UnnamedNamespaceDefinition);

    Cpp_NamespaceDefinition_UnnamedNamespaceDefinition * const &self = static_cast <Cpp_NamespaceDefinition_UnnamedNamespaceDefinition*> (_self);

    self->unnamedNamespaceDefinition = static_cast <Cpp_UnnamedNamespaceDefinition*> (___el);
}

static ParserElement*
cpp_NamespaceDefinition_UnnamedNamespaceDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NamespaceDefinition_UnnamedNamespaceDefinition))) Cpp_NamespaceDefinition_UnnamedNamespaceDefinition;
}

static Ref<Grammar>
create_cpp_NamespaceDefinition_UnnamedNamespaceDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NamespaceDefinition_UnnamedNamespaceDefinition_creation_func));
    grammar->name = String::forData ("NamespaceDefinition_UnnamedNamespaceDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnnamedNamespaceDefinition ();
        entry->assignment_func = cpp_NamespaceDefinition_UnnamedNamespaceDefinition_set_unnamedNamespaceDefinition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamespaceDefinition_UnnamedNamespaceDefinition (Cpp_NamespaceDefinition_UnnamedNamespaceDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamespaceDefinition_UnnamedNamespaceDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NamespaceDefinition_UnnamedNamespaceDefinition\n");

    dump_cpp_UnnamedNamespaceDefinition (el->unnamedNamespaceDefinition, nest_level + 1);

}

static Ref<Grammar>
create_cpp_NamespaceDefinition ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("NamespaceDefinition");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NamespaceDefinition_NamedNamespaceDefinition ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_NamespaceDefinition_UnnamedNamespaceDefinition ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamespaceDefinition (Cpp_NamespaceDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamespaceDefinition\n");
        return;
    }

    switch (el->namespace_definition_type) {
        case Cpp_NamespaceDefinition::_NamedNamespaceDefinition:
            dump_cpp_NamespaceDefinition_NamedNamespaceDefinition (static_cast <Cpp_NamespaceDefinition_NamedNamespaceDefinition const *> (el), nest_level + 0);
            break;
        case Cpp_NamespaceDefinition::_UnnamedNamespaceDefinition:
            dump_cpp_NamespaceDefinition_UnnamedNamespaceDefinition (static_cast <Cpp_NamespaceDefinition_UnnamedNamespaceDefinition const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_NamedNamespaceDefinition_accept_func (
        Cpp_NamedNamespaceDefinition *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_NamedNamespaceDefinition_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_NamedNamespaceDefinition_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_NamedNamespaceDefinition);

    Cpp_NamedNamespaceDefinition * const &el = static_cast <Cpp_NamedNamespaceDefinition*> (_el);

    cpp_NamedNamespaceDefinition_accept_func (el, parser_control, data);
}

static void
cpp_NamedNamespaceDefinition_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamedNamespaceDefinition ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_NamedNamespaceDefinition * const &self = static_cast <Cpp_NamedNamespaceDefinition*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

bool cpp_NamedNamespaceDefinition_Identifier_accept_func (
        Cpp_NamedNamespaceDefinition *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_NamedNamespaceDefinition_Identifier_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_NamedNamespaceDefinition);

    Cpp_NamedNamespaceDefinition * const &el =
            static_cast <Cpp_NamedNamespaceDefinition*> (_el);

    return cpp_NamedNamespaceDefinition_Identifier_accept_func (el, parser_control, data);
}

static void
cpp_NamedNamespaceDefinition_set_namespaceBody (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamedNamespaceDefinition ||
             __el->cpp_element_type != CppElement::_NamespaceBody);

    Cpp_NamedNamespaceDefinition * const &self = static_cast <Cpp_NamedNamespaceDefinition*> (__self);

    self->namespaceBody = static_cast <Cpp_NamespaceBody*> (___el);
}

static ParserElement*
cpp_NamedNamespaceDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NamedNamespaceDefinition))) Cpp_NamedNamespaceDefinition;
}

static Ref<Grammar>
create_cpp_NamedNamespaceDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NamedNamespaceDefinition_creation_func));
    grammar->name = String::forData ("NamedNamespaceDefinition");
    grammar->accept_func = __pargen_cpp_NamedNamespaceDefinition_accept_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("namespace"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_NamedNamespaceDefinition_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_NamedNamespaceDefinition_Identifier_accept_func;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceBody ();
        entry->assignment_func = cpp_NamedNamespaceDefinition_set_namespaceBody;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamedNamespaceDefinition (Cpp_NamedNamespaceDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamedNamespaceDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NamedNamespaceDefinition\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: namespace\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    dump_cpp_NamespaceBody (el->namespaceBody, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static void
cpp_UnnamedNamespaceDefinition_set_namespaceBody (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UnnamedNamespaceDefinition ||
             __el->cpp_element_type != CppElement::_NamespaceBody);

    Cpp_UnnamedNamespaceDefinition * const &self = static_cast <Cpp_UnnamedNamespaceDefinition*> (__self);

    self->namespaceBody = static_cast <Cpp_NamespaceBody*> (___el);
}

static ParserElement*
cpp_UnnamedNamespaceDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UnnamedNamespaceDefinition))) Cpp_UnnamedNamespaceDefinition;
}

static Ref<Grammar>
create_cpp_UnnamedNamespaceDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UnnamedNamespaceDefinition_creation_func));
    grammar->name = String::forData ("UnnamedNamespaceDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("namespace"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceBody ();
        entry->assignment_func = cpp_UnnamedNamespaceDefinition_set_namespaceBody;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UnnamedNamespaceDefinition (Cpp_UnnamedNamespaceDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UnnamedNamespaceDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UnnamedNamespaceDefinition\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: namespace\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    dump_cpp_NamespaceBody (el->namespaceBody, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static void
cpp_NamespaceBody_set_declarationSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamespaceBody ||
             __el->cpp_element_type != CppElement::_DeclarationSeq);

    Cpp_NamespaceBody * const &self = static_cast <Cpp_NamespaceBody*> (__self);

    self->declarationSeq = static_cast <Cpp_DeclarationSeq*> (___el);
}

static ParserElement*
cpp_NamespaceBody_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NamespaceBody))) Cpp_NamespaceBody;
}

static Ref<Grammar>
create_cpp_NamespaceBody ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NamespaceBody_creation_func));
    grammar->name = String::forData ("NamespaceBody");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_DeclarationSeq ();
        entry->assignment_func = cpp_NamespaceBody_set_declarationSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamespaceBody (Cpp_NamespaceBody const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamespaceBody\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NamespaceBody\n");

    dump_cpp_DeclarationSeq (el->declarationSeq, nest_level + 1);

}

static void
cpp_NamespaceAliasDefinition_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamespaceAliasDefinition ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_NamespaceAliasDefinition * const &self = static_cast <Cpp_NamespaceAliasDefinition*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_NamespaceAliasDefinition_set_qualifiedNamespaceSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_NamespaceAliasDefinition ||
             __el->cpp_element_type != CppElement::_QualifiedNamespaceSpecifier);

    Cpp_NamespaceAliasDefinition * const &self = static_cast <Cpp_NamespaceAliasDefinition*> (__self);

    self->qualifiedNamespaceSpecifier = static_cast <Cpp_QualifiedNamespaceSpecifier*> (___el);
}

static ParserElement*
cpp_NamespaceAliasDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_NamespaceAliasDefinition))) Cpp_NamespaceAliasDefinition;
}

static Ref<Grammar>
create_cpp_NamespaceAliasDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_NamespaceAliasDefinition_creation_func));
    grammar->name = String::forData ("NamespaceAliasDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("namespace"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_NamespaceAliasDefinition_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_QualifiedNamespaceSpecifier ();
        entry->assignment_func = cpp_NamespaceAliasDefinition_set_qualifiedNamespaceSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_NamespaceAliasDefinition (Cpp_NamespaceAliasDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) NamespaceAliasDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("NamespaceAliasDefinition\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: namespace\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_QualifiedNamespaceSpecifier (el->qualifiedNamespaceSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_QualifiedNamespaceSpecifier_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedNamespaceSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_QualifiedNamespaceSpecifier * const &self = static_cast <Cpp_QualifiedNamespaceSpecifier*> (__self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_QualifiedNamespaceSpecifier_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedNamespaceSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_QualifiedNamespaceSpecifier * const &self = static_cast <Cpp_QualifiedNamespaceSpecifier*> (__self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_QualifiedNamespaceSpecifier_set_namespaceName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_QualifiedNamespaceSpecifier ||
             __el->cpp_element_type != CppElement::_NamespaceName);

    Cpp_QualifiedNamespaceSpecifier * const &self = static_cast <Cpp_QualifiedNamespaceSpecifier*> (__self);

    self->namespaceName = static_cast <Cpp_NamespaceName*> (___el);
}

static ParserElement*
cpp_QualifiedNamespaceSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_QualifiedNamespaceSpecifier))) Cpp_QualifiedNamespaceSpecifier;
}

static Ref<Grammar>
create_cpp_QualifiedNamespaceSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_QualifiedNamespaceSpecifier_creation_func));
    grammar->name = String::forData ("QualifiedNamespaceSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_QualifiedNamespaceSpecifier_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_QualifiedNamespaceSpecifier_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceName ();
        entry->assignment_func = cpp_QualifiedNamespaceSpecifier_set_namespaceName;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_QualifiedNamespaceSpecifier (Cpp_QualifiedNamespaceSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) QualifiedNamespaceSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("QualifiedNamespaceSpecifier\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_NamespaceName (el->namespaceName, nest_level + 1);

}

static void
cpp_UsingDeclaration_NestedNameSpecifier_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDeclaration ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_UsingDeclaration * const &_self = static_cast <Cpp_UsingDeclaration*> (__self);

    abortIf (_self->using_declaration_type != Cpp_UsingDeclaration::_NestedNameSpecifier);

    Cpp_UsingDeclaration_NestedNameSpecifier * const &self = static_cast <Cpp_UsingDeclaration_NestedNameSpecifier*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_UsingDeclaration_NestedNameSpecifier_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDeclaration ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_UsingDeclaration * const &_self = static_cast <Cpp_UsingDeclaration*> (__self);

    abortIf (_self->using_declaration_type != Cpp_UsingDeclaration::_NestedNameSpecifier);

    Cpp_UsingDeclaration_NestedNameSpecifier * const &self = static_cast <Cpp_UsingDeclaration_NestedNameSpecifier*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_UsingDeclaration_NestedNameSpecifier_set_unqualifiedId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDeclaration ||
             __el->cpp_element_type != CppElement::_UnqualifiedId);

    Cpp_UsingDeclaration * const &_self = static_cast <Cpp_UsingDeclaration*> (__self);

    abortIf (_self->using_declaration_type != Cpp_UsingDeclaration::_NestedNameSpecifier);

    Cpp_UsingDeclaration_NestedNameSpecifier * const &self = static_cast <Cpp_UsingDeclaration_NestedNameSpecifier*> (_self);

    self->unqualifiedId = static_cast <Cpp_UnqualifiedId*> (___el);
}

static ParserElement*
cpp_UsingDeclaration_NestedNameSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UsingDeclaration_NestedNameSpecifier))) Cpp_UsingDeclaration_NestedNameSpecifier;
}

static Ref<Grammar>
create_cpp_UsingDeclaration_NestedNameSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UsingDeclaration_NestedNameSpecifier_creation_func));
    grammar->name = String::forData ("UsingDeclaration_NestedNameSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("using"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_UsingDeclaration_NestedNameSpecifier_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_UsingDeclaration_NestedNameSpecifier_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId ();
        entry->assignment_func = cpp_UsingDeclaration_NestedNameSpecifier_set_unqualifiedId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UsingDeclaration_NestedNameSpecifier (Cpp_UsingDeclaration_NestedNameSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UsingDeclaration_NestedNameSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UsingDeclaration_NestedNameSpecifier\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: using\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: typename_opt\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_UnqualifiedId (el->unqualifiedId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_UsingDeclaration_Using_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDeclaration ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_UsingDeclaration * const &_self = static_cast <Cpp_UsingDeclaration*> (__self);

    abortIf (_self->using_declaration_type != Cpp_UsingDeclaration::_Using);

    Cpp_UsingDeclaration_Using * const &self = static_cast <Cpp_UsingDeclaration_Using*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_UsingDeclaration_Using_set_unqualifiedId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDeclaration ||
             __el->cpp_element_type != CppElement::_UnqualifiedId);

    Cpp_UsingDeclaration * const &_self = static_cast <Cpp_UsingDeclaration*> (__self);

    abortIf (_self->using_declaration_type != Cpp_UsingDeclaration::_Using);

    Cpp_UsingDeclaration_Using * const &self = static_cast <Cpp_UsingDeclaration_Using*> (_self);

    self->unqualifiedId = static_cast <Cpp_UnqualifiedId*> (___el);
}

static ParserElement*
cpp_UsingDeclaration_Using_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UsingDeclaration_Using))) Cpp_UsingDeclaration_Using;
}

static Ref<Grammar>
create_cpp_UsingDeclaration_Using ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UsingDeclaration_Using_creation_func));
    grammar->name = String::forData ("UsingDeclaration_Using");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("using"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_UsingDeclaration_Using_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId ();
        entry->assignment_func = cpp_UsingDeclaration_Using_set_unqualifiedId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UsingDeclaration_Using (Cpp_UsingDeclaration_Using const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UsingDeclaration_Using\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UsingDeclaration_Using\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: using\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_UnqualifiedId (el->unqualifiedId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static Ref<Grammar>
create_cpp_UsingDeclaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("UsingDeclaration");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UsingDeclaration_NestedNameSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_UsingDeclaration_Using ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UsingDeclaration (Cpp_UsingDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UsingDeclaration\n");
        return;
    }

    switch (el->using_declaration_type) {
        case Cpp_UsingDeclaration::_NestedNameSpecifier:
            dump_cpp_UsingDeclaration_NestedNameSpecifier (static_cast <Cpp_UsingDeclaration_NestedNameSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_UsingDeclaration::_Using:
            dump_cpp_UsingDeclaration_Using (static_cast <Cpp_UsingDeclaration_Using const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_UsingDirective_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDirective ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_UsingDirective * const &self = static_cast <Cpp_UsingDirective*> (__self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_UsingDirective_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDirective ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_UsingDirective * const &self = static_cast <Cpp_UsingDirective*> (__self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_UsingDirective_set_namespaceName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_UsingDirective ||
             __el->cpp_element_type != CppElement::_NamespaceName);

    Cpp_UsingDirective * const &self = static_cast <Cpp_UsingDirective*> (__self);

    self->namespaceName = static_cast <Cpp_NamespaceName*> (___el);
}

static ParserElement*
cpp_UsingDirective_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_UsingDirective))) Cpp_UsingDirective;
}

static Ref<Grammar>
create_cpp_UsingDirective ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_UsingDirective_creation_func));
    grammar->name = String::forData ("UsingDirective");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("using"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("namespace"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_UsingDirective_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_UsingDirective_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_NamespaceName ();
        entry->assignment_func = cpp_UsingDirective_set_namespaceName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_UsingDirective (Cpp_UsingDirective const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) UsingDirective\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("UsingDirective\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: using\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: namespace\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_NamespaceName (el->namespaceName, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_AsmDefinition_set_stringLiteral (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_AsmDefinition ||
             __el->cpp_element_type != CppElement::_StringLiteral);

    Cpp_AsmDefinition * const &self = static_cast <Cpp_AsmDefinition*> (__self);

    self->stringLiteral = static_cast <Cpp_StringLiteral*> (___el);
}

static ParserElement*
cpp_AsmDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AsmDefinition))) Cpp_AsmDefinition;
}

static Ref<Grammar>
create_cpp_AsmDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AsmDefinition_creation_func));
    grammar->name = String::forData ("AsmDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("asm"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_StringLiteral ();
        entry->assignment_func = cpp_AsmDefinition_set_stringLiteral;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AsmDefinition (Cpp_AsmDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AsmDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AsmDefinition\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: asm\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_StringLiteral (el->stringLiteral, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_LinkageSpecification_DeclarationSeq_set_stringLiteral (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LinkageSpecification ||
             __el->cpp_element_type != CppElement::_StringLiteral);

    Cpp_LinkageSpecification * const &_self = static_cast <Cpp_LinkageSpecification*> (__self);

    abortIf (_self->linkage_specification_type != Cpp_LinkageSpecification::_DeclarationSeq);

    Cpp_LinkageSpecification_DeclarationSeq * const &self = static_cast <Cpp_LinkageSpecification_DeclarationSeq*> (_self);

    self->stringLiteral = static_cast <Cpp_StringLiteral*> (___el);
}

static void
cpp_LinkageSpecification_DeclarationSeq_set_declarationSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LinkageSpecification ||
             __el->cpp_element_type != CppElement::_DeclarationSeq);

    Cpp_LinkageSpecification * const &_self = static_cast <Cpp_LinkageSpecification*> (__self);

    abortIf (_self->linkage_specification_type != Cpp_LinkageSpecification::_DeclarationSeq);

    Cpp_LinkageSpecification_DeclarationSeq * const &self = static_cast <Cpp_LinkageSpecification_DeclarationSeq*> (_self);

    self->declarationSeq = static_cast <Cpp_DeclarationSeq*> (___el);
}

static ParserElement*
cpp_LinkageSpecification_DeclarationSeq_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LinkageSpecification_DeclarationSeq))) Cpp_LinkageSpecification_DeclarationSeq;
}

static Ref<Grammar>
create_cpp_LinkageSpecification_DeclarationSeq ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LinkageSpecification_DeclarationSeq_creation_func));
    grammar->name = String::forData ("LinkageSpecification_DeclarationSeq");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("extern"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_StringLiteral ();
        entry->assignment_func = cpp_LinkageSpecification_DeclarationSeq_set_stringLiteral;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_LinkageSpecification ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 2;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_DeclarationSeq ();
        entry->assignment_func = cpp_LinkageSpecification_DeclarationSeq_set_declarationSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LinkageSpecification_DeclarationSeq (Cpp_LinkageSpecification_DeclarationSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LinkageSpecification_DeclarationSeq\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LinkageSpecification_DeclarationSeq\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: extern\n");
    dump_cpp_StringLiteral (el->stringLiteral, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    dump_cpp_DeclarationSeq (el->declarationSeq, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static void
cpp_LinkageSpecification_Declaration_set_stringLiteral (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LinkageSpecification ||
             __el->cpp_element_type != CppElement::_StringLiteral);

    Cpp_LinkageSpecification * const &_self = static_cast <Cpp_LinkageSpecification*> (__self);

    abortIf (_self->linkage_specification_type != Cpp_LinkageSpecification::_Declaration);

    Cpp_LinkageSpecification_Declaration * const &self = static_cast <Cpp_LinkageSpecification_Declaration*> (_self);

    self->stringLiteral = static_cast <Cpp_StringLiteral*> (___el);
}

static void
cpp_LinkageSpecification_Declaration_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_LinkageSpecification ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_LinkageSpecification * const &_self = static_cast <Cpp_LinkageSpecification*> (__self);

    abortIf (_self->linkage_specification_type != Cpp_LinkageSpecification::_Declaration);

    Cpp_LinkageSpecification_Declaration * const &self = static_cast <Cpp_LinkageSpecification_Declaration*> (_self);

    self->declaration = static_cast <Cpp_Declaration*> (___el);
}

static ParserElement*
cpp_LinkageSpecification_Declaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_LinkageSpecification_Declaration))) Cpp_LinkageSpecification_Declaration;
}

static Ref<Grammar>
create_cpp_LinkageSpecification_Declaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_LinkageSpecification_Declaration_creation_func));
    grammar->name = String::forData ("LinkageSpecification_Declaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("extern"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_StringLiteral ();
        entry->assignment_func = cpp_LinkageSpecification_Declaration_set_stringLiteral;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_LinkageSpecification_Declaration_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LinkageSpecification_Declaration (Cpp_LinkageSpecification_Declaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LinkageSpecification_Declaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("LinkageSpecification_Declaration\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: extern\n");
    dump_cpp_StringLiteral (el->stringLiteral, nest_level + 1);

    dump_cpp_Declaration (el->declaration, nest_level + 1);

}

static Ref<Grammar>
create_cpp_LinkageSpecification ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("LinkageSpecification");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LinkageSpecification_DeclarationSeq ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_LinkageSpecification_Declaration ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_LinkageSpecification (Cpp_LinkageSpecification const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) LinkageSpecification\n");
        return;
    }

    switch (el->linkage_specification_type) {
        case Cpp_LinkageSpecification::_DeclarationSeq:
            dump_cpp_LinkageSpecification_DeclarationSeq (static_cast <Cpp_LinkageSpecification_DeclarationSeq const *> (el), nest_level + 0);
            break;
        case Cpp_LinkageSpecification::_Declaration:
            dump_cpp_LinkageSpecification_Declaration (static_cast <Cpp_LinkageSpecification_Declaration const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_InitDeclaratorList_InitDeclaratorList_set_initDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitDeclaratorList ||
             __el->cpp_element_type != CppElement::_InitDeclarator);

    Cpp_InitDeclaratorList * const &_self = static_cast <Cpp_InitDeclaratorList*> (__self);

    abortIf (_self->init_declarator_list_type != Cpp_InitDeclaratorList::_InitDeclaratorList);

    Cpp_InitDeclaratorList_InitDeclaratorList * const &self = static_cast <Cpp_InitDeclaratorList_InitDeclaratorList*> (_self);

    self->initDeclarator = static_cast <Cpp_InitDeclarator*> (___el);
}

bool cpp_reset_declarator (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_reset_declarator (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_reset_declarator (_el, parser_control, data);
}

static void
cpp_InitDeclaratorList_InitDeclaratorList_set_initDeclaratorList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitDeclaratorList ||
             __el->cpp_element_type != CppElement::_InitDeclaratorList);

    Cpp_InitDeclaratorList * const &_self = static_cast <Cpp_InitDeclaratorList*> (__self);

    abortIf (_self->init_declarator_list_type != Cpp_InitDeclaratorList::_InitDeclaratorList);

    Cpp_InitDeclaratorList_InitDeclaratorList * const &self = static_cast <Cpp_InitDeclaratorList_InitDeclaratorList*> (_self);

    self->initDeclaratorList = static_cast <Cpp_InitDeclaratorList*> (___el);
}

static ParserElement*
cpp_InitDeclaratorList_InitDeclaratorList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitDeclaratorList_InitDeclaratorList))) Cpp_InitDeclaratorList_InitDeclaratorList;
}

static Ref<Grammar>
create_cpp_InitDeclaratorList_InitDeclaratorList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitDeclaratorList_InitDeclaratorList_creation_func));
    grammar->name = String::forData ("InitDeclaratorList_InitDeclaratorList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitDeclarator ();
        entry->assignment_func = cpp_InitDeclaratorList_InitDeclaratorList_set_initDeclarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_InitDeclaratorList ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_reset_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitDeclaratorList ();
        entry->assignment_func = cpp_InitDeclaratorList_InitDeclaratorList_set_initDeclaratorList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitDeclaratorList_InitDeclaratorList (Cpp_InitDeclaratorList_InitDeclaratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitDeclaratorList_InitDeclaratorList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitDeclaratorList_InitDeclaratorList\n");

    dump_cpp_InitDeclarator (el->initDeclarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_InitDeclaratorList (el->initDeclaratorList, nest_level + 1);

}

static void
cpp_InitDeclaratorList_InitDeclarator_set_initDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitDeclaratorList ||
             __el->cpp_element_type != CppElement::_InitDeclarator);

    Cpp_InitDeclaratorList * const &_self = static_cast <Cpp_InitDeclaratorList*> (__self);

    abortIf (_self->init_declarator_list_type != Cpp_InitDeclaratorList::_InitDeclarator);

    Cpp_InitDeclaratorList_InitDeclarator * const &self = static_cast <Cpp_InitDeclaratorList_InitDeclarator*> (_self);

    self->initDeclarator = static_cast <Cpp_InitDeclarator*> (___el);
}

static ParserElement*
cpp_InitDeclaratorList_InitDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitDeclaratorList_InitDeclarator))) Cpp_InitDeclaratorList_InitDeclarator;
}

static Ref<Grammar>
create_cpp_InitDeclaratorList_InitDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitDeclaratorList_InitDeclarator_creation_func));
    grammar->name = String::forData ("InitDeclaratorList_InitDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitDeclarator ();
        entry->assignment_func = cpp_InitDeclaratorList_InitDeclarator_set_initDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitDeclaratorList_InitDeclarator (Cpp_InitDeclaratorList_InitDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitDeclaratorList_InitDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitDeclaratorList_InitDeclarator\n");

    dump_cpp_InitDeclarator (el->initDeclarator, nest_level + 1);

}

static Ref<Grammar>
create_cpp_InitDeclaratorList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("InitDeclaratorList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InitDeclaratorList_InitDeclaratorList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InitDeclaratorList_InitDeclarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitDeclaratorList (Cpp_InitDeclaratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitDeclaratorList\n");
        return;
    }

    switch (el->init_declarator_list_type) {
        case Cpp_InitDeclaratorList::_InitDeclaratorList:
            dump_cpp_InitDeclaratorList_InitDeclaratorList (static_cast <Cpp_InitDeclaratorList_InitDeclaratorList const *> (el), nest_level + 0);
            break;
        case Cpp_InitDeclaratorList::_InitDeclarator:
            dump_cpp_InitDeclaratorList_InitDeclarator (static_cast <Cpp_InitDeclaratorList_InitDeclarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_InitDeclarator_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitDeclarator ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_InitDeclarator * const &self = static_cast <Cpp_InitDeclarator*> (__self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

bool cpp_begin_initializer (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_initializer (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_initializer (_el, parser_control, data);
}

static void
cpp_InitDeclarator_set_initializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitDeclarator ||
             __el->cpp_element_type != CppElement::_Initializer);

    Cpp_InitDeclarator * const &self = static_cast <Cpp_InitDeclarator*> (__self);

    self->initializer = static_cast <Cpp_Initializer*> (___el);
}

bool cpp_accept_init_declarator_initializer (
        Cpp_InitDeclarator *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_init_declarator_initializer (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_InitDeclarator);

    Cpp_InitDeclarator * const &el =
            static_cast <Cpp_InitDeclarator*> (_el);

    return cpp_accept_init_declarator_initializer (el, parser_control, data);
}

static ParserElement*
cpp_InitDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitDeclarator))) Cpp_InitDeclarator;
}

static Ref<Grammar>
create_cpp_InitDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitDeclarator_creation_func));
    grammar->name = String::forData ("InitDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_Declarator ();
        entry->assignment_func = cpp_InitDeclarator_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_initializer;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Initializer ();
        entry->assignment_func = cpp_InitDeclarator_set_initializer;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_init_declarator_initializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitDeclarator (Cpp_InitDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitDeclarator\n");

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    dump_cpp_Initializer (el->initializer, nest_level + 1);

}

static void
cpp_Declarator_PtrOperator_set_ptrOperator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declarator ||
             __el->cpp_element_type != CppElement::_PtrOperator);

    Cpp_Declarator * const &_self = static_cast <Cpp_Declarator*> (__self);

    abortIf (_self->declarator_type != Cpp_Declarator::_PtrOperator);

    Cpp_Declarator_PtrOperator * const &self = static_cast <Cpp_Declarator_PtrOperator*> (_self);

    self->ptrOperator = static_cast <Cpp_PtrOperator*> (___el);
}

static void
cpp_Declarator_PtrOperator_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declarator ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_Declarator * const &_self = static_cast <Cpp_Declarator*> (__self);

    abortIf (_self->declarator_type != Cpp_Declarator::_PtrOperator);

    Cpp_Declarator_PtrOperator * const &self = static_cast <Cpp_Declarator_PtrOperator*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static ParserElement*
cpp_Declarator_PtrOperator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declarator_PtrOperator))) Cpp_Declarator_PtrOperator;
}

static Ref<Grammar>
create_cpp_Declarator_PtrOperator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declarator_PtrOperator_creation_func));
    grammar->name = String::forData ("Declarator_PtrOperator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PtrOperator ();
        entry->assignment_func = cpp_Declarator_PtrOperator_set_ptrOperator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Declarator ();
        entry->assignment_func = cpp_Declarator_PtrOperator_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declarator_PtrOperator (Cpp_Declarator_PtrOperator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declarator_PtrOperator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declarator_PtrOperator\n");

    dump_cpp_PtrOperator (el->ptrOperator, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

}

static void
cpp_Declarator_DirectDeclarator_set_directDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Declarator ||
             __el->cpp_element_type != CppElement::_DirectDeclarator);

    Cpp_Declarator * const &_self = static_cast <Cpp_Declarator*> (__self);

    abortIf (_self->declarator_type != Cpp_Declarator::_DirectDeclarator);

    Cpp_Declarator_DirectDeclarator * const &self = static_cast <Cpp_Declarator_DirectDeclarator*> (_self);

    self->directDeclarator = static_cast <Cpp_DirectDeclarator*> (___el);
}

static ParserElement*
cpp_Declarator_DirectDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Declarator_DirectDeclarator))) Cpp_Declarator_DirectDeclarator;
}

static Ref<Grammar>
create_cpp_Declarator_DirectDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Declarator_DirectDeclarator_creation_func));
    grammar->name = String::forData ("Declarator_DirectDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DirectDeclarator ();
        entry->assignment_func = cpp_Declarator_DirectDeclarator_set_directDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declarator_DirectDeclarator (Cpp_Declarator_DirectDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declarator_DirectDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Declarator_DirectDeclarator\n");

    dump_cpp_DirectDeclarator (el->directDeclarator, nest_level + 1);

}

static Ref<Grammar>
create_cpp_Declarator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Declarator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declarator_PtrOperator ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Declarator_DirectDeclarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Declarator (Cpp_Declarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Declarator\n");
        return;
    }

    switch (el->declarator_type) {
        case Cpp_Declarator::_PtrOperator:
            dump_cpp_Declarator_PtrOperator (static_cast <Cpp_Declarator_PtrOperator const *> (el), nest_level + 0);
            break;
        case Cpp_Declarator::_DirectDeclarator:
            dump_cpp_Declarator_DirectDeclarator (static_cast <Cpp_Declarator_DirectDeclarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_Scruffy_Declarator_Common_begin_func (
        void *data);

static Ref<Grammar>
create_cpp_Scruffy_Declarator_Common ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_Declarator_Common");
    grammar->aliased_grammar = create_cpp_Declarator ();

    grammar->begin_func = cpp_Scruffy_Declarator_Common_begin_func;

    return grammar.ptr ();
}

bool cpp_Scruffy_Declarator_match_func (
        Cpp_Declarator *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_Scruffy_Declarator_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Declarator);

    Cpp_Declarator * const &el = static_cast <Cpp_Declarator*> (_el);

    return cpp_Scruffy_Declarator_match_func (el, parser_control, data);
}

static Ref<Grammar>
create_cpp_Scruffy_Declarator ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_Declarator");
    grammar->aliased_grammar = create_cpp_Scruffy_Declarator_Common ();

    grammar->match_func = __pargen_cpp_Scruffy_Declarator_match_func;

    return grammar.ptr ();
}

bool cpp_Scruffy_AbstractDeclarator_match_func (
        Cpp_Declarator *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_Scruffy_AbstractDeclarator_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Declarator);

    Cpp_Declarator * const &el = static_cast <Cpp_Declarator*> (_el);

    return cpp_Scruffy_AbstractDeclarator_match_func (el, parser_control, data);
}

static Ref<Grammar>
create_cpp_Scruffy_AbstractDeclarator ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_AbstractDeclarator");
    grammar->aliased_grammar = create_cpp_Scruffy_Declarator_Common ();

    grammar->match_func = __pargen_cpp_Scruffy_AbstractDeclarator_match_func;

    return grammar.ptr ();
}

bool cpp_Scruffy_AnyDeclarator_match_func (
        Cpp_Declarator *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_Scruffy_AnyDeclarator_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Declarator);

    Cpp_Declarator * const &el = static_cast <Cpp_Declarator*> (_el);

    return cpp_Scruffy_AnyDeclarator_match_func (el, parser_control, data);
}

static Ref<Grammar>
create_cpp_Scruffy_AnyDeclarator ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_AnyDeclarator");
    grammar->aliased_grammar = create_cpp_Scruffy_Declarator_Common ();

    grammar->match_func = __pargen_cpp_Scruffy_AnyDeclarator_match_func;

    return grammar.ptr ();
}

static void
cpp_DirectDeclarator_Identifier_set_declaratorId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_DeclaratorId);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Identifier);

    Cpp_DirectDeclarator_Identifier * const &self = static_cast <Cpp_DirectDeclarator_Identifier*> (_self);

    self->declaratorId = static_cast <Cpp_DeclaratorId*> (___el);
}

static ParserElement*
cpp_DirectDeclarator_Identifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DirectDeclarator_Identifier))) Cpp_DirectDeclarator_Identifier;
}

static Ref<Grammar>
create_cpp_DirectDeclarator_Identifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DirectDeclarator_Identifier_creation_func));
    grammar->name = String::forData ("DirectDeclarator_Identifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_DeclaratorId ();
        entry->assignment_func = cpp_DirectDeclarator_Identifier_set_declaratorId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectDeclarator_Identifier (Cpp_DirectDeclarator_Identifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectDeclarator_Identifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DirectDeclarator_Identifier\n");

    dump_cpp_DeclaratorId (el->declaratorId, nest_level + 1);

}

static void
cpp_DirectDeclarator_Function_set_directDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_DirectDeclarator);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Function);

    Cpp_DirectDeclarator_Function * const &self = static_cast <Cpp_DirectDeclarator_Function*> (_self);

    self->directDeclarator = static_cast <Cpp_DirectDeclarator*> (___el);
}

bool cpp_begin_function_parameters (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_function_parameters (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_function_parameters (_el, parser_control, data);
}

static void
cpp_DirectDeclarator_Function_set_parameterDeclarationClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_ParameterDeclarationClause);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Function);

    Cpp_DirectDeclarator_Function * const &self = static_cast <Cpp_DirectDeclarator_Function*> (_self);

    self->parameterDeclarationClause = static_cast <Cpp_ParameterDeclarationClause*> (___el);
}

static void
cpp_DirectDeclarator_Function_set_cvQualifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_CvQualifierSeq);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Function);

    Cpp_DirectDeclarator_Function * const &self = static_cast <Cpp_DirectDeclarator_Function*> (_self);

    self->cvQualifierSeq = static_cast <Cpp_CvQualifierSeq*> (___el);
}

static void
cpp_DirectDeclarator_Function_set_exceptionSpecification (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_ExceptionSpecification);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Function);

    Cpp_DirectDeclarator_Function * const &self = static_cast <Cpp_DirectDeclarator_Function*> (_self);

    self->exceptionSpecification = static_cast <Cpp_ExceptionSpecification*> (___el);
}

bool cpp_accept_function_declarator (
        Cpp_DirectDeclarator *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_function_declarator (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DirectDeclarator);

    Cpp_DirectDeclarator * const &el =
            static_cast <Cpp_DirectDeclarator*> (_el);

    return cpp_accept_function_declarator (el, parser_control, data);
}

static ParserElement*
cpp_DirectDeclarator_Function_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DirectDeclarator_Function))) Cpp_DirectDeclarator_Function;
}

static Ref<Grammar>
create_cpp_DirectDeclarator_Function ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DirectDeclarator_Function_creation_func));
    grammar->name = String::forData ("DirectDeclarator_Function");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_DirectDeclarator ();
        entry->assignment_func = cpp_DirectDeclarator_Function_set_directDeclarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_function_parameters;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationClause ();
        entry->assignment_func = cpp_DirectDeclarator_Function_set_parameterDeclarationClause;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_CvQualifierSeq ();
        entry->assignment_func = cpp_DirectDeclarator_Function_set_cvQualifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExceptionSpecification ();
        entry->assignment_func = cpp_DirectDeclarator_Function_set_exceptionSpecification;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_function_declarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectDeclarator_Function (Cpp_DirectDeclarator_Function const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectDeclarator_Function\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DirectDeclarator_Function\n");

    dump_cpp_DirectDeclarator (el->directDeclarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ParameterDeclarationClause (el->parameterDeclarationClause, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_CvQualifierSeq (el->cvQualifierSeq, nest_level + 1);

    dump_cpp_ExceptionSpecification (el->exceptionSpecification, nest_level + 1);

}

static void
cpp_DirectDeclarator_Array_set_directDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_DirectDeclarator);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Array);

    Cpp_DirectDeclarator_Array * const &self = static_cast <Cpp_DirectDeclarator_Array*> (_self);

    self->directDeclarator = static_cast <Cpp_DirectDeclarator*> (___el);
}

static void
cpp_DirectDeclarator_Array_set_constantExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_ConstantExpression);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Array);

    Cpp_DirectDeclarator_Array * const &self = static_cast <Cpp_DirectDeclarator_Array*> (_self);

    self->constantExpression = static_cast <Cpp_ConstantExpression*> (___el);
}

static ParserElement*
cpp_DirectDeclarator_Array_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DirectDeclarator_Array))) Cpp_DirectDeclarator_Array;
}

static Ref<Grammar>
create_cpp_DirectDeclarator_Array ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DirectDeclarator_Array_creation_func));
    grammar->name = String::forData ("DirectDeclarator_Array");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_DirectDeclarator ();
        entry->assignment_func = cpp_DirectDeclarator_Array_set_directDeclarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ConstantExpression ();
        entry->assignment_func = cpp_DirectDeclarator_Array_set_constantExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectDeclarator_Array (Cpp_DirectDeclarator_Array const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectDeclarator_Array\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DirectDeclarator_Array\n");

    dump_cpp_DirectDeclarator (el->directDeclarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    dump_cpp_ConstantExpression (el->constantExpression, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static void
cpp_DirectDeclarator_Declarator_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DirectDeclarator ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_DirectDeclarator * const &_self = static_cast <Cpp_DirectDeclarator*> (__self);

    abortIf (_self->direct_declarator_type != Cpp_DirectDeclarator::_Declarator);

    Cpp_DirectDeclarator_Declarator * const &self = static_cast <Cpp_DirectDeclarator_Declarator*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static ParserElement*
cpp_DirectDeclarator_Declarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DirectDeclarator_Declarator))) Cpp_DirectDeclarator_Declarator;
}

static Ref<Grammar>
create_cpp_DirectDeclarator_Declarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DirectDeclarator_Declarator_creation_func));
    grammar->name = String::forData ("DirectDeclarator_Declarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declarator ();
        entry->assignment_func = cpp_DirectDeclarator_Declarator_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectDeclarator_Declarator (Cpp_DirectDeclarator_Declarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectDeclarator_Declarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DirectDeclarator_Declarator\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Declarator (el->declarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static Ref<Grammar>
create_cpp_DirectDeclarator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("DirectDeclarator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DirectDeclarator_Identifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DirectDeclarator_Function ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DirectDeclarator_Array ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_DirectDeclarator_Declarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DirectDeclarator (Cpp_DirectDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DirectDeclarator\n");
        return;
    }

    switch (el->direct_declarator_type) {
        case Cpp_DirectDeclarator::_Identifier:
            dump_cpp_DirectDeclarator_Identifier (static_cast <Cpp_DirectDeclarator_Identifier const *> (el), nest_level + 0);
            break;
        case Cpp_DirectDeclarator::_Function:
            dump_cpp_DirectDeclarator_Function (static_cast <Cpp_DirectDeclarator_Function const *> (el), nest_level + 0);
            break;
        case Cpp_DirectDeclarator::_Array:
            dump_cpp_DirectDeclarator_Array (static_cast <Cpp_DirectDeclarator_Array const *> (el), nest_level + 0);
            break;
        case Cpp_DirectDeclarator::_Declarator:
            dump_cpp_DirectDeclarator_Declarator (static_cast <Cpp_DirectDeclarator_Declarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_PtrOperator_Pointer_set_cvQualifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PtrOperator ||
             __el->cpp_element_type != CppElement::_CvQualifierSeq);

    Cpp_PtrOperator * const &_self = static_cast <Cpp_PtrOperator*> (__self);

    abortIf (_self->ptr_operator_type != Cpp_PtrOperator::_Pointer);

    Cpp_PtrOperator_Pointer * const &self = static_cast <Cpp_PtrOperator_Pointer*> (_self);

    self->cvQualifierSeq = static_cast <Cpp_CvQualifierSeq*> (___el);
}

static ParserElement*
cpp_PtrOperator_Pointer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PtrOperator_Pointer))) Cpp_PtrOperator_Pointer;
}

static Ref<Grammar>
create_cpp_PtrOperator_Pointer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PtrOperator_Pointer_creation_func));
    grammar->name = String::forData ("PtrOperator_Pointer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_CvQualifierSeq ();
        entry->assignment_func = cpp_PtrOperator_Pointer_set_cvQualifierSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PtrOperator_Pointer (Cpp_PtrOperator_Pointer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PtrOperator_Pointer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PtrOperator_Pointer\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: *\n");
    dump_cpp_CvQualifierSeq (el->cvQualifierSeq, nest_level + 1);

}

static ParserElement*
cpp_PtrOperator_Reference_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PtrOperator_Reference))) Cpp_PtrOperator_Reference;
}

static Ref<Grammar>
create_cpp_PtrOperator_Reference ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PtrOperator_Reference_creation_func));
    grammar->name = String::forData ("PtrOperator_Reference");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PtrOperator_Reference (Cpp_PtrOperator_Reference const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PtrOperator_Reference\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PtrOperator_Reference\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: &\n");
}

static void
cpp_PtrOperator_PointerToMember_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PtrOperator ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_PtrOperator * const &_self = static_cast <Cpp_PtrOperator*> (__self);

    abortIf (_self->ptr_operator_type != Cpp_PtrOperator::_PointerToMember);

    Cpp_PtrOperator_PointerToMember * const &self = static_cast <Cpp_PtrOperator_PointerToMember*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_PtrOperator_PointerToMember_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PtrOperator ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_PtrOperator * const &_self = static_cast <Cpp_PtrOperator*> (__self);

    abortIf (_self->ptr_operator_type != Cpp_PtrOperator::_PointerToMember);

    Cpp_PtrOperator_PointerToMember * const &self = static_cast <Cpp_PtrOperator_PointerToMember*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_PtrOperator_PointerToMember_set_cvQualifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_PtrOperator ||
             __el->cpp_element_type != CppElement::_CvQualifierSeq);

    Cpp_PtrOperator * const &_self = static_cast <Cpp_PtrOperator*> (__self);

    abortIf (_self->ptr_operator_type != Cpp_PtrOperator::_PointerToMember);

    Cpp_PtrOperator_PointerToMember * const &self = static_cast <Cpp_PtrOperator_PointerToMember*> (_self);

    self->cvQualifierSeq = static_cast <Cpp_CvQualifierSeq*> (___el);
}

static ParserElement*
cpp_PtrOperator_PointerToMember_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PtrOperator_PointerToMember))) Cpp_PtrOperator_PointerToMember;
}

static Ref<Grammar>
create_cpp_PtrOperator_PointerToMember ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PtrOperator_PointerToMember_creation_func));
    grammar->name = String::forData ("PtrOperator_PointerToMember");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_PtrOperator_PointerToMember_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_PtrOperator_PointerToMember_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_CvQualifierSeq ();
        entry->assignment_func = cpp_PtrOperator_PointerToMember_set_cvQualifierSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PtrOperator_PointerToMember (Cpp_PtrOperator_PointerToMember const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PtrOperator_PointerToMember\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PtrOperator_PointerToMember\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: *\n");
    dump_cpp_CvQualifierSeq (el->cvQualifierSeq, nest_level + 1);

}

static Ref<Grammar>
create_cpp_PtrOperator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("PtrOperator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PtrOperator_Pointer ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PtrOperator_Reference ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_PtrOperator_PointerToMember ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PtrOperator (Cpp_PtrOperator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PtrOperator\n");
        return;
    }

    switch (el->ptr_operator_type) {
        case Cpp_PtrOperator::_Pointer:
            dump_cpp_PtrOperator_Pointer (static_cast <Cpp_PtrOperator_Pointer const *> (el), nest_level + 0);
            break;
        case Cpp_PtrOperator::_Reference:
            dump_cpp_PtrOperator_Reference (static_cast <Cpp_PtrOperator_Reference const *> (el), nest_level + 0);
            break;
        case Cpp_PtrOperator::_PointerToMember:
            dump_cpp_PtrOperator_PointerToMember (static_cast <Cpp_PtrOperator_PointerToMember const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_CvQualifierSeq_set_cvQualifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CvQualifierSeq ||
             __el->cpp_element_type != CppElement::_CvQualifier);

    Cpp_CvQualifierSeq * const &self = static_cast <Cpp_CvQualifierSeq*> (__self);

    self->cvQualifier = static_cast <Cpp_CvQualifier*> (___el);
}

static void
cpp_CvQualifierSeq_set_cvQualifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CvQualifierSeq ||
             __el->cpp_element_type != CppElement::_CvQualifierSeq);

    Cpp_CvQualifierSeq * const &self = static_cast <Cpp_CvQualifierSeq*> (__self);

    self->cvQualifierSeq = static_cast <Cpp_CvQualifierSeq*> (___el);
}

static ParserElement*
cpp_CvQualifierSeq_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CvQualifierSeq))) Cpp_CvQualifierSeq;
}

static Ref<Grammar>
create_cpp_CvQualifierSeq ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CvQualifierSeq_creation_func));
    grammar->name = String::forData ("CvQualifierSeq");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CvQualifier ();
        entry->assignment_func = cpp_CvQualifierSeq_set_cvQualifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_CvQualifierSeq ();
        entry->assignment_func = cpp_CvQualifierSeq_set_cvQualifierSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CvQualifierSeq (Cpp_CvQualifierSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CvQualifierSeq\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CvQualifierSeq\n");

    dump_cpp_CvQualifier (el->cvQualifier, nest_level + 1);

    dump_cpp_CvQualifierSeq (el->cvQualifierSeq, nest_level + 1);

}

static ParserElement*
cpp_CvQualifier_Const_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CvQualifier_Const))) Cpp_CvQualifier_Const;
}

static Ref<Grammar>
create_cpp_CvQualifier_Const ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CvQualifier_Const_creation_func));
    grammar->name = String::forData ("CvQualifier_Const");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("const"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CvQualifier_Const (Cpp_CvQualifier_Const const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CvQualifier_Const\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CvQualifier_Const\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: const\n");
}

static ParserElement*
cpp_CvQualifier_Volatile_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CvQualifier_Volatile))) Cpp_CvQualifier_Volatile;
}

static Ref<Grammar>
create_cpp_CvQualifier_Volatile ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CvQualifier_Volatile_creation_func));
    grammar->name = String::forData ("CvQualifier_Volatile");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("volatile"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CvQualifier_Volatile (Cpp_CvQualifier_Volatile const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CvQualifier_Volatile\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CvQualifier_Volatile\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: volatile\n");
}

static Ref<Grammar>
create_cpp_CvQualifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("CvQualifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_CvQualifier_Const ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_CvQualifier_Volatile ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CvQualifier (Cpp_CvQualifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CvQualifier\n");
        return;
    }

    switch (el->cv_qualifier_type) {
        case Cpp_CvQualifier::_Const:
            dump_cpp_CvQualifier_Const (static_cast <Cpp_CvQualifier_Const const *> (el), nest_level + 0);
            break;
        case Cpp_CvQualifier::_Volatile:
            dump_cpp_CvQualifier_Volatile (static_cast <Cpp_CvQualifier_Volatile const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_DeclaratorId_accept_func (
        Cpp_DeclaratorId *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_DeclaratorId_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_DeclaratorId_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_DeclaratorId);

    Cpp_DeclaratorId * const &el = static_cast <Cpp_DeclaratorId*> (_el);

    cpp_DeclaratorId_accept_func (el, parser_control, data);
}

static void
cpp_DeclaratorId_set_idExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_DeclaratorId ||
             __el->cpp_element_type != CppElement::_IdExpression);

    Cpp_DeclaratorId * const &self = static_cast <Cpp_DeclaratorId*> (__self);

    self->idExpression = static_cast <Cpp_IdExpression*> (___el);
}

static ParserElement*
cpp_DeclaratorId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_DeclaratorId))) Cpp_DeclaratorId;
}

static Ref<Grammar>
create_cpp_DeclaratorId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_DeclaratorId_creation_func));
    grammar->name = String::forData ("DeclaratorId");
    grammar->accept_func = __pargen_cpp_DeclaratorId_accept_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IdExpression ();
        entry->assignment_func = cpp_DeclaratorId_set_idExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_DeclaratorId (Cpp_DeclaratorId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) DeclaratorId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("DeclaratorId\n");

    dump_cpp_IdExpression (el->idExpression, nest_level + 1);

}

static void
cpp_TypeId_set_typeSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeId ||
             __el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_TypeId * const &self = static_cast <Cpp_TypeId*> (__self);

    self->typeSpecifierSeq = static_cast <Cpp_TypeSpecifierSeq*> (___el);
}

static void
cpp_TypeId_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeId ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_TypeId * const &self = static_cast <Cpp_TypeId*> (__self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

bool cpp_accept_declaration (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_declaration (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_accept_declaration (_el, parser_control, data);
}

bool cpp_accept_typeid (
        Cpp_TypeId *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_typeid (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TypeId);

    Cpp_TypeId * const &el =
            static_cast <Cpp_TypeId*> (_el);

    return cpp_accept_typeid (el, parser_control, data);
}

static ParserElement*
cpp_TypeId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeId))) Cpp_TypeId;
}

static Ref<Grammar>
create_cpp_TypeId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeId_creation_func));
    grammar->name = String::forData ("TypeId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_TypeSpecifierSeq ();
        entry->assignment_func = cpp_TypeId_set_typeSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_AbstractDeclarator ();
        entry->assignment_func = cpp_TypeId_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_typeid;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeId (Cpp_TypeId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeId\n");

    dump_cpp_TypeSpecifierSeq (el->typeSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

}

static void
cpp_TypeSpecifierSeq_set_typeSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifierSeq ||
             __el->cpp_element_type != CppElement::_TypeSpecifier);

    Cpp_TypeSpecifierSeq * const &self = static_cast <Cpp_TypeSpecifierSeq*> (__self);

    self->typeSpecifier = static_cast <Cpp_TypeSpecifier*> (___el);
}

static void
cpp_TypeSpecifierSeq_set_typeSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeSpecifierSeq ||
             __el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_TypeSpecifierSeq * const &self = static_cast <Cpp_TypeSpecifierSeq*> (__self);

    self->typeSpecifierSeq = static_cast <Cpp_TypeSpecifierSeq*> (___el);
}

static ParserElement*
cpp_TypeSpecifierSeq_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeSpecifierSeq))) Cpp_TypeSpecifierSeq;
}

static Ref<Grammar>
create_cpp_TypeSpecifierSeq ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeSpecifierSeq_creation_func));
    grammar->name = String::forData ("TypeSpecifierSeq");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeSpecifier ();
        entry->assignment_func = cpp_TypeSpecifierSeq_set_typeSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_TypeSpecifierSeq ();
        entry->assignment_func = cpp_TypeSpecifierSeq_set_typeSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeSpecifierSeq (Cpp_TypeSpecifierSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeSpecifierSeq\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeSpecifierSeq\n");

    dump_cpp_TypeSpecifier (el->typeSpecifier, nest_level + 1);

    dump_cpp_TypeSpecifierSeq (el->typeSpecifierSeq, nest_level + 1);

}

void cpp_Scruffy_TypeSpecifierSeq_begin_func (
        void *data);

void cpp_Scruffy_TypeSpecifierSeq_accept_func (
        Cpp_TypeSpecifierSeq *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_Scruffy_TypeSpecifierSeq_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_Scruffy_TypeSpecifierSeq_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_TypeSpecifierSeq * const &el = static_cast <Cpp_TypeSpecifierSeq*> (_el);

    cpp_Scruffy_TypeSpecifierSeq_accept_func (el, parser_control, data);
}

static Ref<Grammar>
create_cpp_Scruffy_TypeSpecifierSeq ()
{
    static Ref<Grammar_Alias> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Alias);
    grammar->name = String::forData ("Scruffy_TypeSpecifierSeq");
    grammar->aliased_grammar = create_cpp_TypeSpecifierSeq ();

    grammar->begin_func = cpp_Scruffy_TypeSpecifierSeq_begin_func;
    grammar->accept_func = __pargen_cpp_Scruffy_TypeSpecifierSeq_accept_func;

    return grammar.ptr ();
}

static void
cpp_ParameterDeclarationClause_Ellipsis_set_parameterDeclarationList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclarationClause ||
             __el->cpp_element_type != CppElement::_ParameterDeclarationList);

    Cpp_ParameterDeclarationClause * const &_self = static_cast <Cpp_ParameterDeclarationClause*> (__self);

    abortIf (_self->parameter_declaration_clause_type != Cpp_ParameterDeclarationClause::_Ellipsis);

    Cpp_ParameterDeclarationClause_Ellipsis * const &self = static_cast <Cpp_ParameterDeclarationClause_Ellipsis*> (_self);

    self->parameterDeclarationList = static_cast <Cpp_ParameterDeclarationList*> (___el);
}

static ParserElement*
cpp_ParameterDeclarationClause_Ellipsis_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ParameterDeclarationClause_Ellipsis))) Cpp_ParameterDeclarationClause_Ellipsis;
}

static Ref<Grammar>
create_cpp_ParameterDeclarationClause_Ellipsis ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ParameterDeclarationClause_Ellipsis_creation_func));
    grammar->name = String::forData ("ParameterDeclarationClause_Ellipsis");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationList ();
        entry->assignment_func = cpp_ParameterDeclarationClause_Ellipsis_set_parameterDeclarationList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_ParameterDeclarationClause ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("..."));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclarationClause_Ellipsis (Cpp_ParameterDeclarationClause_Ellipsis const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclarationClause_Ellipsis\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ParameterDeclarationClause_Ellipsis\n");

    dump_cpp_ParameterDeclarationList (el->parameterDeclarationList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ...\n");
}

static void
cpp_ParameterDeclarationClause_ParameterDeclarationList_set_parameterDeclarationList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclarationClause ||
             __el->cpp_element_type != CppElement::_ParameterDeclarationList);

    Cpp_ParameterDeclarationClause * const &_self = static_cast <Cpp_ParameterDeclarationClause*> (__self);

    abortIf (_self->parameter_declaration_clause_type != Cpp_ParameterDeclarationClause::_ParameterDeclarationList);

    Cpp_ParameterDeclarationClause_ParameterDeclarationList * const &self = static_cast <Cpp_ParameterDeclarationClause_ParameterDeclarationList*> (_self);

    self->parameterDeclarationList = static_cast <Cpp_ParameterDeclarationList*> (___el);
}

static ParserElement*
cpp_ParameterDeclarationClause_ParameterDeclarationList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ParameterDeclarationClause_ParameterDeclarationList))) Cpp_ParameterDeclarationClause_ParameterDeclarationList;
}

static Ref<Grammar>
create_cpp_ParameterDeclarationClause_ParameterDeclarationList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ParameterDeclarationClause_ParameterDeclarationList_creation_func));
    grammar->name = String::forData ("ParameterDeclarationClause_ParameterDeclarationList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ParameterDeclarationList ();
        entry->assignment_func = cpp_ParameterDeclarationClause_ParameterDeclarationList_set_parameterDeclarationList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("..."));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclarationClause_ParameterDeclarationList (Cpp_ParameterDeclarationClause_ParameterDeclarationList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclarationClause_ParameterDeclarationList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ParameterDeclarationClause_ParameterDeclarationList\n");

    dump_cpp_ParameterDeclarationList (el->parameterDeclarationList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ..._opt\n");
}

static Ref<Grammar>
create_cpp_ParameterDeclarationClause ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ParameterDeclarationClause");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationClause_Ellipsis ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationClause_ParameterDeclarationList ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclarationClause (Cpp_ParameterDeclarationClause const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclarationClause\n");
        return;
    }

    switch (el->parameter_declaration_clause_type) {
        case Cpp_ParameterDeclarationClause::_Ellipsis:
            dump_cpp_ParameterDeclarationClause_Ellipsis (static_cast <Cpp_ParameterDeclarationClause_Ellipsis const *> (el), nest_level + 0);
            break;
        case Cpp_ParameterDeclarationClause::_ParameterDeclarationList:
            dump_cpp_ParameterDeclarationClause_ParameterDeclarationList (static_cast <Cpp_ParameterDeclarationClause_ParameterDeclarationList const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ParameterDeclarationList_ParameterDeclarationList_set_parameterDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclarationList ||
             __el->cpp_element_type != CppElement::_ParameterDeclaration);

    Cpp_ParameterDeclarationList * const &_self = static_cast <Cpp_ParameterDeclarationList*> (__self);

    abortIf (_self->parameter_declaration_list_type != Cpp_ParameterDeclarationList::_ParameterDeclarationList);

    Cpp_ParameterDeclarationList_ParameterDeclarationList * const &self = static_cast <Cpp_ParameterDeclarationList_ParameterDeclarationList*> (_self);

    self->parameterDeclaration = static_cast <Cpp_ParameterDeclaration*> (___el);
}

static void
cpp_ParameterDeclarationList_ParameterDeclarationList_set_parameterDeclarationList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclarationList ||
             __el->cpp_element_type != CppElement::_ParameterDeclarationList);

    Cpp_ParameterDeclarationList * const &_self = static_cast <Cpp_ParameterDeclarationList*> (__self);

    abortIf (_self->parameter_declaration_list_type != Cpp_ParameterDeclarationList::_ParameterDeclarationList);

    Cpp_ParameterDeclarationList_ParameterDeclarationList * const &self = static_cast <Cpp_ParameterDeclarationList_ParameterDeclarationList*> (_self);

    self->parameterDeclarationList = static_cast <Cpp_ParameterDeclarationList*> (___el);
}

static ParserElement*
cpp_ParameterDeclarationList_ParameterDeclarationList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ParameterDeclarationList_ParameterDeclarationList))) Cpp_ParameterDeclarationList_ParameterDeclarationList;
}

static Ref<Grammar>
create_cpp_ParameterDeclarationList_ParameterDeclarationList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ParameterDeclarationList_ParameterDeclarationList_creation_func));
    grammar->name = String::forData ("ParameterDeclarationList_ParameterDeclarationList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclaration ();
        entry->assignment_func = cpp_ParameterDeclarationList_ParameterDeclarationList_set_parameterDeclaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_ParameterDeclarationList ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationList ();
        entry->assignment_func = cpp_ParameterDeclarationList_ParameterDeclarationList_set_parameterDeclarationList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclarationList_ParameterDeclarationList (Cpp_ParameterDeclarationList_ParameterDeclarationList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclarationList_ParameterDeclarationList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ParameterDeclarationList_ParameterDeclarationList\n");

    dump_cpp_ParameterDeclaration (el->parameterDeclaration, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_ParameterDeclarationList (el->parameterDeclarationList, nest_level + 1);

}

static void
cpp_ParameterDeclarationList_ParameterDeclaration_set_parameterDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclarationList ||
             __el->cpp_element_type != CppElement::_ParameterDeclaration);

    Cpp_ParameterDeclarationList * const &_self = static_cast <Cpp_ParameterDeclarationList*> (__self);

    abortIf (_self->parameter_declaration_list_type != Cpp_ParameterDeclarationList::_ParameterDeclaration);

    Cpp_ParameterDeclarationList_ParameterDeclaration * const &self = static_cast <Cpp_ParameterDeclarationList_ParameterDeclaration*> (_self);

    self->parameterDeclaration = static_cast <Cpp_ParameterDeclaration*> (___el);
}

static ParserElement*
cpp_ParameterDeclarationList_ParameterDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ParameterDeclarationList_ParameterDeclaration))) Cpp_ParameterDeclarationList_ParameterDeclaration;
}

static Ref<Grammar>
create_cpp_ParameterDeclarationList_ParameterDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ParameterDeclarationList_ParameterDeclaration_creation_func));
    grammar->name = String::forData ("ParameterDeclarationList_ParameterDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclaration ();
        entry->assignment_func = cpp_ParameterDeclarationList_ParameterDeclaration_set_parameterDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclarationList_ParameterDeclaration (Cpp_ParameterDeclarationList_ParameterDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclarationList_ParameterDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ParameterDeclarationList_ParameterDeclaration\n");

    dump_cpp_ParameterDeclaration (el->parameterDeclaration, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ParameterDeclarationList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ParameterDeclarationList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationList_ParameterDeclarationList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclarationList_ParameterDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclarationList (Cpp_ParameterDeclarationList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclarationList\n");
        return;
    }

    switch (el->parameter_declaration_list_type) {
        case Cpp_ParameterDeclarationList::_ParameterDeclarationList:
            dump_cpp_ParameterDeclarationList_ParameterDeclarationList (static_cast <Cpp_ParameterDeclarationList_ParameterDeclarationList const *> (el), nest_level + 0);
            break;
        case Cpp_ParameterDeclarationList::_ParameterDeclaration:
            dump_cpp_ParameterDeclarationList_ParameterDeclaration (static_cast <Cpp_ParameterDeclarationList_ParameterDeclaration const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_begin_parameter_declaration (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_parameter_declaration (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_parameter_declaration (_el, parser_control, data);
}

static void
cpp_ParameterDeclaration_DeclaratorAssignment_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclaration ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_ParameterDeclaration * const &_self = static_cast <Cpp_ParameterDeclaration*> (__self);

    abortIf (_self->parameter_declaration_type != Cpp_ParameterDeclaration::_DeclaratorAssignment);

    Cpp_ParameterDeclaration_DeclaratorAssignment * const &self = static_cast <Cpp_ParameterDeclaration_DeclaratorAssignment*> (_self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static void
cpp_ParameterDeclaration_DeclaratorAssignment_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclaration ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_ParameterDeclaration * const &_self = static_cast <Cpp_ParameterDeclaration*> (__self);

    abortIf (_self->parameter_declaration_type != Cpp_ParameterDeclaration::_DeclaratorAssignment);

    Cpp_ParameterDeclaration_DeclaratorAssignment * const &self = static_cast <Cpp_ParameterDeclaration_DeclaratorAssignment*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static void
cpp_ParameterDeclaration_DeclaratorAssignment_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclaration ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_ParameterDeclaration * const &_self = static_cast <Cpp_ParameterDeclaration*> (__self);

    abortIf (_self->parameter_declaration_type != Cpp_ParameterDeclaration::_DeclaratorAssignment);

    Cpp_ParameterDeclaration_DeclaratorAssignment * const &self = static_cast <Cpp_ParameterDeclaration_DeclaratorAssignment*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

bool cpp_accept_parameter_declaration_declarator_assignment (
        Cpp_ParameterDeclaration *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_parameter_declaration_declarator_assignment (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ParameterDeclaration);

    Cpp_ParameterDeclaration * const &el =
            static_cast <Cpp_ParameterDeclaration*> (_el);

    return cpp_accept_parameter_declaration_declarator_assignment (el, parser_control, data);
}

static ParserElement*
cpp_ParameterDeclaration_DeclaratorAssignment_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ParameterDeclaration_DeclaratorAssignment))) Cpp_ParameterDeclaration_DeclaratorAssignment;
}

static Ref<Grammar>
create_cpp_ParameterDeclaration_DeclaratorAssignment ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ParameterDeclaration_DeclaratorAssignment_creation_func));
    grammar->name = String::forData ("ParameterDeclaration_DeclaratorAssignment");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_parameter_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_DeclSpecifierSeq ();
        entry->assignment_func = cpp_ParameterDeclaration_DeclaratorAssignment_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_AnyDeclarator ();
        entry->assignment_func = cpp_ParameterDeclaration_DeclaratorAssignment_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_ParameterDeclaration ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 4;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_ParameterDeclaration_DeclaratorAssignment_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_parameter_declaration_declarator_assignment;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclaration_DeclaratorAssignment (Cpp_ParameterDeclaration_DeclaratorAssignment const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclaration_DeclaratorAssignment\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ParameterDeclaration_DeclaratorAssignment\n");

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_ParameterDeclaration_Declarator_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclaration ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_ParameterDeclaration * const &_self = static_cast <Cpp_ParameterDeclaration*> (__self);

    abortIf (_self->parameter_declaration_type != Cpp_ParameterDeclaration::_Declarator);

    Cpp_ParameterDeclaration_Declarator * const &self = static_cast <Cpp_ParameterDeclaration_Declarator*> (_self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static void
cpp_ParameterDeclaration_Declarator_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ParameterDeclaration ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_ParameterDeclaration * const &_self = static_cast <Cpp_ParameterDeclaration*> (__self);

    abortIf (_self->parameter_declaration_type != Cpp_ParameterDeclaration::_Declarator);

    Cpp_ParameterDeclaration_Declarator * const &self = static_cast <Cpp_ParameterDeclaration_Declarator*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static ParserElement*
cpp_ParameterDeclaration_Declarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ParameterDeclaration_Declarator))) Cpp_ParameterDeclaration_Declarator;
}

static Ref<Grammar>
create_cpp_ParameterDeclaration_Declarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ParameterDeclaration_Declarator_creation_func));
    grammar->name = String::forData ("ParameterDeclaration_Declarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_parameter_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_DeclSpecifierSeq ();
        entry->assignment_func = cpp_ParameterDeclaration_Declarator_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_AnyDeclarator ();
        entry->assignment_func = cpp_ParameterDeclaration_Declarator_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclaration_Declarator (Cpp_ParameterDeclaration_Declarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclaration_Declarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ParameterDeclaration_Declarator\n");

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ParameterDeclaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ParameterDeclaration");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclaration_DeclaratorAssignment ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclaration_Declarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ParameterDeclaration (Cpp_ParameterDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ParameterDeclaration\n");
        return;
    }

    switch (el->parameter_declaration_type) {
        case Cpp_ParameterDeclaration::_DeclaratorAssignment:
            dump_cpp_ParameterDeclaration_DeclaratorAssignment (static_cast <Cpp_ParameterDeclaration_DeclaratorAssignment const *> (el), nest_level + 0);
            break;
        case Cpp_ParameterDeclaration::_Declarator:
            dump_cpp_ParameterDeclaration_Declarator (static_cast <Cpp_ParameterDeclaration_Declarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_FunctionDefinition_FunctionBody_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionBody);

    Cpp_FunctionDefinition_FunctionBody * const &self = static_cast <Cpp_FunctionDefinition_FunctionBody*> (_self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static void
cpp_FunctionDefinition_FunctionBody_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionBody);

    Cpp_FunctionDefinition_FunctionBody * const &self = static_cast <Cpp_FunctionDefinition_FunctionBody*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static void
cpp_FunctionDefinition_FunctionBody_set_ctorInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_CtorInitializer);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionBody);

    Cpp_FunctionDefinition_FunctionBody * const &self = static_cast <Cpp_FunctionDefinition_FunctionBody*> (_self);

    self->ctorInitializer = static_cast <Cpp_CtorInitializer*> (___el);
}

static void
cpp_FunctionDefinition_FunctionBody_set_functionBody (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_FunctionBody);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionBody);

    Cpp_FunctionDefinition_FunctionBody * const &self = static_cast <Cpp_FunctionDefinition_FunctionBody*> (_self);

    self->functionBody = static_cast <Cpp_FunctionBody*> (___el);
}

bool cpp_function_definition_match_func (
        Cpp_FunctionDefinition *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_function_definition_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_FunctionDefinition);

    Cpp_FunctionDefinition * const &el =
            static_cast <Cpp_FunctionDefinition*> (_el);

    return cpp_function_definition_match_func (el, parser_control, data);
}

static ParserElement*
cpp_FunctionDefinition_FunctionBody_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionDefinition_FunctionBody))) Cpp_FunctionDefinition_FunctionBody;
}

static Ref<Grammar>
create_cpp_FunctionDefinition_FunctionBody ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionDefinition_FunctionBody_creation_func));
    grammar->name = String::forData ("FunctionDefinition_FunctionBody");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_constructor_ambiguity;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_DeclSpecifierSeq ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionBody_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_Declarator ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionBody_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_FunctionDefinition ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 4;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_CtorInitializer ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionBody_set_ctorInitializer;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_FunctionBody ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionBody_set_functionBody;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_function_definition_match_func;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionDefinition_FunctionBody (Cpp_FunctionDefinition_FunctionBody const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionDefinition_FunctionBody\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionDefinition_FunctionBody\n");

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    dump_cpp_CtorInitializer (el->ctorInitializer, nest_level + 1);

    dump_cpp_FunctionBody (el->functionBody, nest_level + 1);

}

static void
cpp_FunctionDefinition_FunctionTryBlock_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionTryBlock);

    Cpp_FunctionDefinition_FunctionTryBlock * const &self = static_cast <Cpp_FunctionDefinition_FunctionTryBlock*> (_self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static void
cpp_FunctionDefinition_FunctionTryBlock_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionTryBlock);

    Cpp_FunctionDefinition_FunctionTryBlock * const &self = static_cast <Cpp_FunctionDefinition_FunctionTryBlock*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static void
cpp_FunctionDefinition_FunctionTryBlock_set_functionTryBlock (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionDefinition ||
             __el->cpp_element_type != CppElement::_FunctionTryBlock);

    Cpp_FunctionDefinition * const &_self = static_cast <Cpp_FunctionDefinition*> (__self);

    abortIf (_self->function_definition_type != Cpp_FunctionDefinition::_FunctionTryBlock);

    Cpp_FunctionDefinition_FunctionTryBlock * const &self = static_cast <Cpp_FunctionDefinition_FunctionTryBlock*> (_self);

    self->functionTryBlock = static_cast <Cpp_FunctionTryBlock*> (___el);
}

static ParserElement*
cpp_FunctionDefinition_FunctionTryBlock_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionDefinition_FunctionTryBlock))) Cpp_FunctionDefinition_FunctionTryBlock;
}

static Ref<Grammar>
create_cpp_FunctionDefinition_FunctionTryBlock ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionDefinition_FunctionTryBlock_creation_func));
    grammar->name = String::forData ("FunctionDefinition_FunctionTryBlock");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_constructor_ambiguity;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_DeclSpecifierSeq ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionTryBlock_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_Declarator ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionTryBlock_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_statement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_FunctionTryBlock ();
        entry->assignment_func = cpp_FunctionDefinition_FunctionTryBlock_set_functionTryBlock;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_function_definition_match_func;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionDefinition_FunctionTryBlock (Cpp_FunctionDefinition_FunctionTryBlock const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionDefinition_FunctionTryBlock\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionDefinition_FunctionTryBlock\n");

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    dump_cpp_FunctionTryBlock (el->functionTryBlock, nest_level + 1);

}

static Ref<Grammar>
create_cpp_FunctionDefinition ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("FunctionDefinition");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionDefinition_FunctionBody ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionDefinition_FunctionTryBlock ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionDefinition (Cpp_FunctionDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionDefinition\n");
        return;
    }

    switch (el->function_definition_type) {
        case Cpp_FunctionDefinition::_FunctionBody:
            dump_cpp_FunctionDefinition_FunctionBody (static_cast <Cpp_FunctionDefinition_FunctionBody const *> (el), nest_level + 0);
            break;
        case Cpp_FunctionDefinition::_FunctionTryBlock:
            dump_cpp_FunctionDefinition_FunctionTryBlock (static_cast <Cpp_FunctionDefinition_FunctionTryBlock const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_FunctionBody_Default_set_compoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionBody ||
             __el->cpp_element_type != CppElement::_CompoundStatement);

    Cpp_FunctionBody * const &_self = static_cast <Cpp_FunctionBody*> (__self);

    abortIf (_self->function_body_type != Cpp_FunctionBody::_Default);

    Cpp_FunctionBody_Default * const &self = static_cast <Cpp_FunctionBody_Default*> (_self);

    self->compoundStatement = static_cast <Cpp_CompoundStatement*> (___el);
}

static ParserElement*
cpp_FunctionBody_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionBody_Default))) Cpp_FunctionBody_Default;
}

static Ref<Grammar>
create_cpp_FunctionBody_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionBody_Default_creation_func));
    grammar->name = String::forData ("FunctionBody_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CompoundStatement ();
        entry->assignment_func = cpp_FunctionBody_Default_set_compoundStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionBody_Default (Cpp_FunctionBody_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionBody_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionBody_Default\n");

    dump_cpp_CompoundStatement (el->compoundStatement, nest_level + 1);

}

static void
cpp_FunctionBody_ClassPrematch_set_scruffy_SkipCompoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionBody ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipCompoundStatement);

    Cpp_FunctionBody * const &_self = static_cast <Cpp_FunctionBody*> (__self);

    abortIf (_self->function_body_type != Cpp_FunctionBody::_ClassPrematch);

    Cpp_FunctionBody_ClassPrematch * const &self = static_cast <Cpp_FunctionBody_ClassPrematch*> (_self);

    self->scruffy_SkipCompoundStatement = static_cast <Cpp_Scruffy_SkipCompoundStatement*> (___el);
}

static ParserElement*
cpp_FunctionBody_ClassPrematch_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionBody_ClassPrematch))) Cpp_FunctionBody_ClassPrematch;
}

static Ref<Grammar>
create_cpp_FunctionBody_ClassPrematch ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionBody_ClassPrematch_creation_func));
    grammar->name = String::forData ("FunctionBody_ClassPrematch");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipCompoundStatement ();
        entry->assignment_func = cpp_FunctionBody_ClassPrematch_set_scruffy_SkipCompoundStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionBody_ClassPrematch (Cpp_FunctionBody_ClassPrematch const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionBody_ClassPrematch\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionBody_ClassPrematch\n");

    dump_cpp_Scruffy_SkipCompoundStatement (el->scruffy_SkipCompoundStatement, nest_level + 1);

}

static Ref<Grammar>
create_cpp_FunctionBody ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("FunctionBody");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionBody_Default ();
    entry->variants.append (String::forData ("default"));
    entry->variants.append (String::forData ("AST"));

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_FunctionBody_ClassPrematch ();
    entry->variants.append (String::forData ("class-prematch"));

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionBody (Cpp_FunctionBody const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionBody\n");
        return;
    }

    switch (el->function_body_type) {
        case Cpp_FunctionBody::_Default:
            dump_cpp_FunctionBody_Default (static_cast <Cpp_FunctionBody_Default const *> (el), nest_level + 0);
            break;
        case Cpp_FunctionBody::_ClassPrematch:
            dump_cpp_FunctionBody_ClassPrematch (static_cast <Cpp_FunctionBody_ClassPrematch const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Initializer_InitializerClause_set_initializerClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Initializer ||
             __el->cpp_element_type != CppElement::_InitializerClause);

    Cpp_Initializer * const &_self = static_cast <Cpp_Initializer*> (__self);

    abortIf (_self->initializer_type != Cpp_Initializer::_InitializerClause);

    Cpp_Initializer_InitializerClause * const &self = static_cast <Cpp_Initializer_InitializerClause*> (_self);

    self->initializerClause = static_cast <Cpp_InitializerClause*> (___el);
}

bool cpp_accept_initializer (
        Cpp_Initializer *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_initializer (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_Initializer);

    Cpp_Initializer * const &el =
            static_cast <Cpp_Initializer*> (_el);

    return cpp_accept_initializer (el, parser_control, data);
}

static ParserElement*
cpp_Initializer_InitializerClause_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Initializer_InitializerClause))) Cpp_Initializer_InitializerClause;
}

static Ref<Grammar>
create_cpp_Initializer_InitializerClause ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Initializer_InitializerClause_creation_func));
    grammar->name = String::forData ("Initializer_InitializerClause");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitializerClause ();
        entry->assignment_func = cpp_Initializer_InitializerClause_set_initializerClause;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_initializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Initializer_InitializerClause (Cpp_Initializer_InitializerClause const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Initializer_InitializerClause\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Initializer_InitializerClause\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_InitializerClause (el->initializerClause, nest_level + 1);

}

static void
cpp_Initializer_ExpressionList_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Initializer ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_Initializer * const &_self = static_cast <Cpp_Initializer*> (__self);

    abortIf (_self->initializer_type != Cpp_Initializer::_ExpressionList);

    Cpp_Initializer_ExpressionList * const &self = static_cast <Cpp_Initializer_ExpressionList*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

static ParserElement*
cpp_Initializer_ExpressionList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Initializer_ExpressionList))) Cpp_Initializer_ExpressionList;
}

static Ref<Grammar>
create_cpp_Initializer_ExpressionList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Initializer_ExpressionList_creation_func));
    grammar->name = String::forData ("Initializer_ExpressionList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_Initializer_ExpressionList_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_initializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Initializer_ExpressionList (Cpp_Initializer_ExpressionList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Initializer_ExpressionList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Initializer_ExpressionList\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static Ref<Grammar>
create_cpp_Initializer ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Initializer");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Initializer_InitializerClause ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Initializer_ExpressionList ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Initializer (Cpp_Initializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Initializer\n");
        return;
    }

    switch (el->initializer_type) {
        case Cpp_Initializer::_InitializerClause:
            dump_cpp_Initializer_InitializerClause (static_cast <Cpp_Initializer_InitializerClause const *> (el), nest_level + 0);
            break;
        case Cpp_Initializer::_ExpressionList:
            dump_cpp_Initializer_ExpressionList (static_cast <Cpp_Initializer_ExpressionList const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_InitializerClause_Expression_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitializerClause ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_InitializerClause * const &_self = static_cast <Cpp_InitializerClause*> (__self);

    abortIf (_self->initializer_clause_type != Cpp_InitializerClause::_Expression);

    Cpp_InitializerClause_Expression * const &self = static_cast <Cpp_InitializerClause_Expression*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static ParserElement*
cpp_InitializerClause_Expression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitializerClause_Expression))) Cpp_InitializerClause_Expression;
}

static Ref<Grammar>
create_cpp_InitializerClause_Expression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitializerClause_Expression_creation_func));
    grammar->name = String::forData ("InitializerClause_Expression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_InitializerClause_Expression_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitializerClause_Expression (Cpp_InitializerClause_Expression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitializerClause_Expression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitializerClause_Expression\n");

    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_InitializerClause_InitializerList_set_initializerList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitializerClause ||
             __el->cpp_element_type != CppElement::_InitializerList);

    Cpp_InitializerClause * const &_self = static_cast <Cpp_InitializerClause*> (__self);

    abortIf (_self->initializer_clause_type != Cpp_InitializerClause::_InitializerList);

    Cpp_InitializerClause_InitializerList * const &self = static_cast <Cpp_InitializerClause_InitializerList*> (_self);

    self->initializerList = static_cast <Cpp_InitializerList*> (___el);
}

static ParserElement*
cpp_InitializerClause_InitializerList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitializerClause_InitializerList))) Cpp_InitializerClause_InitializerList;
}

static Ref<Grammar>
create_cpp_InitializerClause_InitializerList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitializerClause_InitializerList_creation_func));
    grammar->name = String::forData ("InitializerClause_InitializerList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitializerList ();
        entry->assignment_func = cpp_InitializerClause_InitializerList_set_initializerList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitializerClause_InitializerList (Cpp_InitializerClause_InitializerList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitializerClause_InitializerList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitializerClause_InitializerList\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    dump_cpp_InitializerList (el->initializerList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,_opt\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static Ref<Grammar>
create_cpp_InitializerClause ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("InitializerClause");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InitializerClause_Expression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InitializerClause_InitializerList ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitializerClause (Cpp_InitializerClause const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitializerClause\n");
        return;
    }

    switch (el->initializer_clause_type) {
        case Cpp_InitializerClause::_Expression:
            dump_cpp_InitializerClause_Expression (static_cast <Cpp_InitializerClause_Expression const *> (el), nest_level + 0);
            break;
        case Cpp_InitializerClause::_InitializerList:
            dump_cpp_InitializerClause_InitializerList (static_cast <Cpp_InitializerClause_InitializerList const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_InitializerList_InitializerClause_set_initializerClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitializerList ||
             __el->cpp_element_type != CppElement::_InitializerClause);

    Cpp_InitializerList * const &_self = static_cast <Cpp_InitializerList*> (__self);

    abortIf (_self->initializer_list_type != Cpp_InitializerList::_InitializerClause);

    Cpp_InitializerList_InitializerClause * const &self = static_cast <Cpp_InitializerList_InitializerClause*> (_self);

    self->initializerClause = static_cast <Cpp_InitializerClause*> (___el);
}

static ParserElement*
cpp_InitializerList_InitializerClause_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitializerList_InitializerClause))) Cpp_InitializerList_InitializerClause;
}

static Ref<Grammar>
create_cpp_InitializerList_InitializerClause ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitializerList_InitializerClause_creation_func));
    grammar->name = String::forData ("InitializerList_InitializerClause");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitializerClause ();
        entry->assignment_func = cpp_InitializerList_InitializerClause_set_initializerClause;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitializerList_InitializerClause (Cpp_InitializerList_InitializerClause const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitializerList_InitializerClause\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitializerList_InitializerClause\n");

    dump_cpp_InitializerClause (el->initializerClause, nest_level + 1);

}

static void
cpp_InitializerList_InitializerList_set_initializerClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitializerList ||
             __el->cpp_element_type != CppElement::_InitializerClause);

    Cpp_InitializerList * const &_self = static_cast <Cpp_InitializerList*> (__self);

    abortIf (_self->initializer_list_type != Cpp_InitializerList::_InitializerList);

    Cpp_InitializerList_InitializerList * const &self = static_cast <Cpp_InitializerList_InitializerList*> (_self);

    self->initializerClause = static_cast <Cpp_InitializerClause*> (___el);
}

static void
cpp_InitializerList_InitializerList_set_initializerList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_InitializerList ||
             __el->cpp_element_type != CppElement::_InitializerList);

    Cpp_InitializerList * const &_self = static_cast <Cpp_InitializerList*> (__self);

    abortIf (_self->initializer_list_type != Cpp_InitializerList::_InitializerList);

    Cpp_InitializerList_InitializerList * const &self = static_cast <Cpp_InitializerList_InitializerList*> (_self);

    self->initializerList = static_cast <Cpp_InitializerList*> (___el);
}

static ParserElement*
cpp_InitializerList_InitializerList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_InitializerList_InitializerList))) Cpp_InitializerList_InitializerList;
}

static Ref<Grammar>
create_cpp_InitializerList_InitializerList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_InitializerList_InitializerList_creation_func));
    grammar->name = String::forData ("InitializerList_InitializerList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitializerClause ();
        entry->assignment_func = cpp_InitializerList_InitializerList_set_initializerClause;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_InitializerList ();
        entry->assignment_func = cpp_InitializerList_InitializerList_set_initializerList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitializerList_InitializerList (Cpp_InitializerList_InitializerList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitializerList_InitializerList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("InitializerList_InitializerList\n");

    dump_cpp_InitializerClause (el->initializerClause, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_InitializerList (el->initializerList, nest_level + 1);

}

static Ref<Grammar>
create_cpp_InitializerList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("InitializerList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InitializerList_InitializerClause ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_InitializerList_InitializerList ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_InitializerList (Cpp_InitializerList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) InitializerList\n");
        return;
    }

    switch (el->initializer_list_type) {
        case Cpp_InitializerList::_InitializerClause:
            dump_cpp_InitializerList_InitializerClause (static_cast <Cpp_InitializerList_InitializerClause const *> (el), nest_level + 0);
            break;
        case Cpp_InitializerList::_InitializerList:
            dump_cpp_InitializerList_InitializerList (static_cast <Cpp_InitializerList_InitializerList const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ClassName_ClassNameIdentifier_set_classNameIdentifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassName ||
             __el->cpp_element_type != CppElement::_ClassNameIdentifier);

    Cpp_ClassName * const &_self = static_cast <Cpp_ClassName*> (__self);

    abortIf (_self->class_name_type != Cpp_ClassName::_ClassNameIdentifier);

    Cpp_ClassName_ClassNameIdentifier * const &self = static_cast <Cpp_ClassName_ClassNameIdentifier*> (_self);

    self->classNameIdentifier = static_cast <Cpp_ClassNameIdentifier*> (___el);
}

static ParserElement*
cpp_ClassName_ClassNameIdentifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassName_ClassNameIdentifier))) Cpp_ClassName_ClassNameIdentifier;
}

static Ref<Grammar>
create_cpp_ClassName_ClassNameIdentifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassName_ClassNameIdentifier_creation_func));
    grammar->name = String::forData ("ClassName_ClassNameIdentifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassNameIdentifier ();
        entry->assignment_func = cpp_ClassName_ClassNameIdentifier_set_classNameIdentifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassName_ClassNameIdentifier (Cpp_ClassName_ClassNameIdentifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassName_ClassNameIdentifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassName_ClassNameIdentifier\n");

    dump_cpp_ClassNameIdentifier (el->classNameIdentifier, nest_level + 1);

}

static void
cpp_ClassName_TemplateId_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassName ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_ClassName * const &_self = static_cast <Cpp_ClassName*> (__self);

    abortIf (_self->class_name_type != Cpp_ClassName::_TemplateId);

    Cpp_ClassName_TemplateId * const &self = static_cast <Cpp_ClassName_TemplateId*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static ParserElement*
cpp_ClassName_TemplateId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassName_TemplateId))) Cpp_ClassName_TemplateId;
}

static Ref<Grammar>
create_cpp_ClassName_TemplateId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassName_TemplateId_creation_func));
    grammar->name = String::forData ("ClassName_TemplateId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_ClassName_TemplateId_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassName_TemplateId (Cpp_ClassName_TemplateId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassName_TemplateId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassName_TemplateId\n");

    dump_cpp_TemplateId (el->templateId, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ClassName ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ClassName");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassName_ClassNameIdentifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassName_TemplateId ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassName (Cpp_ClassName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassName\n");
        return;
    }

    switch (el->class_name_type) {
        case Cpp_ClassName::_ClassNameIdentifier:
            dump_cpp_ClassName_ClassNameIdentifier (static_cast <Cpp_ClassName_ClassNameIdentifier const *> (el), nest_level + 0);
            break;
        case Cpp_ClassName::_TemplateId:
            dump_cpp_ClassName_TemplateId (static_cast <Cpp_ClassName_TemplateId const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_ClassNameIdentifier_match_func (
        Cpp_ClassNameIdentifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_ClassNameIdentifier_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ClassNameIdentifier);

    Cpp_ClassNameIdentifier * const &el = static_cast <Cpp_ClassNameIdentifier*> (_el);

    return cpp_ClassNameIdentifier_match_func (el, parser_control, data);
}

static void
cpp_ClassNameIdentifier_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassNameIdentifier ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_ClassNameIdentifier * const &self = static_cast <Cpp_ClassNameIdentifier*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_ClassNameIdentifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassNameIdentifier))) Cpp_ClassNameIdentifier;
}

static Ref<Grammar>
create_cpp_ClassNameIdentifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassNameIdentifier_creation_func));
    grammar->name = String::forData ("ClassNameIdentifier");
    grammar->match_func = __pargen_cpp_ClassNameIdentifier_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_ClassNameIdentifier_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassNameIdentifier (Cpp_ClassNameIdentifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassNameIdentifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassNameIdentifier\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_ClassSpecifier_set_classHead (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassSpecifier ||
             __el->cpp_element_type != CppElement::_ClassHead);

    Cpp_ClassSpecifier * const &self = static_cast <Cpp_ClassSpecifier*> (__self);

    self->classHead = static_cast <Cpp_ClassHead*> (___el);
}

bool cpp_accept_class_head (
        Cpp_ClassSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_class_head (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ClassSpecifier);

    Cpp_ClassSpecifier * const &el =
            static_cast <Cpp_ClassSpecifier*> (_el);

    return cpp_accept_class_head (el, parser_control, data);
}

bool cpp_begin_class_definition (
        Cpp_ClassSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_class_definition (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ClassSpecifier);

    Cpp_ClassSpecifier * const &el =
            static_cast <Cpp_ClassSpecifier*> (_el);

    return cpp_begin_class_definition (el, parser_control, data);
}

static void
cpp_ClassSpecifier_set_memberSpecification (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassSpecifier ||
             __el->cpp_element_type != CppElement::_MemberSpecification);

    Cpp_ClassSpecifier * const &self = static_cast <Cpp_ClassSpecifier*> (__self);

    self->memberSpecification = static_cast <Cpp_MemberSpecification*> (___el);
}

bool cpp_accept_class_definition (
        Cpp_ClassSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_class_definition (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ClassSpecifier);

    Cpp_ClassSpecifier * const &el =
            static_cast <Cpp_ClassSpecifier*> (_el);

    return cpp_accept_class_definition (el, parser_control, data);
}

static ParserElement*
cpp_ClassSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassSpecifier))) Cpp_ClassSpecifier;
}

static Ref<Grammar>
create_cpp_ClassSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassSpecifier_creation_func));
    grammar->name = String::forData ("ClassSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassHead ();
        entry->assignment_func = cpp_ClassSpecifier_set_classHead;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_class_head;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("{"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_class_definition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_MemberSpecification ();
        entry->assignment_func = cpp_ClassSpecifier_set_memberSpecification;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("}"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_class_definition;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassSpecifier (Cpp_ClassSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassSpecifier\n");

    dump_cpp_ClassHead (el->classHead, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: {\n");
    dump_cpp_MemberSpecification (el->memberSpecification, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: }\n");
}

static void
cpp_ClassHead_Simple_set_classKey (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_ClassKey);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_Simple);

    Cpp_ClassHead_Simple * const &self = static_cast <Cpp_ClassHead_Simple*> (_self);

    self->classKey = static_cast <Cpp_ClassKey*> (___el);
}

static void
cpp_ClassHead_Simple_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_Simple);

    Cpp_ClassHead_Simple * const &self = static_cast <Cpp_ClassHead_Simple*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

bool cpp_accept_class_identifier (
        Cpp_ClassHead *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_class_identifier (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ClassHead);

    Cpp_ClassHead * const &el =
            static_cast <Cpp_ClassHead*> (_el);

    return cpp_accept_class_identifier (el, parser_control, data);
}

static void
cpp_ClassHead_Simple_set_baseClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_BaseClause);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_Simple);

    Cpp_ClassHead_Simple * const &self = static_cast <Cpp_ClassHead_Simple*> (_self);

    self->baseClause = static_cast <Cpp_BaseClause*> (___el);
}

static ParserElement*
cpp_ClassHead_Simple_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassHead_Simple))) Cpp_ClassHead_Simple;
}

static Ref<Grammar>
create_cpp_ClassHead_Simple ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassHead_Simple_creation_func));
    grammar->name = String::forData ("ClassHead_Simple");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassKey ();
        entry->assignment_func = cpp_ClassHead_Simple_set_classKey;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_ClassHead_Simple_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_class_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_BaseClause ();
        entry->assignment_func = cpp_ClassHead_Simple_set_baseClause;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassHead_Simple (Cpp_ClassHead_Simple const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassHead_Simple\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassHead_Simple\n");

    dump_cpp_ClassKey (el->classKey, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

    dump_cpp_BaseClause (el->baseClause, nest_level + 1);

}

static void
cpp_ClassHead_NestedNameSpecifier_set_classKey (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_ClassKey);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_NestedNameSpecifier);

    Cpp_ClassHead_NestedNameSpecifier * const &self = static_cast <Cpp_ClassHead_NestedNameSpecifier*> (_self);

    self->classKey = static_cast <Cpp_ClassKey*> (___el);
}

static void
cpp_ClassHead_NestedNameSpecifier_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_NestedNameSpecifier);

    Cpp_ClassHead_NestedNameSpecifier * const &self = static_cast <Cpp_ClassHead_NestedNameSpecifier*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_ClassHead_NestedNameSpecifier_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_NestedNameSpecifier);

    Cpp_ClassHead_NestedNameSpecifier * const &self = static_cast <Cpp_ClassHead_NestedNameSpecifier*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_ClassHead_NestedNameSpecifier_set_baseClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_BaseClause);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_NestedNameSpecifier);

    Cpp_ClassHead_NestedNameSpecifier * const &self = static_cast <Cpp_ClassHead_NestedNameSpecifier*> (_self);

    self->baseClause = static_cast <Cpp_BaseClause*> (___el);
}

static ParserElement*
cpp_ClassHead_NestedNameSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassHead_NestedNameSpecifier))) Cpp_ClassHead_NestedNameSpecifier;
}

static Ref<Grammar>
create_cpp_ClassHead_NestedNameSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassHead_NestedNameSpecifier_creation_func));
    grammar->name = String::forData ("ClassHead_NestedNameSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassKey ();
        entry->assignment_func = cpp_ClassHead_NestedNameSpecifier_set_classKey;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_ClassHead_NestedNameSpecifier_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_ClassHead_NestedNameSpecifier_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_class_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_BaseClause ();
        entry->assignment_func = cpp_ClassHead_NestedNameSpecifier_set_baseClause;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassHead_NestedNameSpecifier (Cpp_ClassHead_NestedNameSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassHead_NestedNameSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassHead_NestedNameSpecifier\n");

    dump_cpp_ClassKey (el->classKey, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_Identifier (el->identifier, nest_level + 1);

    dump_cpp_BaseClause (el->baseClause, nest_level + 1);

}

static void
cpp_ClassHead_TemplateId_set_classKey (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_ClassKey);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_TemplateId);

    Cpp_ClassHead_TemplateId * const &self = static_cast <Cpp_ClassHead_TemplateId*> (_self);

    self->classKey = static_cast <Cpp_ClassKey*> (___el);
}

static void
cpp_ClassHead_TemplateId_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_TemplateId);

    Cpp_ClassHead_TemplateId * const &self = static_cast <Cpp_ClassHead_TemplateId*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_ClassHead_TemplateId_set_templateId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_TemplateId);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_TemplateId);

    Cpp_ClassHead_TemplateId * const &self = static_cast <Cpp_ClassHead_TemplateId*> (_self);

    self->templateId = static_cast <Cpp_TemplateId*> (___el);
}

static void
cpp_ClassHead_TemplateId_set_baseClause (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ClassHead ||
             __el->cpp_element_type != CppElement::_BaseClause);

    Cpp_ClassHead * const &_self = static_cast <Cpp_ClassHead*> (__self);

    abortIf (_self->class_head_type != Cpp_ClassHead::_TemplateId);

    Cpp_ClassHead_TemplateId * const &self = static_cast <Cpp_ClassHead_TemplateId*> (_self);

    self->baseClause = static_cast <Cpp_BaseClause*> (___el);
}

static ParserElement*
cpp_ClassHead_TemplateId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassHead_TemplateId))) Cpp_ClassHead_TemplateId;
}

static Ref<Grammar>
create_cpp_ClassHead_TemplateId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassHead_TemplateId_creation_func));
    grammar->name = String::forData ("ClassHead_TemplateId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassKey ();
        entry->assignment_func = cpp_ClassHead_TemplateId_set_classKey;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_ClassHead_TemplateId_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateId ();
        entry->assignment_func = cpp_ClassHead_TemplateId_set_templateId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_class_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_BaseClause ();
        entry->assignment_func = cpp_ClassHead_TemplateId_set_baseClause;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassHead_TemplateId (Cpp_ClassHead_TemplateId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassHead_TemplateId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassHead_TemplateId\n");

    dump_cpp_ClassKey (el->classKey, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_TemplateId (el->templateId, nest_level + 1);

    dump_cpp_BaseClause (el->baseClause, nest_level + 1);

}

static Ref<Grammar>
create_cpp_ClassHead ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ClassHead");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassHead_Simple ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassHead_NestedNameSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassHead_TemplateId ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassHead (Cpp_ClassHead const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassHead\n");
        return;
    }

    switch (el->class_head_type) {
        case Cpp_ClassHead::_Simple:
            dump_cpp_ClassHead_Simple (static_cast <Cpp_ClassHead_Simple const *> (el), nest_level + 0);
            break;
        case Cpp_ClassHead::_NestedNameSpecifier:
            dump_cpp_ClassHead_NestedNameSpecifier (static_cast <Cpp_ClassHead_NestedNameSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_ClassHead::_TemplateId:
            dump_cpp_ClassHead_TemplateId (static_cast <Cpp_ClassHead_TemplateId const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static ParserElement*
cpp_ClassKey_Class_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassKey_Class))) Cpp_ClassKey_Class;
}

static Ref<Grammar>
create_cpp_ClassKey_Class ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassKey_Class_creation_func));
    grammar->name = String::forData ("ClassKey_Class");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("class"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassKey_Class (Cpp_ClassKey_Class const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassKey_Class\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassKey_Class\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: class\n");
}

static ParserElement*
cpp_ClassKey_Struct_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassKey_Struct))) Cpp_ClassKey_Struct;
}

static Ref<Grammar>
create_cpp_ClassKey_Struct ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassKey_Struct_creation_func));
    grammar->name = String::forData ("ClassKey_Struct");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("struct"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassKey_Struct (Cpp_ClassKey_Struct const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassKey_Struct\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassKey_Struct\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: struct\n");
}

static ParserElement*
cpp_ClassKey_Union_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ClassKey_Union))) Cpp_ClassKey_Union;
}

static Ref<Grammar>
create_cpp_ClassKey_Union ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ClassKey_Union_creation_func));
    grammar->name = String::forData ("ClassKey_Union");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("union"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassKey_Union (Cpp_ClassKey_Union const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassKey_Union\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ClassKey_Union\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: union\n");
}

static Ref<Grammar>
create_cpp_ClassKey ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ClassKey");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassKey_Class ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassKey_Struct ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ClassKey_Union ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ClassKey (Cpp_ClassKey const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ClassKey\n");
        return;
    }

    switch (el->class_key_type) {
        case Cpp_ClassKey::_Class:
            dump_cpp_ClassKey_Class (static_cast <Cpp_ClassKey_Class const *> (el), nest_level + 0);
            break;
        case Cpp_ClassKey::_Struct:
            dump_cpp_ClassKey_Struct (static_cast <Cpp_ClassKey_Struct const *> (el), nest_level + 0);
            break;
        case Cpp_ClassKey::_Union:
            dump_cpp_ClassKey_Union (static_cast <Cpp_ClassKey_Union const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MemberSpecification_MemberDeclaration_set_memberDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberSpecification ||
             __el->cpp_element_type != CppElement::_MemberDeclaration);

    Cpp_MemberSpecification * const &_self = static_cast <Cpp_MemberSpecification*> (__self);

    abortIf (_self->member_specification_type != Cpp_MemberSpecification::_MemberDeclaration);

    Cpp_MemberSpecification_MemberDeclaration * const &self = static_cast <Cpp_MemberSpecification_MemberDeclaration*> (_self);

    self->memberDeclaration = static_cast <Cpp_MemberDeclaration*> (___el);
}

static void
cpp_MemberSpecification_MemberDeclaration_set_memberSpecification (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberSpecification ||
             __el->cpp_element_type != CppElement::_MemberSpecification);

    Cpp_MemberSpecification * const &_self = static_cast <Cpp_MemberSpecification*> (__self);

    abortIf (_self->member_specification_type != Cpp_MemberSpecification::_MemberDeclaration);

    Cpp_MemberSpecification_MemberDeclaration * const &self = static_cast <Cpp_MemberSpecification_MemberDeclaration*> (_self);

    self->memberSpecification = static_cast <Cpp_MemberSpecification*> (___el);
}

static ParserElement*
cpp_MemberSpecification_MemberDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberSpecification_MemberDeclaration))) Cpp_MemberSpecification_MemberDeclaration;
}

static Ref<Grammar>
create_cpp_MemberSpecification_MemberDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberSpecification_MemberDeclaration_creation_func));
    grammar->name = String::forData ("MemberSpecification_MemberDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaration ();
        entry->assignment_func = cpp_MemberSpecification_MemberDeclaration_set_memberDeclaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_MemberSpecification ();
        entry->assignment_func = cpp_MemberSpecification_MemberDeclaration_set_memberSpecification;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberSpecification_MemberDeclaration (Cpp_MemberSpecification_MemberDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberSpecification_MemberDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberSpecification_MemberDeclaration\n");

    dump_cpp_MemberDeclaration (el->memberDeclaration, nest_level + 1);

    dump_cpp_MemberSpecification (el->memberSpecification, nest_level + 1);

}

static void
cpp_MemberSpecification_AccessSpecifier_set_accessSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberSpecification ||
             __el->cpp_element_type != CppElement::_AccessSpecifier);

    Cpp_MemberSpecification * const &_self = static_cast <Cpp_MemberSpecification*> (__self);

    abortIf (_self->member_specification_type != Cpp_MemberSpecification::_AccessSpecifier);

    Cpp_MemberSpecification_AccessSpecifier * const &self = static_cast <Cpp_MemberSpecification_AccessSpecifier*> (_self);

    self->accessSpecifier = static_cast <Cpp_AccessSpecifier*> (___el);
}

static void
cpp_MemberSpecification_AccessSpecifier_set_memberSpecification (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberSpecification ||
             __el->cpp_element_type != CppElement::_MemberSpecification);

    Cpp_MemberSpecification * const &_self = static_cast <Cpp_MemberSpecification*> (__self);

    abortIf (_self->member_specification_type != Cpp_MemberSpecification::_AccessSpecifier);

    Cpp_MemberSpecification_AccessSpecifier * const &self = static_cast <Cpp_MemberSpecification_AccessSpecifier*> (_self);

    self->memberSpecification = static_cast <Cpp_MemberSpecification*> (___el);
}

static ParserElement*
cpp_MemberSpecification_AccessSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberSpecification_AccessSpecifier))) Cpp_MemberSpecification_AccessSpecifier;
}

static Ref<Grammar>
create_cpp_MemberSpecification_AccessSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberSpecification_AccessSpecifier_creation_func));
    grammar->name = String::forData ("MemberSpecification_AccessSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AccessSpecifier ();
        entry->assignment_func = cpp_MemberSpecification_AccessSpecifier_set_accessSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_MemberSpecification ();
        entry->assignment_func = cpp_MemberSpecification_AccessSpecifier_set_memberSpecification;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberSpecification_AccessSpecifier (Cpp_MemberSpecification_AccessSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberSpecification_AccessSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberSpecification_AccessSpecifier\n");

    dump_cpp_AccessSpecifier (el->accessSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_MemberSpecification (el->memberSpecification, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MemberSpecification ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemberSpecification");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberSpecification_MemberDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberSpecification_AccessSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberSpecification (Cpp_MemberSpecification const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberSpecification\n");
        return;
    }

    switch (el->member_specification_type) {
        case Cpp_MemberSpecification::_MemberDeclaration:
            dump_cpp_MemberSpecification_MemberDeclaration (static_cast <Cpp_MemberSpecification_MemberDeclaration const *> (el), nest_level + 0);
            break;
        case Cpp_MemberSpecification::_AccessSpecifier:
            dump_cpp_MemberSpecification_AccessSpecifier (static_cast <Cpp_MemberSpecification_AccessSpecifier const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MemberDeclaration_MemberDeclaratorList_set_declSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_DeclSpecifierSeq);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_MemberDeclaratorList);

    Cpp_MemberDeclaration_MemberDeclaratorList * const &self = static_cast <Cpp_MemberDeclaration_MemberDeclaratorList*> (_self);

    self->declSpecifierSeq = static_cast <Cpp_DeclSpecifierSeq*> (___el);
}

static void
cpp_MemberDeclaration_MemberDeclaratorList_set_memberDeclaratorList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_MemberDeclaratorList);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_MemberDeclaratorList);

    Cpp_MemberDeclaration_MemberDeclaratorList * const &self = static_cast <Cpp_MemberDeclaration_MemberDeclaratorList*> (_self);

    self->memberDeclaratorList = static_cast <Cpp_MemberDeclaratorList*> (___el);
}

static ParserElement*
cpp_MemberDeclaration_MemberDeclaratorList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaration_MemberDeclaratorList))) Cpp_MemberDeclaration_MemberDeclaratorList;
}

static Ref<Grammar>
create_cpp_MemberDeclaration_MemberDeclaratorList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaration_MemberDeclaratorList_creation_func));
    grammar->name = String::forData ("MemberDeclaration_MemberDeclaratorList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_constructor_ambiguity;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_DeclSpecifierSeq ();
        entry->assignment_func = cpp_MemberDeclaration_MemberDeclaratorList_set_declSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_MemberDeclaratorList ();
        entry->assignment_func = cpp_MemberDeclaration_MemberDeclaratorList_set_memberDeclaratorList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaration_MemberDeclaratorList (Cpp_MemberDeclaration_MemberDeclaratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaration_MemberDeclaratorList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaration_MemberDeclaratorList\n");

    dump_cpp_DeclSpecifierSeq (el->declSpecifierSeq, nest_level + 1);

    dump_cpp_MemberDeclaratorList (el->memberDeclaratorList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_MemberDeclaration_FunctionDefinition_set_functionDefinition (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_FunctionDefinition);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_FunctionDefinition);

    Cpp_MemberDeclaration_FunctionDefinition * const &self = static_cast <Cpp_MemberDeclaration_FunctionDefinition*> (_self);

    self->functionDefinition = static_cast <Cpp_FunctionDefinition*> (___el);
}

static ParserElement*
cpp_MemberDeclaration_FunctionDefinition_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaration_FunctionDefinition))) Cpp_MemberDeclaration_FunctionDefinition;
}

static Ref<Grammar>
create_cpp_MemberDeclaration_FunctionDefinition ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaration_FunctionDefinition_creation_func));
    grammar->name = String::forData ("MemberDeclaration_FunctionDefinition");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_FunctionDefinition ();
        entry->assignment_func = cpp_MemberDeclaration_FunctionDefinition_set_functionDefinition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaration_FunctionDefinition (Cpp_MemberDeclaration_FunctionDefinition const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaration_FunctionDefinition\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaration_FunctionDefinition\n");

    dump_cpp_FunctionDefinition (el->functionDefinition, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;_opt\n");
}

static void
cpp_MemberDeclaration_NestedNameSpecifier_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_NestedNameSpecifier);

    Cpp_MemberDeclaration_NestedNameSpecifier * const &self = static_cast <Cpp_MemberDeclaration_NestedNameSpecifier*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_MemberDeclaration_NestedNameSpecifier_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_NestedNameSpecifier);

    Cpp_MemberDeclaration_NestedNameSpecifier * const &self = static_cast <Cpp_MemberDeclaration_NestedNameSpecifier*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_MemberDeclaration_NestedNameSpecifier_set_unqualifiedId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_UnqualifiedId);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_NestedNameSpecifier);

    Cpp_MemberDeclaration_NestedNameSpecifier * const &self = static_cast <Cpp_MemberDeclaration_NestedNameSpecifier*> (_self);

    self->unqualifiedId = static_cast <Cpp_UnqualifiedId*> (___el);
}

static ParserElement*
cpp_MemberDeclaration_NestedNameSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaration_NestedNameSpecifier))) Cpp_MemberDeclaration_NestedNameSpecifier;
}

static Ref<Grammar>
create_cpp_MemberDeclaration_NestedNameSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaration_NestedNameSpecifier_creation_func));
    grammar->name = String::forData ("MemberDeclaration_NestedNameSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_MemberDeclaration_NestedNameSpecifier_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_MemberDeclaration_NestedNameSpecifier_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UnqualifiedId ();
        entry->assignment_func = cpp_MemberDeclaration_NestedNameSpecifier_set_unqualifiedId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (";"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaration_NestedNameSpecifier (Cpp_MemberDeclaration_NestedNameSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaration_NestedNameSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaration_NestedNameSpecifier\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: template_opt\n");
    dump_cpp_UnqualifiedId (el->unqualifiedId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ;\n");
}

static void
cpp_MemberDeclaration_UsingDeclaration_set_usingDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_UsingDeclaration);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_UsingDeclaration);

    Cpp_MemberDeclaration_UsingDeclaration * const &self = static_cast <Cpp_MemberDeclaration_UsingDeclaration*> (_self);

    self->usingDeclaration = static_cast <Cpp_UsingDeclaration*> (___el);
}

static ParserElement*
cpp_MemberDeclaration_UsingDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaration_UsingDeclaration))) Cpp_MemberDeclaration_UsingDeclaration;
}

static Ref<Grammar>
create_cpp_MemberDeclaration_UsingDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaration_UsingDeclaration_creation_func));
    grammar->name = String::forData ("MemberDeclaration_UsingDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_UsingDeclaration ();
        entry->assignment_func = cpp_MemberDeclaration_UsingDeclaration_set_usingDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaration_UsingDeclaration (Cpp_MemberDeclaration_UsingDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaration_UsingDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaration_UsingDeclaration\n");

    dump_cpp_UsingDeclaration (el->usingDeclaration, nest_level + 1);

}

static void
cpp_MemberDeclaration_TemplateDeclaration_set_templateDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaration ||
             __el->cpp_element_type != CppElement::_TemplateDeclaration);

    Cpp_MemberDeclaration * const &_self = static_cast <Cpp_MemberDeclaration*> (__self);

    abortIf (_self->member_declaration_type != Cpp_MemberDeclaration::_TemplateDeclaration);

    Cpp_MemberDeclaration_TemplateDeclaration * const &self = static_cast <Cpp_MemberDeclaration_TemplateDeclaration*> (_self);

    self->templateDeclaration = static_cast <Cpp_TemplateDeclaration*> (___el);
}

static ParserElement*
cpp_MemberDeclaration_TemplateDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaration_TemplateDeclaration))) Cpp_MemberDeclaration_TemplateDeclaration;
}

static Ref<Grammar>
create_cpp_MemberDeclaration_TemplateDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaration_TemplateDeclaration_creation_func));
    grammar->name = String::forData ("MemberDeclaration_TemplateDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateDeclaration ();
        entry->assignment_func = cpp_MemberDeclaration_TemplateDeclaration_set_templateDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaration_TemplateDeclaration (Cpp_MemberDeclaration_TemplateDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaration_TemplateDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaration_TemplateDeclaration\n");

    dump_cpp_TemplateDeclaration (el->templateDeclaration, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MemberDeclaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemberDeclaration");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaration_MemberDeclaratorList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaration_FunctionDefinition ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaration_NestedNameSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaration_UsingDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaration_TemplateDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaration (Cpp_MemberDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaration\n");
        return;
    }

    switch (el->member_declaration_type) {
        case Cpp_MemberDeclaration::_MemberDeclaratorList:
            dump_cpp_MemberDeclaration_MemberDeclaratorList (static_cast <Cpp_MemberDeclaration_MemberDeclaratorList const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclaration::_FunctionDefinition:
            dump_cpp_MemberDeclaration_FunctionDefinition (static_cast <Cpp_MemberDeclaration_FunctionDefinition const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclaration::_NestedNameSpecifier:
            dump_cpp_MemberDeclaration_NestedNameSpecifier (static_cast <Cpp_MemberDeclaration_NestedNameSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclaration::_UsingDeclaration:
            dump_cpp_MemberDeclaration_UsingDeclaration (static_cast <Cpp_MemberDeclaration_UsingDeclaration const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclaration::_TemplateDeclaration:
            dump_cpp_MemberDeclaration_TemplateDeclaration (static_cast <Cpp_MemberDeclaration_TemplateDeclaration const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MemberDeclaratorList_MemberDeclaratorList_set_memberDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaratorList ||
             __el->cpp_element_type != CppElement::_MemberDeclarator);

    Cpp_MemberDeclaratorList * const &_self = static_cast <Cpp_MemberDeclaratorList*> (__self);

    abortIf (_self->member_declarator_list_type != Cpp_MemberDeclaratorList::_MemberDeclaratorList);

    Cpp_MemberDeclaratorList_MemberDeclaratorList * const &self = static_cast <Cpp_MemberDeclaratorList_MemberDeclaratorList*> (_self);

    self->memberDeclarator = static_cast <Cpp_MemberDeclarator*> (___el);
}

static void
cpp_MemberDeclaratorList_MemberDeclaratorList_set_memberDeclaratorList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaratorList ||
             __el->cpp_element_type != CppElement::_MemberDeclaratorList);

    Cpp_MemberDeclaratorList * const &_self = static_cast <Cpp_MemberDeclaratorList*> (__self);

    abortIf (_self->member_declarator_list_type != Cpp_MemberDeclaratorList::_MemberDeclaratorList);

    Cpp_MemberDeclaratorList_MemberDeclaratorList * const &self = static_cast <Cpp_MemberDeclaratorList_MemberDeclaratorList*> (_self);

    self->memberDeclaratorList = static_cast <Cpp_MemberDeclaratorList*> (___el);
}

static ParserElement*
cpp_MemberDeclaratorList_MemberDeclaratorList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaratorList_MemberDeclaratorList))) Cpp_MemberDeclaratorList_MemberDeclaratorList;
}

static Ref<Grammar>
create_cpp_MemberDeclaratorList_MemberDeclaratorList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaratorList_MemberDeclaratorList_creation_func));
    grammar->name = String::forData ("MemberDeclaratorList_MemberDeclaratorList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclarator ();
        entry->assignment_func = cpp_MemberDeclaratorList_MemberDeclaratorList_set_memberDeclarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_MemberDeclaratorList ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_reset_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaratorList ();
        entry->assignment_func = cpp_MemberDeclaratorList_MemberDeclaratorList_set_memberDeclaratorList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaratorList_MemberDeclaratorList (Cpp_MemberDeclaratorList_MemberDeclaratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaratorList_MemberDeclaratorList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaratorList_MemberDeclaratorList\n");

    dump_cpp_MemberDeclarator (el->memberDeclarator, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_MemberDeclaratorList (el->memberDeclaratorList, nest_level + 1);

}

static void
cpp_MemberDeclaratorList_MemberDeclarator_set_memberDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclaratorList ||
             __el->cpp_element_type != CppElement::_MemberDeclarator);

    Cpp_MemberDeclaratorList * const &_self = static_cast <Cpp_MemberDeclaratorList*> (__self);

    abortIf (_self->member_declarator_list_type != Cpp_MemberDeclaratorList::_MemberDeclarator);

    Cpp_MemberDeclaratorList_MemberDeclarator * const &self = static_cast <Cpp_MemberDeclaratorList_MemberDeclarator*> (_self);

    self->memberDeclarator = static_cast <Cpp_MemberDeclarator*> (___el);
}

static ParserElement*
cpp_MemberDeclaratorList_MemberDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclaratorList_MemberDeclarator))) Cpp_MemberDeclaratorList_MemberDeclarator;
}

static Ref<Grammar>
create_cpp_MemberDeclaratorList_MemberDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclaratorList_MemberDeclarator_creation_func));
    grammar->name = String::forData ("MemberDeclaratorList_MemberDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclarator ();
        entry->assignment_func = cpp_MemberDeclaratorList_MemberDeclarator_set_memberDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaratorList_MemberDeclarator (Cpp_MemberDeclaratorList_MemberDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaratorList_MemberDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclaratorList_MemberDeclarator\n");

    dump_cpp_MemberDeclarator (el->memberDeclarator, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MemberDeclaratorList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemberDeclaratorList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaratorList_MemberDeclaratorList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclaratorList_MemberDeclarator ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclaratorList (Cpp_MemberDeclaratorList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclaratorList\n");
        return;
    }

    switch (el->member_declarator_list_type) {
        case Cpp_MemberDeclaratorList::_MemberDeclaratorList:
            dump_cpp_MemberDeclaratorList_MemberDeclaratorList (static_cast <Cpp_MemberDeclaratorList_MemberDeclaratorList const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclaratorList::_MemberDeclarator:
            dump_cpp_MemberDeclaratorList_MemberDeclarator (static_cast <Cpp_MemberDeclaratorList_MemberDeclarator const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MemberDeclarator_PureSpecifier_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclarator ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_MemberDeclarator * const &_self = static_cast <Cpp_MemberDeclarator*> (__self);

    abortIf (_self->member_declarator_type != Cpp_MemberDeclarator::_PureSpecifier);

    Cpp_MemberDeclarator_PureSpecifier * const &self = static_cast <Cpp_MemberDeclarator_PureSpecifier*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static void
cpp_MemberDeclarator_PureSpecifier_set_pureSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclarator ||
             __el->cpp_element_type != CppElement::_PureSpecifier);

    Cpp_MemberDeclarator * const &_self = static_cast <Cpp_MemberDeclarator*> (__self);

    abortIf (_self->member_declarator_type != Cpp_MemberDeclarator::_PureSpecifier);

    Cpp_MemberDeclarator_PureSpecifier * const &self = static_cast <Cpp_MemberDeclarator_PureSpecifier*> (_self);

    self->pureSpecifier = static_cast <Cpp_PureSpecifier*> (___el);
}

static ParserElement*
cpp_MemberDeclarator_PureSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclarator_PureSpecifier))) Cpp_MemberDeclarator_PureSpecifier;
}

static Ref<Grammar>
create_cpp_MemberDeclarator_PureSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclarator_PureSpecifier_creation_func));
    grammar->name = String::forData ("MemberDeclarator_PureSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_Declarator ();
        entry->assignment_func = cpp_MemberDeclarator_PureSpecifier_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_PureSpecifier ();
        entry->assignment_func = cpp_MemberDeclarator_PureSpecifier_set_pureSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclarator_PureSpecifier (Cpp_MemberDeclarator_PureSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclarator_PureSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclarator_PureSpecifier\n");

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    dump_cpp_PureSpecifier (el->pureSpecifier, nest_level + 1);

}

static void
cpp_MemberDeclarator_ConstantInitializer_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclarator ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_MemberDeclarator * const &_self = static_cast <Cpp_MemberDeclarator*> (__self);

    abortIf (_self->member_declarator_type != Cpp_MemberDeclarator::_ConstantInitializer);

    Cpp_MemberDeclarator_ConstantInitializer * const &self = static_cast <Cpp_MemberDeclarator_ConstantInitializer*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static void
cpp_MemberDeclarator_ConstantInitializer_set_constantInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclarator ||
             __el->cpp_element_type != CppElement::_ConstantInitializer);

    Cpp_MemberDeclarator * const &_self = static_cast <Cpp_MemberDeclarator*> (__self);

    abortIf (_self->member_declarator_type != Cpp_MemberDeclarator::_ConstantInitializer);

    Cpp_MemberDeclarator_ConstantInitializer * const &self = static_cast <Cpp_MemberDeclarator_ConstantInitializer*> (_self);

    self->constantInitializer = static_cast <Cpp_ConstantInitializer*> (___el);
}

static ParserElement*
cpp_MemberDeclarator_ConstantInitializer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclarator_ConstantInitializer))) Cpp_MemberDeclarator_ConstantInitializer;
}

static Ref<Grammar>
create_cpp_MemberDeclarator_ConstantInitializer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclarator_ConstantInitializer_creation_func));
    grammar->name = String::forData ("MemberDeclarator_ConstantInitializer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_Declarator ();
        entry->assignment_func = cpp_MemberDeclarator_ConstantInitializer_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ConstantInitializer ();
        entry->assignment_func = cpp_MemberDeclarator_ConstantInitializer_set_constantInitializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclarator_ConstantInitializer (Cpp_MemberDeclarator_ConstantInitializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclarator_ConstantInitializer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclarator_ConstantInitializer\n");

    dump_cpp_Declarator (el->declarator, nest_level + 1);

    dump_cpp_ConstantInitializer (el->constantInitializer, nest_level + 1);

}

static void
cpp_MemberDeclarator_Bitfield_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclarator ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_MemberDeclarator * const &_self = static_cast <Cpp_MemberDeclarator*> (__self);

    abortIf (_self->member_declarator_type != Cpp_MemberDeclarator::_Bitfield);

    Cpp_MemberDeclarator_Bitfield * const &self = static_cast <Cpp_MemberDeclarator_Bitfield*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_MemberDeclarator_Bitfield_set_constantExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemberDeclarator ||
             __el->cpp_element_type != CppElement::_ConstantExpression);

    Cpp_MemberDeclarator * const &_self = static_cast <Cpp_MemberDeclarator*> (__self);

    abortIf (_self->member_declarator_type != Cpp_MemberDeclarator::_Bitfield);

    Cpp_MemberDeclarator_Bitfield * const &self = static_cast <Cpp_MemberDeclarator_Bitfield*> (_self);

    self->constantExpression = static_cast <Cpp_ConstantExpression*> (___el);
}

static ParserElement*
cpp_MemberDeclarator_Bitfield_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemberDeclarator_Bitfield))) Cpp_MemberDeclarator_Bitfield;
}

static Ref<Grammar>
create_cpp_MemberDeclarator_Bitfield ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemberDeclarator_Bitfield_creation_func));
    grammar->name = String::forData ("MemberDeclarator_Bitfield");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_MemberDeclarator_Bitfield_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConstantExpression ();
        entry->assignment_func = cpp_MemberDeclarator_Bitfield_set_constantExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclarator_Bitfield (Cpp_MemberDeclarator_Bitfield const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclarator_Bitfield\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemberDeclarator_Bitfield\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_ConstantExpression (el->constantExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MemberDeclarator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemberDeclarator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclarator_PureSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclarator_ConstantInitializer ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemberDeclarator_Bitfield ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemberDeclarator (Cpp_MemberDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemberDeclarator\n");
        return;
    }

    switch (el->member_declarator_type) {
        case Cpp_MemberDeclarator::_PureSpecifier:
            dump_cpp_MemberDeclarator_PureSpecifier (static_cast <Cpp_MemberDeclarator_PureSpecifier const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclarator::_ConstantInitializer:
            dump_cpp_MemberDeclarator_ConstantInitializer (static_cast <Cpp_MemberDeclarator_ConstantInitializer const *> (el), nest_level + 0);
            break;
        case Cpp_MemberDeclarator::_Bitfield:
            dump_cpp_MemberDeclarator_Bitfield (static_cast <Cpp_MemberDeclarator_Bitfield const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static ParserElement*
cpp_PureSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_PureSpecifier))) Cpp_PureSpecifier;
}

static Ref<Grammar>
create_cpp_PureSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_PureSpecifier_creation_func));
    grammar->name = String::forData ("PureSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("0"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_PureSpecifier (Cpp_PureSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) PureSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("PureSpecifier\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: 0\n");
}

static void
cpp_ConstantInitializer_set_constantExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConstantInitializer ||
             __el->cpp_element_type != CppElement::_ConstantExpression);

    Cpp_ConstantInitializer * const &self = static_cast <Cpp_ConstantInitializer*> (__self);

    self->constantExpression = static_cast <Cpp_ConstantExpression*> (___el);
}

static ParserElement*
cpp_ConstantInitializer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConstantInitializer))) Cpp_ConstantInitializer;
}

static Ref<Grammar>
create_cpp_ConstantInitializer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConstantInitializer_creation_func));
    grammar->name = String::forData ("ConstantInitializer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConstantExpression ();
        entry->assignment_func = cpp_ConstantInitializer_set_constantExpression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConstantInitializer (Cpp_ConstantInitializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConstantInitializer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConstantInitializer\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_ConstantExpression (el->constantExpression, nest_level + 1);

}

static void
cpp_BaseClause_set_baseSpecifierList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseClause ||
             __el->cpp_element_type != CppElement::_BaseSpecifierList);

    Cpp_BaseClause * const &self = static_cast <Cpp_BaseClause*> (__self);

    self->baseSpecifierList = static_cast <Cpp_BaseSpecifierList*> (___el);
}

static ParserElement*
cpp_BaseClause_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BaseClause))) Cpp_BaseClause;
}

static Ref<Grammar>
create_cpp_BaseClause ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BaseClause_creation_func));
    grammar->name = String::forData ("BaseClause");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifierList ();
        entry->assignment_func = cpp_BaseClause_set_baseSpecifierList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseClause (Cpp_BaseClause const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseClause\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BaseClause\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_BaseSpecifierList (el->baseSpecifierList, nest_level + 1);

}

static void
cpp_BaseSpecifierList_BaseSpecifierList_set_baseSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifierList ||
             __el->cpp_element_type != CppElement::_BaseSpecifier);

    Cpp_BaseSpecifierList * const &_self = static_cast <Cpp_BaseSpecifierList*> (__self);

    abortIf (_self->base_specifier_list_type != Cpp_BaseSpecifierList::_BaseSpecifierList);

    Cpp_BaseSpecifierList_BaseSpecifierList * const &self = static_cast <Cpp_BaseSpecifierList_BaseSpecifierList*> (_self);

    self->baseSpecifier = static_cast <Cpp_BaseSpecifier*> (___el);
}

static void
cpp_BaseSpecifierList_BaseSpecifierList_set_baseSpecifierList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifierList ||
             __el->cpp_element_type != CppElement::_BaseSpecifierList);

    Cpp_BaseSpecifierList * const &_self = static_cast <Cpp_BaseSpecifierList*> (__self);

    abortIf (_self->base_specifier_list_type != Cpp_BaseSpecifierList::_BaseSpecifierList);

    Cpp_BaseSpecifierList_BaseSpecifierList * const &self = static_cast <Cpp_BaseSpecifierList_BaseSpecifierList*> (_self);

    self->baseSpecifierList = static_cast <Cpp_BaseSpecifierList*> (___el);
}

static ParserElement*
cpp_BaseSpecifierList_BaseSpecifierList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BaseSpecifierList_BaseSpecifierList))) Cpp_BaseSpecifierList_BaseSpecifierList;
}

static Ref<Grammar>
create_cpp_BaseSpecifierList_BaseSpecifierList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BaseSpecifierList_BaseSpecifierList_creation_func));
    grammar->name = String::forData ("BaseSpecifierList_BaseSpecifierList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifier ();
        entry->assignment_func = cpp_BaseSpecifierList_BaseSpecifierList_set_baseSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifierList ();
        entry->assignment_func = cpp_BaseSpecifierList_BaseSpecifierList_set_baseSpecifierList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifierList_BaseSpecifierList (Cpp_BaseSpecifierList_BaseSpecifierList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifierList_BaseSpecifierList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BaseSpecifierList_BaseSpecifierList\n");

    dump_cpp_BaseSpecifier (el->baseSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_BaseSpecifierList (el->baseSpecifierList, nest_level + 1);

}

static void
cpp_BaseSpecifierList_BaseSpecifier_set_baseSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifierList ||
             __el->cpp_element_type != CppElement::_BaseSpecifier);

    Cpp_BaseSpecifierList * const &_self = static_cast <Cpp_BaseSpecifierList*> (__self);

    abortIf (_self->base_specifier_list_type != Cpp_BaseSpecifierList::_BaseSpecifier);

    Cpp_BaseSpecifierList_BaseSpecifier * const &self = static_cast <Cpp_BaseSpecifierList_BaseSpecifier*> (_self);

    self->baseSpecifier = static_cast <Cpp_BaseSpecifier*> (___el);
}

static ParserElement*
cpp_BaseSpecifierList_BaseSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BaseSpecifierList_BaseSpecifier))) Cpp_BaseSpecifierList_BaseSpecifier;
}

static Ref<Grammar>
create_cpp_BaseSpecifierList_BaseSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BaseSpecifierList_BaseSpecifier_creation_func));
    grammar->name = String::forData ("BaseSpecifierList_BaseSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifier ();
        entry->assignment_func = cpp_BaseSpecifierList_BaseSpecifier_set_baseSpecifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifierList_BaseSpecifier (Cpp_BaseSpecifierList_BaseSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifierList_BaseSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BaseSpecifierList_BaseSpecifier\n");

    dump_cpp_BaseSpecifier (el->baseSpecifier, nest_level + 1);

}

static Ref<Grammar>
create_cpp_BaseSpecifierList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("BaseSpecifierList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifierList_BaseSpecifierList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifierList_BaseSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifierList (Cpp_BaseSpecifierList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifierList\n");
        return;
    }

    switch (el->base_specifier_list_type) {
        case Cpp_BaseSpecifierList::_BaseSpecifierList:
            dump_cpp_BaseSpecifierList_BaseSpecifierList (static_cast <Cpp_BaseSpecifierList_BaseSpecifierList const *> (el), nest_level + 0);
            break;
        case Cpp_BaseSpecifierList::_BaseSpecifier:
            dump_cpp_BaseSpecifierList_BaseSpecifier (static_cast <Cpp_BaseSpecifierList_BaseSpecifier const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

void cpp_BaseSpecifier_accept_func (
        Cpp_BaseSpecifier *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

void __pargen_cpp_BaseSpecifier_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    if (parser_element == NULL) {
        cpp_BaseSpecifier_accept_func (NULL, parser_control, data);
        return;
    }

    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_BaseSpecifier);

    Cpp_BaseSpecifier * const &el = static_cast <Cpp_BaseSpecifier*> (_el);

    cpp_BaseSpecifier_accept_func (el, parser_control, data);
}

static void
cpp_BaseSpecifier_Base_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Base);

    Cpp_BaseSpecifier_Base * const &self = static_cast <Cpp_BaseSpecifier_Base*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_BaseSpecifier_Base_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Base);

    Cpp_BaseSpecifier_Base * const &self = static_cast <Cpp_BaseSpecifier_Base*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_BaseSpecifier_Base_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Base);

    Cpp_BaseSpecifier_Base * const &self = static_cast <Cpp_BaseSpecifier_Base*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_BaseSpecifier_Base_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BaseSpecifier_Base))) Cpp_BaseSpecifier_Base;
}

static Ref<Grammar>
create_cpp_BaseSpecifier_Base ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BaseSpecifier_Base_creation_func));
    grammar->name = String::forData ("BaseSpecifier_Base");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_BaseSpecifier_Base_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_BaseSpecifier_Base_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_BaseSpecifier_Base_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifier_Base (Cpp_BaseSpecifier_Base const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifier_Base\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BaseSpecifier_Base\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_ClassName (el->className, nest_level + 1);

}

static void
cpp_BaseSpecifier_Virtual_set_accessSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_AccessSpecifier);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Virtual);

    Cpp_BaseSpecifier_Virtual * const &self = static_cast <Cpp_BaseSpecifier_Virtual*> (_self);

    self->accessSpecifier = static_cast <Cpp_AccessSpecifier*> (___el);
}

static void
cpp_BaseSpecifier_Virtual_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Virtual);

    Cpp_BaseSpecifier_Virtual * const &self = static_cast <Cpp_BaseSpecifier_Virtual*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_BaseSpecifier_Virtual_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Virtual);

    Cpp_BaseSpecifier_Virtual * const &self = static_cast <Cpp_BaseSpecifier_Virtual*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_BaseSpecifier_Virtual_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_Virtual);

    Cpp_BaseSpecifier_Virtual * const &self = static_cast <Cpp_BaseSpecifier_Virtual*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_BaseSpecifier_Virtual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BaseSpecifier_Virtual))) Cpp_BaseSpecifier_Virtual;
}

static Ref<Grammar>
create_cpp_BaseSpecifier_Virtual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BaseSpecifier_Virtual_creation_func));
    grammar->name = String::forData ("BaseSpecifier_Virtual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("virtual"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_AccessSpecifier ();
        entry->assignment_func = cpp_BaseSpecifier_Virtual_set_accessSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_BaseSpecifier_Virtual_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_BaseSpecifier_Virtual_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_BaseSpecifier_Virtual_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifier_Virtual (Cpp_BaseSpecifier_Virtual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifier_Virtual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BaseSpecifier_Virtual\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: virtual\n");
    dump_cpp_AccessSpecifier (el->accessSpecifier, nest_level + 1);

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_ClassName (el->className, nest_level + 1);

}

static void
cpp_BaseSpecifier_AccessSpecifier_set_accessSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_AccessSpecifier);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_AccessSpecifier);

    Cpp_BaseSpecifier_AccessSpecifier * const &self = static_cast <Cpp_BaseSpecifier_AccessSpecifier*> (_self);

    self->accessSpecifier = static_cast <Cpp_AccessSpecifier*> (___el);
}

static void
cpp_BaseSpecifier_AccessSpecifier_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_AccessSpecifier);

    Cpp_BaseSpecifier_AccessSpecifier * const &self = static_cast <Cpp_BaseSpecifier_AccessSpecifier*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_BaseSpecifier_AccessSpecifier_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_AccessSpecifier);

    Cpp_BaseSpecifier_AccessSpecifier * const &self = static_cast <Cpp_BaseSpecifier_AccessSpecifier*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_BaseSpecifier_AccessSpecifier_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_BaseSpecifier ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_BaseSpecifier * const &_self = static_cast <Cpp_BaseSpecifier*> (__self);

    abortIf (_self->base_specifier_type != Cpp_BaseSpecifier::_AccessSpecifier);

    Cpp_BaseSpecifier_AccessSpecifier * const &self = static_cast <Cpp_BaseSpecifier_AccessSpecifier*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_BaseSpecifier_AccessSpecifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_BaseSpecifier_AccessSpecifier))) Cpp_BaseSpecifier_AccessSpecifier;
}

static Ref<Grammar>
create_cpp_BaseSpecifier_AccessSpecifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_BaseSpecifier_AccessSpecifier_creation_func));
    grammar->name = String::forData ("BaseSpecifier_AccessSpecifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AccessSpecifier ();
        entry->assignment_func = cpp_BaseSpecifier_AccessSpecifier_set_accessSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("virtual"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_BaseSpecifier_AccessSpecifier_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_BaseSpecifier_AccessSpecifier_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_BaseSpecifier_AccessSpecifier_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifier_AccessSpecifier (Cpp_BaseSpecifier_AccessSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifier_AccessSpecifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("BaseSpecifier_AccessSpecifier\n");

    dump_cpp_AccessSpecifier (el->accessSpecifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: virtual_opt\n");
    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_ClassName (el->className, nest_level + 1);

}

static Ref<Grammar>
create_cpp_BaseSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("BaseSpecifier");
    grammar->accept_func = __pargen_cpp_BaseSpecifier_accept_func;

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifier_Base ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifier_Virtual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_BaseSpecifier_AccessSpecifier ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_BaseSpecifier (Cpp_BaseSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) BaseSpecifier\n");
        return;
    }

    switch (el->base_specifier_type) {
        case Cpp_BaseSpecifier::_Base:
            dump_cpp_BaseSpecifier_Base (static_cast <Cpp_BaseSpecifier_Base const *> (el), nest_level + 0);
            break;
        case Cpp_BaseSpecifier::_Virtual:
            dump_cpp_BaseSpecifier_Virtual (static_cast <Cpp_BaseSpecifier_Virtual const *> (el), nest_level + 0);
            break;
        case Cpp_BaseSpecifier::_AccessSpecifier:
            dump_cpp_BaseSpecifier_AccessSpecifier (static_cast <Cpp_BaseSpecifier_AccessSpecifier const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static ParserElement*
cpp_AccessSpecifier_Private_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AccessSpecifier_Private))) Cpp_AccessSpecifier_Private;
}

static Ref<Grammar>
create_cpp_AccessSpecifier_Private ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AccessSpecifier_Private_creation_func));
    grammar->name = String::forData ("AccessSpecifier_Private");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("private"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AccessSpecifier_Private (Cpp_AccessSpecifier_Private const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AccessSpecifier_Private\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AccessSpecifier_Private\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: private\n");
}

static ParserElement*
cpp_AccessSpecifier_Protected_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AccessSpecifier_Protected))) Cpp_AccessSpecifier_Protected;
}

static Ref<Grammar>
create_cpp_AccessSpecifier_Protected ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AccessSpecifier_Protected_creation_func));
    grammar->name = String::forData ("AccessSpecifier_Protected");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("protected"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AccessSpecifier_Protected (Cpp_AccessSpecifier_Protected const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AccessSpecifier_Protected\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AccessSpecifier_Protected\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: protected\n");
}

static ParserElement*
cpp_AccessSpecifier_Public_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_AccessSpecifier_Public))) Cpp_AccessSpecifier_Public;
}

static Ref<Grammar>
create_cpp_AccessSpecifier_Public ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_AccessSpecifier_Public_creation_func));
    grammar->name = String::forData ("AccessSpecifier_Public");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("public"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AccessSpecifier_Public (Cpp_AccessSpecifier_Public const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AccessSpecifier_Public\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("AccessSpecifier_Public\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: public\n");
}

static Ref<Grammar>
create_cpp_AccessSpecifier ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("AccessSpecifier");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AccessSpecifier_Private ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AccessSpecifier_Protected ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_AccessSpecifier_Public ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_AccessSpecifier (Cpp_AccessSpecifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) AccessSpecifier\n");
        return;
    }

    switch (el->access_specifier_type) {
        case Cpp_AccessSpecifier::_Private:
            dump_cpp_AccessSpecifier_Private (static_cast <Cpp_AccessSpecifier_Private const *> (el), nest_level + 0);
            break;
        case Cpp_AccessSpecifier::_Protected:
            dump_cpp_AccessSpecifier_Protected (static_cast <Cpp_AccessSpecifier_Protected const *> (el), nest_level + 0);
            break;
        case Cpp_AccessSpecifier::_Public:
            dump_cpp_AccessSpecifier_Public (static_cast <Cpp_AccessSpecifier_Public const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ConversionFunctionId_set_conversionTypeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConversionFunctionId ||
             __el->cpp_element_type != CppElement::_ConversionTypeId);

    Cpp_ConversionFunctionId * const &self = static_cast <Cpp_ConversionFunctionId*> (__self);

    self->conversionTypeId = static_cast <Cpp_ConversionTypeId*> (___el);
}

static ParserElement*
cpp_ConversionFunctionId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConversionFunctionId))) Cpp_ConversionFunctionId;
}

static Ref<Grammar>
create_cpp_ConversionFunctionId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConversionFunctionId_creation_func));
    grammar->name = String::forData ("ConversionFunctionId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("operator"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ConversionTypeId ();
        entry->assignment_func = cpp_ConversionFunctionId_set_conversionTypeId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConversionFunctionId (Cpp_ConversionFunctionId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConversionFunctionId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConversionFunctionId\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: operator\n");
    dump_cpp_ConversionTypeId (el->conversionTypeId, nest_level + 1);

}

static void
cpp_ConversionTypeId_set_typeSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConversionTypeId ||
             __el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_ConversionTypeId * const &self = static_cast <Cpp_ConversionTypeId*> (__self);

    self->typeSpecifierSeq = static_cast <Cpp_TypeSpecifierSeq*> (___el);
}

static void
cpp_ConversionTypeId_set_conversionDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConversionTypeId ||
             __el->cpp_element_type != CppElement::_ConversionDeclarator);

    Cpp_ConversionTypeId * const &self = static_cast <Cpp_ConversionTypeId*> (__self);

    self->conversionDeclarator = static_cast <Cpp_ConversionDeclarator*> (___el);
}

static ParserElement*
cpp_ConversionTypeId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConversionTypeId))) Cpp_ConversionTypeId;
}

static Ref<Grammar>
create_cpp_ConversionTypeId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConversionTypeId_creation_func));
    grammar->name = String::forData ("ConversionTypeId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_TypeSpecifierSeq ();
        entry->assignment_func = cpp_ConversionTypeId_set_typeSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ConversionDeclarator ();
        entry->assignment_func = cpp_ConversionTypeId_set_conversionDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConversionTypeId (Cpp_ConversionTypeId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConversionTypeId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConversionTypeId\n");

    dump_cpp_TypeSpecifierSeq (el->typeSpecifierSeq, nest_level + 1);

    dump_cpp_ConversionDeclarator (el->conversionDeclarator, nest_level + 1);

}

static void
cpp_ConversionDeclarator_set_ptrOperator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConversionDeclarator ||
             __el->cpp_element_type != CppElement::_PtrOperator);

    Cpp_ConversionDeclarator * const &self = static_cast <Cpp_ConversionDeclarator*> (__self);

    self->ptrOperator = static_cast <Cpp_PtrOperator*> (___el);
}

static void
cpp_ConversionDeclarator_set_conversionDeclarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ConversionDeclarator ||
             __el->cpp_element_type != CppElement::_ConversionDeclarator);

    Cpp_ConversionDeclarator * const &self = static_cast <Cpp_ConversionDeclarator*> (__self);

    self->conversionDeclarator = static_cast <Cpp_ConversionDeclarator*> (___el);
}

static ParserElement*
cpp_ConversionDeclarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ConversionDeclarator))) Cpp_ConversionDeclarator;
}

static Ref<Grammar>
create_cpp_ConversionDeclarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ConversionDeclarator_creation_func));
    grammar->name = String::forData ("ConversionDeclarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_PtrOperator ();
        entry->assignment_func = cpp_ConversionDeclarator_set_ptrOperator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ConversionDeclarator ();
        entry->assignment_func = cpp_ConversionDeclarator_set_conversionDeclarator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ConversionDeclarator (Cpp_ConversionDeclarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ConversionDeclarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ConversionDeclarator\n");

    dump_cpp_PtrOperator (el->ptrOperator, nest_level + 1);

    dump_cpp_ConversionDeclarator (el->conversionDeclarator, nest_level + 1);

}

static void
cpp_CtorInitializer_set_memInitializerList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_CtorInitializer ||
             __el->cpp_element_type != CppElement::_MemInitializerList);

    Cpp_CtorInitializer * const &self = static_cast <Cpp_CtorInitializer*> (__self);

    self->memInitializerList = static_cast <Cpp_MemInitializerList*> (___el);
}

static ParserElement*
cpp_CtorInitializer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_CtorInitializer))) Cpp_CtorInitializer;
}

static Ref<Grammar>
create_cpp_CtorInitializer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_CtorInitializer_creation_func));
    grammar->name = String::forData ("CtorInitializer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (":"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerList ();
        entry->assignment_func = cpp_CtorInitializer_set_memInitializerList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_CtorInitializer (Cpp_CtorInitializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) CtorInitializer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("CtorInitializer\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: :\n");
    dump_cpp_MemInitializerList (el->memInitializerList, nest_level + 1);

}

static void
cpp_MemInitializerList_List_set_memInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerList ||
             __el->cpp_element_type != CppElement::_MemInitializer);

    Cpp_MemInitializerList * const &_self = static_cast <Cpp_MemInitializerList*> (__self);

    abortIf (_self->mem_initializer_list_type != Cpp_MemInitializerList::_List);

    Cpp_MemInitializerList_List * const &self = static_cast <Cpp_MemInitializerList_List*> (_self);

    self->memInitializer = static_cast <Cpp_MemInitializer*> (___el);
}

static void
cpp_MemInitializerList_List_set_memInitializerList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerList ||
             __el->cpp_element_type != CppElement::_MemInitializerList);

    Cpp_MemInitializerList * const &_self = static_cast <Cpp_MemInitializerList*> (__self);

    abortIf (_self->mem_initializer_list_type != Cpp_MemInitializerList::_List);

    Cpp_MemInitializerList_List * const &self = static_cast <Cpp_MemInitializerList_List*> (_self);

    self->memInitializerList = static_cast <Cpp_MemInitializerList*> (___el);
}

static ParserElement*
cpp_MemInitializerList_List_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemInitializerList_List))) Cpp_MemInitializerList_List;
}

static Ref<Grammar>
create_cpp_MemInitializerList_List ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemInitializerList_List_creation_func));
    grammar->name = String::forData ("MemInitializerList_List");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemInitializer ();
        entry->assignment_func = cpp_MemInitializerList_List_set_memInitializer;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerList ();
        entry->assignment_func = cpp_MemInitializerList_List_set_memInitializerList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializerList_List (Cpp_MemInitializerList_List const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializerList_List\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemInitializerList_List\n");

    dump_cpp_MemInitializer (el->memInitializer, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_MemInitializerList (el->memInitializerList, nest_level + 1);

}

static void
cpp_MemInitializerList_MemInitializer_set_memInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerList ||
             __el->cpp_element_type != CppElement::_MemInitializer);

    Cpp_MemInitializerList * const &_self = static_cast <Cpp_MemInitializerList*> (__self);

    abortIf (_self->mem_initializer_list_type != Cpp_MemInitializerList::_MemInitializer);

    Cpp_MemInitializerList_MemInitializer * const &self = static_cast <Cpp_MemInitializerList_MemInitializer*> (_self);

    self->memInitializer = static_cast <Cpp_MemInitializer*> (___el);
}

static ParserElement*
cpp_MemInitializerList_MemInitializer_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemInitializerList_MemInitializer))) Cpp_MemInitializerList_MemInitializer;
}

static Ref<Grammar>
create_cpp_MemInitializerList_MemInitializer ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemInitializerList_MemInitializer_creation_func));
    grammar->name = String::forData ("MemInitializerList_MemInitializer");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemInitializer ();
        entry->assignment_func = cpp_MemInitializerList_MemInitializer_set_memInitializer;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializerList_MemInitializer (Cpp_MemInitializerList_MemInitializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializerList_MemInitializer\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemInitializerList_MemInitializer\n");

    dump_cpp_MemInitializer (el->memInitializer, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MemInitializerList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemInitializerList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerList_List ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerList_MemInitializer ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializerList (Cpp_MemInitializerList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializerList\n");
        return;
    }

    switch (el->mem_initializer_list_type) {
        case Cpp_MemInitializerList::_List:
            dump_cpp_MemInitializerList_List (static_cast <Cpp_MemInitializerList_List const *> (el), nest_level + 0);
            break;
        case Cpp_MemInitializerList::_MemInitializer:
            dump_cpp_MemInitializerList_MemInitializer (static_cast <Cpp_MemInitializerList_MemInitializer const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MemInitializer_Default_set_memInitializerId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializer ||
             __el->cpp_element_type != CppElement::_MemInitializerId);

    Cpp_MemInitializer * const &_self = static_cast <Cpp_MemInitializer*> (__self);

    abortIf (_self->mem_initializer_type != Cpp_MemInitializer::_Default);

    Cpp_MemInitializer_Default * const &self = static_cast <Cpp_MemInitializer_Default*> (_self);

    self->memInitializerId = static_cast <Cpp_MemInitializerId*> (___el);
}

static void
cpp_MemInitializer_Default_set_expressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializer ||
             __el->cpp_element_type != CppElement::_ExpressionList);

    Cpp_MemInitializer * const &_self = static_cast <Cpp_MemInitializer*> (__self);

    abortIf (_self->mem_initializer_type != Cpp_MemInitializer::_Default);

    Cpp_MemInitializer_Default * const &self = static_cast <Cpp_MemInitializer_Default*> (_self);

    self->expressionList = static_cast <Cpp_ExpressionList*> (___el);
}

bool cpp_MemInitializer_accept_func (
        Cpp_MemInitializer *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_MemInitializer_accept_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_MemInitializer);

    Cpp_MemInitializer * const &el =
            static_cast <Cpp_MemInitializer*> (_el);

    return cpp_MemInitializer_accept_func (el, parser_control, data);
}

static ParserElement*
cpp_MemInitializer_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemInitializer_Default))) Cpp_MemInitializer_Default;
}

static Ref<Grammar>
create_cpp_MemInitializer_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemInitializer_Default_creation_func));
    grammar->name = String::forData ("MemInitializer_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerId ();
        entry->assignment_func = cpp_MemInitializer_Default_set_memInitializerId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_ExpressionList ();
        entry->assignment_func = cpp_MemInitializer_Default_set_expressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_MemInitializer_accept_func;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializer_Default (Cpp_MemInitializer_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializer_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemInitializer_Default\n");

    dump_cpp_MemInitializerId (el->memInitializerId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExpressionList (el->expressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_MemInitializer_ClassPrematch_set_scruffy_NotAParenthesis (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializer ||
             __el->cpp_element_type != CppElement::_Scruffy_NotAParenthesis);

    Cpp_MemInitializer * const &_self = static_cast <Cpp_MemInitializer*> (__self);

    abortIf (_self->mem_initializer_type != Cpp_MemInitializer::_ClassPrematch);

    Cpp_MemInitializer_ClassPrematch * const &self = static_cast <Cpp_MemInitializer_ClassPrematch*> (_self);

    self->scruffy_NotAParenthesiss.append (static_cast <Cpp_Scruffy_NotAParenthesis*> (__el));
}

static void
cpp_MemInitializer_ClassPrematch_set_scruffy_SkipExpressionList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializer ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipExpressionList);

    Cpp_MemInitializer * const &_self = static_cast <Cpp_MemInitializer*> (__self);

    abortIf (_self->mem_initializer_type != Cpp_MemInitializer::_ClassPrematch);

    Cpp_MemInitializer_ClassPrematch * const &self = static_cast <Cpp_MemInitializer_ClassPrematch*> (_self);

    self->scruffy_SkipExpressionList = static_cast <Cpp_Scruffy_SkipExpressionList*> (___el);
}

static ParserElement*
cpp_MemInitializer_ClassPrematch_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemInitializer_ClassPrematch))) Cpp_MemInitializer_ClassPrematch;
}

static Ref<Grammar>
create_cpp_MemInitializer_ClassPrematch ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemInitializer_ClassPrematch_creation_func));
    grammar->name = String::forData ("MemInitializer_ClassPrematch");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Sequence;
        entry->grammar = create_cpp_Scruffy_NotAParenthesis ();
        entry->assignment_func = cpp_MemInitializer_ClassPrematch_set_scruffy_NotAParenthesis;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_SkipExpressionList ();
        entry->assignment_func = cpp_MemInitializer_ClassPrematch_set_scruffy_SkipExpressionList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializer_ClassPrematch (Cpp_MemInitializer_ClassPrematch const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializer_ClassPrematch\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemInitializer_ClassPrematch\n");

    {
        List<Cpp_Scruffy_NotAParenthesis*>::DataIterator sub_iter (el->scruffy_NotAParenthesiss);
        while (!sub_iter.done ()) {
            Cpp_Scruffy_NotAParenthesis *&sub = sub_iter.next ();
            dump_cpp_Scruffy_NotAParenthesis (sub, nest_level + 1);
        }
    }

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Scruffy_SkipExpressionList (el->scruffy_SkipExpressionList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static Ref<Grammar>
create_cpp_MemInitializer ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemInitializer");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemInitializer_Default ();
    entry->variants.append (String::forData ("default"));
    entry->variants.append (String::forData ("AST"));

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemInitializer_ClassPrematch ();
    entry->variants.append (String::forData ("class-prematch"));

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializer (Cpp_MemInitializer const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializer\n");
        return;
    }

    switch (el->mem_initializer_type) {
        case Cpp_MemInitializer::_Default:
            dump_cpp_MemInitializer_Default (static_cast <Cpp_MemInitializer_Default const *> (el), nest_level + 0);
            break;
        case Cpp_MemInitializer::_ClassPrematch:
            dump_cpp_MemInitializer_ClassPrematch (static_cast <Cpp_MemInitializer_ClassPrematch const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_MemInitializerId_ClassName_set_scruffy_GlobalScope (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerId ||
             __el->cpp_element_type != CppElement::_Scruffy_GlobalScope);

    Cpp_MemInitializerId * const &_self = static_cast <Cpp_MemInitializerId*> (__self);

    abortIf (_self->mem_initializer_id_type != Cpp_MemInitializerId::_ClassName);

    Cpp_MemInitializerId_ClassName * const &self = static_cast <Cpp_MemInitializerId_ClassName*> (_self);

    self->scruffy_GlobalScope = static_cast <Cpp_Scruffy_GlobalScope*> (___el);
}

static void
cpp_MemInitializerId_ClassName_set_nestedNameSpecifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerId ||
             __el->cpp_element_type != CppElement::_NestedNameSpecifier);

    Cpp_MemInitializerId * const &_self = static_cast <Cpp_MemInitializerId*> (__self);

    abortIf (_self->mem_initializer_id_type != Cpp_MemInitializerId::_ClassName);

    Cpp_MemInitializerId_ClassName * const &self = static_cast <Cpp_MemInitializerId_ClassName*> (_self);

    self->nestedNameSpecifier = static_cast <Cpp_NestedNameSpecifier*> (___el);
}

static void
cpp_MemInitializerId_ClassName_set_className (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerId ||
             __el->cpp_element_type != CppElement::_ClassName);

    Cpp_MemInitializerId * const &_self = static_cast <Cpp_MemInitializerId*> (__self);

    abortIf (_self->mem_initializer_id_type != Cpp_MemInitializerId::_ClassName);

    Cpp_MemInitializerId_ClassName * const &self = static_cast <Cpp_MemInitializerId_ClassName*> (_self);

    self->className = static_cast <Cpp_ClassName*> (___el);
}

static ParserElement*
cpp_MemInitializerId_ClassName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemInitializerId_ClassName))) Cpp_MemInitializerId_ClassName;
}

static Ref<Grammar>
create_cpp_MemInitializerId_ClassName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemInitializerId_ClassName_creation_func));
    grammar->name = String::forData ("MemInitializerId_ClassName");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_nested_name;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_GlobalScope ();
        entry->assignment_func = cpp_MemInitializerId_ClassName_set_scruffy_GlobalScope;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_NestedNameSpecifier ();
        entry->assignment_func = cpp_MemInitializerId_ClassName_set_nestedNameSpecifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ClassName ();
        entry->assignment_func = cpp_MemInitializerId_ClassName_set_className;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializerId_ClassName (Cpp_MemInitializerId_ClassName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializerId_ClassName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemInitializerId_ClassName\n");

    dump_cpp_Scruffy_GlobalScope (el->scruffy_GlobalScope, nest_level + 1);

    dump_cpp_NestedNameSpecifier (el->nestedNameSpecifier, nest_level + 1);

    dump_cpp_ClassName (el->className, nest_level + 1);

}

static void
cpp_MemInitializerId_Identifier_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_MemInitializerId ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_MemInitializerId * const &_self = static_cast <Cpp_MemInitializerId*> (__self);

    abortIf (_self->mem_initializer_id_type != Cpp_MemInitializerId::_Identifier);

    Cpp_MemInitializerId_Identifier * const &self = static_cast <Cpp_MemInitializerId_Identifier*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_MemInitializerId_Identifier_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_MemInitializerId_Identifier))) Cpp_MemInitializerId_Identifier;
}

static Ref<Grammar>
create_cpp_MemInitializerId_Identifier ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_MemInitializerId_Identifier_creation_func));
    grammar->name = String::forData ("MemInitializerId_Identifier");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_MemInitializerId_Identifier_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializerId_Identifier (Cpp_MemInitializerId_Identifier const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializerId_Identifier\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("MemInitializerId_Identifier\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static Ref<Grammar>
create_cpp_MemInitializerId ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("MemInitializerId");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerId_ClassName ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_MemInitializerId_Identifier ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_MemInitializerId (Cpp_MemInitializerId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) MemInitializerId\n");
        return;
    }

    switch (el->mem_initializer_id_type) {
        case Cpp_MemInitializerId::_ClassName:
            dump_cpp_MemInitializerId_ClassName (static_cast <Cpp_MemInitializerId_ClassName const *> (el), nest_level + 0);
            break;
        case Cpp_MemInitializerId::_Identifier:
            dump_cpp_MemInitializerId_Identifier (static_cast <Cpp_MemInitializerId_Identifier const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_OperatorFunctionId_set__operator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_OperatorFunctionId ||
             __el->cpp_element_type != CppElement::_Operator);

    Cpp_OperatorFunctionId * const &self = static_cast <Cpp_OperatorFunctionId*> (__self);

    self->_operator = static_cast <Cpp_Operator*> (___el);
}

static ParserElement*
cpp_OperatorFunctionId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_OperatorFunctionId))) Cpp_OperatorFunctionId;
}

static Ref<Grammar>
create_cpp_OperatorFunctionId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_OperatorFunctionId_creation_func));
    grammar->name = String::forData ("OperatorFunctionId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("operator"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Operator ();
        entry->assignment_func = cpp_OperatorFunctionId_set__operator;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_OperatorFunctionId (Cpp_OperatorFunctionId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) OperatorFunctionId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("OperatorFunctionId\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: operator\n");
    dump_cpp_Operator (el->_operator, nest_level + 1);

}

static ParserElement*
cpp_Operator_New_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_New))) Cpp_Operator_New;
}

static Ref<Grammar>
create_cpp_Operator_New ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_New_creation_func));
    grammar->name = String::forData ("Operator_New");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("new"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_New (Cpp_Operator_New const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_New\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_New\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: new\n");
}

static ParserElement*
cpp_Operator_Delete_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Delete))) Cpp_Operator_Delete;
}

static Ref<Grammar>
create_cpp_Operator_Delete ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Delete_creation_func));
    grammar->name = String::forData ("Operator_Delete");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("delete"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Delete (Cpp_Operator_Delete const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Delete\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Delete\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: delete\n");
}

static ParserElement*
cpp_Operator_NewArray_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_NewArray))) Cpp_Operator_NewArray;
}

static Ref<Grammar>
create_cpp_Operator_NewArray ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_NewArray_creation_func));
    grammar->name = String::forData ("Operator_NewArray");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("new"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_NewArray (Cpp_Operator_NewArray const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_NewArray\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_NewArray\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: new\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static ParserElement*
cpp_Operator_DeleteArray_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_DeleteArray))) Cpp_Operator_DeleteArray;
}

static Ref<Grammar>
create_cpp_Operator_DeleteArray ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_DeleteArray_creation_func));
    grammar->name = String::forData ("Operator_DeleteArray");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("delete"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_DeleteArray (Cpp_Operator_DeleteArray const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_DeleteArray\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_DeleteArray\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: delete\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static ParserElement*
cpp_Operator_Plus_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Plus))) Cpp_Operator_Plus;
}

static Ref<Grammar>
create_cpp_Operator_Plus ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Plus_creation_func));
    grammar->name = String::forData ("Operator_Plus");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("+"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Plus (Cpp_Operator_Plus const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Plus\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Plus\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: +\n");
}

static ParserElement*
cpp_Operator_Minus_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Minus))) Cpp_Operator_Minus;
}

static Ref<Grammar>
create_cpp_Operator_Minus ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Minus_creation_func));
    grammar->name = String::forData ("Operator_Minus");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("-"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Minus (Cpp_Operator_Minus const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Minus\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Minus\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: -\n");
}

static ParserElement*
cpp_Operator_Multiply_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Multiply))) Cpp_Operator_Multiply;
}

static Ref<Grammar>
create_cpp_Operator_Multiply ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Multiply_creation_func));
    grammar->name = String::forData ("Operator_Multiply");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Multiply (Cpp_Operator_Multiply const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Multiply\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Multiply\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: *\n");
}

static ParserElement*
cpp_Operator_Divide_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Divide))) Cpp_Operator_Divide;
}

static Ref<Grammar>
create_cpp_Operator_Divide ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Divide_creation_func));
    grammar->name = String::forData ("Operator_Divide");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("/"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Divide (Cpp_Operator_Divide const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Divide\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Divide\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: /\n");
}

static ParserElement*
cpp_Operator_Remainder_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Remainder))) Cpp_Operator_Remainder;
}

static Ref<Grammar>
create_cpp_Operator_Remainder ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Remainder_creation_func));
    grammar->name = String::forData ("Operator_Remainder");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("%"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Remainder (Cpp_Operator_Remainder const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Remainder\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Remainder\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: %\n");
}

static ParserElement*
cpp_Operator_ExclusiveOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_ExclusiveOr))) Cpp_Operator_ExclusiveOr;
}

static Ref<Grammar>
create_cpp_Operator_ExclusiveOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_ExclusiveOr_creation_func));
    grammar->name = String::forData ("Operator_ExclusiveOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("^"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_ExclusiveOr (Cpp_Operator_ExclusiveOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_ExclusiveOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_ExclusiveOr\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ^\n");
}

static ParserElement*
cpp_Operator_And_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_And))) Cpp_Operator_And;
}

static Ref<Grammar>
create_cpp_Operator_And ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_And_creation_func));
    grammar->name = String::forData ("Operator_And");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_And (Cpp_Operator_And const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_And\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_And\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: &\n");
}

static ParserElement*
cpp_Operator_InclusiveOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_InclusiveOr))) Cpp_Operator_InclusiveOr;
}

static Ref<Grammar>
create_cpp_Operator_InclusiveOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_InclusiveOr_creation_func));
    grammar->name = String::forData ("Operator_InclusiveOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("|"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_InclusiveOr (Cpp_Operator_InclusiveOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_InclusiveOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_InclusiveOr\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: |\n");
}

static ParserElement*
cpp_Operator_Inversion_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Inversion))) Cpp_Operator_Inversion;
}

static Ref<Grammar>
create_cpp_Operator_Inversion ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Inversion_creation_func));
    grammar->name = String::forData ("Operator_Inversion");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("~"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Inversion (Cpp_Operator_Inversion const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Inversion\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Inversion\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ~\n");
}

static ParserElement*
cpp_Operator_Not_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Not))) Cpp_Operator_Not;
}

static Ref<Grammar>
create_cpp_Operator_Not ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Not_creation_func));
    grammar->name = String::forData ("Operator_Not");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("!"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Not (Cpp_Operator_Not const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Not\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Not\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: !\n");
}

static ParserElement*
cpp_Operator_Set_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Set))) Cpp_Operator_Set;
}

static Ref<Grammar>
create_cpp_Operator_Set ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Set_creation_func));
    grammar->name = String::forData ("Operator_Set");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Set (Cpp_Operator_Set const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Set\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Set\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
}

static ParserElement*
cpp_Operator_Less_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Less))) Cpp_Operator_Less;
}

static Ref<Grammar>
create_cpp_Operator_Less ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Less_creation_func));
    grammar->name = String::forData ("Operator_Less");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Less (Cpp_Operator_Less const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Less\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Less\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
}

static ParserElement*
cpp_Operator_Greater_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Greater))) Cpp_Operator_Greater;
}

static Ref<Grammar>
create_cpp_Operator_Greater ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Greater_creation_func));
    grammar->name = String::forData ("Operator_Greater");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Greater (Cpp_Operator_Greater const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Greater\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Greater\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
}

static ParserElement*
cpp_Operator_PlusSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_PlusSet))) Cpp_Operator_PlusSet;
}

static Ref<Grammar>
create_cpp_Operator_PlusSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_PlusSet_creation_func));
    grammar->name = String::forData ("Operator_PlusSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("+="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_PlusSet (Cpp_Operator_PlusSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_PlusSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_PlusSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: +=\n");
}

static ParserElement*
cpp_Operator_MinusSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_MinusSet))) Cpp_Operator_MinusSet;
}

static Ref<Grammar>
create_cpp_Operator_MinusSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_MinusSet_creation_func));
    grammar->name = String::forData ("Operator_MinusSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("-="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_MinusSet (Cpp_Operator_MinusSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_MinusSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_MinusSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: -=\n");
}

static ParserElement*
cpp_Operator_MultiplySet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_MultiplySet))) Cpp_Operator_MultiplySet;
}

static Ref<Grammar>
create_cpp_Operator_MultiplySet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_MultiplySet_creation_func));
    grammar->name = String::forData ("Operator_MultiplySet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("*="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_MultiplySet (Cpp_Operator_MultiplySet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_MultiplySet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_MultiplySet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: *=\n");
}

static ParserElement*
cpp_Operator_DivideSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_DivideSet))) Cpp_Operator_DivideSet;
}

static Ref<Grammar>
create_cpp_Operator_DivideSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_DivideSet_creation_func));
    grammar->name = String::forData ("Operator_DivideSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("/="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_DivideSet (Cpp_Operator_DivideSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_DivideSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_DivideSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: /=\n");
}

static ParserElement*
cpp_Operator_RemainderSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_RemainderSet))) Cpp_Operator_RemainderSet;
}

static Ref<Grammar>
create_cpp_Operator_RemainderSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_RemainderSet_creation_func));
    grammar->name = String::forData ("Operator_RemainderSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("%="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_RemainderSet (Cpp_Operator_RemainderSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_RemainderSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_RemainderSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: %=\n");
}

static ParserElement*
cpp_Operator_ExclusiveOrSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_ExclusiveOrSet))) Cpp_Operator_ExclusiveOrSet;
}

static Ref<Grammar>
create_cpp_Operator_ExclusiveOrSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_ExclusiveOrSet_creation_func));
    grammar->name = String::forData ("Operator_ExclusiveOrSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("^="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_ExclusiveOrSet (Cpp_Operator_ExclusiveOrSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_ExclusiveOrSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_ExclusiveOrSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ^=\n");
}

static ParserElement*
cpp_Operator_AndSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_AndSet))) Cpp_Operator_AndSet;
}

static Ref<Grammar>
create_cpp_Operator_AndSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_AndSet_creation_func));
    grammar->name = String::forData ("Operator_AndSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_AndSet (Cpp_Operator_AndSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_AndSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_AndSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: &=\n");
}

static ParserElement*
cpp_Operator_InclusiveOrSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_InclusiveOrSet))) Cpp_Operator_InclusiveOrSet;
}

static Ref<Grammar>
create_cpp_Operator_InclusiveOrSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_InclusiveOrSet_creation_func));
    grammar->name = String::forData ("Operator_InclusiveOrSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("|="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_InclusiveOrSet (Cpp_Operator_InclusiveOrSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_InclusiveOrSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_InclusiveOrSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: |=\n");
}

static ParserElement*
cpp_Operator_LeftShift_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_LeftShift))) Cpp_Operator_LeftShift;
}

static Ref<Grammar>
create_cpp_Operator_LeftShift ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_LeftShift_creation_func));
    grammar->name = String::forData ("Operator_LeftShift");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_LeftShift (Cpp_Operator_LeftShift const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_LeftShift\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_LeftShift\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: <<\n");
}

static ParserElement*
cpp_Operator_RightShift_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_RightShift))) Cpp_Operator_RightShift;
}

static Ref<Grammar>
create_cpp_Operator_RightShift ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_RightShift_creation_func));
    grammar->name = String::forData ("Operator_RightShift");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">>"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_RightShift (Cpp_Operator_RightShift const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_RightShift\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_RightShift\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: >>\n");
}

static ParserElement*
cpp_Operator_RightShiftSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_RightShiftSet))) Cpp_Operator_RightShiftSet;
}

static Ref<Grammar>
create_cpp_Operator_RightShiftSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_RightShiftSet_creation_func));
    grammar->name = String::forData ("Operator_RightShiftSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">>="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_RightShiftSet (Cpp_Operator_RightShiftSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_RightShiftSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_RightShiftSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: >>=\n");
}

static ParserElement*
cpp_Operator_LeftShiftSet_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_LeftShiftSet))) Cpp_Operator_LeftShiftSet;
}

static Ref<Grammar>
create_cpp_Operator_LeftShiftSet ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_LeftShiftSet_creation_func));
    grammar->name = String::forData ("Operator_LeftShiftSet");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<<="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_LeftShiftSet (Cpp_Operator_LeftShiftSet const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_LeftShiftSet\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_LeftShiftSet\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: <<=\n");
}

static ParserElement*
cpp_Operator_Equal_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Equal))) Cpp_Operator_Equal;
}

static Ref<Grammar>
create_cpp_Operator_Equal ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Equal_creation_func));
    grammar->name = String::forData ("Operator_Equal");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("=="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Equal (Cpp_Operator_Equal const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Equal\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Equal\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ==\n");
}

static ParserElement*
cpp_Operator_NotEqual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_NotEqual))) Cpp_Operator_NotEqual;
}

static Ref<Grammar>
create_cpp_Operator_NotEqual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_NotEqual_creation_func));
    grammar->name = String::forData ("Operator_NotEqual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("!="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_NotEqual (Cpp_Operator_NotEqual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_NotEqual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_NotEqual\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: !=\n");
}

static ParserElement*
cpp_Operator_LessOrEqual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_LessOrEqual))) Cpp_Operator_LessOrEqual;
}

static Ref<Grammar>
create_cpp_Operator_LessOrEqual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_LessOrEqual_creation_func));
    grammar->name = String::forData ("Operator_LessOrEqual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_LessOrEqual (Cpp_Operator_LessOrEqual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_LessOrEqual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_LessOrEqual\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: <=\n");
}

static ParserElement*
cpp_Operator_GreaterOrEqual_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_GreaterOrEqual))) Cpp_Operator_GreaterOrEqual;
}

static Ref<Grammar>
create_cpp_Operator_GreaterOrEqual ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_GreaterOrEqual_creation_func));
    grammar->name = String::forData ("Operator_GreaterOrEqual");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_GreaterOrEqual (Cpp_Operator_GreaterOrEqual const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_GreaterOrEqual\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_GreaterOrEqual\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: >=\n");
}

static ParserElement*
cpp_Operator_LogicalAnd_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_LogicalAnd))) Cpp_Operator_LogicalAnd;
}

static Ref<Grammar>
create_cpp_Operator_LogicalAnd ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_LogicalAnd_creation_func));
    grammar->name = String::forData ("Operator_LogicalAnd");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("&&"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_LogicalAnd (Cpp_Operator_LogicalAnd const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_LogicalAnd\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_LogicalAnd\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: &&\n");
}

static ParserElement*
cpp_Operator_LogicalOr_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_LogicalOr))) Cpp_Operator_LogicalOr;
}

static Ref<Grammar>
create_cpp_Operator_LogicalOr ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_LogicalOr_creation_func));
    grammar->name = String::forData ("Operator_LogicalOr");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("||"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_LogicalOr (Cpp_Operator_LogicalOr const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_LogicalOr\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_LogicalOr\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ||\n");
}

static ParserElement*
cpp_Operator_Increment_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Increment))) Cpp_Operator_Increment;
}

static Ref<Grammar>
create_cpp_Operator_Increment ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Increment_creation_func));
    grammar->name = String::forData ("Operator_Increment");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("++"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Increment (Cpp_Operator_Increment const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Increment\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Increment\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ++\n");
}

static ParserElement*
cpp_Operator_Decrement_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Decrement))) Cpp_Operator_Decrement;
}

static Ref<Grammar>
create_cpp_Operator_Decrement ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Decrement_creation_func));
    grammar->name = String::forData ("Operator_Decrement");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("--"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Decrement (Cpp_Operator_Decrement const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Decrement\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Decrement\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: --\n");
}

static ParserElement*
cpp_Operator_Comma_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Comma))) Cpp_Operator_Comma;
}

static Ref<Grammar>
create_cpp_Operator_Comma ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Comma_creation_func));
    grammar->name = String::forData ("Operator_Comma");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Comma (Cpp_Operator_Comma const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Comma\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Comma\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
}

static ParserElement*
cpp_Operator_MemberDereference_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_MemberDereference))) Cpp_Operator_MemberDereference;
}

static Ref<Grammar>
create_cpp_Operator_MemberDereference ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_MemberDereference_creation_func));
    grammar->name = String::forData ("Operator_MemberDereference");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("->*"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_MemberDereference (Cpp_Operator_MemberDereference const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_MemberDereference\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_MemberDereference\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ->*\n");
}

static ParserElement*
cpp_Operator_Dereference_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Dereference))) Cpp_Operator_Dereference;
}

static Ref<Grammar>
create_cpp_Operator_Dereference ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Dereference_creation_func));
    grammar->name = String::forData ("Operator_Dereference");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("->"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Dereference (Cpp_Operator_Dereference const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Dereference\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Dereference\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ->\n");
}

static ParserElement*
cpp_Operator_FunctionCall_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_FunctionCall))) Cpp_Operator_FunctionCall;
}

static Ref<Grammar>
create_cpp_Operator_FunctionCall ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_FunctionCall_creation_func));
    grammar->name = String::forData ("Operator_FunctionCall");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_FunctionCall (Cpp_Operator_FunctionCall const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_FunctionCall\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_FunctionCall\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static ParserElement*
cpp_Operator_Subscripting_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Operator_Subscripting))) Cpp_Operator_Subscripting;
}

static Ref<Grammar>
create_cpp_Operator_Subscripting ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Operator_Subscripting_creation_func));
    grammar->name = String::forData ("Operator_Subscripting");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("["));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("]"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator_Subscripting (Cpp_Operator_Subscripting const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator_Subscripting\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Operator_Subscripting\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: [\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: ]\n");
}

static Ref<Grammar>
create_cpp_Operator ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Operator");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_New ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Delete ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_NewArray ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_DeleteArray ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Plus ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Minus ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Multiply ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Divide ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Remainder ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_ExclusiveOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_And ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_InclusiveOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Inversion ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Not ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Set ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Less ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Greater ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_PlusSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_MinusSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_MultiplySet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_DivideSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_RemainderSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_ExclusiveOrSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_AndSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_InclusiveOrSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_LeftShift ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_RightShift ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_RightShiftSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_LeftShiftSet ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Equal ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_NotEqual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_LessOrEqual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_GreaterOrEqual ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_LogicalAnd ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_LogicalOr ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Increment ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Decrement ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Comma ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_MemberDereference ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Dereference ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_FunctionCall ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Operator_Subscripting ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Operator (Cpp_Operator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Operator\n");
        return;
    }

    switch (el->operator_type) {
        case Cpp_Operator::_New:
            dump_cpp_Operator_New (static_cast <Cpp_Operator_New const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Delete:
            dump_cpp_Operator_Delete (static_cast <Cpp_Operator_Delete const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_NewArray:
            dump_cpp_Operator_NewArray (static_cast <Cpp_Operator_NewArray const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_DeleteArray:
            dump_cpp_Operator_DeleteArray (static_cast <Cpp_Operator_DeleteArray const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Plus:
            dump_cpp_Operator_Plus (static_cast <Cpp_Operator_Plus const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Minus:
            dump_cpp_Operator_Minus (static_cast <Cpp_Operator_Minus const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Multiply:
            dump_cpp_Operator_Multiply (static_cast <Cpp_Operator_Multiply const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Divide:
            dump_cpp_Operator_Divide (static_cast <Cpp_Operator_Divide const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Remainder:
            dump_cpp_Operator_Remainder (static_cast <Cpp_Operator_Remainder const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_ExclusiveOr:
            dump_cpp_Operator_ExclusiveOr (static_cast <Cpp_Operator_ExclusiveOr const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_And:
            dump_cpp_Operator_And (static_cast <Cpp_Operator_And const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_InclusiveOr:
            dump_cpp_Operator_InclusiveOr (static_cast <Cpp_Operator_InclusiveOr const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Inversion:
            dump_cpp_Operator_Inversion (static_cast <Cpp_Operator_Inversion const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Not:
            dump_cpp_Operator_Not (static_cast <Cpp_Operator_Not const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Set:
            dump_cpp_Operator_Set (static_cast <Cpp_Operator_Set const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Less:
            dump_cpp_Operator_Less (static_cast <Cpp_Operator_Less const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Greater:
            dump_cpp_Operator_Greater (static_cast <Cpp_Operator_Greater const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_PlusSet:
            dump_cpp_Operator_PlusSet (static_cast <Cpp_Operator_PlusSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_MinusSet:
            dump_cpp_Operator_MinusSet (static_cast <Cpp_Operator_MinusSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_MultiplySet:
            dump_cpp_Operator_MultiplySet (static_cast <Cpp_Operator_MultiplySet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_DivideSet:
            dump_cpp_Operator_DivideSet (static_cast <Cpp_Operator_DivideSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_RemainderSet:
            dump_cpp_Operator_RemainderSet (static_cast <Cpp_Operator_RemainderSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_ExclusiveOrSet:
            dump_cpp_Operator_ExclusiveOrSet (static_cast <Cpp_Operator_ExclusiveOrSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_AndSet:
            dump_cpp_Operator_AndSet (static_cast <Cpp_Operator_AndSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_InclusiveOrSet:
            dump_cpp_Operator_InclusiveOrSet (static_cast <Cpp_Operator_InclusiveOrSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_LeftShift:
            dump_cpp_Operator_LeftShift (static_cast <Cpp_Operator_LeftShift const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_RightShift:
            dump_cpp_Operator_RightShift (static_cast <Cpp_Operator_RightShift const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_RightShiftSet:
            dump_cpp_Operator_RightShiftSet (static_cast <Cpp_Operator_RightShiftSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_LeftShiftSet:
            dump_cpp_Operator_LeftShiftSet (static_cast <Cpp_Operator_LeftShiftSet const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Equal:
            dump_cpp_Operator_Equal (static_cast <Cpp_Operator_Equal const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_NotEqual:
            dump_cpp_Operator_NotEqual (static_cast <Cpp_Operator_NotEqual const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_LessOrEqual:
            dump_cpp_Operator_LessOrEqual (static_cast <Cpp_Operator_LessOrEqual const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_GreaterOrEqual:
            dump_cpp_Operator_GreaterOrEqual (static_cast <Cpp_Operator_GreaterOrEqual const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_LogicalAnd:
            dump_cpp_Operator_LogicalAnd (static_cast <Cpp_Operator_LogicalAnd const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_LogicalOr:
            dump_cpp_Operator_LogicalOr (static_cast <Cpp_Operator_LogicalOr const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Increment:
            dump_cpp_Operator_Increment (static_cast <Cpp_Operator_Increment const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Decrement:
            dump_cpp_Operator_Decrement (static_cast <Cpp_Operator_Decrement const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Comma:
            dump_cpp_Operator_Comma (static_cast <Cpp_Operator_Comma const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_MemberDereference:
            dump_cpp_Operator_MemberDereference (static_cast <Cpp_Operator_MemberDereference const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Dereference:
            dump_cpp_Operator_Dereference (static_cast <Cpp_Operator_Dereference const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_FunctionCall:
            dump_cpp_Operator_FunctionCall (static_cast <Cpp_Operator_FunctionCall const *> (el), nest_level + 0);
            break;
        case Cpp_Operator::_Subscripting:
            dump_cpp_Operator_Subscripting (static_cast <Cpp_Operator_Subscripting const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

bool cpp_begin_template_declaration (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_template_declaration (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_begin_template_declaration (_el, parser_control, data);
}

static void
cpp_TemplateDeclaration_set_templateParameterList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateDeclaration ||
             __el->cpp_element_type != CppElement::_TemplateParameterList);

    Cpp_TemplateDeclaration * const &self = static_cast <Cpp_TemplateDeclaration*> (__self);

    self->templateParameterList = static_cast <Cpp_TemplateParameterList*> (___el);
}

bool cpp_accept_template_parameter_list (
        CppElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_template_parameter_list (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);

    return cpp_accept_template_parameter_list (_el, parser_control, data);
}

static void
cpp_TemplateDeclaration_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateDeclaration ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_TemplateDeclaration * const &self = static_cast <Cpp_TemplateDeclaration*> (__self);

    self->declaration = static_cast <Cpp_Declaration*> (___el);
}

bool cpp_accept_template_declaration (
        Cpp_TemplateDeclaration *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_template_declaration (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TemplateDeclaration);

    Cpp_TemplateDeclaration * const &el =
            static_cast <Cpp_TemplateDeclaration*> (_el);

    return cpp_accept_template_declaration (el, parser_control, data);
}

static ParserElement*
cpp_TemplateDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateDeclaration))) Cpp_TemplateDeclaration;
}

static Ref<Grammar>
create_cpp_TemplateDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateDeclaration_creation_func));
    grammar->name = String::forData ("TemplateDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("export"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_template_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameterList ();
        entry->assignment_func = cpp_TemplateDeclaration_set_templateParameterList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_parameter_list;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_TemplateDeclaration_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateDeclaration (Cpp_TemplateDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateDeclaration\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: export_opt\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TemplateParameterList (el->templateParameterList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    dump_cpp_Declaration (el->declaration, nest_level + 1);

}

static void
cpp_TemplateParameterList_TemplateParameterList_set_templateParameter (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateParameterList ||
             __el->cpp_element_type != CppElement::_TemplateParameter);

    Cpp_TemplateParameterList * const &_self = static_cast <Cpp_TemplateParameterList*> (__self);

    abortIf (_self->template_parameter_list_type != Cpp_TemplateParameterList::_TemplateParameterList);

    Cpp_TemplateParameterList_TemplateParameterList * const &self = static_cast <Cpp_TemplateParameterList_TemplateParameterList*> (_self);

    self->templateParameter = static_cast <Cpp_TemplateParameter*> (___el);
}

static void
cpp_TemplateParameterList_TemplateParameterList_set_templateParameterList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateParameterList ||
             __el->cpp_element_type != CppElement::_TemplateParameterList);

    Cpp_TemplateParameterList * const &_self = static_cast <Cpp_TemplateParameterList*> (__self);

    abortIf (_self->template_parameter_list_type != Cpp_TemplateParameterList::_TemplateParameterList);

    Cpp_TemplateParameterList_TemplateParameterList * const &self = static_cast <Cpp_TemplateParameterList_TemplateParameterList*> (_self);

    self->templateParameterList = static_cast <Cpp_TemplateParameterList*> (___el);
}

static ParserElement*
cpp_TemplateParameterList_TemplateParameterList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateParameterList_TemplateParameterList))) Cpp_TemplateParameterList_TemplateParameterList;
}

static Ref<Grammar>
create_cpp_TemplateParameterList_TemplateParameterList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateParameterList_TemplateParameterList_creation_func));
    grammar->name = String::forData ("TemplateParameterList_TemplateParameterList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameter ();
        entry->assignment_func = cpp_TemplateParameterList_TemplateParameterList_set_templateParameter;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameterList ();
        entry->assignment_func = cpp_TemplateParameterList_TemplateParameterList_set_templateParameterList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateParameterList_TemplateParameterList (Cpp_TemplateParameterList_TemplateParameterList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateParameterList_TemplateParameterList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateParameterList_TemplateParameterList\n");

    dump_cpp_TemplateParameter (el->templateParameter, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_TemplateParameterList (el->templateParameterList, nest_level + 1);

}

static void
cpp_TemplateParameterList_TemplateParameter_set_templateParameter (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateParameterList ||
             __el->cpp_element_type != CppElement::_TemplateParameter);

    Cpp_TemplateParameterList * const &_self = static_cast <Cpp_TemplateParameterList*> (__self);

    abortIf (_self->template_parameter_list_type != Cpp_TemplateParameterList::_TemplateParameter);

    Cpp_TemplateParameterList_TemplateParameter * const &self = static_cast <Cpp_TemplateParameterList_TemplateParameter*> (_self);

    self->templateParameter = static_cast <Cpp_TemplateParameter*> (___el);
}

static ParserElement*
cpp_TemplateParameterList_TemplateParameter_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateParameterList_TemplateParameter))) Cpp_TemplateParameterList_TemplateParameter;
}

static Ref<Grammar>
create_cpp_TemplateParameterList_TemplateParameter ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateParameterList_TemplateParameter_creation_func));
    grammar->name = String::forData ("TemplateParameterList_TemplateParameter");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameter ();
        entry->assignment_func = cpp_TemplateParameterList_TemplateParameter_set_templateParameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateParameterList_TemplateParameter (Cpp_TemplateParameterList_TemplateParameter const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateParameterList_TemplateParameter\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateParameterList_TemplateParameter\n");

    dump_cpp_TemplateParameter (el->templateParameter, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TemplateParameterList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TemplateParameterList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameterList_TemplateParameterList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameterList_TemplateParameter ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateParameterList (Cpp_TemplateParameterList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateParameterList\n");
        return;
    }

    switch (el->template_parameter_list_type) {
        case Cpp_TemplateParameterList::_TemplateParameterList:
            dump_cpp_TemplateParameterList_TemplateParameterList (static_cast <Cpp_TemplateParameterList_TemplateParameterList const *> (el), nest_level + 0);
            break;
        case Cpp_TemplateParameterList::_TemplateParameter:
            dump_cpp_TemplateParameterList_TemplateParameter (static_cast <Cpp_TemplateParameterList_TemplateParameter const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_TemplateParameter_TypeParameter_set_typeParameter (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateParameter ||
             __el->cpp_element_type != CppElement::_TypeParameter);

    Cpp_TemplateParameter * const &_self = static_cast <Cpp_TemplateParameter*> (__self);

    abortIf (_self->template_parameter_type != Cpp_TemplateParameter::_TypeParameter);

    Cpp_TemplateParameter_TypeParameter * const &self = static_cast <Cpp_TemplateParameter_TypeParameter*> (_self);

    self->typeParameter = static_cast <Cpp_TypeParameter*> (___el);
}

static ParserElement*
cpp_TemplateParameter_TypeParameter_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateParameter_TypeParameter))) Cpp_TemplateParameter_TypeParameter;
}

static Ref<Grammar>
create_cpp_TemplateParameter_TypeParameter ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateParameter_TypeParameter_creation_func));
    grammar->name = String::forData ("TemplateParameter_TypeParameter");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter ();
        entry->assignment_func = cpp_TemplateParameter_TypeParameter_set_typeParameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateParameter_TypeParameter (Cpp_TemplateParameter_TypeParameter const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateParameter_TypeParameter\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateParameter_TypeParameter\n");

    dump_cpp_TypeParameter (el->typeParameter, nest_level + 1);

}

static void
cpp_TemplateParameter_ParameterDeclaration_set_parameterDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateParameter ||
             __el->cpp_element_type != CppElement::_ParameterDeclaration);

    Cpp_TemplateParameter * const &_self = static_cast <Cpp_TemplateParameter*> (__self);

    abortIf (_self->template_parameter_type != Cpp_TemplateParameter::_ParameterDeclaration);

    Cpp_TemplateParameter_ParameterDeclaration * const &self = static_cast <Cpp_TemplateParameter_ParameterDeclaration*> (_self);

    self->parameterDeclaration = static_cast <Cpp_ParameterDeclaration*> (___el);
}

static ParserElement*
cpp_TemplateParameter_ParameterDeclaration_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateParameter_ParameterDeclaration))) Cpp_TemplateParameter_ParameterDeclaration;
}

static Ref<Grammar>
create_cpp_TemplateParameter_ParameterDeclaration ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateParameter_ParameterDeclaration_creation_func));
    grammar->name = String::forData ("TemplateParameter_ParameterDeclaration");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ParameterDeclaration ();
        entry->assignment_func = cpp_TemplateParameter_ParameterDeclaration_set_parameterDeclaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateParameter_ParameterDeclaration (Cpp_TemplateParameter_ParameterDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateParameter_ParameterDeclaration\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateParameter_ParameterDeclaration\n");

    dump_cpp_ParameterDeclaration (el->parameterDeclaration, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TemplateParameter ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TemplateParameter");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameter_TypeParameter ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameter_ParameterDeclaration ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateParameter (Cpp_TemplateParameter const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateParameter\n");
        return;
    }

    switch (el->template_parameter_type) {
        case Cpp_TemplateParameter::_TypeParameter:
            dump_cpp_TemplateParameter_TypeParameter (static_cast <Cpp_TemplateParameter_TypeParameter const *> (el), nest_level + 0);
            break;
        case Cpp_TemplateParameter::_ParameterDeclaration:
            dump_cpp_TemplateParameter_ParameterDeclaration (static_cast <Cpp_TemplateParameter_ParameterDeclaration const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_TypeParameter_Class_Default_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Class_Default);

    Cpp_TypeParameter_Class_Default * const &self = static_cast <Cpp_TypeParameter_Class_Default*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_TypeParameter_Class_Default_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Class_Default);

    Cpp_TypeParameter_Class_Default * const &self = static_cast <Cpp_TypeParameter_Class_Default*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

bool cpp_accept_type_template_parameter (
        Cpp_TypeParameter *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_type_template_parameter (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TypeParameter);

    Cpp_TypeParameter * const &el =
            static_cast <Cpp_TypeParameter*> (_el);

    return cpp_accept_type_template_parameter (el, parser_control, data);
}

static ParserElement*
cpp_TypeParameter_Class_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeParameter_Class_Default))) Cpp_TypeParameter_Class_Default;
}

static Ref<Grammar>
create_cpp_TypeParameter_Class_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeParameter_Class_Default_creation_func));
    grammar->name = String::forData ("TypeParameter_Class_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("class"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypeParameter_Class_Default_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_TypeParameter_Class_Default_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_template_parameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter_Class_Default (Cpp_TypeParameter_Class_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter_Class_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeParameter_Class_Default\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: class\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

}

static void
cpp_TypeParameter_Class_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Class);

    Cpp_TypeParameter_Class * const &self = static_cast <Cpp_TypeParameter_Class*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_TypeParameter_Class_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeParameter_Class))) Cpp_TypeParameter_Class;
}

static Ref<Grammar>
create_cpp_TypeParameter_Class ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeParameter_Class_creation_func));
    grammar->name = String::forData ("TypeParameter_Class");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("class"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypeParameter_Class_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_template_parameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter_Class (Cpp_TypeParameter_Class const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter_Class\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeParameter_Class\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: class\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_TypeParameter_Typename_Default_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Typename_Default);

    Cpp_TypeParameter_Typename_Default * const &self = static_cast <Cpp_TypeParameter_Typename_Default*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_TypeParameter_Typename_Default_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Typename_Default);

    Cpp_TypeParameter_Typename_Default * const &self = static_cast <Cpp_TypeParameter_Typename_Default*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static ParserElement*
cpp_TypeParameter_Typename_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeParameter_Typename_Default))) Cpp_TypeParameter_Typename_Default;
}

static Ref<Grammar>
create_cpp_TypeParameter_Typename_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeParameter_Typename_Default_creation_func));
    grammar->name = String::forData ("TypeParameter_Typename_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypeParameter_Typename_Default_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_TypeParameter_Typename_Default_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_template_parameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter_Typename_Default (Cpp_TypeParameter_Typename_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter_Typename_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeParameter_Typename_Default\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typename\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_TypeId (el->typeId, nest_level + 1);

}

static void
cpp_TypeParameter_Typename_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Typename);

    Cpp_TypeParameter_Typename * const &self = static_cast <Cpp_TypeParameter_Typename*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_TypeParameter_Typename_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeParameter_Typename))) Cpp_TypeParameter_Typename;
}

static Ref<Grammar>
create_cpp_TypeParameter_Typename ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeParameter_Typename_creation_func));
    grammar->name = String::forData ("TypeParameter_Typename");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("typename"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypeParameter_Typename_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_template_parameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter_Typename (Cpp_TypeParameter_Typename const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter_Typename\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeParameter_Typename\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: typename\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_TypeParameter_Template_Default_set_templateParameterList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_TemplateParameterList);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Template_Default);

    Cpp_TypeParameter_Template_Default * const &self = static_cast <Cpp_TypeParameter_Template_Default*> (_self);

    self->templateParameterList = static_cast <Cpp_TemplateParameterList*> (___el);
}

static void
cpp_TypeParameter_Template_Default_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Template_Default);

    Cpp_TypeParameter_Template_Default * const &self = static_cast <Cpp_TypeParameter_Template_Default*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static void
cpp_TypeParameter_Template_Default_set_idExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_IdExpression);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Template_Default);

    Cpp_TypeParameter_Template_Default * const &self = static_cast <Cpp_TypeParameter_Template_Default*> (_self);

    self->idExpression = static_cast <Cpp_IdExpression*> (___el);
}

static ParserElement*
cpp_TypeParameter_Template_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeParameter_Template_Default))) Cpp_TypeParameter_Template_Default;
}

static Ref<Grammar>
create_cpp_TypeParameter_Template_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeParameter_Template_Default_creation_func));
    grammar->name = String::forData ("TypeParameter_Template_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_template_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameterList ();
        entry->assignment_func = cpp_TypeParameter_Template_Default_set_templateParameterList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_parameter_list;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("class"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypeParameter_Template_Default_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("="));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IdExpression ();
        entry->assignment_func = cpp_TypeParameter_Template_Default_set_idExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_template_parameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter_Template_Default (Cpp_TypeParameter_Template_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter_Template_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeParameter_Template_Default\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TemplateParameterList (el->templateParameterList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: class\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: =\n");
    dump_cpp_IdExpression (el->idExpression, nest_level + 1);

}

static void
cpp_TypeParameter_Template_set_templateParameterList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_TemplateParameterList);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Template);

    Cpp_TypeParameter_Template * const &self = static_cast <Cpp_TypeParameter_Template*> (_self);

    self->templateParameterList = static_cast <Cpp_TemplateParameterList*> (___el);
}

static void
cpp_TypeParameter_Template_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeParameter ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TypeParameter * const &_self = static_cast <Cpp_TypeParameter*> (__self);

    abortIf (_self->type_parameter_type != Cpp_TypeParameter::_Template);

    Cpp_TypeParameter_Template * const &self = static_cast <Cpp_TypeParameter_Template*> (_self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_TypeParameter_Template_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeParameter_Template))) Cpp_TypeParameter_Template;
}

static Ref<Grammar>
create_cpp_TypeParameter_Template ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeParameter_Template_creation_func));
    grammar->name = String::forData ("TypeParameter_Template");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_template_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateParameterList ();
        entry->assignment_func = cpp_TypeParameter_Template_set_templateParameterList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_parameter_list;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("class"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TypeParameter_Template_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_type_template_parameter;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter_Template (Cpp_TypeParameter_Template const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter_Template\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeParameter_Template\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TemplateParameterList (el->templateParameterList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: class\n");
    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TypeParameter ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TypeParameter");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter_Class_Default ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter_Class ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter_Typename_Default ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter_Typename ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter_Template_Default ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeParameter_Template ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeParameter (Cpp_TypeParameter const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeParameter\n");
        return;
    }

    switch (el->type_parameter_type) {
        case Cpp_TypeParameter::_Class_Default:
            dump_cpp_TypeParameter_Class_Default (static_cast <Cpp_TypeParameter_Class_Default const *> (el), nest_level + 0);
            break;
        case Cpp_TypeParameter::_Class:
            dump_cpp_TypeParameter_Class (static_cast <Cpp_TypeParameter_Class const *> (el), nest_level + 0);
            break;
        case Cpp_TypeParameter::_Typename_Default:
            dump_cpp_TypeParameter_Typename_Default (static_cast <Cpp_TypeParameter_Typename_Default const *> (el), nest_level + 0);
            break;
        case Cpp_TypeParameter::_Typename:
            dump_cpp_TypeParameter_Typename (static_cast <Cpp_TypeParameter_Typename const *> (el), nest_level + 0);
            break;
        case Cpp_TypeParameter::_Template_Default:
            dump_cpp_TypeParameter_Template_Default (static_cast <Cpp_TypeParameter_Template_Default const *> (el), nest_level + 0);
            break;
        case Cpp_TypeParameter::_Template:
            dump_cpp_TypeParameter_Template (static_cast <Cpp_TypeParameter_Template const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_TemplateId_set_templateName (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateId ||
             __el->cpp_element_type != CppElement::_TemplateName);

    Cpp_TemplateId * const &self = static_cast <Cpp_TemplateId*> (__self);

    self->templateName = static_cast <Cpp_TemplateName*> (___el);
}

bool cpp_begin_template_id (
        Cpp_TemplateId *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_begin_template_id (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TemplateId);

    Cpp_TemplateId * const &el =
            static_cast <Cpp_TemplateId*> (_el);

    return cpp_begin_template_id (el, parser_control, data);
}

static void
cpp_TemplateId_set_templateArgumentList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateId ||
             __el->cpp_element_type != CppElement::_TemplateArgumentList);

    Cpp_TemplateId * const &self = static_cast <Cpp_TemplateId*> (__self);

    self->templateArgumentList = static_cast <Cpp_TemplateArgumentList*> (___el);
}

bool cpp_accept_template_id (
        Cpp_TemplateId *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_template_id (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TemplateId);

    Cpp_TemplateId * const &el =
            static_cast <Cpp_TemplateId*> (_el);

    return cpp_accept_template_id (el, parser_control, data);
}

static ParserElement*
cpp_TemplateId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateId))) Cpp_TemplateId;
}

static Ref<Grammar>
create_cpp_TemplateId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateId_creation_func));
    grammar->name = String::forData ("TemplateId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateName ();
        entry->assignment_func = cpp_TemplateId_set_templateName;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_template_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_TemplateArgumentList ();
        entry->assignment_func = cpp_TemplateId_set_templateArgumentList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_id;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateId (Cpp_TemplateId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateId\n");

    dump_cpp_TemplateName (el->templateName, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    dump_cpp_TemplateArgumentList (el->templateArgumentList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
}

bool cpp_TemplateName_match_func (
        Cpp_TemplateName *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_TemplateName_match_func (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el = static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TemplateName);

    Cpp_TemplateName * const &el = static_cast <Cpp_TemplateName*> (_el);

    return cpp_TemplateName_match_func (el, parser_control, data);
}

static void
cpp_TemplateName_set_identifier (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateName ||
             __el->cpp_element_type != CppElement::_Identifier);

    Cpp_TemplateName * const &self = static_cast <Cpp_TemplateName*> (__self);

    self->identifier = static_cast <Cpp_Identifier*> (___el);
}

static ParserElement*
cpp_TemplateName_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateName))) Cpp_TemplateName;
}

static Ref<Grammar>
create_cpp_TemplateName ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateName_creation_func));
    grammar->name = String::forData ("TemplateName");
    grammar->match_func = __pargen_cpp_TemplateName_match_func;

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Identifier ();
        entry->assignment_func = cpp_TemplateName_set_identifier;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateName (Cpp_TemplateName const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateName\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateName\n");

    dump_cpp_Identifier (el->identifier, nest_level + 1);

}

static void
cpp_TemplateArgumentList_TemplateArgumentList_set_templateArgument (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateArgumentList ||
             __el->cpp_element_type != CppElement::_TemplateArgument);

    Cpp_TemplateArgumentList * const &_self = static_cast <Cpp_TemplateArgumentList*> (__self);

    abortIf (_self->template_argument_list_type != Cpp_TemplateArgumentList::_TemplateArgumentList);

    Cpp_TemplateArgumentList_TemplateArgumentList * const &self = static_cast <Cpp_TemplateArgumentList_TemplateArgumentList*> (_self);

    self->templateArgument = static_cast <Cpp_TemplateArgument*> (___el);
}

static void
cpp_TemplateArgumentList_TemplateArgumentList_set_templateArgumentList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateArgumentList ||
             __el->cpp_element_type != CppElement::_TemplateArgumentList);

    Cpp_TemplateArgumentList * const &_self = static_cast <Cpp_TemplateArgumentList*> (__self);

    abortIf (_self->template_argument_list_type != Cpp_TemplateArgumentList::_TemplateArgumentList);

    Cpp_TemplateArgumentList_TemplateArgumentList * const &self = static_cast <Cpp_TemplateArgumentList_TemplateArgumentList*> (_self);

    self->templateArgumentList = static_cast <Cpp_TemplateArgumentList*> (___el);
}

static ParserElement*
cpp_TemplateArgumentList_TemplateArgumentList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateArgumentList_TemplateArgumentList))) Cpp_TemplateArgumentList_TemplateArgumentList;
}

static Ref<Grammar>
create_cpp_TemplateArgumentList_TemplateArgumentList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateArgumentList_TemplateArgumentList_creation_func));
    grammar->name = String::forData ("TemplateArgumentList_TemplateArgumentList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgument ();
        entry->assignment_func = cpp_TemplateArgumentList_TemplateArgumentList_set_templateArgument;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->is_jump = true;
        entry->jump_grammar = create_cpp_TemplateArgumentList ();
        entry->jump_switch_grammar_index = 1;
        entry->jump_compound_grammar_index = 1;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgumentList ();
        entry->assignment_func = cpp_TemplateArgumentList_TemplateArgumentList_set_templateArgumentList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgumentList_TemplateArgumentList (Cpp_TemplateArgumentList_TemplateArgumentList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgumentList_TemplateArgumentList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateArgumentList_TemplateArgumentList\n");

    dump_cpp_TemplateArgument (el->templateArgument, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_TemplateArgumentList (el->templateArgumentList, nest_level + 1);

}

static void
cpp_TemplateArgumentList_TemplateArgument_set_templateArgument (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateArgumentList ||
             __el->cpp_element_type != CppElement::_TemplateArgument);

    Cpp_TemplateArgumentList * const &_self = static_cast <Cpp_TemplateArgumentList*> (__self);

    abortIf (_self->template_argument_list_type != Cpp_TemplateArgumentList::_TemplateArgument);

    Cpp_TemplateArgumentList_TemplateArgument * const &self = static_cast <Cpp_TemplateArgumentList_TemplateArgument*> (_self);

    self->templateArgument = static_cast <Cpp_TemplateArgument*> (___el);
}

static ParserElement*
cpp_TemplateArgumentList_TemplateArgument_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateArgumentList_TemplateArgument))) Cpp_TemplateArgumentList_TemplateArgument;
}

static Ref<Grammar>
create_cpp_TemplateArgumentList_TemplateArgument ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateArgumentList_TemplateArgument_creation_func));
    grammar->name = String::forData ("TemplateArgumentList_TemplateArgument");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgument ();
        entry->assignment_func = cpp_TemplateArgumentList_TemplateArgument_set_templateArgument;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgumentList_TemplateArgument (Cpp_TemplateArgumentList_TemplateArgument const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgumentList_TemplateArgument\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateArgumentList_TemplateArgument\n");

    dump_cpp_TemplateArgument (el->templateArgument, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TemplateArgumentList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TemplateArgumentList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgumentList_TemplateArgumentList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgumentList_TemplateArgument ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgumentList (Cpp_TemplateArgumentList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgumentList\n");
        return;
    }

    switch (el->template_argument_list_type) {
        case Cpp_TemplateArgumentList::_TemplateArgumentList:
            dump_cpp_TemplateArgumentList_TemplateArgumentList (static_cast <Cpp_TemplateArgumentList_TemplateArgumentList const *> (el), nest_level + 0);
            break;
        case Cpp_TemplateArgumentList::_TemplateArgument:
            dump_cpp_TemplateArgumentList_TemplateArgument (static_cast <Cpp_TemplateArgumentList_TemplateArgument const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_TemplateArgument_TypeId_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateArgument ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_TemplateArgument * const &_self = static_cast <Cpp_TemplateArgument*> (__self);

    abortIf (_self->template_argument_type != Cpp_TemplateArgument::_TypeId);

    Cpp_TemplateArgument_TypeId * const &self = static_cast <Cpp_TemplateArgument_TypeId*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

bool cpp_accept_template_argument (
        Cpp_TemplateArgument *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_template_argument (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_TemplateArgument);

    Cpp_TemplateArgument * const &el =
            static_cast <Cpp_TemplateArgument*> (_el);

    return cpp_accept_template_argument (el, parser_control, data);
}

static ParserElement*
cpp_TemplateArgument_TypeId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateArgument_TypeId))) Cpp_TemplateArgument_TypeId;
}

static Ref<Grammar>
create_cpp_TemplateArgument_TypeId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateArgument_TypeId_creation_func));
    grammar->name = String::forData ("TemplateArgument_TypeId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_TemplateArgument_TypeId_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_argument;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgument_TypeId (Cpp_TemplateArgument_TypeId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgument_TypeId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateArgument_TypeId\n");

    dump_cpp_TypeId (el->typeId, nest_level + 1);

}

static void
cpp_TemplateArgument_AssignmentExpression_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateArgument ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_TemplateArgument * const &_self = static_cast <Cpp_TemplateArgument*> (__self);

    abortIf (_self->template_argument_type != Cpp_TemplateArgument::_AssignmentExpression);

    Cpp_TemplateArgument_AssignmentExpression * const &self = static_cast <Cpp_TemplateArgument_AssignmentExpression*> (_self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

static ParserElement*
cpp_TemplateArgument_AssignmentExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateArgument_AssignmentExpression))) Cpp_TemplateArgument_AssignmentExpression;
}

static Ref<Grammar>
create_cpp_TemplateArgument_AssignmentExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateArgument_AssignmentExpression_creation_func));
    grammar->name = String::forData ("TemplateArgument_AssignmentExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_TemplateArgument_AssignmentExpression_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_argument;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgument_AssignmentExpression (Cpp_TemplateArgument_AssignmentExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgument_AssignmentExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateArgument_AssignmentExpression\n");

    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_TemplateArgument_IdExpression_set_idExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TemplateArgument ||
             __el->cpp_element_type != CppElement::_IdExpression);

    Cpp_TemplateArgument * const &_self = static_cast <Cpp_TemplateArgument*> (__self);

    abortIf (_self->template_argument_type != Cpp_TemplateArgument::_IdExpression);

    Cpp_TemplateArgument_IdExpression * const &self = static_cast <Cpp_TemplateArgument_IdExpression*> (_self);

    self->idExpression = static_cast <Cpp_IdExpression*> (___el);
}

static ParserElement*
cpp_TemplateArgument_IdExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TemplateArgument_IdExpression))) Cpp_TemplateArgument_IdExpression;
}

static Ref<Grammar>
create_cpp_TemplateArgument_IdExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TemplateArgument_IdExpression_creation_func));
    grammar->name = String::forData ("TemplateArgument_IdExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_id;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_IdExpression ();
        entry->assignment_func = cpp_TemplateArgument_IdExpression_set_idExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_template_argument;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgument_IdExpression (Cpp_TemplateArgument_IdExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgument_IdExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TemplateArgument_IdExpression\n");

    dump_cpp_IdExpression (el->idExpression, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TemplateArgument ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TemplateArgument");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgument_TypeId ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgument_AssignmentExpression ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TemplateArgument_IdExpression ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TemplateArgument (Cpp_TemplateArgument const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TemplateArgument\n");
        return;
    }

    switch (el->template_argument_type) {
        case Cpp_TemplateArgument::_TypeId:
            dump_cpp_TemplateArgument_TypeId (static_cast <Cpp_TemplateArgument_TypeId const *> (el), nest_level + 0);
            break;
        case Cpp_TemplateArgument::_AssignmentExpression:
            dump_cpp_TemplateArgument_AssignmentExpression (static_cast <Cpp_TemplateArgument_AssignmentExpression const *> (el), nest_level + 0);
            break;
        case Cpp_TemplateArgument::_IdExpression:
            dump_cpp_TemplateArgument_IdExpression (static_cast <Cpp_TemplateArgument_IdExpression const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ExplicitInstantiation_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExplicitInstantiation ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_ExplicitInstantiation * const &self = static_cast <Cpp_ExplicitInstantiation*> (__self);

    self->declaration = static_cast <Cpp_Declaration*> (___el);
}

static ParserElement*
cpp_ExplicitInstantiation_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExplicitInstantiation))) Cpp_ExplicitInstantiation;
}

static Ref<Grammar>
create_cpp_ExplicitInstantiation ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExplicitInstantiation_creation_func));
    grammar->name = String::forData ("ExplicitInstantiation");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_ExplicitInstantiation_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExplicitInstantiation (Cpp_ExplicitInstantiation const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExplicitInstantiation\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExplicitInstantiation\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    dump_cpp_Declaration (el->declaration, nest_level + 1);

}

static void
cpp_ExplicitSpecialization_set_declaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExplicitSpecialization ||
             __el->cpp_element_type != CppElement::_Declaration);

    Cpp_ExplicitSpecialization * const &self = static_cast <Cpp_ExplicitSpecialization*> (__self);

    self->declaration = static_cast <Cpp_Declaration*> (___el);
}

static ParserElement*
cpp_ExplicitSpecialization_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExplicitSpecialization))) Cpp_ExplicitSpecialization;
}

static Ref<Grammar>
create_cpp_ExplicitSpecialization ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExplicitSpecialization_creation_func));
    grammar->name = String::forData ("ExplicitSpecialization");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("template"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("<"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (">"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Declaration ();
        entry->assignment_func = cpp_ExplicitSpecialization_set_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExplicitSpecialization (Cpp_ExplicitSpecialization const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExplicitSpecialization\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExplicitSpecialization\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: template\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: <\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: >\n");
    dump_cpp_Declaration (el->declaration, nest_level + 1);

}

static void
cpp_TryBlock_set_compoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TryBlock ||
             __el->cpp_element_type != CppElement::_CompoundStatement);

    Cpp_TryBlock * const &self = static_cast <Cpp_TryBlock*> (__self);

    self->compoundStatement = static_cast <Cpp_CompoundStatement*> (___el);
}

static void
cpp_TryBlock_set_handlerSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TryBlock ||
             __el->cpp_element_type != CppElement::_HandlerSeq);

    Cpp_TryBlock * const &self = static_cast <Cpp_TryBlock*> (__self);

    self->handlerSeq = static_cast <Cpp_HandlerSeq*> (___el);
}

static ParserElement*
cpp_TryBlock_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TryBlock))) Cpp_TryBlock;
}

static Ref<Grammar>
create_cpp_TryBlock ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TryBlock_creation_func));
    grammar->name = String::forData ("TryBlock");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("try"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CompoundStatement ();
        entry->assignment_func = cpp_TryBlock_set_compoundStatement;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_HandlerSeq ();
        entry->assignment_func = cpp_TryBlock_set_handlerSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TryBlock (Cpp_TryBlock const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TryBlock\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TryBlock\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: try\n");
    dump_cpp_CompoundStatement (el->compoundStatement, nest_level + 1);

    dump_cpp_HandlerSeq (el->handlerSeq, nest_level + 1);

}

static void
cpp_FunctionTryBlock_set_ctorInitializer (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionTryBlock ||
             __el->cpp_element_type != CppElement::_CtorInitializer);

    Cpp_FunctionTryBlock * const &self = static_cast <Cpp_FunctionTryBlock*> (__self);

    self->ctorInitializer = static_cast <Cpp_CtorInitializer*> (___el);
}

static void
cpp_FunctionTryBlock_set_functionBody (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionTryBlock ||
             __el->cpp_element_type != CppElement::_FunctionBody);

    Cpp_FunctionTryBlock * const &self = static_cast <Cpp_FunctionTryBlock*> (__self);

    self->functionBody = static_cast <Cpp_FunctionBody*> (___el);
}

static void
cpp_FunctionTryBlock_set_handlerSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_FunctionTryBlock ||
             __el->cpp_element_type != CppElement::_HandlerSeq);

    Cpp_FunctionTryBlock * const &self = static_cast <Cpp_FunctionTryBlock*> (__self);

    self->handlerSeq = static_cast <Cpp_HandlerSeq*> (___el);
}

static ParserElement*
cpp_FunctionTryBlock_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_FunctionTryBlock))) Cpp_FunctionTryBlock;
}

static Ref<Grammar>
create_cpp_FunctionTryBlock ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_FunctionTryBlock_creation_func));
    grammar->name = String::forData ("FunctionTryBlock");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("try"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_function_definition;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_CtorInitializer ();
        entry->assignment_func = cpp_FunctionTryBlock_set_ctorInitializer;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_FunctionBody ();
        entry->assignment_func = cpp_FunctionTryBlock_set_functionBody;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_HandlerSeq ();
        entry->assignment_func = cpp_FunctionTryBlock_set_handlerSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_FunctionTryBlock (Cpp_FunctionTryBlock const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) FunctionTryBlock\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("FunctionTryBlock\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: try\n");
    dump_cpp_CtorInitializer (el->ctorInitializer, nest_level + 1);

    dump_cpp_FunctionBody (el->functionBody, nest_level + 1);

    dump_cpp_HandlerSeq (el->handlerSeq, nest_level + 1);

}

static void
cpp_HandlerSeq_set_handler (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_HandlerSeq ||
             __el->cpp_element_type != CppElement::_Handler);

    Cpp_HandlerSeq * const &self = static_cast <Cpp_HandlerSeq*> (__self);

    self->handler = static_cast <Cpp_Handler*> (___el);
}

static void
cpp_HandlerSeq_set_handlerSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_HandlerSeq ||
             __el->cpp_element_type != CppElement::_HandlerSeq);

    Cpp_HandlerSeq * const &self = static_cast <Cpp_HandlerSeq*> (__self);

    self->handlerSeq = static_cast <Cpp_HandlerSeq*> (___el);
}

static ParserElement*
cpp_HandlerSeq_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_HandlerSeq))) Cpp_HandlerSeq;
}

static Ref<Grammar>
create_cpp_HandlerSeq ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_HandlerSeq_creation_func));
    grammar->name = String::forData ("HandlerSeq");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Handler ();
        entry->assignment_func = cpp_HandlerSeq_set_handler;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_HandlerSeq ();
        entry->assignment_func = cpp_HandlerSeq_set_handlerSeq;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_HandlerSeq (Cpp_HandlerSeq const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) HandlerSeq\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("HandlerSeq\n");

    dump_cpp_Handler (el->handler, nest_level + 1);

    dump_cpp_HandlerSeq (el->handlerSeq, nest_level + 1);

}

static void
cpp_Handler_Default_set_exceptionDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Handler ||
             __el->cpp_element_type != CppElement::_ExceptionDeclaration);

    Cpp_Handler * const &_self = static_cast <Cpp_Handler*> (__self);

    abortIf (_self->handler_type != Cpp_Handler::_Default);

    Cpp_Handler_Default * const &self = static_cast <Cpp_Handler_Default*> (_self);

    self->exceptionDeclaration = static_cast <Cpp_ExceptionDeclaration*> (___el);
}

static void
cpp_Handler_Default_set_compoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Handler ||
             __el->cpp_element_type != CppElement::_CompoundStatement);

    Cpp_Handler * const &_self = static_cast <Cpp_Handler*> (__self);

    abortIf (_self->handler_type != Cpp_Handler::_Default);

    Cpp_Handler_Default * const &self = static_cast <Cpp_Handler_Default*> (_self);

    self->compoundStatement = static_cast <Cpp_CompoundStatement*> (___el);
}

static ParserElement*
cpp_Handler_Default_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Handler_Default))) Cpp_Handler_Default;
}

static Ref<Grammar>
create_cpp_Handler_Default ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Handler_Default_creation_func));
    grammar->name = String::forData ("Handler_Default");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("catch"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_ExceptionDeclaration ();
        entry->assignment_func = cpp_Handler_Default_set_exceptionDeclaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_CompoundStatement ();
        entry->assignment_func = cpp_Handler_Default_set_compoundStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Handler_Default (Cpp_Handler_Default const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Handler_Default\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Handler_Default\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: catch\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_ExceptionDeclaration (el->exceptionDeclaration, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_CompoundStatement (el->compoundStatement, nest_level + 1);

}

static void
cpp_Handler_ClassPrematch_set_scruffy_SkipExceptionDeclaration (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Handler ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipExceptionDeclaration);

    Cpp_Handler * const &_self = static_cast <Cpp_Handler*> (__self);

    abortIf (_self->handler_type != Cpp_Handler::_ClassPrematch);

    Cpp_Handler_ClassPrematch * const &self = static_cast <Cpp_Handler_ClassPrematch*> (_self);

    self->scruffy_SkipExceptionDeclaration = static_cast <Cpp_Scruffy_SkipExceptionDeclaration*> (___el);
}

static void
cpp_Handler_ClassPrematch_set_scruffy_SkipCompoundStatement (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Handler ||
             __el->cpp_element_type != CppElement::_Scruffy_SkipCompoundStatement);

    Cpp_Handler * const &_self = static_cast <Cpp_Handler*> (__self);

    abortIf (_self->handler_type != Cpp_Handler::_ClassPrematch);

    Cpp_Handler_ClassPrematch * const &self = static_cast <Cpp_Handler_ClassPrematch*> (_self);

    self->scruffy_SkipCompoundStatement = static_cast <Cpp_Scruffy_SkipCompoundStatement*> (___el);
}

static ParserElement*
cpp_Handler_ClassPrematch_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Handler_ClassPrematch))) Cpp_Handler_ClassPrematch;
}

static Ref<Grammar>
create_cpp_Handler_ClassPrematch ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Handler_ClassPrematch_creation_func));
    grammar->name = String::forData ("Handler_ClassPrematch");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("catch"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipExceptionDeclaration ();
        entry->assignment_func = cpp_Handler_ClassPrematch_set_scruffy_SkipExceptionDeclaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_SkipCompoundStatement ();
        entry->assignment_func = cpp_Handler_ClassPrematch_set_scruffy_SkipCompoundStatement;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Handler_ClassPrematch (Cpp_Handler_ClassPrematch const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Handler_ClassPrematch\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("Handler_ClassPrematch\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: catch\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_Scruffy_SkipExceptionDeclaration (el->scruffy_SkipExceptionDeclaration, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
    dump_cpp_Scruffy_SkipCompoundStatement (el->scruffy_SkipCompoundStatement, nest_level + 1);

}

static Ref<Grammar>
create_cpp_Handler ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("Handler");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Handler_Default ();
    entry->variants.append (String::forData ("default"));
    entry->variants.append (String::forData ("AST"));

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_Handler_ClassPrematch ();
    entry->variants.append (String::forData ("class-prematch"));

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_Handler (Cpp_Handler const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) Handler\n");
        return;
    }

    switch (el->handler_type) {
        case Cpp_Handler::_Default:
            dump_cpp_Handler_Default (static_cast <Cpp_Handler_Default const *> (el), nest_level + 0);
            break;
        case Cpp_Handler::_ClassPrematch:
            dump_cpp_Handler_ClassPrematch (static_cast <Cpp_Handler_ClassPrematch const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ExceptionDeclaration_Declarator_set_typeSpecifierSeq (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExceptionDeclaration ||
             __el->cpp_element_type != CppElement::_TypeSpecifierSeq);

    Cpp_ExceptionDeclaration * const &_self = static_cast <Cpp_ExceptionDeclaration*> (__self);

    abortIf (_self->exception_declaration_type != Cpp_ExceptionDeclaration::_Declarator);

    Cpp_ExceptionDeclaration_Declarator * const &self = static_cast <Cpp_ExceptionDeclaration_Declarator*> (_self);

    self->typeSpecifierSeq = static_cast <Cpp_TypeSpecifierSeq*> (___el);
}

static void
cpp_ExceptionDeclaration_Declarator_set_declarator (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExceptionDeclaration ||
             __el->cpp_element_type != CppElement::_Declarator);

    Cpp_ExceptionDeclaration * const &_self = static_cast <Cpp_ExceptionDeclaration*> (__self);

    abortIf (_self->exception_declaration_type != Cpp_ExceptionDeclaration::_Declarator);

    Cpp_ExceptionDeclaration_Declarator * const &self = static_cast <Cpp_ExceptionDeclaration_Declarator*> (_self);

    self->declarator = static_cast <Cpp_Declarator*> (___el);
}

static ParserElement*
cpp_ExceptionDeclaration_Declarator_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExceptionDeclaration_Declarator))) Cpp_ExceptionDeclaration_Declarator;
}

static Ref<Grammar>
create_cpp_ExceptionDeclaration_Declarator ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExceptionDeclaration_Declarator_creation_func));
    grammar->name = String::forData ("ExceptionDeclaration_Declarator");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_declaration;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_Scruffy_TypeSpecifierSeq ();
        entry->assignment_func = cpp_ExceptionDeclaration_Declarator_set_typeSpecifierSeq;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_Scruffy_AnyDeclarator ();
        entry->assignment_func = cpp_ExceptionDeclaration_Declarator_set_declarator;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_declaration;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExceptionDeclaration_Declarator (Cpp_ExceptionDeclaration_Declarator const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExceptionDeclaration_Declarator\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExceptionDeclaration_Declarator\n");

    dump_cpp_TypeSpecifierSeq (el->typeSpecifierSeq, nest_level + 1);

    dump_cpp_Declarator (el->declarator, nest_level + 1);

}

static ParserElement*
cpp_ExceptionDeclaration_Ellipsis_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExceptionDeclaration_Ellipsis))) Cpp_ExceptionDeclaration_Ellipsis;
}

static Ref<Grammar>
create_cpp_ExceptionDeclaration_Ellipsis ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExceptionDeclaration_Ellipsis_creation_func));
    grammar->name = String::forData ("ExceptionDeclaration_Ellipsis");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("..."));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExceptionDeclaration_Ellipsis (Cpp_ExceptionDeclaration_Ellipsis const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExceptionDeclaration_Ellipsis\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExceptionDeclaration_Ellipsis\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: ...\n");
}

static Ref<Grammar>
create_cpp_ExceptionDeclaration ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("ExceptionDeclaration");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ExceptionDeclaration_Declarator ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_ExceptionDeclaration_Ellipsis ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExceptionDeclaration (Cpp_ExceptionDeclaration const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExceptionDeclaration\n");
        return;
    }

    switch (el->exception_declaration_type) {
        case Cpp_ExceptionDeclaration::_Declarator:
            dump_cpp_ExceptionDeclaration_Declarator (static_cast <Cpp_ExceptionDeclaration_Declarator const *> (el), nest_level + 0);
            break;
        case Cpp_ExceptionDeclaration::_Ellipsis:
            dump_cpp_ExceptionDeclaration_Ellipsis (static_cast <Cpp_ExceptionDeclaration_Ellipsis const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_ThrowExpression_set_assignmentExpression (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ThrowExpression ||
             __el->cpp_element_type != CppElement::_AssignmentExpression);

    Cpp_ThrowExpression * const &self = static_cast <Cpp_ThrowExpression*> (__self);

    self->assignmentExpression = static_cast <Cpp_AssignmentExpression*> (___el);
}

bool cpp_accept_throw_expression (
        Cpp_ThrowExpression *parser_element,
        Pargen::ParserControl *parser_control,
        void *data);

bool __pargen_cpp_accept_throw_expression (
        ParserElement *parser_element,
        Pargen::ParserControl *parser_control,
        void *data)
{
    CppElement * const &_el =
            static_cast <CppElement*> (parser_element);
    abortIf (_el->cpp_element_type != CppElement::_ThrowExpression);

    Cpp_ThrowExpression * const &el =
            static_cast <Cpp_ThrowExpression*> (_el);

    return cpp_accept_throw_expression (el, parser_control, data);
}

static ParserElement*
cpp_ThrowExpression_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ThrowExpression))) Cpp_ThrowExpression;
}

static Ref<Grammar>
create_cpp_ThrowExpression ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ThrowExpression_creation_func));
    grammar->name = String::forData ("ThrowExpression");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("throw"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_expression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_AssignmentExpression ();
        entry->assignment_func = cpp_ThrowExpression_set_assignmentExpression;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_accept_throw_expression;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ThrowExpression (Cpp_ThrowExpression const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ThrowExpression\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ThrowExpression\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: throw\n");
    dump_cpp_AssignmentExpression (el->assignmentExpression, nest_level + 1);

}

static void
cpp_ExceptionSpecification_set_typeIdList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_ExceptionSpecification ||
             __el->cpp_element_type != CppElement::_TypeIdList);

    Cpp_ExceptionSpecification * const &self = static_cast <Cpp_ExceptionSpecification*> (__self);

    self->typeIdList = static_cast <Cpp_TypeIdList*> (___el);
}

static ParserElement*
cpp_ExceptionSpecification_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_ExceptionSpecification))) Cpp_ExceptionSpecification;
}

static Ref<Grammar>
create_cpp_ExceptionSpecification ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_ExceptionSpecification_creation_func));
    grammar->name = String::forData ("ExceptionSpecification");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("throw"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken ("("));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->flags |= CompoundGrammarEntry::Optional;
        entry->grammar = create_cpp_TypeIdList ();
        entry->assignment_func = cpp_ExceptionSpecification_set_typeIdList;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (")"));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_ExceptionSpecification (Cpp_ExceptionSpecification const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) ExceptionSpecification\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("ExceptionSpecification\n");

    print_tab (errf, nest_level + 1);
    errf->print ("token: throw\n");
    print_tab (errf, nest_level + 1);
    errf->print ("token: (\n");
    dump_cpp_TypeIdList (el->typeIdList, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: )\n");
}

static void
cpp_TypeIdList_TypeIdList_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeIdList ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_TypeIdList * const &_self = static_cast <Cpp_TypeIdList*> (__self);

    abortIf (_self->type_id_list_type != Cpp_TypeIdList::_TypeIdList);

    Cpp_TypeIdList_TypeIdList * const &self = static_cast <Cpp_TypeIdList_TypeIdList*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static void
cpp_TypeIdList_TypeIdList_set_typeIdList (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeIdList ||
             __el->cpp_element_type != CppElement::_TypeIdList);

    Cpp_TypeIdList * const &_self = static_cast <Cpp_TypeIdList*> (__self);

    abortIf (_self->type_id_list_type != Cpp_TypeIdList::_TypeIdList);

    Cpp_TypeIdList_TypeIdList * const &self = static_cast <Cpp_TypeIdList_TypeIdList*> (_self);

    self->typeIdList = static_cast <Cpp_TypeIdList*> (___el);
}

static ParserElement*
cpp_TypeIdList_TypeIdList_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeIdList_TypeIdList))) Cpp_TypeIdList_TypeIdList;
}

static Ref<Grammar>
create_cpp_TypeIdList_TypeIdList ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeIdList_TypeIdList_creation_func));
    grammar->name = String::forData ("TypeIdList_TypeIdList");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_TypeIdList_TypeIdList_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());

        Ref<Grammar_Immediate_SingleToken> grammar__immediate = grab (new Grammar_Immediate_SingleToken (","));
        entry->grammar = grammar__immediate;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeIdList ();
        entry->assignment_func = cpp_TypeIdList_TypeIdList_set_typeIdList;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeIdList_TypeIdList (Cpp_TypeIdList_TypeIdList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeIdList_TypeIdList\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeIdList_TypeIdList\n");

    dump_cpp_TypeId (el->typeId, nest_level + 1);

    print_tab (errf, nest_level + 1);
    errf->print ("token: ,\n");
    dump_cpp_TypeIdList (el->typeIdList, nest_level + 1);

}

static void
cpp_TypeIdList_TypeId_set_typeId (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_TypeIdList ||
             __el->cpp_element_type != CppElement::_TypeId);

    Cpp_TypeIdList * const &_self = static_cast <Cpp_TypeIdList*> (__self);

    abortIf (_self->type_id_list_type != Cpp_TypeIdList::_TypeId);

    Cpp_TypeIdList_TypeId * const &self = static_cast <Cpp_TypeIdList_TypeId*> (_self);

    self->typeId = static_cast <Cpp_TypeId*> (___el);
}

static ParserElement*
cpp_TypeIdList_TypeId_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_TypeIdList_TypeId))) Cpp_TypeIdList_TypeId;
}

static Ref<Grammar>
create_cpp_TypeIdList_TypeId ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_TypeIdList_TypeId_creation_func));
    grammar->name = String::forData ("TypeIdList_TypeId");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->inline_match_func = __pargen_cpp_begin_typeid;
        grammar->grammar_entries.append (entry);
    }

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TypeId ();
        entry->assignment_func = cpp_TypeIdList_TypeId_set_typeId;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeIdList_TypeId (Cpp_TypeIdList_TypeId const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeIdList_TypeId\n");
        return;
    }

    print_tab (errf, nest_level);
    errf->print ("TypeIdList_TypeId\n");

    dump_cpp_TypeId (el->typeId, nest_level + 1);

}

static Ref<Grammar>
create_cpp_TypeIdList ()
{
    static Ref<Grammar_Switch> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Switch ());
    grammar->name = String::forData ("TypeIdList");

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeIdList_TypeIdList ();

        grammar->grammar_entries.append (entry);
    }

    {
        Ref<SwitchGrammarEntry> entry = grab (new SwitchGrammarEntry ());
        entry->grammar = create_cpp_TypeIdList_TypeId ();

        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_TypeIdList (Cpp_TypeIdList const *el, Size nest_level)
{
    if (el == NULL) {
        print_tab (errf, nest_level);
        errf->print ("(null) TypeIdList\n");
        return;
    }

    switch (el->type_id_list_type) {
        case Cpp_TypeIdList::_TypeIdList:
            dump_cpp_TypeIdList_TypeIdList (static_cast <Cpp_TypeIdList_TypeIdList const *> (el), nest_level + 0);
            break;
        case Cpp_TypeIdList::_TypeId:
            dump_cpp_TypeIdList_TypeId (static_cast <Cpp_TypeIdList_TypeId const *> (el), nest_level + 0);
            break;
        default:
            abortIfReached ();
    }
}

static void
cpp_Grammar_set_translationUnit (
        ParserElement *___self,
        ParserElement *___el)
{
    if (___self == NULL ||
        ___el == NULL)
    {
        return;
    }

    CppElement * const &__self = static_cast <CppElement*> (___self);
    CppElement * const &__el = static_cast <CppElement*> (___el);

    abortIf (__self->cpp_element_type != CppElement::_Grammar ||
             __el->cpp_element_type != CppElement::_TranslationUnit);

    Cpp_Grammar * const &self = static_cast <Cpp_Grammar*> (__self);

    self->translationUnit = static_cast <Cpp_TranslationUnit*> (___el);
}

static ParserElement*
cpp_Grammar_creation_func (VStack * const vstack /* non-null */)
{
    return new (vstack->push_malign (sizeof (Cpp_Grammar))) Cpp_Grammar;
}

Ref<Grammar>
create_cpp_grammar ()
{
    static Ref<Grammar_Compound> grammar;
    if (!grammar.isNull ())
        return grammar.ptr ();

    grammar = grab (new Grammar_Compound (cpp_Grammar_creation_func));
    grammar->name = String::forData ("Grammar");

    {
        Ref<CompoundGrammarEntry> entry = grab (new CompoundGrammarEntry ());
        entry->grammar = create_cpp_TranslationUnit ();
        entry->assignment_func = cpp_Grammar_set_translationUnit;
        grammar->grammar_entries.append (entry);
    }

    return grammar.ptr ();
}

void
dump_cpp_grammar (Cpp_Grammar const *el)
{
    if (el == NULL) {
        print_tab (errf, 0);
        errf->print ("(null) Grammar\n");
        return;
    }

    errf->print ("Grammar\n");

    dump_cpp_TranslationUnit (el->translationUnit, 1);

}

}

